/*
	The GTKWave electronic waveform viewer
    Copyright (C) 2004 Andrew Bardsley
	Copyright (C) 2000-2003 Department of Computer Science
	The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
	Copyright (C) 1999-2003 Tony Bybell

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

	`addtraces.c'
	Add traces dialogue box

	$Id: addtrace.c,v 1.21 2004/05/14 22:29:57 janinl Exp $
*/

#include "addtrace.h"
#include "log.h"
#include "rc.h"
#include "support.h"
#include "utils.h"
#include "gui_if.h"
#include "analyzer.h"
#include "help.h"
#include "strace.h"
#include "tracepane.h"
#include "signalpane.h"

static GtkWidget *add_traces_dialog = NULL;

/* MUST MATCH NUMBER OF ENTRIES IN regexp_type OPT-MENU */
static char *signals_regexps[] = {
    "\\(\\[.*\\]\\)*$",         /* WRange */
    "\\>.\\([0-9]\\)*$",        /* WStrand */
    "\\(\\[.*\\]\\)*$",         /* Range */
    "\\>.\\([0-9]\\)*$",        /* Strand */
    ""                          /* None */
};

#define SIGNALS_RE_LAST_WORD_INDEX 1

gboolean wave_add_traces_dialog_is_active (void)
{
    return add_traces_dialog != NULL;
}

static int search_regexp_num_rows = 0;
static char *search_box_regexp_text = NULL;
static guint signals_regexps_type = 0;
static int search_hier_num_rows = 0;
static SignalTree *search_hier_selected_tree = NULL;
static SignalTree *search_hier_current_tree = NULL;
static SignalTreeChain *search_hier_tree_chain = NULL;
static gint signals_page = 0;

static char *leastsig_hiername (char *nam)
{
    if (nam)
    {
        char ch, *t;

        for (t = nam; (ch = *(t++));)
        {
            if (ch == hier_delimiter)
                nam = t;
        }
    }
    return nam;
}

static void search_regexp_selector (void)
{
    int i;
    GSList *symc = NULL, *symc_ptr;
    GtkCList *clist = GTK_CLIST (lookup_widget (add_traces_dialog, "regexp_list"));

    for (i = 0; i < search_regexp_num_rows; i++)
    {
        Symbol *s = (Symbol *) gtk_clist_get_row_data (clist, i);

        if (!s->selected)
            continue;

        wave_insert_node_trace (s->n, NULL);
    }

    symc_ptr = symc;
    while (symc_ptr)
    {
        ((Symbol *) (symc_ptr->data))->selected = TRUE;
        symc_ptr = g_slist_next (symc_ptr);
    }
    g_slist_free (symc);

	wave_signal_pane_adjust_after_insertion (wave_window);
}

static void search_hier_selector (void)
{
    int i, hi, lo;
    SignalTree *t;

    /* when nothing selected... */
    if (!search_hier_selected_tree)
        return;

    /* tree_fetch_high */
    for (t = search_hier_selected_tree; t->child; t = t->child);
    hi = t->which;

    /* tree_fetch_low */
    for (t = search_hier_selected_tree; t->child;)
        for (t = t->child; t->next; t = t->next);
    lo = t->which;

    for (i = lo; i <= hi; i++)
    {
        Symbol *s = wave_symbols[i];

        if (s->vec_root)
        {
            s->vec_root->selected = wave_flags.autocoal_flag;
        }
    }

    for (i = lo; i <= hi; i++)
    {
        Symbol *s = wave_symbols[i];

        wave_insert_node_trace (s->n, NULL);
    }
	wave_signal_pane_adjust_after_insertion (wave_window);
}

static void search_hier_refresh_tree (SignalTree * t)
{
    SignalTree *t2;
    GtkCList *cl = GTK_CLIST (lookup_widget (add_traces_dialog, "hierarchy_clist"));
    GtkEntry *entry = GTK_ENTRY (lookup_widget (add_traces_dialog, "hierarchy_entry"));
    int len;
    int row;
    int pixlen = 0, maxpixlen = 0;
    static char *dotdot = "..";
    SignalTreeChain *tc = NULL;

    gtk_clist_freeze (cl);
    gtk_clist_clear (cl);

    search_hier_num_rows = 0;

    if (t != treeroot)
    {
        maxpixlen = gdk_string_measure (signalfont, (gchar *) dotdot);
    }

    if (!wave_flags.search_hier_group_flag)
    {
        char *tmp;

        for (t2 = t; t2; t2 = t2->next)
        {
            if (t2->child)
            {
                tmp = wave_alloca (strlen (t2->name) + 5);
                strcpy (tmp, "(+) ");
                strcpy (tmp + 4, t2->name);
            } else if (t2->which != -1 && wave_symbols[t2->which]->vec_root)
            {
                if (wave_flags.autocoal_flag)
                {
                    char *tmp2, *tmp3;

                    if (wave_symbols[t2->which]->vec_root != wave_symbols[t2->which])
                    {
                        continue;
                    }

                    tmp2 = makename_chain (wave_symbols[t2->which]);
                    tmp3 = leastsig_hiername (tmp2);
                    tmp = wave_alloca (strlen (tmp3) + 4);
                    strcpy (tmp, "[] ");
                    strcpy (tmp + 3, tmp3);
                    g_free (tmp2);
                } else
                {
                    tmp = wave_alloca (strlen (t2->name) + 4);
                    strcpy (tmp, "[] ");
                    strcpy (tmp + 3, t2->name);
                }
            } else
            {
                tmp = t2->name;
            }
            row = gtk_clist_prepend (cl, (gchar **) & tmp);
            pixlen = gdk_string_measure (signalfont, (gchar *) tmp);
            if (pixlen > maxpixlen)
            {
                maxpixlen = pixlen;
            }
            gtk_clist_set_row_data (cl, row, t2);
            search_hier_num_rows++;
        }
    } else
    {
        char *tmp, *tmp2, *tmp3;

        for (t2 = t; t2; t2 = t2->next)
        {
            if (t2->child)
                continue;

            if (t2->which != -1 && wave_symbols[t2->which]->vec_root)
            {
                if (wave_flags.autocoal_flag)
                {
                    if (wave_symbols[t2->which]->vec_root != wave_symbols[t2->which])
                    {
                        continue;
                    }

                    tmp2 = makename_chain (wave_symbols[t2->which]);
                    tmp3 = leastsig_hiername (tmp2);
                    tmp = wave_alloca (strlen (tmp3) + 4);
                    strcpy (tmp, "[] ");
                    strcpy (tmp + 3, tmp3);
                    g_free (tmp2);
                } else
                {
                    tmp = wave_alloca (strlen (t2->name) + 4);
                    strcpy (tmp, "[] ");
                    strcpy (tmp + 3, t2->name);
                }
            } else
            {
                tmp = t2->name;
            }

            row = gtk_clist_prepend (cl, (gchar **) & tmp);
            pixlen = gdk_string_measure (signalfont, (gchar *) tmp);
            if (maxpixlen < pixlen)
                maxpixlen = pixlen;
            gtk_clist_set_row_data (cl, row, t2);
            search_hier_num_rows++;
        }

        for (t2 = t; t2; t2 = t2->next)
        {
            if (t2->child)
            {
                tmp = wave_alloca (strlen (t2->name) + 5);
                strcpy (tmp, "(+) ");
                strcpy (tmp + 4, t2->name);
                row = gtk_clist_prepend (cl, (gchar **) & tmp);
                pixlen = gdk_string_measure (signalfont, (gchar *) tmp);
                if (maxpixlen < pixlen)
                    maxpixlen = pixlen;
                gtk_clist_set_row_data (cl, row, t2);
                search_hier_num_rows++;
            }
        }
    }

    if (t != treeroot)
    {
        row = gtk_clist_prepend (cl, (gchar **) & dotdot);
        gtk_clist_set_row_data (cl, row, NULL);
        search_hier_num_rows++;
    }

    if (maxpixlen)
        gtk_clist_set_column_width (cl, 0, maxpixlen);
    gtk_clist_thaw (cl);

    tc = search_hier_tree_chain;
    if (tc)
    {
        char *buf, *ptr;

        len = 1;
        for (; tc; tc = tc->next)
        {
            len += strlen (tc->label->name);
            if (tc->next)
                len++;
        }

        buf = ptr = g_new (char, len);

        /*
         * This is faster than repeatedly using strcat(); the result
         * is to concatenate the label->name fields from the chain
         * with hier_delimiter as a separator.
         */
        for (tc = search_hier_tree_chain; tc; tc = tc->next)
        {
            char *lptr;

            for (lptr = tc->label->name; (*ptr = *(lptr++)); ptr++);
            if (tc->next)
                *(ptr++) = hier_delimiter;
        }

        gtk_entry_set_text (entry, buf);
        g_free (buf);
    } else
    {
        gtk_entry_set_text (entry, "");
    }
}
void search_insert_callback (GtkButton * button, gpointer user_data)
{
    Traces tcache;

    memcpy (&tcache, &traces, sizeof (Traces));
    traces.total = 0;
    traces.first = traces.last = NULL;

    switch (signals_page)
    {
    case 0:
        search_hier_selector ();
        break;
    case 1:
        search_regexp_selector ();
        break;
    default:
        break;
    }

    traces.buffercount = traces.total;
    traces.buffer = traces.first;
    traces.bufferlast = traces.last;
    traces.first = tcache.first;
    traces.last = tcache.last;
    traces.total = tcache.total;

    paste_buffer (NULL);

    traces.buffercount = tcache.buffercount;
    traces.buffer = tcache.buffer;
    traces.bufferlast = tcache.bufferlast;

    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
	wave_signal_pane_adjust_after_insertion (wave_window);
}

void search_replace_callback (GtkButton * button, gpointer user_data)
{
    Traces tcache;
    TraceEnt *tfirst = NULL, *tlast = NULL;

    memcpy (&tcache, &traces, sizeof (Traces));
    traces.total = 0;
    traces.first = traces.last = NULL;

    switch (signals_page)
    {
    case 0:
        search_hier_selector ();
        break;
    case 1:
        search_regexp_selector ();
        break;
    default:
        break;
    }

    tfirst = traces.first;
    tlast = traces.last;        /* cache for selecting */

    traces.buffercount = traces.total;
    traces.buffer = traces.first;
    traces.bufferlast = traces.last;
    traces.first = tcache.first;
    traces.last = tcache.last;
    traces.total = tcache.total;

    paste_buffer (NULL);

    traces.buffercount = tcache.buffercount;
    traces.buffer = tcache.buffer;
    traces.bufferlast = tcache.bufferlast;

    cut_buffer ();

    for (; tfirst; tfirst = tfirst->next)
    {
        trace_config_set_flags (tfirst->config, TRACE_FLAGS_SELECTED, TRUE);
        if (tfirst == tlast)
            break;
    }

    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
	wave_signal_pane_adjust_after_insertion (wave_window);
}

void search_close_callback (GtkButton * button, gpointer user_data)
{
    search_delete_callback (NULL, NULL, NULL);
}

void on_traces_notebook_switch_page (GtkNotebook * notebook, GtkNotebookPage * page, gint page_num, gpointer user_data)
{
    signals_page = page_num;
}

void on_regexp_type_clicked (GtkButton * button, gpointer user_data)
{
    signals_regexps_type = wave_get_option_menu_index (GTK_OPTION_MENU (button));
}

/* wave_init_add_traces_dialog : make (or just raise) the `Add Traces' dialogue box */
void wave_init_add_traces_dialog (void)
{
    if (add_traces_dialog)
    {
        /* Gtk 1.2 has no other way to do this.  :^( */
        gdk_window_raise (add_traces_dialog->window);
        return;
    }
    add_traces_dialog = create_add_traces ();
    gtk_widget_show (add_traces_dialog);

    if (search_box_regexp_text)
    {
        gtk_entry_set_text (GTK_ENTRY (lookup_widget (add_traces_dialog, "regexp_entry")), search_box_regexp_text);
        search_regexp_entry_activate_callback (NULL, NULL);
    } else
        search_regexp_num_rows = 0;

    search_hier_current_tree = treeroot; /* treeroot defined in tree.h */
    search_hier_selected_tree = NULL;
    search_hier_tree_chain = NULL;
    search_hier_refresh_tree (search_hier_current_tree);
    signals_page = 0;
}

gboolean search_delete_callback (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
    gtk_widget_destroy (add_traces_dialog);
    add_traces_dialog = NULL;
    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
    return TRUE;
}

void search_hier_select_row_callback (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    GtkCList *cl = GTK_CLIST (lookup_widget (add_traces_dialog, "hierarchy_clist"));
    SignalTree *t;

    t = (SignalTree *) gtk_clist_get_row_data (cl, row);
    if (t)
    {
        search_hier_selected_tree = t;
        if (t->child)
        {
            SignalTreeChain *tc2, *tc = search_hier_tree_chain;

            if (tc)
            {
                while (tc->next)
                    tc = tc->next;

                tc2 = g_new (SignalTreeChain, 1);

                tc2->label = t;
                tc2->tree = search_hier_current_tree;
				tc2->next = NULL;
                tc->next = tc2;
            } else
            {
                search_hier_tree_chain = g_new (SignalTreeChain, 1);

                search_hier_tree_chain->tree = search_hier_current_tree;
                search_hier_tree_chain->label = t;
                search_hier_tree_chain->next = NULL;
            }

            search_hier_current_tree = t->child;
			search_hier_refresh_tree (search_hier_current_tree);
        }
    } else
    {
        SignalTreeChain *tc = search_hier_tree_chain;

        search_hier_selected_tree = NULL;
        if (tc)
        {
            for (;; tc = tc->next)
            {
                if (!tc->next)
                {
                    g_free (tc);
                    search_hier_tree_chain = NULL;
                    search_hier_current_tree = treeroot;
                    break;
                } else if (!tc->next->next)
                {
                    search_hier_current_tree = tc->next->tree;
                    g_free (tc->next);
                    tc->next = NULL;
                    break;
                }
            }
            search_hier_refresh_tree (search_hier_current_tree);
        }
    }
}

void search_hier_unselect_row_callback (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    GtkCList *cl = GTK_CLIST (lookup_widget (add_traces_dialog, "hierarchy_clist"));
    SignalTree *t;

    t = (SignalTree *) gtk_clist_get_row_data (cl, row);
    search_hier_selected_tree = NULL;

    if (t)
    {
        // WAVE_DEBUG (wave_log (FALSE, "Unselected: %s\n", t->config->name));
    }
}

#define WAVE_MAX_CLIST_LENGTH (10000)

void search_regexp_entry_activate_callback (GtkEditable * editable, gpointer user_data)
{
    GtkCList *cl;
    GtkWidget *entry = GTK_WIDGET (lookup_widget (add_traces_dialog, "regexp_entry"));
    gchar *entry_text;
    char *entry_suffixed, *re;
    int i, row, len, rows = 0, pixlen = 0, maxpixlen = 0;

    entry_text = gtk_entry_get_text (GTK_ENTRY (entry));
    len = strlen (entry_text);

    if (search_box_regexp_text)
    {
        search_box_regexp_text = g_realloc (search_box_regexp_text, len + 1);
    } else
    {
        search_box_regexp_text = g_new (char, len + 1);
    }
    memcpy (search_box_regexp_text, entry_text, len + 1);

    rows = 0;

    cl = GTK_CLIST (lookup_widget (add_traces_dialog, "regexp_list"));
    gtk_clist_freeze (cl);
    gtk_clist_clear (cl);

    /* MAKE REGEXP */
    entry_suffixed = re =
      wave_alloca (len + strlen (signals_regexps[signals_regexps_type]) +
      (signals_regexps_type <= SIGNALS_RE_LAST_WORD_INDEX ? 3 : 1));
    if (signals_regexps_type <= SIGNALS_RE_LAST_WORD_INDEX)
    {
        memcpy (re, "\\<", 2);  /* match on word boundary */
        re += 2;
    }
    memcpy (re, entry_text, len);
    re += len;
    strcpy (re, signals_regexps[signals_regexps_type]);
    wave_regex_compile (entry_suffixed);

    for (i = 0; i < wave_num_symbols; i++)
    {
        wave_symbols[i]->selected = FALSE;
    }

    /* gtk_widget_set_sensitive (entry, FALSE); */
    for (i = 0; i < wave_num_symbols; i++)
    {
        if (wave_regex_match (wave_symbols[i]->name))
        {
            if (!wave_symbols[i]->vec_root)
            {
                row = gtk_clist_append (cl, (gchar **) & wave_symbols[i]->name);
                pixlen = gdk_string_measure (signalfont, (gchar *) wave_symbols[i]->name);
            } else
            {
                char *s = g_strdup_printf ("[] %s", wave_symbols[i]->name);

                row = gtk_clist_append (cl, (gchar **) & s);
                pixlen = gdk_string_measure (signalfont, (gchar *) s);
                g_free (s);
            }

            if (pixlen > maxpixlen)
                maxpixlen = pixlen;
            gtk_clist_set_row_data (cl, row, wave_symbols[i]);
            if (rows == WAVE_MAX_CLIST_LENGTH)
                break;
            rows++;
        }
    }
    /* gtk_widget_set_sensitive (entry, TRUE); */

    search_regexp_num_rows = rows;

    if (maxpixlen)
        gtk_clist_set_column_width (cl, 0, maxpixlen);
    gtk_clist_thaw (cl);

    if (rows >= WAVE_MAX_CLIST_LENGTH)
    {
        wave_log (FALSE, "Limiting results to first %d entries.\n", rows);
    } else if (rows == 0)
    {
        wave_log (FALSE, "No entries match that regular expression.\n");
    }
}

void search_regexp_match_all_callback (GtkButton * button, gpointer user_data)
{
    gtk_entry_set_text (GTK_ENTRY (lookup_widget (add_traces_dialog, "regexp_entry")), ".*");
    search_regexp_entry_activate_callback (NULL, NULL);
}

void search_regexp_unselect_all_callback (GtkButton * button, gpointer user_data)
{
    gtk_clist_unselect_all (GTK_CLIST (lookup_widget (add_traces_dialog, "regexp_list")));
}

void search_regexp_select_all_callback (GtkButton * button, gpointer user_data)
{
    gtk_clist_select_all (GTK_CLIST (lookup_widget (add_traces_dialog, "regexp_list")));
}

void
search_regexp_list_select_row_callback (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    Symbol *s = (Symbol *) gtk_clist_get_row_data (GTK_CLIST (clist), row);

    s->selected = TRUE;
}

void
search_regexp_list_unselect_row_callback (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    Symbol *s = (Symbol *) gtk_clist_get_row_data (GTK_CLIST (clist), row);

    s->selected = FALSE;
}

void search_append_callback (GtkButton * button, gpointer user_data)
{
    switch (signals_page)
    {
    case 0:
        search_hier_selector ();
        break;
    case 1:
        search_regexp_selector ();
        break;
    default:
        break;
    }

    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
}

void on_hierarchy_entry_activate (GtkEditable * editable, gpointer user_data)
{
    /* Does nothing for now */
}

/* wave_show_all_traces : replace the traces in the display with a defaultly
	confifured set of all the traces */
static void wave_show_all_traces (void)
{
    wave_remove_all_signals_from_trace ();
    wave_add_all_signals_to_trace ();
    wave_flags.do_initial_zoom_fit = TRUE;
    wave_max_signal_length (wave_window);
	wave_signal_pane_adjust_after_insertion (wave_window);
    wave_redraw (wave_window);
}

static GtkWidget *wave_show_all_traces_confirm_dialog = NULL;

extern void on_show_all_traces_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (traces.first)           /* Are there any traces in the display? */
    {
        if (wave_show_all_traces_confirm_dialog)
        {
            wave_show_dialog (wave_show_all_traces_confirm_dialog);
        } else
        {
            wave_show_all_traces_confirm_dialog = create_show_all_confirm_dialog ();
            gtk_widget_show (wave_show_all_traces_confirm_dialog);
        }
    } else
    {
        wave_show_all_traces ();
    }
}

void on_show_all_confirm_ok_clicked (GtkButton * button, gpointer user_data)
{
    on_show_all_confirm_dialog_delete_event (widget_top_parent (GTK_WIDGET (button)), NULL, NULL);
    wave_show_all_traces ();
}

void on_show_all_confirm_cancel_clicked (GtkButton * button, gpointer user_data)
{
    on_show_all_confirm_dialog_delete_event (widget_top_parent (GTK_WIDGET (button)), NULL, NULL);
}

gboolean on_show_all_confirm_dialog_delete_event (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
    gtk_widget_destroy (widget);
    wave_show_all_traces_confirm_dialog = NULL;
    return FALSE;
}

/* wave_remove_all_traces_from_display : remove all the traces from the display */
void wave_remove_all_signals_from_trace (void)
{
    TraceEnt *trace_iter = traces.first;

    while (trace_iter)
    {
        TraceEnt *next = trace_iter->next;

        remove_trace (trace_iter, TRUE);
        trace_iter = next;
    }
}

/* wave_add_all_signals_to_trace : add all signals from the input file to the trace view */
void wave_add_all_signals_to_trace (void)
{
    int i;

    for (i = 0; i < wave_num_symbols; i++)
    {
        Symbol *s = wave_symbols[i];
        Symbol *t = s->vec_root;

        wave_insert_node_trace (s->n, NULL);
    }
}

void on_traces_add_traces_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_help ("Add Signals",
        "provides an easy means of adding traces to the display. Various functions are provided in the Signal Search requester which allow searching using POSIX regular expressions and bundling (coalescing individual bits into a single vector). Note that this function is disabled when Pattern Search is active"))
        return;

    /* don't mess with sigs when strace or dnd active */
    if (straces)
    {
        strace_error ();
        return;
    }
    if (wave_dnd_state)
    {
        wave_dnd_error ();
        return;
    }

    wave_init_add_traces_dialog ();
}

extern void on_show_prefixes_activate (GtkMenuItem * menuitem, gpointer user_data)
{
	gboolean active = GTK_CHECK_MENU_ITEM (menuitem)->active;
	wave_show_prefixes_state = active;
    wave_redraw (wave_window);
}
