/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `analog.c'
    Analogue drawing functions and GUI elements

    $Id: analog.c,v 1.10 2003/11/10 12:12:04 bardslea Exp $
*/

#include "analog.h"
#include "rc.h"
#include "gui_if.h"
#include "support.h"
#include "currenttime.h"
#include "strace.h"
#include "help.h"
#include "log.h"
#include "signalpane.h"
#include "tracepane.h"

static void analog_trace1 (WaveDrawingContext *context, Node *node, int *trans, int type, int lasttype,
  int x0, int y0, int y1, int yu)
{
    /* Do nothing? */
}

static void analog_trace2 (WaveDrawingContext *context, Node *node, WaveEvent * ev, int *trans, int type,
	int x, int x0, int y0, int x1, int y1, int yu, int ytext)
{
    if (ev->value.vector)
    {
        double val = *((double *) ev->value.vector);
        AnalogDrawerContext *c = context->trace->drawing_info->context;
        double range;
        int y;
        TraceState state;

        if (val < c->from)
        {
            state = TRACE_STATE_CLIP;
            val = c->from;
        } else if (val > c->to)
        {
            state = TRACE_STATE_CLIP;
            val = c->to;
        } else
        {
            state = TRACE_STATE_ANALOG;
        }
        val -= c->from;
        range = c->to - c->from;
        y = y0 - (int) ((y0 - y1) * (val / range));
        if (*trans > y0 || *trans < y1 || *trans == y)
        {
            WAVE_DRAW_STATE_LINE (context, state, x0, y, x1, y);
        } else if (wave_flags.roundcap_vectors_flag)
        {
            /* AB, FIXME, make this nicer
               WAVE_DRAW_STATE_LINE (context, state, x0, *trans, x1, y);
             */
            WAVE_DRAW_STATE_LINE (context, state, x0 - 2, *trans, x0, *trans);
            WAVE_DRAW_STATE_LINE (context, state, x0 - 2, *trans, x0, y);
            WAVE_DRAW_STATE_LINE (context, state, x0, y, x1 - 2, y);
        } else
        {
            WAVE_DRAW_STATE_LINE (context, state, x0, *trans, x0, y);
            WAVE_DRAW_STATE_LINE (context, state, x0, y, x1, y);
        }
        *trans = y;
    } else
    {
        *trans = -1;
    }
}

static gboolean analog_compat (TraceEnt * t)
{
    /*
     * FIXME: should also try to detect if the trace holds float values.
     */
    return (t->n.nd->ext != NULL);
}

static void analog_save (FILE * f, void *context)
{
    AnalogDrawerContext *c = context;

    fprintf (f, "~ %f %f\n", c->from, c->to);
}

TraceDrawer analog_trace_drawer = {
    "analog", analog_compat,
    NULL, NULL, NULL,
    analog_trace1, analog_trace2,
    analog_save,
};

void *make_analog_trace_drawer_context (double from, double to)
{
    AnalogDrawerContext *ret;

    ret = g_new (AnalogDrawerContext, 1);
    ret->from = from;
    ret->to = to;

    return (void *) ret;
}

gboolean set_analog_trace_drawer (TraceEnt * t, double from, double to)
{
    if (!analog_compat (t))
    {
        /*
         * Wouldn't work anyway.
         */
        return FALSE;
    }
    unset_trace_drawer (t);
    t->drawing_info->drawer = &analog_trace_drawer;
    t->drawing_info->context = make_analog_trace_drawer_context (from, to);
    return TRUE;
}

gboolean get_analog_trace_config (TraceEnt * t, double *from_ptr, double *to_ptr)
{
    AnalogDrawerContext *c;

    if (t->drawing_info->drawer != &analog_trace_drawer)
    {
        return FALSE;
    }
    c = t->drawing_info->context;
    *from_ptr = c->from;
    *to_ptr = c->to;
    return TRUE;
}

/* Analogue trace GUI elements */

static double analog_range_from = 0.0, analog_range_to = 0.0;
static TraceEnt *analog_tr = NULL;
static GtkWidget *analog_dialog = NULL;

void on_set_analog_range_from_changed (GtkEditable * editable, gpointer user_data)
{
    analog_range_from = (double) gtk_spin_button_get_value_as_float (GTK_SPIN_BUTTON (editable));
}

void on_set_analog_range_to_changed (GtkEditable * editable, gpointer user_data)
{
    analog_range_to = (double) gtk_spin_button_get_value_as_float (GTK_SPIN_BUTTON (editable));
}

void on_set_analog_range_cancel_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_destroy (analog_dialog);
    analog_dialog = NULL;
}

void on_set_analog_range_ok_clicked (GtkButton * button, gpointer user_data)
{
    set_analog_trace_drawer (analog_tr, analog_range_from, analog_range_to);
    gtk_widget_destroy (analog_dialog);
    analog_dialog = NULL;
    wave_redraw (wave_window);
}

void on_edit_display_format_analog_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    TraceEnt *t;

    if (wave_help ("Edit->Display Format->Analog", "formats the first selected as an analog trace, first popping up a dialog box to ask what the range of the trace should be"))
        return;

    /* don't mess with sigs when strace or wave_dnd active */
	if (straces)
	{
		strace_error ();
		return;
	}
	if (wave_dnd_state)
	{
		wave_dnd_error ();
		return;
	}

    WAVE_DEBUG (wave_log (FALSE, "WAVE_DEBUG: Display Format = Analog\n"));

    for (t = traces.first; t; t = t->next)
    {
        if (TRACE_IS_SELECTED (t->config))
        {
            if (analog_dialog)
                return;
            analog_dialog = create_set_analog_range_dialog ();
            if (!get_analog_trace_config (t, &analog_range_from, &analog_range_to))
            {
                analog_range_from = 0.0;
                analog_range_to = 1.0;
            }
            analog_tr = t;
            gtk_spin_button_set_value (GTK_SPIN_BUTTON (lookup_widget (analog_dialog, "set_analog_range_from")), analog_range_from);
            gtk_spin_button_set_value (GTK_SPIN_BUTTON (lookup_widget (analog_dialog, "set_analog_range_to")), analog_range_to);
            gtk_widget_show (analog_dialog);
            return;
        }
    }
    wave_log (TRUE, "Please select a trace to modify first.\n");
}
