/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `analyser.c'
    Signal/vector trace list handling

    $Id: analyzer.c,v 1.56 2003/11/13 11:05:37 bardslea Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "utils.h"
#include "reader.h"
#include "bsearch.h"
#include "log.h"
#include "draw.h"
#include "signalpane.h"

#ifdef _MSC_VER
#define strcasecmp _stricmp
#endif

Times tims = { 0, 0, 0, 0, 0, 0, 0 };
Traces traces = { 0, NULL, NULL, NULL, NULL, 0 };

/* wave_new_trace_config : like `new_trace_config' but use the correct default
	decor and group */
static TraceConfig *wave_new_trace_config (void)
{
    TraceConfig *ret = new_trace_config ();

    ret->group = wave_config->default_group;
    ret->decor = wave_config->default_decor;

    return ret;
}

/* 
    Add a trace to the display...
*/
static void add_trace (TraceEnt * t)
{
    if (traces.first == NULL)
    {
        t->next = t->prev = NULL;
        traces.first = traces.last = t;
    } else
    {
        t->next = NULL;
        t->prev = traces.last;
        traces.last->next = t;
        traces.last = t;
    }
    traces.total++;
}

/* wave_new_trace_ent : make an empty TraceEnt */
static TraceEnt *wave_new_trace_ent (void)
{
	TraceEnt *ret = g_new (TraceEnt, 1);

	ret->next = ret->prev = NULL;
	ret->asciivalue = NULL;
	ret->config = NULL;
	ret->drawing_info = NULL;
	ret->n.nd = NULL;

	return ret;
}

/* wave_insert_separator_trace : insert a separator trace into the
	wave window.  This currently handles blank traces, comment traces
	and extensions for vertically expanded traces.  Makes a new trace
	config if `trace_config' is NULL.  If `insert_position' is not NULL
	then insert after that trace rather than following the usu. insertion rules */
TraceEnt *wave_insert_separator_trace (char *name, TraceConfig * trace_config, TraceEnt *insert_position)
{
    TraceEnt *t = wave_new_trace_ent ();

    if (!trace_config)
    {
        t->config = wave_new_trace_config ();

        if (name)
        {
            g_free (t->config->name);
            t->config->name = g_strdup (name);
        }
    } else
        t->config = trace_config;

    trace_config_set_flags (t->config, TRACE_FLAGS_SEPARATOR, TRUE);

    if (!traces.first)
    {
        traces.first = traces.last = t;
        traces.total = 1;
        return t;
    } else
    {
        TraceEnt *temp_buffer = traces.buffer;
        TraceEnt *temp_bufferlast = traces.bufferlast;
        int temp_buffercount = traces.buffercount;

        traces.buffer = traces.bufferlast = t;
        traces.buffercount = 1;
        paste_buffer (insert_position);

        traces.buffer = temp_buffer;
        traces.bufferlast = temp_bufferlast;
        traces.buffercount = temp_buffercount;

        return t;
    }
	wave_signal_pane_adjust_after_insertion (wave_window);
}

/* wave_trace_update_config_from_drawing_info: update the config elements
	of the given TraceEnt from the suggestions provided in that TraceEnt's
	drawing_info */
static void wave_trace_update_config_from_drawing_info (TraceEnt *trace)
{
	TraceDrawingInfo *drawing_info = trace->drawing_info;
	TraceConfig *config = trace->config;

	if (! config || ! drawing_info)
		return;

	if (drawing_info->recommended_height != 0)
		config->height = drawing_info->recommended_height;
}

/* wave_insert_node_trace : insert a trace for node nd */
TraceEnt *wave_insert_node_trace (Node * nd, TraceConfig * config)
{
    TraceEnt *t;

    if (!nd)
        return NULL;            /* passed it a null node ptr by mistake */

    if (nd->import_info)
        wave_reader_import_trace (nd);

    t = wave_new_trace_ent ();

    if (! config)
    {
        t->config = wave_new_trace_config ();
        t->config->name = g_strdup (nd->nname);
    } else
		t->config = config;

    wave_window->signals.window_width_dirty = TRUE;

    if (nd->ext)
    {                           /* expansion vectors */
        int n;

        n = nd->ext->msi - nd->ext->lsi;
        if (n < 0)
        {
            n = -n;
        }
        n++;

		if (! config) /* Assume that this was correctly set if a config was offered */
		{
			trace_config_set_flags (t->config, TRACE_FLAGS_RJUSTIFIED, TRUE);

			/* FIXME, make a way to put suggested config stuff in nodes */
			if (nd->trace_format == TRACE_FORMAT_NOFORMAT)
				t->config->format = (n > 3 || n < -3 ? TRACE_FORMAT_HEX : TRACE_FORMAT_BIN);
			else t->config->format = nd->trace_format;
		}
    } else
    {
		if (! config)
		{
			if (nd->trace_format == TRACE_FORMAT_NOFORMAT)
				t->config->format = TRACE_FORMAT_BIN; /* binary */
			else t->config->format = nd->trace_format;
		}
    }
    t->n.nd = nd;
    t->drawing_info = nd->drawing_info;
    /* If we didn't have other preferences, update the config from
    	the drawing info suggestions provided by the reader */
	if (! config) wave_trace_update_config_from_drawing_info (t);
    add_trace (t);

	{ /* Add extension traces for height != 1 traces */
	   int extension_count = t->config->height - 1;

        /* Insert extension traces to the correct height */
        while (extension_count != 0)
        {
            TraceEnt *extension = wave_insert_separator_trace (NULL, NULL, NULL);

            trace_config_set_flags (extension->config, TRACE_FLAGS_EXTENSION, TRUE);

            extension_count--;
        }
    }

    return t;
}

/*
    Free up a trace's mallocs...
*/
static void free_trace (TraceEnt * t)
{
    if (t->asciivalue)
        g_free (t->asciivalue);

    g_free (t);
}

/* remove_trace : remove a trace from the display and optionally 
    deallocate its memory usage */
void remove_trace (TraceEnt * t, int dofree)
{
    traces.total--;
    if (t == traces.first)
    {
        traces.first = t->next;
        if (t->next)
            t->next->prev = NULL;
        else
            traces.last = NULL;
    } else
    {
        t->prev->next = t->next;
        if (t->next)
            t->next->prev = t->prev;
        else
            traces.last = t->prev;
    }

    if (dofree)
    {
        free_trace (t);
    }
}

/*
    Deallocate the cut/paste buffer...
*/
void free_cut_buffer (void)
{
    TraceEnt *t, *t2;

    t = traces.buffer;

    while (t)
    {
        t2 = t->next;
        free_trace (t);
        t = t2;
    }

    traces.buffer = traces.bufferlast = NULL;
    traces.buffercount = 0;
}

/*
    Cut selected traces from the main screen
    and throw them in the cut buffer.  If anything's
    in the cut buffer, deallocate it first...
*/
TraceEnt *cut_buffer (void)
{
    TraceEnt *t, *tnext;
    TraceEnt *first = NULL, *current = NULL;

    wave_window->signals.shift_click_trace = NULL;   /* so shift-clicking doesn't explode */

    t = traces.first;
    while (t)
    {
        if (TRACE_IS_SELECTED (t->config))
            break;
        t = t->next;
    }
    if (!t)
    {
        return NULL;            /* keeps a double cut from blowing out the buffer */
    }

    wave_window->signals.window_width_dirty = TRUE;

    free_cut_buffer ();
    t = traces.first;

    while (t)
    {
        tnext = t->next;
        if (TRACE_IS_SELECTED (t->config))
        {
            traces.bufferlast = t;
            traces.buffercount++;

            trace_config_set_flags (t->config, TRACE_FLAGS_SELECTED, FALSE);
            remove_trace (t, 0);
            if (!current)
            {
                first = t;
                t->prev = NULL;
            } else
            {
                current->next = t;
                t->prev = current;
            }
            current = t;
            t->next = NULL;
        }
        t = tnext;
    }

    traces.buffer = first;
    return first;
}

/*
    Paste the cut buffer into the main display one and
    mark the cut buffer empty...
	If tinsert is not NULL then insert after it rather than following the
	`after all selections or after all traces' rule.
*/
TraceEnt *paste_buffer (TraceEnt *tinsert)
{
    TraceEnt *t, *tinsertnext;

    if (!traces.buffer)
        return NULL;

    wave_window->signals.window_width_dirty = TRUE;

	if (! tinsert)
	{
		if (! traces.first)
		{
			t = traces.last = traces.first = traces.buffer;
			while (t)
			{
				traces.last = t;
				traces.total++;
				t = t->next;
			}

			traces.buffer = traces.bufferlast = NULL;
			traces.buffercount = 0;

			return traces.first;
		} else
			t = traces.first;

		while (t)
		{
			if (TRACE_IS_SELECTED (t->config))
				tinsert = t;
			t = t->next;
		}
	}

    if (! tinsert)
        tinsert = traces.last;

    tinsertnext = tinsert->next;
    tinsert->next = traces.buffer;
    traces.buffer->prev = tinsert;
    traces.bufferlast->next = tinsertnext;
    traces.total += traces.buffercount;

    if (! tinsertnext)
        traces.last = traces.bufferlast;
    else
        tinsertnext->prev = traces.bufferlast;

    traces.buffer = traces.bufferlast = NULL;
    traces.buffercount = 0;

    return traces.first;
}

/*
    Prepend the cut buffer into the main display one and
    mark the cut buffer empty...
*/
TraceEnt *prepend_buffer (void)
{
    TraceEnt *t, *prev;

    if (!traces.buffer)
        return NULL;

    wave_window->signals.window_width_dirty = TRUE;

    t = traces.buffer;

    while (t)
    {
        prev = t;
        trace_config_set_flags (t->config, TRACE_FLAGS_SEPARATOR, TRUE);
        traces.total++;
        t = t->next;
    }

    if ((prev->next = traces.first))
    {
        /* traces.last current value is ok as it stays the same */
        traces.first->prev = prev; /* but we need the reverse link back up */
    } else
    {
        traces.last = prev;
    }

    traces.first = traces.buffer;

    traces.buffer = traces.bufferlast = NULL;
    traces.buffercount = 0;

    return traces.first;
}

/*
    reversal of traces
*/
int reverse_traces (void)
{
    TraceEnt *t;
    TraceEnt **tsort, **tsort_pnt;
    int i;

    if (!traces.total)
        return 0;

    t = traces.first;
    tsort = tsort_pnt = wave_alloca (sizeof (TraceEnt *) * traces.total);
    for (i = 0; i < traces.total; i++)
    {
        if (!t)
        {
            fprintf (stderr, "INTERNAL ERROR: traces.total vs traversal mismatch");
            exit (EXIT_FAILURE);
        }

        *(tsort_pnt++) = t;

        t = t->next;
    }

    traces.first = *(--tsort_pnt);

    for (i = traces.total - 1; i >= 0; i--)
    {
        t = *tsort_pnt;

        if (i == traces.total - 1)
        {
            t->prev = NULL;
        } else
        {
            t->prev = *(tsort_pnt + 1);
        }

        if (i)
        {
            t->next = *(--tsort_pnt);
        }
    }

    traces.last = *tsort;
    traces.last->next = NULL;

    return 1;
}

/*
    sort on tracename pointers (alpha/caseins alpha/sig sort)
*/
static int tracenamecompare (const void *s1, const void *s2)
{
    char *str1, *str2;

    str1 = (*((TraceEnt **) s1))->config->name;
    str2 = (*((TraceEnt **) s2))->config->name;

    if ((!str1) || (!*str1))
    {                           /* force blank lines to go to bottom */
        if ((!str2) || (!*str2))
        {
            return 0;
        } else
        {
            return 1;
        }
    } else if ((!str2) || (!*str2))
    {
        return -1;              /* str1==str2==zero case is covered above */
    }

    return strcmp (str1, str2);
}

static int traceinamecompare (const void *s1, const void *s2)
{
    char *str1, *str2;

    str1 = (*((TraceEnt **) s1))->config->name;
    str2 = (*((TraceEnt **) s2))->config->name;

    if ((!str1) || (!*str1))
    {                           /* force blank lines to go to bottom */
        if ((!str2) || (!*str2))
        {
            return 0;
        } else
        {
            return 1;
        }
    } else if ((!str2) || (!*str2))
    {
        return -1;              /* str1==str2==zero case is covered above */
    }

    return strcasecmp (str1, str2);
}

static int tracesignamecompare (const void *s1, const void *s2)
{
    char *str1, *str2;

    str1 = (*((TraceEnt **) s1))->config->name;
    str2 = (*((TraceEnt **) s2))->config->name;

    if ((!str1) || (!*str1))
    {                           /* force blank lines to go to bottom */
        if ((!str2) || (!*str2))
        {
            return 0;
        } else
        {
            return 1;
        }
    } else if ((!str2) || (!*str2))
    {
        return -1;              /* str1==str2==zero case is covered above */
    }

    return sigcmp (str1, str2);
}

/*
    alphabetisation of traces
*/
int alphebetise_traces (int mode)
{
    TraceEnt *t, *prev = NULL;
    TraceEnt **tsort, **tsort_pnt;
    char *subst, ch;
    int i;
    int (*cptr) (const void *, const void *);

    if (!traces.total)
        return 0;

    t = traces.first;
    tsort = tsort_pnt = wave_alloca (sizeof (TraceEnt *) * traces.total);
    for (i = 0; i < traces.total; i++)
    {
        if (!t)
        {
            fprintf (stderr, "INTERNAL ERROR: traces.total vs traversal mismatch!  Exiting.\n");
            exit (255);
        }
        *(tsort_pnt++) = t;

        if ((subst = t->config->name))
            while ((ch = (*subst)))
            {
                if (ch == hier_delimiter)
                {
                    *subst = 0x01;
                }               /* forces sort at hier boundaries */
                subst++;
            }

        t = t->next;
    }

    switch (mode)
    {
    case 0:
        cptr = traceinamecompare;
        break;
    case 1:
        cptr = tracenamecompare;
        break;
    default:
        cptr = tracesignamecompare;
        break;
    }

    /* Note this has rotten performance on sorted lists */
    qsort (tsort, traces.total, sizeof (TraceEnt *), cptr);

    tsort_pnt = tsort;
    for (i = 0; i < traces.total; i++)
    {
        t = *(tsort_pnt++);

        if (!i)
        {
            traces.first = t;
            t->prev = NULL;
        } else
        {
            prev->next = t;
            t->prev = prev;
        }

        prev = t;

        if ((subst = t->config->name))
            while ((ch = (*subst)))
            {
                if (ch == 0x01)
                {
                    *subst = hier_delimiter;
                }               /* restore hier */
                subst++;
            }
    }

    traces.last = prev;
    prev->next = NULL;

    return 1;
}
