/*
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `analyser.h'
    Signal/vector trace list handling

    $Id: analyzer.h,v 1.57 2003/11/11 21:10:43 bardslea Exp $
*/

#ifndef WAVE_ANALYZER_H
#define WAVE_ANALYZER_H

#include <gtk/gtk.h>
#include <stdlib.h>
#include "alloc.h"
#include "fileio.h"
#include "config/global.h"
#include "format/format.h"
#include "format/data.h"

/* Global configuration */
extern WaveConfig *wave_config;

typedef struct
{
    GtkWidget *window;
    GtkWidget *pbar;
    GtkAdjustment *adj;
    int timer;                  /* might be used later.. */
    gfloat value, oldvalue;
}
SearchProgressData;

typedef enum
{
	HANDSHAKE_STATE_QUIESCENT = 'q',
	HANDSHAKE_STATE_REQUP = 'r',
	HANDSHAKE_STATE_ACKUP = 'a',
	HANDSHAKE_STATE_REQDOWN = 'z'
}
HandshakeState;

typedef enum
{
	NODE_FLAGS_DIGITAL = (1 << 0),	/* value is a simple digital 0XZ1 value in v.val */
	NODE_FLAGS_VECTOR = (1 << 1),	/* vector of digital chars */
	NODE_FLAGS_STRING = (1 << 2),	/* value is a string in v.vector */
	NODE_FLAGS_REAL = (1 << 3),		/* value is a real in *((double *)v.vector) */
    NODE_FLAGS_SYNC_HANDSHAKE = (1 << 4),	/* Handshake, no data */
    NODE_FLAGS_DATA_HANDSHAKE = (1 << 5) 	/* Handshake with data, vector[0] is the control char */
}
NodeFlags;

#define NODE_FLAGS_VECTOR_MASK \
	(NODE_FLAGS_VECTOR | \
	NODE_FLAGS_STRING | \
	NODE_FLAGS_REAL | \
	NODE_FLAGS_DATA_HANDSHAKE)

#define NODE_FLAGS_VAL_MASK \
	(NODE_FLAGS_DIGITAL | \
	NODE_FLAGS_SYNC_HANDSHAKE)

#define NODE_FLAGS_TYPE_MASK \
	(NODE_FLAGS_VECTOR_MASK | \
	NODE_FLAGS_VAL_MASK)

typedef struct
{
	char val;	/* Control state */
	FormatData *data;
}
HandshakeValue;

#define HANDSHAKE_VALUE(val) ((HandshakeValue *)(val))

#define MAX_HISTENT_TIME (~( (ULL_LITERAL (-1)) << (sizeof(TimeType) * 8 - 1)))

struct Node_;

typedef struct
{
    int msi, lsi;
}
ExtNode;

/* Forward declarations */
struct TraceDrawingInfo_;
struct WaveEventIterator_;

typedef struct Node_
{
    char *nname;                /* ascii name of node */
	struct WaveReader_ *reader;	/* reader this node came from */
    ExtNode *ext;               /* extension to node for vectors */

	struct WaveEventIterator_ *iterator; /* display event iterator, remember to do a seek
									before using the iterator to ensure that events are
									upto date after insertion.  The event_list for inserting
									events can be found as iterator->event_list */

    void *import_info;          /* format specific node info.  This will only be set for
                                   nodes which need to be imported to be used. */
	struct TraceDrawingInfo_ *drawing_info; /* Extra information on how this node should appear
								   when displayed.  Can be NULL */
	Format format;				/* FIXME, there *really* needs to be a suggested TraceConfig
									entry section in nodes */
	TraceFormat trace_format; 	/* Ditto */

	NodeFlags flags;			/* Node's data type flags */	

	unsigned ref_count;			/* Reference count for aliasing */
}
Node;

typedef struct
{
    TimeType first;             /* beginning time of trace */
    TimeType last;              /* end time of trace */
    TimeType start;             /* beginning time of trace on screen */
    TimeType end;               /* ending time of trace on screen */
    TimeType marker;
    TimeType prevmarker;        /* from last drawmarker()           */
    TimeType lmbcache;          /* initial marker pos                   */
    TimeType timecache;         /* to get around floating pt limitation */
    TimeType laststart;         /* caches last set value                */

    gdouble zoom;               /* current zoom  */
    gdouble prevzoom;           /* for zoom undo */
}
Times;

typedef struct TraceEnt_
{
    struct TraceEnt_ *next;     /* doubly linked list of traces */
    struct TraceEnt_ *prev;

    char *asciivalue;           /* value that marker points to */
    union
    {
        Node *nd;               /* what makes up this trace */
    }
    n;
    TraceConfig *config;        /* configuration data */
    struct TraceDrawingInfo_ *drawing_info;
}
TraceEnt;

typedef struct
{
    int total;                  /* total number of traces */
    TraceEnt *first;            /* ptr. to first trace in list */
    TraceEnt *last;             /* end of list of traces */
    TraceEnt *buffer;           /* cut/copy buffer of traces */
    TraceEnt *bufferlast;       /* last element of bufferchain */
    int buffercount;            /* number of traces in buffer */
}
Traces;

/* wave_insert_separator_trace : insert a separator trace into the
	wave window.  This currently handles blank traces, comment traces
	and extensions for vertically expanded traces.  Makes a new trace
	config if `trace_config' is NULL.  If `insert_position' is not NULL
	then insert after that trace rather than following the usu. insertion rules */
extern TraceEnt *wave_insert_separator_trace (char *name, TraceConfig * trace_config, TraceEnt *insert_position);

/* wave_insert_node_trace : insert a trace for node nd */
extern TraceEnt *wave_insert_node_trace (Node * nd, TraceConfig * config);

extern TraceEnt *cut_buffer (void);
extern void free_cut_buffer (void);
extern TraceEnt *paste_buffer (TraceEnt *tinsert);
extern TraceEnt *prepend_buffer (void);
extern int alphebetise_traces (int mode);
extern int reverse_traces (void);

extern Times tims;
extern Traces traces;

/* wave_new_node : allocate a new Node, but only fill with a name, an
	indication to which reader this Node is connected and the flags on the node
	(usually just the data type the node will carry) */
extern Node *wave_new_node (char *name, struct WaveReader_ *reader, NodeFlags flags);

/* wave_alias_node : alias a node by sharing all it's features, `dst' must
	be a pre-allocated node whose elements get overwritten */
extern void wave_alias_node (Node *dst, Node *src);

/* wave_new_ext_node : allocate a new ExtNode and fill it */
extern ExtNode *wave_new_ext_node (int msi, int lsi);

/* remove_trace : remove a trace from the display and optionally 
    deallocate its memory usage */
extern void remove_trace (TraceEnt * t, int dofree);

#endif
