/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `bitvec.c'
    Bitvector handling

    $Id: bitvec.c,v 1.47 2003/08/05 22:55:03 bardslea Exp $
*/

#include "analyzer.h"
#include "utils.h"
#include "bsearch.h"
#include <stdlib.h>
#include "reader.h"
#include "draw.h"
#include "event.h"

/* wave_new_node : allocate a new Node, but only fill with a name, an
	indication to which reader this Node is connected and the flags on the node
	(usually just the data type the node will carry) */
Node *wave_new_node (char *name, WaveReader *reader, NodeFlags flags)
{
    Node *ret = g_new (Node, 1);
    struct WaveEventList_ *event_list = wave_new_event_list (ret);

    ret->nname = name;
    ret->reader = reader;
    ret->flags = flags;
    ret->ext = NULL;
    ret->import_info = NULL;
    ret->drawing_info = NULL;
    ret->iterator = wave_new_event_iterator (event_list);

    ret->format = NULL;
    ret->trace_format = TRACE_FORMAT_NOFORMAT;

    ret->ref_count = 1;

    return ret;
}

/* wave_alias_node : alias a node by sharing all it's features, `dst' must
	be a pre-allocated node whose elements get overwritten */
void wave_alias_node (Node *dst, Node *src)
{
	src->ref_count ++; /* The reference count is used to decide when to deallocate
		the elements of the Node, not the Node structure itself (which is replicated
		for each Node) */

	/* Add stuff to this list as required */
	dst->reader = src->reader;
	if (dst->iterator)
		wave_free_event_iterator (dst->iterator);
	dst->iterator = src->iterator;
	dst->ext = src->ext;
	dst->trace_format = src->trace_format;
	dst->ref_count = src->ref_count;
}

/* wave_new_ext_node : allocate a new ExtNode and fill it */
ExtNode *wave_new_ext_node (int msi, int lsi)
{
    ExtNode *ret = g_new (ExtNode, 1);

    ret->msi = msi;
    ret->lsi = lsi;

    return ret;
}

/*
    compares two facilities a la strcmp but preserves
    numbers for comparisons

    there are two flavors..the slow and accurate to any 
    arbitrary number of digits version (first) and the
    fast one good to 2**31-1.  we default to the faster
    version since there's probably no real need to
    process ints larger than two billion anyway...
*/

#ifdef WAVE_USE_SIGCMP_INFINITE_PRECISION
int sigcmp (char *s1, char *s2)
{
    char *n1, *n2;
    unsigned char c1, c2;
    int len1, len2;

    for (;;)
    {
        c1 = (unsigned char) *s1;
        c2 = (unsigned char) *s2;

        if ((c1 == 0) && (c2 == 0))
            return 0;
        if ((c1 >= '0') && (c1 <= '9') && (c2 >= '0') && (c2 <= '9'))
        {
            n1 = s1;
            n2 = s2;
            len1 = len2 = 0;

            do
            {
                len1++;
                c1 = (unsigned char) *(n1++);
            }
            while ((c1 >= '0') && (c1 <= '9'));
            if (!c1)
                n1--;

            do
            {
                len2++;
                c2 = (unsigned char) *(n2++);
            }
            while ((c2 >= '0') && (c2 <= '9'));
            if (!c2)
                n2--;

            do
            {
                if (len1 == len2)
                {
                    c1 = (unsigned char) *(s1++);
                    len1--;
                    c2 = (unsigned char) *(s2++);
                    len2--;
                } else if (len1 < len2)
                {
                    c1 = '0';
                    c2 = (unsigned char) *(s2++);
                    len2--;
                } else
                {
                    c1 = (unsigned char) *(s1++);
                    len1--;
                    c2 = '0';
                }

                if (c1 != c2)
                    return (int) c1 - (int) c2;
            }
            while (len1);

            s1 = n1;
            s2 = n2;
            continue;
        } else if (c1 != c2)
        {
            return (int) c1 - (int) c2;
        }

        s1++;
        s2++;
    }
}
#else
int sigcmp (char *s1, char *s2)
{
    unsigned char c1, c2;
    int u1, u2;

    for (;;)
    {
        c1 = (unsigned char) *(s1++);
        c2 = (unsigned char) *(s2++);

        if (!c1 && !c2)
            return 0;
        if (c1 <= '9' && c2 <= '9' && c2 >= '0' && c1 >= '0')
        {
            u1 = (int) (c1 & 15);
            u2 = (int) (c2 & 15);

            while (((c2 = (unsigned char) *s2) >= '0') && (c2 <= '9'))
            {
                u2 *= 10;
                u2 += (unsigned int) (c2 & 15);
                s2++;
            }

            while (((c2 = (unsigned char) *s1) >= '0') && (c2 <= '9'))
            {
                u1 *= 10;
                u1 += (unsigned int) (c2 & 15);
                s1++;
            }

            if (u1 == u2)
                continue;
            else
                return (int) u1 - (int) u2;
        } else if (c1 != c2)
        {
            return (int) c1 - (int) c2;
        }
    }
}
#endif

/* 
    heapsort algorithm.  this typically outperforms quicksort.
*/
static Symbol **hp;
static void heapify (int i, int heap_size)
{
    int l, r;
    unsigned int largest;
    Symbol *t;
    int maxele = heap_size / 2 - 1; /* points to where heapswaps don't matter anymore */

    for (;;)
    {
        l = 2 * i + 1;
        r = l + 1;

        if (l < heap_size && sigcmp (hp[l]->name, hp[i]->name) > 0)
        {
            largest = l;
        } else
        {
            largest = i;
        }
        if (r < heap_size && sigcmp (hp[r]->name, hp[largest]->name) > 0)
        {
            largest = r;
        }

        if (i != largest)
        {
            t = hp[i];
            hp[i] = hp[largest];
            hp[largest] = t;

            if (largest <= maxele)
            {
                i = largest;
            } else
            {
                break;
            }
        } else
        {
            break;
        }
    }
}

void wave_heapsort (Symbol ** a, int num)
{
    int i;
    int indx = num - 1;
    Symbol *t;

    hp = a;

    for (i = (num / 2 - 1); i > 0; i--)
    {                           /* build symbols into having heap property */
        heapify (i, num);
    }

    for (;;)
    {
        if (indx)
            heapify (0, indx + 1);
        WAVE_DEBUG (printf ("%s\n", a[0]->name));

        if (indx != 0)
        {
            t = a[0];           /* sort in place by doing REVERSE sort and */
            a[0] = a[indx];     /* swapping the front and back of the tree. */
            a[indx--] = t;
        } else
        {
            break;
        }
    }
}

/*
    Malloc/Create a name from a range of signals starting from vec_root...currently the single
    bit facility_name[x] case never gets hit, but may be used in the
    future...
*/
char *makename_chain (Symbol * sym)
{
    int i;
    Symbol *symhi = NULL, *symlo = NULL;
    char hier_delimiter2;
    char *name = NULL;
    char *s1, *s2;
    int root1len = 0, root2len = 0;
    int l1, l2;

        hier_delimiter2 = '[';

    if (!wave_flags.autocoal_rev_flag)
    {                           /* normal case for MTI */
        symhi = sym;
        while (sym)
        {
            symlo = sym;
            sym = sym->vec_chain;
        }
    } else
    {                           /* for verilog XL */
        symlo = sym;
        while (sym)
        {
            symhi = sym;
            sym = sym->vec_chain;
        }
    }

    s1 = symhi->n->nname;
    s2 = symlo->n->nname;

    l1 = strlen (s1);

    for (i = l1 - 1; i >= 0; i--)
    {
        if (s1[i] == hier_delimiter2)
        {
            root1len = i + 1;
            break;
        }
    }

    l2 = strlen (s2);
    for (i = l2 - 1; i >= 0; i--)
    {
        if (s2[i] == hier_delimiter2)
        {
            root2len = i + 1;
            break;
        }
    }

    if (root1len != root2len || !root1len || !root2len || g_strncasecmp (s1, s2, root1len))
    {
        if (symlo != symhi)
        	name = g_strdup ("<Vector>");
        else name = g_strdup (s1);
    } else
    {
        int add1, add2;

        add1 = l1 - root1len;
        add2 = l2 - root2len;
            add1--;
            add2--;

        if (symlo != symhi)
        {
            unsigned char fixup1 = 0, fixup2 = 0;

                fixup1 = *(s1 + l1 - 1);
                *(s1 + l1 - 1) = 0;
                fixup2 = *(s2 + l2 - 1);
                *(s2 + l2 - 1) = 0;

            name = g_strdup_printf ("%s[%s:%s]", s1, s1 + root1len, s2 + root2len);

                *(s1 + l1 - 1) = fixup1;
                *(s2 + l2 - 1) = fixup2;
        } else
        {
            unsigned char fixup1 = 0;

                fixup1 = *(s1 + l1 - 1);
                *(s1 + l1 - 1) = 0;

            name = g_strdup_printf ("%s[%s]", s1, s1 + root1len);

                *(s1 + l1 - 1) = fixup1;
        }
    }

    return name;
}
