/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `color.c'
    Colour handling

    $Id: color.c,v 1.30 2003/11/09 22:21:33 bardslea Exp $
*/

#include <assert.h>
#include <glib.h>
#include <math.h>
#include "analyzer.h"
#include "currenttime.h"
#include "color.h"
#include "tracepane.h"

/* wave_rgb_to_gdk_color
   wave_gdk_color_to_rgb : turn an {R, G, B} double triple into
	a GdkColor triple and back again */
void wave_rgb_to_gdk_color (const gdouble * src, GdkColor * dest)
{
    assert (src);
    assert (dest);

    dest->red = ceil (src[0] * 65535.0);
    dest->green = ceil (src[1] * 65535.0);
    dest->blue = ceil (src[2] * 65535.0);
    dest->pixel = (gulong) ((dest->red >> 8) << 24 | (dest->green >> 8) << 16 | (dest->blue >> 8));
}
void wave_gdk_color_to_rgb (const GdkColor * src, gdouble * dest)
{
    assert (src);
    assert (dest);

    dest[0] = src->red / 65535.0;
    dest[1] = src->green / 65535.0;
    dest[2] = src->blue / 65535.0;
}

/* wave_color_to_gdk_color
   wave_gdk_color_to_wave_color : turn into a GDK color from a single word
   	WaveColor and back again.  Note that this depends on the GdkColor being
   	populated (ie. not returned by gdk_gc_get_values) */
GdkColor wave_color_to_gdk_color (WaveColor color)
{
    GdkColor ret;

    ret.pixel = (gulong) (color.red << 24 | color.green << 16 | color.blue);
    ret.red = (color.red << 8) + color.red;
    ret.green = (color.green << 8) + color.green;
    ret.blue = (color.blue << 8) + color.blue;

    return ret;
}

WaveColor wave_gdk_color_to_wave_color (GdkColor * color)
{
    WaveColor ret;

    ret.red = color->red >> 8;
    ret.green = color->green >> 8;
    ret.blue = color->blue >> 8;

    return ret;
}

/* wave_color_to_gc : make a new graphics context with the
	wave colour `color' as the foreground colour */
GdkGC *wave_color_to_gc (GtkWidget * widget, WaveColor color)
{
    GtkStyle *style;
    GdkGCValuesMask gc_values_mask;
    GdkGCValues gc_values;
    GdkGC *gc;
    GdkColor gdk_color = wave_color_to_gdk_color (color);

    if (!widget)
        return 0;

    style = widget->style;

    gc_values_mask = GDK_GC_FOREGROUND;
    gc_values.foreground = gdk_color;

    gdk_color_alloc (style->colormap, &gc_values.foreground);
    gc = gdk_gc_new_with_values (widget->window, &gc_values, gc_values_mask);

    return gc;
}

static GdkGC **new_gc_table (int size)
{
    int i;
    GdkGC **ret = g_new (GdkGC *, size);

    for (i = 0; i < size; i++)
        ret[i] = NULL;

    return ret;
}

static void free_gc_table (GdkGC ** gcs, int size)
{
    int i;

    for (i = 0; i < size; i++)
        if (gcs[i])
            gdk_gc_unref (gcs[i]);
    g_free (gcs);
}

/* wave_free_state_colors_gcs : free the GCs (well, unref) contained in the gcs
	field of the given TraceStateColors structure.  This is a prelude to
	deallocating the structure itself */
void wave_free_state_colors_gcs (TraceStateColors * state_colors)
{
    if (state_colors->gcs)
        free_gc_table (state_colors->gcs, NB_TRACE_STATES);
    state_colors->gcs = NULL;
}

/* wave_free_pane_colors_gcs : " " for pane colours */
void wave_free_pane_colors_gcs (TracePaneColors * pane_colors)
{
    if (pane_colors->gcs)
        free_gc_table (pane_colors->gcs, NB_TRACE_PANE_FEATURES);
    pane_colors->gcs = NULL;
}

/* wave_get_state_color_gc : get the GC to draw a particular state in a
	particular trace.  Make the GC from the colour in the state_colors if
	it's not already there */
GdkGC *wave_get_state_color_gc (TraceStateColors * state_colors, TraceState state, WaveWindow *window)
{
    GdkGC *found_gc = NULL;

    if (!state_colors->gcs)
        state_colors->gcs = new_gc_table (NB_TRACE_STATES);

    found_gc = ((GdkGC **) state_colors->gcs)[state];

    if (!found_gc)
    {
        found_gc = wave_color_to_gc (window->traces.drawing_area, state_colors->colors[state]);
        ((GdkGC **) state_colors->gcs)[state] = found_gc;
    }

    return found_gc;
}

/* wave_get_pane_color_gc : get the GC to draw a particular pane feature.
	Make the GC from the color in the pane_colors if it's not already there */
GdkGC *wave_get_pane_color_gc (TracePaneColors * pane_colors, TracePaneFeature feature, WaveWindow *window)
{
    GdkGC *found_gc = NULL;

    if (!pane_colors->gcs)
        pane_colors->gcs = new_gc_table (NB_TRACE_PANE_FEATURES);

    found_gc = ((GdkGC **) pane_colors->gcs)[feature];

    if (!found_gc)
    {
        found_gc = wave_color_to_gc (window->traces.drawing_area, pane_colors->colors[feature]);
        ((GdkGC **) pane_colors->gcs)[feature] = found_gc;
    }

    return found_gc;
}

/* wave_gdk_color_is_light : returns true if the given color is too light to
	be easily seen on a white background */
gboolean wave_gdk_color_is_light (GdkColor * color)
{
    return color->red > 0x8000 && color->green > 0x8000 && color->blue > 0x8000;
}
