/*
	The GTKWave electronic waveform viewer
	Copyright (C) 2000-2003 Department of Computer Science
	The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

	`colorsel.c'
	Colour selection dialogue box

	$Id: colorsel.c,v 1.30 2003/11/06 16:15:45 bardslea Exp $
*/

#include <gtk/gtk.h>
#include "colorsel.h"
#include "color.h"
#include "analyzer.h"
#include "currenttime.h"
#include "config/global.h"
#include "math.h"
#include "support.h"
#include "gui_if.h"
#include "log.h"
#include "tracepane.h"

/* Colour selection dialogue widgets/drawables */
static GtkWidget *color_selection_dialog = NULL;
static GdkPixmap *color_examples_pixmap;
static GtkWidget *color_selector_frame;
static GtkWidget *custom_color_selection_dialog = NULL;

/* Examples pane colours */
static GdkColor *color_examples_state_colors;
static GdkGC **color_examples_state_gcs;
static GdkColor *color_examples_pane_colors;
static GdkGC **color_examples_pane_gcs;

/* True for black-on-background examples pane indicator graphics,
	False for white-on-background */
static gboolean color_examples_black_indicators;

/* ...decor_name_ignore_change : TRUE if an outstanding `changed' event on the decor name entry
	is to be ignored */
static gboolean traces_page_decor_name_ignore_change = FALSE;
static gboolean group_page_decor_name_ignore_change = FALSE;
static gboolean decor_page_decor_name_ignore_change = FALSE;

/* ...private_colors_ignore_change : TRUE if an outstanding click on a private colours setting
	button is to be ignores */
static gboolean group_page_private_colors_ignore_change = FALSE;

/* new_decor_count : counter to auto generate decor names */
static int new_decor_count = 1;

/* Pages of the "Change Colors of..." notebook */
typedef enum
{
    COLOR_SELECTOR_NAMED_DECOR,
    COLOR_SELECTOR_TRACE_GROUP,
    COLOR_SELECTOR_SELECTED_TRACES
}
ColorSelectorChangeColorsPages;

/* Draw the single signal digital trace */
static void paint_color_selection_single_digital (GdkPixmap * pixmap,
  int y, int x, int xstep, int ystep, int font_height, GdkGC ** state_gcs)
{
    /* Unloaded trace */
    gdk_draw_rectangle (pixmap, state_gcs[TRACE_STATE_UNLOADED], TRUE, x, y - 2 * ystep, xstep, 2 * ystep);
    {
        GdkPoint unloaded_outline[] = { {x, y}, {x + xstep, y}, {x + xstep, y - 2 * ystep},
        {x, y - 2 * ystep}
        };

        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_X], unloaded_outline, 4);
    }
    gdk_draw_string (pixmap, wavefont, state_gcs[TRACE_STATE_FONT], x + 4, y - (ystep - font_height / 2), "?");

    /* Low */
    x += xstep;
    gdk_draw_line (pixmap, state_gcs[TRACE_STATE_LOW], x + 1, y, x + xstep, y);
    /* High */
    x += xstep;
    gdk_draw_line (pixmap, state_gcs[TRACE_STATE_TRANS], x, y - 1, x, y - 2 * ystep);
    gdk_draw_line (pixmap, state_gcs[TRACE_STATE_HIGH], x, y - 2 * ystep, x + xstep, y - 2 * ystep);

    /* Mid */
    x += xstep;
    gdk_draw_line (pixmap, state_gcs[TRACE_STATE_TRANS], x, (y + 1) - 2 * ystep, x, y - ystep);
    gdk_draw_line (pixmap, state_gcs[TRACE_STATE_MID], x, y - ystep, x + xstep, y - ystep);

    /* Unknown */
    x += xstep;
    gdk_draw_rectangle (pixmap, state_gcs[TRACE_STATE_XFILL], TRUE, x, y - 2 * ystep, xstep, 2 * ystep);
    {
        GdkPoint unloaded_outline[] = { {x + xstep, y}, {x, y}, {x, y - 2 * ystep},
        {x + xstep, y - 2 * ystep}
        };

        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_X], unloaded_outline, 4);
    }
    gdk_draw_string (pixmap, wavefont, state_gcs[TRACE_STATE_FONT], x + 4, y - (ystep - font_height / 2), "X");
}

/* Draw the bus digital trace */
static void paint_color_selection_bus_digital (GdkPixmap * pixmap,
  int y, int startx, int xstep, int ystep, int xvector, int font_height, GdkGC ** state_gcs)
{
    /* Three values */
    {
        int x = startx + xstep;
        GdkPoint first_value_outline[] = { {x + xstep - xvector, y}, {x + xvector, y},
        {x, y - ystep}, {x + xvector, y - 2 * ystep},
        {x + xstep - xvector, y - 2 * ystep}
        };
        GdkPoint third_value_outline[] = { {x + 2 * xstep + xvector, y}, {x + 3 * xstep - xvector, y},
        {x + 3 * xstep, y - ystep}, {x + 3 * xstep - xvector, y - 2 * ystep},
        {x + 2 * xstep + xvector, y - 2 * ystep}
        };

        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_VBOX], first_value_outline, 5);
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VBOX], x + xstep + xvector, y, x + 2 * xstep - xvector, y);
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VBOX], x + xstep + xvector, y - 2 * ystep,
          x + 2 * xstep - xvector, y - 2 * ystep);
        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_VBOX], third_value_outline, 5);

        x += xstep;
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VTRANS], x - xvector, y, x + xvector, y - 2 * ystep);
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VTRANS], x + xvector, y, x - xvector, y - 2 * ystep);

        x += xstep;
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VTRANS], x - xvector, y, x + xvector, y - 2 * ystep);
        gdk_draw_line (pixmap, state_gcs[TRACE_STATE_VTRANS], x + xvector, y, x - xvector, y - 2 * ystep);
    }

    /* Unloaded trace */
    {
        int x = startx;
        GdkPoint unloaded_outline[] = { {x, y}, {x + xstep - xvector, y},
        {x + xstep, y - ystep}, {x + xstep - xvector, y - 2 * ystep},
        {x, y - 2 * ystep}
        };

        gdk_draw_polygon (pixmap, state_gcs[TRACE_STATE_UNLOADED], TRUE, unloaded_outline, 5);
        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_X], unloaded_outline, 5);
        gdk_draw_string (pixmap, wavefont, state_gcs[TRACE_STATE_FONT], x + 4, y - (ystep - font_height / 2), "?");
    }

    /* Unknown trace */
    {
        int x = startx + 4 * xstep;
        GdkPoint x_outline[] = { {x + xstep, y}, {x + xvector, y},
        {x, y - ystep}, {x + xvector, y - 2 * ystep},
        {x + xstep, y - 2 * ystep}
        };

        gdk_draw_polygon (pixmap, state_gcs[TRACE_STATE_XFILL], TRUE, x_outline, 5);
        gdk_draw_lines (pixmap, state_gcs[TRACE_STATE_X], x_outline, 5);
        gdk_draw_string (pixmap, wavefont, state_gcs[TRACE_STATE_FONT], x + 4, y - (ystep - font_height / 2), "X");
    }
}

/* Analogue trace */
static void paint_color_selection_analog (GdkPixmap * pixmap,
  int y, int x, int xstep, int ystep, int font_height, GdkGC ** state_gcs)
{                               /* Draw an analogue trace */
    int max_x = 5 * xstep;
    double x_small_step = max_x / 50.0;
    double angle_step = M_PI / 5.0;
    double angle = angle_step;
    double x_iter = x;
    double amplitude = ystep * 2.5;
    int lasty;
    gboolean clipped, last_was_clipped = FALSE;
    GdkGC *analog_gc = state_gcs[TRACE_STATE_ANALOG];
    GdkGC *clip_gc = state_gcs[TRACE_STATE_CLIP];
    GdkGC *gc;

    y -= ystep;
    lasty = y;

    while (x_iter < (x + max_x))
    {
        int newy = y + sin (angle) * amplitude;

        gc = analog_gc;

        if (newy > (y + ystep))
        {
            newy = y + ystep;
            clipped = TRUE;
        } else if (newy < (y - ystep))
        {
            newy = y - ystep;
            clipped = TRUE;
        } else
            clipped = FALSE;

        gdk_draw_line (pixmap, (clipped && last_was_clipped ? clip_gc : analog_gc),
          x_iter, lasty, x_iter + x_small_step, newy);

        lasty = newy;
        last_was_clipped = clipped;
        amplitude *= 0.96;
        x_iter += x_small_step;
        angle += angle_step;
    }
}

/* ..._y : trace exaples y positions. Note that traces extend from
	{x, y} to {x + 4 * xstep, y - 2 * ystep} */
static int digital_bit_trace_y = 50;
static int digital_bus_trace_y = 80;
static int analog_trace_y = 110;

/* trace_examples_{width,height} : dimensions of examples pane */
static int trace_examples_width = 150;
static int trace_examples_height = 150;

/* Paint the global pane colours */
static void paint_color_selection_pane_features (GdkPixmap * pixmap, int xmargin, int xstep, GdkGC ** pane_gcs)
{
    int margin = 5;
    int vert_grid_x[2] = { xmargin + xstep / 2 - 1, xmargin + 4 * xstep + xstep / 2 + 1 };
    GdkGC *grid_gc = pane_gcs[TRACE_PANE_GRID];
    GdkGC *pane_font_gc = pane_gcs[TRACE_PANE_FONT];
    GdkGC *primary_marker_gc = pane_gcs[TRACE_PANE_PRIMARY_MARKER];
    GdkGC *named_markers_gc = pane_gcs[TRACE_PANE_NAMED_MARKERS];

    /* Ticks */
    {
        double tick_x;
        double tick_step_x = xstep / 2.0;
        int end_x = trace_examples_width - margin;

        for (tick_x = margin; tick_x <= end_x; tick_x += tick_step_x)
            gdk_draw_line (pixmap, grid_gc, tick_x, 18, tick_x, 24);

        gdk_draw_line (pixmap, grid_gc, margin, 18, (tick_x - tick_step_x), 18);
    }

    /* Times */
    gdk_draw_string (pixmap, wavefont, pane_font_gc, vert_grid_x[0] + 3, 16, "ns");
    gdk_draw_string (pixmap, wavefont, pane_font_gc, vert_grid_x[0] -
      (3 + gdk_string_width (wavefont, "10")), 16, "10");
    gdk_draw_string (pixmap, wavefont, pane_font_gc, vert_grid_x[1] + 3, 16, "ns");
    gdk_draw_string (pixmap, wavefont, pane_font_gc, vert_grid_x[1] -
      (3 + gdk_string_width (wavefont, "20")), 16, "20");

    /* Two vertical grid lines */
    gdk_draw_line (pixmap, grid_gc, vert_grid_x[0], margin, vert_grid_x[0], analog_trace_y + 10);
    gdk_draw_line (pixmap, grid_gc, vert_grid_x[1], margin, vert_grid_x[1], analog_trace_y + 10);

    /* Primary marker */
    gdk_draw_line (pixmap, primary_marker_gc, 51, margin, 51, analog_trace_y + 10);

    /* Named marker */
    gdk_draw_line (pixmap, named_markers_gc, 77, 18, 77, analog_trace_y + 10);
    gdk_draw_string (pixmap, wavefont, named_markers_gc, 77 - (gdk_string_width (wavefont, "A") / 2), 16, "A");

    /* Then grid lines under each trace */
    gdk_draw_line (pixmap, grid_gc,
      margin, digital_bit_trace_y + 3, trace_examples_width - margin, digital_bit_trace_y + 3);
    gdk_draw_line (pixmap, grid_gc,
      margin, digital_bus_trace_y + 3, trace_examples_width - margin, digital_bus_trace_y + 3);
    gdk_draw_line (pixmap, grid_gc, margin, analog_trace_y + 3, trace_examples_width - margin, analog_trace_y + 3);
}

/* paint_color_selection_examples : paint the trace appearance
	drawing area in the colour selection dialogue. */
static void paint_color_selection_examples (GtkWidget * pane, GdkPixmap * pixmap,
  GdkGC ** state_gcs, GdkGC ** pane_gcs, gboolean black_indicators)
{
    int xmargin = 12;
    int xstep = 25;
    int ystep = 7;              /* Adjust this to suit font */
    int xvector = 3;            /* x displacement of `X' shape transition arms in vector drawing */
    int font_height = gdk_string_height (wavefont, "X");

    GdkGC *back_gc = pane_gcs[TRACE_PANE_BACKGROUND];

    /*
       GdkGC *contrast_gc = (black_indicators ? pane->style->black_gc : pane->style->white_gc);
     */

    /* Adjust ystep to fit font */
    if (ystep < ((font_height / 2) + 2))
        ystep = (font_height / 2) + 2;

    /* Paint background colour across whole pixmap */
    gdk_draw_rectangle (pixmap, back_gc, TRUE, 0, 0, trace_examples_width, trace_examples_height);

    /* gdk_draw_line (pixmap, state_gcs[9], xmargin, digital_bit_trace_y - 3 * ystep, xmargin + 4 * xstep, digital_bit_trace_y - 3 * ystep); */

    /* Draw the example traces */
    paint_color_selection_pane_features (pixmap, xmargin, xstep, pane_gcs);
    paint_color_selection_single_digital (pixmap, digital_bit_trace_y, xmargin, xstep, ystep, font_height, state_gcs);
    paint_color_selection_bus_digital (pixmap, digital_bus_trace_y, xmargin, xstep, ystep,
      xvector, font_height, state_gcs);
    paint_color_selection_analog (pixmap, analog_trace_y, xmargin, xstep, ystep, font_height, state_gcs);
}

/* update_trace_state_colors_from_example_colors : make a new TraceStateColors from the colours in
	the colour selection dialogue examples pane ... or update the given TraceStateColors if it is
	non-NULL */
static TraceStateColors *update_trace_state_colors_from_example_colors (TraceStateColors * state_colors,
  GdkColor * examples_state_colors)
{
    TraceState state;
    TraceStateColors *ret;

    if (state_colors)
        ret = state_colors;
    else
    {
        ret = g_new (TraceStateColors, 1);
        ret->gcs = NULL;
        ret->colors = g_new (WaveColor, NB_TRACE_STATES);
    }

    for (state = 0; state < NB_TRACE_STATES; state++)
        ret->colors[state] = wave_gdk_color_to_wave_color (examples_state_colors + state);

    return ret;
}

/* update_color_selection_examples_state_colors : update the state colours of the examples pane,
	but don't redraw the pane */
static void update_color_selection_examples_state_colors (TraceStateColors * state_colors)
{
    int i;

    for (i = 0; i < NB_TRACE_STATES; i++)
    {
        color_examples_state_colors[i] = wave_color_to_gdk_color (state_colors->colors[i]);
        gdk_color_alloc (gtk_widget_get_colormap (color_selection_dialog), color_examples_state_colors + i);
        gdk_gc_set_foreground (color_examples_state_gcs[i], color_examples_state_colors + i);
    }
}

/* generate_new_decor_name : generate a unique decor name by counting
	up the new_decor_count until the name is unique */
static char *generate_new_decor_name (GList * decors)
{
    char *name = NULL;

    do
    {
        if (name)
            g_free (name);
        name = g_strdup_printf ("decor%d", new_decor_count);
        new_decor_count++;
    }
    while (find_trace_decor_in_list (decors, name));

    return name;
}

/* The list of pane features and trace states has row data corresponding to
	the feature/state named in that row.  State value row data has values from
	the enumeration TraceStates, pane features are numbers PANE_FEATURES_OFFSET +
	value from enumeration TracePaneFeature */
#define PANE_FEATURES_OFFSET (1000)

/* color_selection_refresh_decor_list : refresh the decor list in either an OptionMenu
	(if optionMenuNCList is TRUE) or a GtkCList (if ...), sets the menu item callback
	to `callback' (if non NULL and optionMenuNCList is TRUE) */
static void color_selection_refresh_decor_list (GtkWidget * list, gboolean optionMenuNCList,
  void (*callback) (GtkMenuItem *, TraceDecor *))
{
    GList *decors = wave_config->decors;
    GtkWidget *menu;

    if (optionMenuNCList)
    {
        menu = gtk_menu_new ();
        gtk_option_menu_remove_menu (GTK_OPTION_MENU (list));
    } else
    {
        gtk_clist_clear (GTK_CLIST (list));
        gtk_clist_freeze (GTK_CLIST (list));
    }

    while (decors)
    {
        TraceDecor *decor = (TraceDecor *) decors->data;

        if (optionMenuNCList)
        {
            GtkWidget *menu_item = gtk_menu_item_new_with_label (decor->name);

            gtk_widget_show (menu_item);
            gtk_object_set_data (GTK_OBJECT (menu_item), "decor", decor);
            gtk_menu_append (GTK_MENU (menu), menu_item);
            if (callback)
                gtk_signal_connect (GTK_OBJECT (menu_item), "activate", GTK_SIGNAL_FUNC (callback), decor);
        } else
        {
            char *text[1];

            text[0] = decor->name;
            gtk_clist_append (GTK_CLIST (list), text);
            gtk_clist_set_row_data (GTK_CLIST (list), GTK_CLIST (list)->rows - 1, decor);
        }

        decors = g_list_next (decors);
    }

    if (optionMenuNCList)
        gtk_option_menu_set_menu (GTK_OPTION_MENU (list), menu);
    else
        gtk_clist_thaw (GTK_CLIST (list));
}

/* color_selection_refresh_group_list : refresh the trace group list in the given GtkOptionMenu,
	attach the given callback (if not NULL) and pass it the TraceGroup selected */
static void color_selection_refresh_group_list (GtkOptionMenu * list, void (*callback) (GtkMenuItem *, TraceGroup *))
{
    GList *groups = wave_config->groups;
    GtkWidget *menu = gtk_menu_new ();

    gtk_option_menu_remove_menu (list);
    while (groups)
    {
        TraceGroup *group = (TraceGroup *) groups->data;

        GtkWidget *menu_item = gtk_menu_item_new_with_label (group->name);

        gtk_widget_show (menu_item);
        gtk_object_set_data (GTK_OBJECT (menu_item), "group", group);
        gtk_menu_append (GTK_MENU (menu), menu_item);

        if (callback)
            gtk_signal_connect (GTK_OBJECT (menu_item), "activate", GTK_SIGNAL_FUNC (callback), group);

        groups = g_list_next (groups);
    }
    gtk_option_menu_set_menu (list, menu);
}

/* color_selection_trace_group_selected : actions when a trace group is selected in
	the `change colors of...' notebook */
void color_selection_trace_group_selected (GtkMenuItem * menu_item, TraceGroup * group)
{
    GtkOptionMenu *group_decor_option_menu = GTK_OPTION_MENU
      (lookup_widget (color_selection_dialog, "trace_group_decor_name"));
    int decor_index = g_list_index (wave_config->decors, group->decor);

    if (group->state_colors)
    {
        group_page_private_colors_ignore_change = TRUE;
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog,
              "trace_group_private_colors")), TRUE);
        group_page_private_colors_ignore_change = FALSE;
    } else
    {
        group_page_decor_name_ignore_change = TRUE;
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog,
              "trace_group_from_decor")), TRUE);
        group_page_decor_name_ignore_change = FALSE;
    }

    gtk_widget_set_sensitive (lookup_widget (color_selection_dialog, "copy_from_preview"), group->state_colors != NULL);

    group_page_decor_name_ignore_change = TRUE;
    gtk_option_menu_set_history (group_decor_option_menu, decor_index);
    group_page_decor_name_ignore_change = FALSE;
}

gboolean color_examples_pane_expose_event (GtkWidget * widget, GdkEventExpose * event, gpointer user_data)
{
    gdk_draw_pixmap (widget->window,
      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
      color_examples_pixmap, event->area.x, event->area.y,
      event->area.x, event->area.y, event->area.width, event->area.height);

    return FALSE;
}

gboolean color_examples_pane_configure_event (GtkWidget * widget, GdkEventConfigure * event, gpointer user_data)
{
    if (!color_examples_pixmap)
    {
        color_examples_pixmap = gdk_pixmap_new (widget->window,
          widget->allocation.width, widget->allocation.height, -1);
    }

    color_examples_black_indicators = wave_gdk_color_is_light (color_examples_pane_colors + TRACE_PANE_BACKGROUND);
    paint_color_selection_examples (widget, color_examples_pixmap, color_examples_state_gcs,
      color_examples_pane_gcs, color_examples_black_indicators);

    return FALSE;
}

gboolean color_selection_dialog_dismiss (void)
{
    gtk_widget_hide (color_selection_dialog);
    gtk_widget_hide (custom_color_selection_dialog);
    return TRUE;
}

gboolean custom_color_selection_dialog_dismiss (void)
{
    gtk_widget_hide (custom_color_selection_dialog);
    return TRUE;
}

/* color_selector_row_to_feature : reports which feature/trace state is referred to by
	row `row' in the colour selector feature table.  *`paneNstate' is TRUE for pane
	colours and FALSE for trace states.  The return value is the feature/trace state
	enumeration value or -1 for an invalid selection. */
static int color_selector_row_to_feature (GtkCList * colors_list, int row, gboolean * paneNstate)
{
    int ret = -1;
    int feature_choice = GPOINTER_TO_INT (gtk_clist_get_row_data (colors_list, row));

    if (feature_choice < NB_TRACE_STATES)
    {
        *paneNstate = FALSE;
        ret = feature_choice;
    } else if (feature_choice >= PANE_FEATURES_OFFSET &&
      feature_choice < (PANE_FEATURES_OFFSET + NB_TRACE_PANE_FEATURES))
    {
        *paneNstate = TRUE;
        ret = feature_choice - PANE_FEATURES_OFFSET;
    }

    return ret;
}

void feature_colors_list_select_row (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    GtkWidget *color_selector = lookup_widget (custom_color_selection_dialog,
      "custom_color_selector");
    gdouble color[3];
    gboolean paneNstate = FALSE;
    int feature = color_selector_row_to_feature (clist, row, &paneNstate);

    /* Set colour selector colour */
    if (feature != -1)
    {
        wave_gdk_color_to_rgb ((paneNstate ? color_examples_pane_colors + feature
            : color_examples_state_colors + feature), color);

        gtk_color_selection_set_color (GTK_COLOR_SELECTION (color_selector), color);
    }
}

/* reduce_color_to_333_color_cube : reduce a colour value to just 0, 1/4, 1/2 or 1,
	based on the closeness of the given value to one of those snap values.  Can
	round up or down. */
static double reduce_color_to_333_color_cube (double color)
{
    return (color > 0.75 ? 1.0 : (color > 0.25 ? 0.5 : 0.0));
}

/* change_example_colors : change colours in the examples pane */
static void change_example_colors (GtkWidget * unused, double *color)
{
    GtkWidget *color_examples_pane = lookup_widget (color_selection_dialog, "color_examples_pane");
    GtkCList *feature_colors_list = GTK_CLIST (lookup_widget (color_selection_dialog, "feature_colors_list"));
    GList *selected_rows = feature_colors_list->selection;

    while (selected_rows)
    {
        GdkColor *gdk_color = NULL;
        GdkGC *gdk_gc = NULL;
        gboolean paneNstate = FALSE;
        int feature_choice = color_selector_row_to_feature (feature_colors_list,
          GPOINTER_TO_INT (selected_rows->data), &paneNstate);

        if (feature_choice != -1)
        {
            if (paneNstate)
            {
                gdk_color = color_examples_pane_colors + feature_choice;
                gdk_gc = color_examples_pane_gcs[feature_choice];
            } else
            {
                gdk_color = color_examples_state_colors + feature_choice;
                gdk_gc = color_examples_state_gcs[feature_choice];
            }
        }
        /* else, it's the separator */

        if (gdk_color)
        {
            wave_rgb_to_gdk_color (color, gdk_color);
            gdk_color_alloc (gtk_widget_get_colormap (color_selection_dialog), gdk_color);
            if (gdk_gc)
                gdk_gc_set_foreground (gdk_gc, gdk_color);
        }

        selected_rows = g_list_next (selected_rows);
    }

    color_examples_black_indicators = wave_gdk_color_is_light (color_examples_pane_colors + TRACE_PANE_BACKGROUND);
    paint_color_selection_examples (color_selection_dialog, color_examples_pixmap,
      color_examples_state_gcs, color_examples_pane_gcs, color_examples_black_indicators);

    gdk_draw_pixmap (color_examples_pane->window,
      color_examples_pane->style->fg_gc[GTK_WIDGET_STATE (color_examples_pane)],
      color_examples_pixmap, 0, 0, 0, 0, 150, 150);
}

/* custom_color_selection_color_changed : change colours in the examples panes */
void custom_color_selection_color_changed (GtkColorSelection * colorselection, gpointer user_data)
{
    gdouble color[3];
    gboolean snap_to_333_color_cube =
      GTK_TOGGLE_BUTTON (lookup_widget (custom_color_selection_dialog, "snap_to_333_color_cube"))->active;

    gtk_color_selection_get_color (colorselection, color);

    /* Do colour resolution reduction if necessary */
    if (snap_to_333_color_cube)
    {
        double old_color;
        gboolean color_changed = FALSE;
        int i;

        for (i = 0; i < 3; i++)
        {
            old_color = color[i];
            color[i] = reduce_color_to_333_color_cube (color[i]);

            /* Float comparison, I know, but this is only an optimisation */
            if (old_color != color[i])
                color_changed = TRUE;
        }

        if (color_changed)
            gtk_color_selection_set_color (colorselection, color);
    }

    change_example_colors (NULL, color);
}

/* color_selection_apply_colors_to_selected_traces : apply colour changes to selected traces according
	to settings in the "change colors of..." notebook */
static void color_selection_apply_colors_to_selected_traces (void)
{
    gboolean use_private_colors =
      GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_private_colors"))->active;
    gboolean use_groups_colors =
      GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_groups_colors"))->active;
    TraceDecor *decor =
      (TraceDecor *)
      wave_get_option_menu_active_data (GTK_OPTION_MENU (lookup_widget (GTK_WIDGET (color_selection_dialog),
          "selected_traces_decor_name")), "decor");
    TraceEnt *trace_iter = traces.first;

    while (trace_iter)
    {
        TraceConfig *config = trace_iter->config;

        if (TRACE_IS_SELECTED (config) && !TRACE_IS_SEPARATOR (config))
        {
            /* Destroy the private colours GCs unless there aren't any */
            if (config->state_colors)
                wave_free_state_colors_gcs (config->state_colors);

            /* Destroy the private colours completely if we aren't using them */
            if (!use_private_colors && config->state_colors)
            {
                free_trace_state_colors (config->state_colors);
                config->state_colors = NULL;
            }

            /* Copy prototype colours into the private colour set */
            if (use_private_colors)
            {
                config->state_colors = update_trace_state_colors_from_example_colors
                  (config->state_colors, color_examples_state_colors);
            } else if (use_groups_colors)
                /* Discard any decor association */
                config->decor = NULL;
            else                /* Use decor colors */
                config->decor = decor;
        }
        trace_iter = trace_iter->next;
    }
}

/* color_selection_change_trace_group_colors : apply the example colours to the
	selected trace group, or just change the trace group's decor */
static void color_selection_change_trace_group_colors (void)
{
    gboolean use_private_colors =
      GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "trace_group_private_colors"))->active;
    TraceGroup *group =
      (TraceGroup *)
      wave_get_option_menu_active_data (GTK_OPTION_MENU (lookup_widget (GTK_WIDGET (color_selection_dialog),
          "trace_group_name")), "group");
    TraceDecor *decor =
      (TraceDecor *)
      wave_get_option_menu_active_data (GTK_OPTION_MENU (lookup_widget (GTK_WIDGET (color_selection_dialog),
          "trace_group_decor_name")), "decor");

    if (group->state_colors)
        wave_free_state_colors_gcs (group->state_colors);

    if (use_private_colors)
    {
        group->state_colors = update_trace_state_colors_from_example_colors
          (group->state_colors, color_examples_state_colors);
    } else
    {
        if (group->state_colors)
            free_trace_state_colors (group->state_colors);
        group->state_colors = NULL;
    }

    /* Always reset the decor */
    group->decor = decor;
}

/* color_selection_change_decor_colors : apply the example colours to the
	selected decor */
static void color_selection_change_decor_colors (void)
{
    GtkWidget *decor_list = lookup_widget (GTK_WIDGET (color_selection_dialog), "decor_list");
    GList *selected_rows = GTK_CLIST (decor_list)->selection;

    if (selected_rows)
    {
        TraceDecor *decor = (TraceDecor *) gtk_clist_get_row_data (GTK_CLIST (decor_list),
          GPOINTER_TO_INT (selected_rows->data));

        wave_free_state_colors_gcs (decor->state_colors);
        decor->state_colors = update_trace_state_colors_from_example_colors
          (decor->state_colors, color_examples_state_colors);
    }
}

/* color_selection_apply_pane_colors : update pane colours from the example colors in the
	color_selection_dialog */
static void color_selection_apply_pane_colors (void)
{
    int feature;

    wave_free_pane_colors_gcs (wave_config->pane_colors);

    for (feature = 0; feature < NB_TRACE_PANE_FEATURES; feature++)
    {
        wave_config->pane_colors->colors[feature] = wave_gdk_color_to_wave_color (color_examples_pane_colors + feature);
    }
}

/* color_selection_dialog_apply : use `change color of...' notebook to decide
	where to copy the selected colours to */
void color_selection_dialog_apply (void)
{
    GtkWidget *notebook = lookup_widget (GTK_WIDGET (color_selection_dialog), "change_colors_of_notebook");
    int notebook_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));

    switch (notebook_page)
    {
    case COLOR_SELECTOR_SELECTED_TRACES:
        color_selection_apply_colors_to_selected_traces ();
        break;
    case COLOR_SELECTOR_TRACE_GROUP:
        color_selection_change_trace_group_colors ();
        break;
    default:
        color_selection_change_decor_colors ();
        break;
    }
    color_selection_apply_pane_colors ();

    gtk_signal_emit_by_name (GTK_OBJECT (wave_window->traces.hslider), "value_changed");
    gtk_signal_emit_by_name (GTK_OBJECT (wave_window->traces.hslider), "changed");
}

/* color_selection_dialog_ok : do apply action then hide the window */
void color_selection_dialog_ok (GtkButton * button, gpointer user_data)
{
    color_selection_dialog_apply ();
    color_selection_dialog_dismiss ();
}

/* color_selection_change_colors_of_notebook_page_switched : renew decor/group lists */
void color_selection_change_colors_of_notebook_page_switched (GtkNotebook * notebook, GtkNotebookPage * page,
  gint page_num, gpointer user_data)
{
    switch (page_num)
    {
    case COLOR_SELECTOR_SELECTED_TRACES:
        {
            GtkOptionMenu *decor_name_option_menu = GTK_OPTION_MENU
              (lookup_widget (color_selection_dialog, "selected_traces_decor_name"));
            int decor_index;
            TraceDecor *decor = wave_get_option_menu_active_data (decor_name_option_menu, "decor");

            gtk_widget_set_sensitive (lookup_widget (color_selection_dialog, "copy_from_preview"),
              GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_private_colors"))->active);

            color_selection_refresh_decor_list (GTK_WIDGET (decor_name_option_menu), TRUE,
              (void (*)(GtkMenuItem *, TraceDecor *)) color_selection_selected_traces_decor_selected);

            /* Reselect the appropriate decor */
            decor_index = g_list_index (wave_config->decors, decor);
            if (decor_index < 0)
                decor_index = 0;

            traces_page_decor_name_ignore_change = TRUE;
            gtk_option_menu_set_history (decor_name_option_menu, decor_index);
            traces_page_decor_name_ignore_change = FALSE;
        }
        break;
    case COLOR_SELECTOR_TRACE_GROUP:
        {
            GtkOptionMenu *group_name_option_menu = GTK_OPTION_MENU
              (lookup_widget (color_selection_dialog, "trace_group_name"));
            GtkOptionMenu *group_decor_option_menu = GTK_OPTION_MENU
              (lookup_widget (color_selection_dialog, "trace_group_decor_name"));
            TraceGroup *group = wave_get_option_menu_active_data (group_name_option_menu, "group");

            color_selection_refresh_decor_list (GTK_WIDGET (group_decor_option_menu), TRUE,
              (void (*)(GtkMenuItem *, TraceDecor *)) color_selection_group_decor_selected);
            color_selection_refresh_group_list (group_name_option_menu, color_selection_trace_group_selected);

            /* Reselect the appropriate group */
            if (g_list_index (wave_config->groups, group) < 0)
                group = wave_config->default_group;

            color_selection_trace_group_selected (NULL, group);
        }
        break;
    case COLOR_SELECTOR_NAMED_DECOR:
        gtk_widget_set_sensitive (lookup_widget (color_selection_dialog, "copy_from_preview"), TRUE);
        break;
    }
}

/* update_example_colors_clicked : update example colours from the colour source listed in the
	"Change Colors of ..." notebook */
void update_example_colors_clicked (GtkButton * button, gpointer user_data)
{
    GtkWidget *notebook = lookup_widget (GTK_WIDGET (button), "change_colors_of_notebook");
    int notebook_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
    TraceStateColors *update_colors = NULL;

    switch (notebook_page)
    {
    case COLOR_SELECTOR_SELECTED_TRACES:
        {                       /* Find the first selected trace and pilfer its colours */
            TraceEnt *trace_iter = traces.first;

            while (trace_iter)
            {
                TraceConfig *config = trace_iter->config;

                if (TRACE_IS_SELECTED (config) && !TRACE_IS_SEPARATOR (config))
                {
                    update_colors = trace_config_get_state_colors (config, wave_config);
                    break;
                }
                trace_iter = trace_iter->next;
            }
        }
        break;
    case COLOR_SELECTOR_TRACE_GROUP:
        {
            TraceGroup *group =
              (TraceGroup *) wave_get_option_menu_active_data (GTK_OPTION_MENU (lookup_widget (color_selection_dialog,
                  "trace_group_name")), "group");
            gboolean use_private_colors =
              GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "trace_group_private_colors"))->active;

            if (use_private_colors)
                update_colors = group->state_colors;
            else
            {
                TraceDecor *decor =
                  (TraceDecor *)
                  wave_get_option_menu_active_data (GTK_OPTION_MENU (lookup_widget (color_selection_dialog,
                      "trace_group_decor_name")), "decor");

                update_colors = decor->state_colors;
            }
        }
        break;
    case COLOR_SELECTOR_NAMED_DECOR:
        {
            GtkWidget *decor_list = lookup_widget (GTK_WIDGET (button), "decor_list");
            GList *selected_rows = GTK_CLIST (decor_list)->selection;

            if (selected_rows)
            {
                TraceDecor *decor = (TraceDecor *) gtk_clist_get_row_data (GTK_CLIST (decor_list),
                  GPOINTER_TO_INT (selected_rows->data));

                update_colors = decor->state_colors;
            }
        }
        break;
    }

    if (update_colors)
    {
        GtkWidget *pane = lookup_widget (color_selection_dialog, "color_examples_pane");

        update_color_selection_examples_state_colors (update_colors);
        paint_color_selection_examples (color_selection_dialog, color_examples_pixmap, color_examples_state_gcs,
          color_examples_pane_gcs, color_examples_black_indicators);
        gdk_draw_pixmap (pane->window,
          pane->style->fg_gc[GTK_WIDGET_STATE (pane)],
          color_examples_pixmap, 0, 0, 0, 0, trace_examples_width, trace_examples_height);
    }
}

/* color_selection_decor_... : decor list handling callbacks */
void color_selection_decor_name_changed (GtkEditable * editable, gpointer user_data)
{
    GtkWidget *list;

    if (decor_page_decor_name_ignore_change)
        return;

    list = lookup_widget (GTK_WIDGET (editable), "decor_list");

    if (GTK_CLIST (list)->selection) /* Is a selected row */
    {
        int row = GPOINTER_TO_INT (GTK_CLIST (list)->selection->data);
        TraceDecor *decor = (TraceDecor *) gtk_clist_get_row_data (GTK_CLIST (list), row);

        char *text = gtk_entry_get_text (GTK_ENTRY (editable));

        g_free (decor->name);
        decor->name = g_strdup (text);
        gtk_clist_set_text (GTK_CLIST (list), row, 0, text);
    }
}

void color_selection_decor_new_clicked (GtkButton * button, gpointer user_data)
{
    GtkWidget *decor_list = lookup_widget (GTK_WIDGET (button), "decor_list");
    int row_number = GTK_CLIST (decor_list)->rows;
    GList *selected_rows = GTK_CLIST (decor_list)->selection;
    char *name = generate_new_decor_name (wave_config->decors);
    TraceDecor *decor = new_trace_decor (name);

    /* If there is a selection, then insert after that selection rather than at the
       end of the list */
    while (selected_rows)
    {
        row_number = GPOINTER_TO_INT (selected_rows->data) + 1;
        selected_rows = g_list_next (selected_rows);
    }

    gtk_clist_insert (GTK_CLIST (decor_list), row_number, &name);
    gtk_clist_set_row_data (GTK_CLIST (decor_list), row_number, decor);
    gtk_clist_select_row (GTK_CLIST (decor_list), row_number, 0);

    /* Insert into decor list, this happens even without apply */
    wave_config->decors = g_list_insert (wave_config->decors, decor, row_number);
}

void color_selection_decor_delete_clicked (GtkButton * button, gpointer user_data)
{
    GtkWidget *decor_list = lookup_widget (GTK_WIDGET (button), "decor_list");
    GList *selected_rows = GTK_CLIST (decor_list)->selection;

    if (selected_rows)
    {
        int row_number = GPOINTER_TO_INT (selected_rows->data);
        TraceDecor *decor = (TraceDecor *) gtk_clist_get_row_data (GTK_CLIST (decor_list), row_number);
        gboolean need_to_redraw = FALSE;

        /* Don't allow default decor (always at entry 0) to be removed */
        if (row_number == 0)
            return;

        gtk_clist_remove (GTK_CLIST (decor_list), row_number);
        wave_config->decors = g_list_remove (wave_config->decors, decor);

        {                       /* Traverse the traces and replace all references to this decor */
            TraceEnt *trace_iter = traces.first;

            while (trace_iter)
            {
                TraceConfig *config = trace_iter->config;

                if (config->decor == decor)
                {
                    need_to_redraw = TRUE;
                    config->decor = wave_config->default_decor;
                }
                trace_iter = trace_iter->next;
            }
        }
        {                       /* Traverse the groups and do the same */
            GList *group_iter = wave_config->groups;

            while (group_iter)
            {
                TraceGroup *group = (TraceGroup *) group_iter->data;

                if (group->decor == decor)
                {
                    need_to_redraw = TRUE;
                    group->decor = wave_config->default_decor;
                }
                group_iter = g_list_next (group_iter);
            }
        }
        gtk_clist_select_row (GTK_CLIST (decor_list), GTK_CLIST (decor_list)->rows - 1, 0);

        if (need_to_redraw)
        {
            gtk_signal_emit_by_name (GTK_OBJECT (wave_window->traces.hslider), "value_changed");
            gtk_signal_emit_by_name (GTK_OBJECT (wave_window->traces.hslider), "changed");
        }
    }
}

void color_selection_decor_row_selected (GtkCList * clist, gint row, gint column, GdkEvent * event, gpointer user_data)
{
    char *text[1];

    gtk_clist_get_text (clist, row, 0, text);

    decor_page_decor_name_ignore_change = TRUE;
    gtk_entry_set_text (GTK_ENTRY (lookup_widget (GTK_WIDGET (clist), "decor_name_entry")), text[0]);
    decor_page_decor_name_ignore_change = FALSE;
}

/* color_selection_group_decor_selected : actions when a decor is selected for
	a trace group */
void color_selection_group_decor_selected (void)
{
    if (group_page_decor_name_ignore_change)
        return;

    if (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "trace_group_from_decor"))->active)
        color_selection_dialog_apply ();
}

/* color_selection_selected_traces_decor_selected : update colours of selected traces from a single
	decor */
void color_selection_selected_traces_decor_selected (void)
{
    if (traces_page_decor_name_ignore_change)
        return;

    if (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_from_decor"))->active)
        color_selection_dialog_apply ();
}

/* color_selection_selected_traces_trace_group_colors_clicked : update colours of selected traces to use colour
	from their preselected trace groups */
void color_selection_selected_traces_trace_group_colors_clicked (void)
{
    if (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_groups_colors"))->active)
        color_selection_dialog_apply ();
}

/* ...private_colors_clicked : update colours when the "Private Colors" buttons are clicked */
void color_selection_selected_traces_private_colors_clicked (void)
{
    gboolean active =
      GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "selected_traces_private_colors"))->active;

    gtk_widget_set_sensitive (lookup_widget (color_selection_dialog, "copy_from_preview"), active);
    if (active)
        color_selection_dialog_apply ();
}
void color_selection_group_private_colors_clicked (void)
{
    gboolean active = GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "trace_group_private_colors"))->active;

    gtk_widget_set_sensitive (lookup_widget (color_selection_dialog, "copy_from_preview"), active);

    if (group_page_private_colors_ignore_change)
        return;

    if (GTK_TOGGLE_BUTTON (lookup_widget (color_selection_dialog, "trace_group_private_colors"))->active)
        color_selection_dialog_apply ();
}

#define PALETTE_COLOR_SQUARE_SIZE (18)

/* set_colors_of_color_cube : fill in the 3x3x3 colour cube palette into the
	given table */
static void set_colors_of_color_cube (GtkTable * table)
{
    int r, g, b;

    int levels = 3;
    double step_size = 65536.0 / (levels - 1);
    double float_step_size = 1.0 / (levels - 1);

    /* gtk_table_resize (table, levels * levels, levels); */

    for (r = 0; r < levels; r++)
    {
        for (g = 0; g < levels; g++)
        {
            for (b = 0; b < levels; b++)
            {
                GtkWidget *button = gtk_button_new_with_label ("");
                GtkWidget *label = GTK_BIN (button)->child;
                GtkRcStyle *rc_style = gtk_rc_style_new ();
                double *float_color = g_new (double, 3);
                int x = b + (levels * g);
                int y = r;

                gtk_widget_set_usize (label, PALETTE_COLOR_SQUARE_SIZE, PALETTE_COLOR_SQUARE_SIZE);
                gtk_table_attach (table, button, x, x + 1, y, y + 1, GTK_FILL, GTK_FILL, 0, 0);

                rc_style->bg[GTK_STATE_NORMAL].red = CLAMP (step_size * r, 0, 65535);
                rc_style->bg[GTK_STATE_NORMAL].green = CLAMP (step_size * g, 0, 65535);
                rc_style->bg[GTK_STATE_NORMAL].blue = CLAMP (step_size * b, 0, 65535);
                float_color[0] = float_step_size * r;
                float_color[1] = float_step_size * g;
                float_color[2] = float_step_size * b;

                rc_style->bg[GTK_STATE_PRELIGHT].red = rc_style->bg[GTK_STATE_NORMAL].red;
                rc_style->bg[GTK_STATE_PRELIGHT].green = rc_style->bg[GTK_STATE_NORMAL].green;
                rc_style->bg[GTK_STATE_PRELIGHT].blue = rc_style->bg[GTK_STATE_NORMAL].blue;

                rc_style->color_flags[GTK_STATE_NORMAL] |= GTK_RC_BG;
                rc_style->color_flags[GTK_STATE_PRELIGHT] |= GTK_RC_BG;
                gtk_widget_modify_style (button, rc_style);
                gtk_rc_style_unref (rc_style);

                gtk_signal_connect (GTK_OBJECT (button), "pressed",
                  GTK_SIGNAL_FUNC (change_example_colors), float_color);

                gtk_widget_show (button);
            }
        }
    }
}

/* open_color_selection_dialog : create the colour selection dialogue box, populate it
	and make some convenience structures to store colours and GCs */
void open_color_selection_dialog (void)
{
    if (!color_selection_dialog)
    {
        GtkWidget *feature_colors_list, *decor_list;
		GtkWidget *notebook;

        color_selection_dialog = create_color_selection_dialog ();
        custom_color_selection_dialog = create_custom_color_selection_dialog ();
        gtk_widget_realize (color_selection_dialog);

        feature_colors_list = lookup_widget (color_selection_dialog, "feature_colors_list");
        color_selector_frame = lookup_widget (color_selection_dialog, "color_selector_frame");
        decor_list = lookup_widget (color_selection_dialog, "decor_list");
		notebook = lookup_widget (color_selection_dialog, "change_colors_of_notebook");

        color_examples_state_gcs = g_new (GdkGC *, NB_TRACE_STATES);
        color_examples_state_colors = g_new (GdkColor, NB_TRACE_STATES);
        color_examples_pane_gcs = g_new (GdkGC *, NB_TRACE_PANE_FEATURES);
        color_examples_pane_colors = g_new (GdkColor, NB_TRACE_PANE_FEATURES);

        set_colors_of_color_cube (GTK_TABLE (lookup_widget (color_selection_dialog, "color_cube_table")));

        /* Fill in state/pane colour names */
        gtk_clist_column_titles_hide (GTK_CLIST (feature_colors_list));
        {
            /* Set all colours to a mid grey which will indicate insensitivity and
               will contrast with and background colour */
            int i;
            gchar *entry[1];
            GtkCList *clist = GTK_CLIST (feature_colors_list);

            /* Make GCs and insert elements into list */
            for (i = 0; wave_trace_state_feature_info[i].name; i++)
            {
                color_examples_state_gcs[i] = gdk_gc_new (color_selection_dialog->window);

                entry[0] = (gchar *) wave_trace_state_feature_info[i].name;
                gtk_clist_append (clist, entry);
                gtk_clist_set_row_data (clist, clist->rows - 1, GINT_TO_POINTER (i));
            }

            /* Update colours using default decor */
            update_color_selection_examples_state_colors (wave_config->default_decor->state_colors);

            /* Separator */
            entry[0] = "";
            gtk_clist_append (clist, entry);
            gtk_clist_set_row_data (clist, clist->rows - 1,
              GINT_TO_POINTER (PANE_FEATURES_OFFSET + NB_TRACE_PANE_FEATURES));

            /* Now add global colours */
            for (i = 0; wave_trace_pane_feature_info[i].name; i++)
            {
                GdkGC *gc;
                GdkGCValues gc_values;

                /* Make new gc with foreground colour */
                gc_values.foreground = wave_color_to_gdk_color (wave_config->pane_colors->colors[i]);
                /* Named marker lines are dashed */
                gc_values.line_style = (i == TRACE_PANE_NAMED_MARKERS ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID);
                gdk_color_alloc (gtk_widget_get_colormap (color_selection_dialog), &gc_values.foreground);
                gc = gdk_gc_new_with_values (color_selection_dialog->window,
                  &gc_values, GDK_GC_FOREGROUND | GDK_GC_LINE_STYLE);

                /* Remember the colour and gc */
                color_examples_pane_colors[i] = gc_values.foreground;
                color_examples_pane_gcs[i] = gc;

                entry[0] = (gchar *) wave_trace_pane_feature_info[i].name;
                gtk_clist_append (clist, entry);
                gtk_clist_set_row_data (clist, clist->rows - 1, GINT_TO_POINTER (i + PANE_FEATURES_OFFSET));
            }
        }
        /* Set decor and group names in the notebook */
        color_selection_refresh_decor_list (lookup_widget (color_selection_dialog,
            "selected_traces_decor_name"), TRUE,
          (void (*)(GtkMenuItem *, TraceDecor *)) color_selection_selected_traces_decor_selected);
        color_selection_refresh_decor_list (lookup_widget (color_selection_dialog,
            "trace_group_decor_name"), TRUE,
          (void (*)(GtkMenuItem *, TraceDecor *)) color_selection_group_decor_selected);
        color_selection_refresh_decor_list (decor_list, FALSE, NULL);
        gtk_clist_column_titles_hide (GTK_CLIST (decor_list));
        color_selection_refresh_group_list (GTK_OPTION_MENU (lookup_widget (color_selection_dialog,
              "trace_group_name")), color_selection_trace_group_selected);
        color_selection_trace_group_selected (NULL, wave_config->default_group);
		color_selection_change_colors_of_notebook_page_switched (GTK_NOTEBOOK (notebook),
			GTK_NOTEBOOK_PAGE (gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), COLOR_SELECTOR_NAMED_DECOR)),
			COLOR_SELECTOR_NAMED_DECOR, NULL);

        gtk_widget_show (color_selection_dialog);
    } else
    {
        /* Open and bring to front */
        gtk_widget_show (color_selection_dialog);
        gdk_window_raise (color_selection_dialog->window);
    }
}

/* color_selection_custom_color_clicked : show the custom colour selection dialogue */
void color_selection_custom_color_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_show (custom_color_selection_dialog);
    gdk_window_raise (custom_color_selection_dialog->window);
}
