/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `command.c'
    Command line interaction

    $Id: command.c,v 1.12 2004/05/17 01:32:29 janinl Exp $
*/

#include "command.h"
#include "str.h"
#include "log.h"
#include "main.h"
#include "analyzer.h"
#include "utils.h"
#include "signalpane.h"
#include "tracepane.h"
#include "navigate.h"
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>

FileDescLineParser *wave_command_parser = NULL;
FILE *wave_command_response_stream = NULL;
WaveErrorContext *wave_command_error_context = NULL;

typedef enum
{
	WAVE_RESPONSE_OK,
	WAVE_RESPONSE_FAIL
}
WaveCommandResponse;

static char *wave_responses[] = {"ok", "end", "fail"};

/* wave_command_response : print the standard response types */
static void wave_command_response (WaveCommandResponse response)
{
	fprintf (wave_command_response_stream, "%s\n", wave_responses[response]);
	fflush (wave_command_response_stream);
}

/* wave_command_help : print help message to response stream */
static void wave_command_help (void)
{
	fprintf (wave_command_response_stream,
		"%s commands\n\n"
		"hello                      Give a jaunty welcome\n"
		"load <filename> [<format>] Load file <filename> (of format <format)\n"
		"add-trace <signal-name> [<trace-name>] Display named trace\n"
		"add-trace-with-key <signal-key-name>   Display key-named trace\n"
		"zoom fit                   Adjust the zoom (more options coming soon)\n"
		"redraw                     Redraw window(s)\n"
		"{quit,exit}                Exit GTKWave\n", wave_about_title);
}

/* wave_interpret_command : interpret and act on command line line */
void wave_interpret_command (char *line)
{
	char *line_ptr = line;
	char *token;
	WaveCommandResponse response = WAVE_RESPONSE_OK;
	SimpleTokenType token_type;

	token_type = simple_parse_token (&line_ptr, &token, wave_command_error_context);	
	/* Allow things like `add_trace' in place of `add-trace' */
	wave_substitute_char (token, '_', '-');

	if (g_strcasecmp (token, "hello") == 0)
	{
		fprintf (wave_command_response_stream, "Hi there, user\n");
	} else if (g_strcasecmp (token, "load") == 0)
	{
		/* Load file: load <filename> [<format>] */
		char *filename = NULL;
		char *format = NULL;

		if (simple_parse_token (&line_ptr, &filename, wave_command_error_context) == SIMPLE_NO_TOKEN)
		{
			wave_log (TRUE, "Invalid load command, usage: load <filename> [<format>]\n");
		} else {
			SimpleTokenType format_token = simple_parse_token (&line_ptr, &format,
				wave_command_error_context);

			wave_load_trace_file (filename, (format_token == SIMPLE_NO_TOKEN ? NULL : format));
		}
	} else if (g_strcasecmp (token, "add-trace") == 0)
	{
		/* Add trace to display: add-trace <signal-name> other args later! */
		char *signal_name = NULL;

		if (simple_parse_token (&line_ptr, &signal_name, wave_command_error_context) == SIMPLE_NO_TOKEN)
		{
			wave_log (TRUE, "Invalid add-trace command, usage: add-trace <signal-name>\n");
		} else {
			Symbol *symbol = symfind (signal_name);

			if (symbol && symbol->n)
			{
				wave_insert_node_trace (symbol->n, NULL);
				wave_signal_pane_adjust_after_insertion (wave_window);
			} else {
				fprintf (wave_command_response_stream, "Couldn't find signal `%s'\n", signal_name);
				response = WAVE_RESPONSE_FAIL;
			}
		}
	} else if (g_strcasecmp (token, "add-trace-with-key") == 0)
	{
		/* Add trace to display: add-trace <signal-key-name> other args later! */
		char *signal_key_name = NULL;

		if (simple_parse_token (&line_ptr, &signal_key_name, wave_command_error_context) == SIMPLE_NO_TOKEN)
		{
			wave_log (TRUE, "Invalid add-trace-with-key command, usage: add-trace <signal-key-name>\n");
		} else {
			Node *node = wave_reader_signal_key_find (wave_reader, signal_key_name);

			if (node)
			{
				wave_insert_node_trace (node, NULL);
				wave_signal_pane_adjust_after_insertion (wave_window);
			} else {
				fprintf (wave_command_response_stream, "Couldn't find signal key `%s'\n", signal_key_name);
				response = WAVE_RESPONSE_FAIL;
			}
		}
	} else if (g_strcasecmp (token, "zoom") == 0)
	{
	  /* zoom <param>: Only param=fit is defined for the moment */
		char *zoom_param = NULL;

		if (simple_parse_token (&line_ptr, &zoom_param, wave_command_error_context) == SIMPLE_NO_TOKEN)
		{
			wave_log (TRUE, "Invalid zoom command, usage: zoom fit\n");
		} else {
		  clicked_zoom_fit (NULL, NULL);
		}
	}
	else if (g_strcasecmp (token, "help") == 0)
		wave_command_help ();
	else if (g_strcasecmp (token, "quit") == 0 || g_strcasecmp (token, "exit") == 0)
		exit (EXIT_SUCCESS);
	else if (g_strcasecmp (token, "redraw") == 0)
	{
		wave_redraw (wave_window);
	} else {
		wave_log (TRUE, "Don't understand command: %s\n", line);
		response = WAVE_RESPONSE_FAIL;
	}

	wave_command_response (response);
}

void wave_command_received_input (gpointer data, int source, GdkInputCondition condition)
{
	char *line;

	line = file_desc_line_parser_read_line (wave_command_parser, FALSE);
	wave_command_error_context->line_num = wave_command_parser->line_num;
	if (line) wave_interpret_command (line);
	/* Try and read as many lines as are in the input buffer */
	while (line)
	{
		line = file_desc_line_parser_read_line (wave_command_parser, TRUE);
		wave_command_error_context->line_num = wave_command_parser->line_num;
		if (line) wave_interpret_command (line);
	}
}

/* wave_command_try_to_open : try to open a wave command
	line on the given file descriptors.  Returns a success value. */
gboolean wave_command_try_to_open (int in_fd, int out_fd)
{
	FileDesc *input_file_desc = file_desc_fdopen (in_fd, "rt");

	if (! input_file_desc) return FALSE;

	wave_command_parser = file_desc_line_parser_new (input_file_desc);
	wave_command_response_stream = fdopen (out_fd, "wt");
	wave_command_error_context = wave_new_error_context ();
	wave_command_error_context->operation_name = "GTKWave command line";

	if (wave_command_parser && wave_command_response_stream)
	{
		gdk_input_add (in_fd, GDK_INPUT_READ, wave_command_received_input, NULL);
		return TRUE;
	} else {
		return FALSE;
	}
}
