/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 Esterel Technologies [http://www.esterel-technologies.com]
    Copyright (C) 2002-2003 The University of Manchester

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `color.c'
    WaveColor and TraceDecor handling, keeping colour handling
    independent of Gdk

    $Id: color.c,v 1.26 2003/03/24 17:31:26 bardslea Exp $
*/

#include <stdlib.h>
#include <string.h>
#include "config/global.h"

/* new_wave_color : allocate a new colour (black!) */
static WaveColor new_wave_color (void)
{
    WaveColor ret;

    ret.red = 0;
    ret.green = 0;
    ret.blue = 0;

    return ret;
}

/* string_to_wave_color : parse a colour value (#RRGGBB) to a new trace colour */
WaveColor string_to_wave_color (const char *str)
{
    WaveColor ret = new_wave_color ();

    if (*str == '#' && strlen (str) == 7)
    {
        unsigned int rgb = strtoul (str + 1, NULL, 16);

        ret.red = rgb >> 16;
        ret.green = (rgb & 0x00FF00) >> 8;
        ret.blue = (rgb & 0x0000FF);
    } else
        config_log (FALSE, "Invalid color %s\n", str);

    return ret;
}

/* wave_color_to_string : produce an #RRGGBB string (newly allocated)
	for the given wave colour */
char *wave_color_to_string (WaveColor color)
{
    return g_strdup_printf ("#%02x%02x%02x", color.red, color.green, color.blue);
}

/* new_trace_decor : allocate an empty decor */
TraceDecor *new_trace_decor (const char *name)
{
    TraceDecor *ret = g_new (TraceDecor, 1);

    ret->name = g_strdup ((name ? name : ""));
    ret->state_colors = new_trace_state_colors ();

    return ret;
}

/* match_decor_name : function for g_list_find_custom to match a char * against a decor's name */
static gint match_decor_name (TraceDecor * decor, char *str)
{
    return strcmp ((decor->name ? decor->name : ""), str);
}

/* find_trace_decor_in_list : find a decor in a list of decors */
TraceDecor *find_trace_decor_in_list (GList * decors, const char *name)
{
    GList *found = g_list_find_custom (decors, (char *) name, (GCompareFunc) match_decor_name);

    return (found ? (TraceDecor *) found->data : NULL);
}

/* {new,free}_trace_state_colors : create/destroy a state colour table with default entries */
TraceStateColors *new_trace_state_colors (void)
{
    TraceStateColors *ret = g_new (TraceStateColors, 1);
    int i;

    ret->colors = g_new (WaveColor, NB_TRACE_STATES);
    ret->gcs = NULL;

    for (i = 0; i < NB_TRACE_STATES; i++)
        ret->colors[i] = wave_trace_state_feature_info[i].default_color;

    return ret;
}

void free_trace_state_colors (TraceStateColors * state_colors)
{
    g_free (state_colors->colors);
    g_free (state_colors);
}

/* copy_trace_state_colors : copy a state colours structure */
TraceStateColors *copy_trace_state_colors (TraceStateColors * state_colors)
{
    TraceStateColors *ret = g_new (TraceStateColors, 1);
    int i;

    ret->colors = g_new (WaveColor, NB_TRACE_STATES);
    ret->gcs = NULL;

    for (i = 0; i < NB_TRACE_STATES; i++)
        ret->colors[i] = state_colors->colors[i];

    return ret;
}

/* new_trace_pane_colors : create a table of pane colours with default entries */
TracePaneColors *new_trace_pane_colors (void)
{
    TracePaneColors *ret = g_new (TracePaneColors, 1);
    int i;

    ret->colors = g_new (WaveColor, NB_TRACE_PANE_FEATURES);
    ret->gcs = NULL;

    for (i = 0; i < NB_TRACE_PANE_FEATURES; i++)
    {
        ret->colors[i] = new_wave_color ();
        ret->colors[i] = wave_trace_pane_feature_info[i].default_color;
    }

    return ret;
}
