/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 Esterel Technologies [http://www.esterel-technologies.com]
    Copyright (C) 2002-2003 The University of Manchester

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `parse.c'
    Saved traces XML file reader

    $Id: parse.c,v 1.33 2003/07/13 15:37:16 bardslea Exp $
*/

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <libxml/parser.h>
#include "config/global.h"

typedef enum
{
    WAVE_PARSE_STATE_CONFIG,
    WAVE_PARSE_STATE_DECORS,
    WAVE_PARSE_STATE_GROUPS,
    WAVE_PARSE_STATE_PANE_COLORS,
    WAVE_PARSE_STATE_TRACES,
    WAVE_PARSE_STATE_MARKERS,
    WAVE_PARSE_STATE_DECOR,
    WAVE_PARSE_STATE_TRACE,
    WAVE_PARSE_STATE_GROUP,
    WAVE_PARSE_STATE_STATE_COLORS,
    WAVE_PARSE_STATE_VECTOR,
    WAVE_PARSE_STATE_ERROR,
    WAVE_PARSE_STATE_NONE
}
WaveParseState;

/* Type to hold name/value pairs.  Only ever contains temporary pointers
	to data held by other structures */
typedef struct
{
    char *name;
    char *value;
}
WaveXmlAttr;

#if 0
/* Is this needed, can we just use the default parser driver AB ? */

int myXmlSAXParseFile (xmlSAXHandlerPtr sax, void *user_data, const char *filename)
{
    int ret = 0;

    xmlParserCtxtPtr ctxt;

    ctxt = xmlCreateFileParserCtxt (filename);
    if (ctxt == NULL)
        return -1;
    ctxt->sax = sax;
    ctxt->userData = user_data;

    xmlParseDocument (ctxt);
    if (ctxt->wellFormed)
        ret = 0;
    else
        ret = -1;
    if (sax != NULL)
        ctxt->sax = NULL;

    xmlFreeParserCtxt (ctxt);

    return ret;
}
#endif

/* unescape_string : short term measure to do entity unescaping until
	I learn more about libxml and do the right thing, remember to
	free the resulting string */
static char *unescape_string (const char *string)
{
    char *ret = g_strdup (string);
    char *from_ptr = ret, *to_ptr = ret;

    while (*from_ptr)
    {
        if (*from_ptr == '&')
        {
            char *semi_ptr = strchr (from_ptr + 1, ';');

            if (!semi_ptr)
                config_log (TRUE, "No closing `;' in XML entity in string `%s'", string);
            else
            {
                int entity_length = semi_ptr - (from_ptr + 1);

                if (from_ptr[1] == '#') /* Number reference */
                {
                    if (entity_length < 2)
                        config_log (TRUE, "Invalid XML numeric character entity in string `%s'", string);
                    else
                    {
                        gboolean is_hex = from_ptr[2] == 'x';
                        char char_value = strtoul (from_ptr + (is_hex ? 3 : 2), NULL,
                          (is_hex ? 16 : 10));

                        *to_ptr = char_value;
                    }
                }

                else if (g_strncasecmp (from_ptr + 1, "amp", entity_length) == 0)
                    *to_ptr = '&';
                else if (g_strncasecmp (from_ptr + 1, "lt", entity_length) == 0)
                    *to_ptr = '<';
                else if (g_strncasecmp (from_ptr + 1, "gt", entity_length) == 0)
                    *to_ptr = '>';
                else if (g_strncasecmp (from_ptr + 1, "quot", entity_length) == 0)
                    *to_ptr = '"';
                else if (g_strncasecmp (from_ptr + 1, "apos", entity_length) == 0)
                    *to_ptr = '\'';
                else
                {
                    config_log (TRUE, "Invalid XML entity in string `%s'", string);
                    *to_ptr = '?';
                }

                from_ptr = semi_ptr + 1;
            }
        } else
        {
            *to_ptr = *from_ptr;
            from_ptr++;
        }
        to_ptr++;
    }

    *to_ptr = '\0';
    return ret;
}

/* {warning,error}_sax : SAX parser error reporting functions */
static void warning_sax (void *user_data, const char *format, ...)
{
    va_list ap;
    char *message;

    va_start (ap, format);
    message = g_strdup_vprintf (format, ap);
    va_end (ap);
    config_log (FALSE, "Warning parsing config file: %s", message);
    g_free (message);
}
static void error_sax (void *user_data, const char *format, ...)
{
    va_list ap;
    char *message;

    va_start (ap, format);
    message = g_strdup_vprintf (format, ap);
    va_end (ap);
    config_log (TRUE, "Warning parsing config file: %s", message);
    g_free (message);
}

/* unexpected_parse_element : report unexpected_parse_element parse tag */
static void unexpected_parse_element (const char *found, const char *expected, gboolean error)
{
    config_log (error, "Unexpected element <%s>, expecting <%s>\n", found, expected);
}

/* parse_yesno : returns TRUE for an argument of "true", FALSE for "false" and raises
	a warning (and returns FALSE) for anything else */
static gboolean parse_yesno (char *str)
{
    if (strcmp (str, "yes") == 0)
        return TRUE;
    else if (strcmp (str, "no") != 0)
        config_log (FALSE, "Invalid boolean value `%s'\n", str);

    return FALSE;
}

/* parse_xml_attribute : get the first parameter off the given attribute list */
static const xmlChar **parse_xml_attribute (WaveXmlAttr * p, const xmlChar ** attributes)
{
    if (!*attributes)           /* run off the end of the parameter list */
        return attributes;

    p->name = (char *) (*attributes);
    attributes++;

    if (!*attributes)
        config_log (TRUE, "No value for `%s' attribute\n", p->name);
    else
        p->value = (char *) (*attributes);

    attributes++;
    return attributes;
}

/* unexpected_attribute : moan about an unexpected attribute */
static void unexpected_attribute (const char *element, WaveXmlAttr * attribute)
{
    config_log (TRUE, "Unexpected <%s> attribute `%s=\"%s\"'\n", element, attribute->name, attribute->value);
}

/* parse_named_color : get the `name' and `color' attributes of a named-color */
static void parse_named_color (const xmlChar ** attributes, xmlChar ** name, xmlChar ** color)
{
    const xmlChar **attributes_iter = attributes;
    WaveXmlAttr attribute;

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, (const xmlChar *) "name") == 0)
            *name = attribute.value;
        else if (xmlStrcmp (attribute.name, (const xmlChar *) "color") == 0)
            *color = attribute.value;
        else
            unexpected_attribute ("named-color", &attribute);
    }
}

/* parse_marker : create a TraceMarker from the `name' and `time' of the given attributes */
TraceMarker *parse_marker (const xmlChar ** attributes, WaveConfig * config)
{
    const xmlChar **attributes_iter = attributes;
    WaveXmlAttr attribute;
    char *name = NULL;
    TimeType time = -1;
    TraceMarker *ret = NULL;

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, (const xmlChar *) "name") == 0)
            name = unescape_string (attribute.value);
        else if (xmlStrcmp (attribute.name, (const xmlChar *) "time") == 0)
            time = wave_parse_time (attribute.value, config->time_dimension);
        else
            unexpected_attribute ("named-color", &attribute);
    }

    if (name && time != -1)
    {
        ret = new_trace_marker (name, time);
        g_free (name);
    }
    return ret;
}

/* parse_named_trace_state_color : given the attribs. of a named-color, fill in the appropriate
	trace state entry with that named-color */
static void parse_named_trace_state_color (TraceStateColors * colors, const xmlChar ** attributes)
{
    xmlChar *name = NULL, *color = NULL;
    TraceState state;

    parse_named_color (attributes, &name, &color);
    if (color && name)
    {
        state = string_to_trace_state (name);
        if (state != -1)
            colors->colors[state] = string_to_wave_color (color);
        else
            config_log (FALSE, "Invalid trace state `%s' in named-color\n", name);

    } else
        config_log (FALSE, "One of `name' or `color' attributes missing in named-color\n");
}

/* parse_named_pane_color : given the attribs. of a named-color, fill in the appropriate
	pane color entry with that named-color */
static void parse_named_pane_color (TracePaneColors * colors, const xmlChar ** attributes)
{
    xmlChar *name = NULL, *color = NULL;
    TracePaneFeature feature;

    parse_named_color (attributes, &name, &color);
    if (color && name)
    {
        feature = string_to_trace_pane_feature (name);
        if (feature != -1)
            colors->colors[feature] = string_to_wave_color (color);
        else
            config_log (FALSE, "Invalid pane feature `%s' in named-color\n", name);

    } else
        config_log (FALSE, "One of `name' or `color' attributes missing in named-color\n");
}

static TraceDecor *parse_trace_decor (const xmlChar ** attributes)
{
    const xmlChar **attributes_iter = attributes;
    TraceDecor *decor;
    WaveXmlAttr attribute;

    if (!attributes)
        config_log (TRUE, "No attributes for `decor'\n");

    decor = new_trace_decor (NULL);

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, (const xmlChar *) "name") == 0)
        {
            g_free (decor->name);
            decor->name = unescape_string (attribute.value);
        } else
            unexpected_attribute ("decor", &attribute);
    }

    return decor;
}

static TraceGroup *parse_trace_group (WaveConfig * config, const xmlChar ** attributes)
{
    const xmlChar **attributes_iter = attributes;
    TraceGroup *group;
    WaveXmlAttr attribute;

    if (!attributes)
        config_log (TRUE, "No attributes for group\n");

    group = new_trace_group ("default");

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, (const xmlChar *) "name") == 0)
        {
            g_free (group->name);
            group->name = unescape_string (attribute.value);
        } else if (xmlStrcmp (attribute.name, (const xmlChar *) "decor") == 0)
        {
            TraceDecor *found = find_trace_decor_in_list (config->decors, attribute.value);

            if (found)
                group->decor = found;
            else
                config_log (FALSE, "Unknown decor `%s' for group `%s'\n", attribute.value, group->name);
        } else
            unexpected_attribute ("decor", &attribute);
    }

    return group;
}

static TraceConfig *parse_trace_separator (WaveConfig * config, const xmlChar ** attributes)
{
    const xmlChar **attributes_iter = attributes;
    WaveXmlAttr attribute;
    TraceConfig *trace_config;

    if (!attributes)
        config_log (TRUE, "No attributes for separator\n");

    trace_config = new_trace_config ();
    trace_config->group = config->default_group;
    trace_config->decor = config->default_decor;
    trace_config_set_flags (trace_config, TRACE_FLAGS_SEPARATOR, TRUE);

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);
        if (xmlStrcmp (attribute.name, "name") == 0)
		{
            g_free (trace_config->name);
            trace_config->name = unescape_string (attribute.value);
		} else if (xmlStrcmp (attribute.name, (const xmlChar *) "group") == 0)
            trace_config->name = g_strdup (attribute.value);
        else if (xmlStrcmp (attribute.name, (const xmlChar *) "selected") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_SELECTED, parse_yesno (attribute.value));
        else if (xmlStrcmp (attribute.name, "height") == 0)
            trace_config->height = atoi (attribute.value);
        else
            unexpected_attribute ("separator", &attribute);
    }

    return trace_config;
}

static TraceConfig *parse_trace (WaveConfig * config, const xmlChar ** attributes)
{
    const xmlChar **attributes_iter = attributes;
    WaveXmlAttr attribute;
    TraceConfig *trace_config;

    if (!attributes)
        config_log (TRUE, "No attributes for `trace'\n");

    trace_config = new_trace_config ();
    trace_config->group = config->default_group;
    trace_config->decor = config->default_decor;

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, "name") == 0)
        {
            g_free (trace_config->name);
            trace_config->name = unescape_string (attribute.value);
        } else if (xmlStrcmp (attribute.name, "decor") == 0)
        {
            TraceDecor *found = find_trace_decor_in_list (config->decors, attribute.value);

            if (found)
                trace_config->decor = found;
            else
                trace_config->decor = new_trace_decor (NULL);
        } else if (xmlStrcmp (attribute.name, "alias_name") == 0)
        {                       /* FIXME, what does alias_name mean? */
            trace_config->alias_name = unescape_string (attribute.value);
            trace_config_set_flags (trace_config, TRACE_FLAGS_ALIAS, TRUE);
        } else if (xmlStrcmp (attribute.name, "mode") == 0)
            trace_config->format = string_to_trace_format (attribute.value);
        else if (xmlStrcmp (attribute.name, "group") == 0)
        {
            TraceGroup *found = find_trace_group_in_list (config->groups, attribute.value);

            if (found)
                trace_config->group = found;
            else
                config_log (FALSE, "Unknown group `%s'\n", attribute.value);
        }

        /* *** Trace display information */

        /* Only display the trace name */
        else if (xmlStrcmp (attribute.name, "hidden") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_HIDDEN, parse_yesno (attribute.value));
        /* For binary signals, flips high and low states */
        else if (xmlStrcmp (attribute.name, "inverted") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_INVERTED, parse_yesno (attribute.value));
        /* Reverse the bit order : 0011 =>1100 */
        else if (xmlStrcmp (attribute.name, "reversed") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_REVERSED, parse_yesno (attribute.value));
        /* Bits are justified : 001bin => 0010bin => 2hex */
        else if (xmlStrcmp (attribute.name, "rjustified") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_RJUSTIFIED, parse_yesno (attribute.value));
        /* The wave is selected */
        else if (xmlStrcmp (attribute.name, "selected") == 0)
            trace_config_set_flags (trace_config, TRACE_FLAGS_SELECTED, parse_yesno (attribute.value));
        else if (xmlStrcmp (attribute.name, "height") == 0)
            trace_config->height = atoi (attribute.value);
        else
            unexpected_attribute ("trace", &attribute);
    }

    return trace_config;
}

/* parse_signal : parse a signal name element */
static TraceNodeBit *parse_signal (const xmlChar ** attributes)
{
    const xmlChar **attributes_iter = attributes;
    WaveXmlAttr attribute;
    char *name = NULL;
    int lrindex = -1;

    while (attributes_iter && *attributes_iter)
    {
        attributes_iter = parse_xml_attribute (&attribute, attributes_iter);

        if (xmlStrcmp (attribute.name, (const xmlChar *) "name") == 0)
            name = g_strdup (attribute.value);
        else if (xmlStrcmp (attribute.name, (const xmlChar *) "lrindex") == 0)
            lrindex = atoi (attribute.value);
        else
            unexpected_attribute ("signal", &attribute);
    }

    return (name ? new_trace_node_bit (name, lrindex) : NULL);
}

/* Temporary handles to objects built in parsing the config file */
static TraceStateColors *temp_state_colors = NULL;
static TraceDecor *temp_decor = NULL;
static TraceConfig *temp_trace = NULL;
static TraceGroup *temp_group = NULL;
static GList *temp_signals = NULL;

static WaveParseState parse_state = WAVE_PARSE_STATE_NONE;
static WaveParseState old_parse_state = WAVE_PARSE_STATE_NONE;

static void parse_start_of_element (void *user_data, const xmlChar * elem_name, const xmlChar ** attrib)
{
    char *name = (char *) elem_name;
    WaveConfig *config = (WaveConfig *) user_data;

#ifdef DEBUG_CONFIG_PARSE
    config_log (TRUE, "OPEN %d => ", parse_state);
#endif

    switch (parse_state)
    {
    case WAVE_PARSE_STATE_NONE:
        if (xmlStrcmp (name, (const xmlChar *) "config") != 0)
            unexpected_parse_element (name, "config", TRUE);
        parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_CONFIG:
        if (xmlStrcmp (name, (const xmlChar *) "decors") == 0)
            parse_state = WAVE_PARSE_STATE_DECORS;
        else if (xmlStrcmp (name, (const xmlChar *) "trace-groups") == 0)
            parse_state = WAVE_PARSE_STATE_GROUPS;
        else if (xmlStrcmp (name, (const xmlChar *) "pane-colors") == 0)
            parse_state = WAVE_PARSE_STATE_PANE_COLORS;
        else if (xmlStrcmp (name, (const xmlChar *) "traces") == 0)
            parse_state = WAVE_PARSE_STATE_TRACES;
        else if (xmlStrcmp (name, (const xmlChar *) "markers") == 0)
            parse_state = WAVE_PARSE_STATE_MARKERS;
        else
            unexpected_parse_element (name, "decors>, <trace-groups>, <pane-colors> or <traces", TRUE);
        break;
    case WAVE_PARSE_STATE_DECORS:
        if (xmlStrcmp (name, (const xmlChar *) "decor") == 0)
        {
            temp_decor = parse_trace_decor (attrib);

            if (temp_decor)
            {
                gboolean is_default = strcmp (temp_decor->name, "default") == 0;
                TraceDecor *found = find_trace_decor_in_list (config->decors, temp_decor->name);

                if (found)
                {
                    if (is_default)
                    {
                        config->decors = g_list_remove (config->decors, found);
                        config->decors = g_list_prepend (config->decors, temp_decor);
                    } else
                        config_log (TRUE, "Trace decor `%s' multiply defined\n", temp_decor->name);
                } else
                    config->decors = g_list_append (config->decors, temp_decor);
                if (is_default)
                    config->default_decor = temp_decor;
            }
        } else
            unexpected_parse_element (name, "decor", TRUE);
        parse_state = WAVE_PARSE_STATE_DECOR;
        break;
    case WAVE_PARSE_STATE_GROUPS:
        if (xmlStrcmp (name, (const xmlChar *) "trace-group") == 0)
        {
            temp_group = parse_trace_group (config, attrib);

            if (temp_group)
            {
                gboolean is_default = strcmp (temp_group->name, "default") == 0;
                TraceGroup *found = find_trace_group_in_list (config->groups, temp_group->name);

                if (found)
                {
                    if (is_default)
                    {
                        config->groups = g_list_remove (config->groups, found);
                        config->groups = g_list_prepend (config->groups, temp_group);
                    } else
                        config_log (TRUE, "Trace group `%s' multiply defined\n", temp_group->name);
                } else
                    config->groups = g_list_append (config->groups, temp_group);
                if (is_default)
                    config->default_group = temp_group;
            }
        } else
            unexpected_parse_element (name, "trace-group", TRUE);
        parse_state = WAVE_PARSE_STATE_GROUP;
        break;
    case WAVE_PARSE_STATE_TRACES:
        if (xmlStrcmp (name, (const xmlChar *) "trace") == 0)
            temp_trace = parse_trace (config, attrib);
        else if (xmlStrcmp (name, (const xmlChar *) "separator") == 0)
            temp_trace = parse_trace_separator (config, attrib);
        else
            unexpected_parse_element (name, "trace> or <separator", TRUE);
        config->traces = g_list_append (config->traces, temp_trace);
        parse_state = WAVE_PARSE_STATE_TRACE;
        break;
    case WAVE_PARSE_STATE_MARKERS:
        if (xmlStrcmp (name, (const xmlChar *) "marker") == 0)
        {
            TraceMarker *marker = parse_marker (attrib, config);

            if (marker)
            {
                TraceMarker *found = find_trace_marker_in_list (config->markers, marker->name);

                if (found)
                {
                	GList *link = g_list_find (config->markers, found);

                    config_log (FALSE, "Trace marker `%s' redefined\n", marker->name);
                    link->data = marker;

					free_trace_marker (found);
                } else
                    config->markers = g_list_append (config->markers, marker);
            }
        } else
            unexpected_parse_element (name, "marker", TRUE);
        break;
    case WAVE_PARSE_STATE_TRACE:
        if (xmlStrcmp (name, (const xmlChar *) "signal") == 0)
        {
            if (temp_trace)
            {
                TraceNodeBit *node_bit = parse_signal (attrib);

                if (node_bit)
                    temp_trace->signals = g_list_prepend (NULL, node_bit);
            }
        } else if (xmlStrcmp (name, (const xmlChar *) "vector") == 0)
        {
            temp_signals = NULL;
            parse_state = WAVE_PARSE_STATE_VECTOR;
        } else if (xmlStrcmp (name, (const xmlChar *) "trace-state-colors") == 0)
        {
            temp_state_colors = new_trace_state_colors ();
            if (temp_trace)
                temp_trace->state_colors = temp_state_colors;
            old_parse_state = WAVE_PARSE_STATE_TRACE;
            parse_state = WAVE_PARSE_STATE_STATE_COLORS;
        } else
            unexpected_parse_element (name, "signal>, <vector> or <trace-state-colors", TRUE);
        break;
    case WAVE_PARSE_STATE_VECTOR:
        if (xmlStrcmp (name, (const xmlChar *) "signal") == 0)
        {
            TraceNodeBit *node_bit = parse_signal (attrib);

            if (node_bit)
                temp_signals = g_list_prepend (temp_signals, node_bit);
        } else
            unexpected_parse_element (name, "signal", TRUE);
        break;
    case WAVE_PARSE_STATE_GROUP:
        if (xmlStrcmp (name, (const xmlChar *) "trace-state-colors") == 0)
        {
            temp_state_colors = new_trace_state_colors ();
            if (temp_group)
                temp_group->state_colors = temp_state_colors;
            old_parse_state = WAVE_PARSE_STATE_GROUP;
            parse_state = WAVE_PARSE_STATE_STATE_COLORS;
        } else
            unexpected_parse_element (name, "trace-state-colors", TRUE);
        break;
    case WAVE_PARSE_STATE_DECOR:
        if (xmlStrcmp (name, (const xmlChar *) "trace-state-colors") == 0)
        {
            temp_state_colors = new_trace_state_colors ();
            if (temp_decor)
                temp_decor->state_colors = temp_state_colors;
            old_parse_state = WAVE_PARSE_STATE_DECOR;
            parse_state = WAVE_PARSE_STATE_STATE_COLORS;
        } else
            unexpected_parse_element (name, "trace-state-colors", TRUE);
        break;
    case WAVE_PARSE_STATE_STATE_COLORS:
        if (xmlStrcmp (name, (const xmlChar *) "named-color") == 0)
            parse_named_trace_state_color (temp_state_colors, attrib);
        else
            unexpected_parse_element (name, "named-color", TRUE);
        break;
    case WAVE_PARSE_STATE_PANE_COLORS:
        if (xmlStrcmp (name, (const xmlChar *) "named-color") == 0)
            parse_named_pane_color (config->pane_colors, attrib);
        else
            unexpected_parse_element (name, "named-color", TRUE);
        break;
    case WAVE_PARSE_STATE_ERROR:
        return;
        break;
    }

#ifdef DEBUG_CONFIG_PARSE
    config_log (TRUE, "%d\n", parse_state);
#endif
}

static void parse_end_of_element (void *user_data, const xmlChar * elem_name)
{
    char *name = (char *) elem_name;

    /* WaveConfig *config = (WaveConfig *) user_data; */

#ifdef DEBUG_CONFIG_PARSE
    config_log (TRUE, "CLOS %d => ", parse_state);
#endif

    switch (parse_state)
    {
    case WAVE_PARSE_STATE_NONE:
        unexpected_parse_element (name, "end of document", TRUE);
        break;
    case WAVE_PARSE_STATE_CONFIG:
        if (xmlStrcmp (name, (const xmlChar *) "config") == 0)
            parse_state = WAVE_PARSE_STATE_NONE;
        break;
    case WAVE_PARSE_STATE_DECORS:
        if (xmlStrcmp (name, (const xmlChar *) "decors") == 0)
            parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_GROUPS:
        if (xmlStrcmp (name, (const xmlChar *) "trace-groups") == 0)
            parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_TRACES:
        if (xmlStrcmp (name, (const xmlChar *) "traces") == 0)
            parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_MARKERS:
        if (xmlStrcmp (name, (const xmlChar *) "markers") == 0)
            parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_DECOR:
        parse_state = WAVE_PARSE_STATE_DECORS;
        break;
    case WAVE_PARSE_STATE_TRACE:
		if (xmlStrcmp (name, (const xmlChar *) "trace") == 0)
		{
			if (temp_signals && temp_trace)
				temp_trace->signals = g_list_reverse (temp_signals);
			temp_signals = NULL;    /* Make sure we don't accidentally pick up some old
									   signals next time around */
			parse_state = WAVE_PARSE_STATE_TRACES;
		} else if (xmlStrcmp (name, (const xmlChar *) "separator") == 0)
			parse_state = WAVE_PARSE_STATE_TRACES;
        break;
    case WAVE_PARSE_STATE_VECTOR:
        if (xmlStrcmp (name, (const xmlChar *) "vector") == 0)
            parse_state = WAVE_PARSE_STATE_TRACE;
        break;
    case WAVE_PARSE_STATE_GROUP:
        parse_state = WAVE_PARSE_STATE_GROUPS;
        break;
    case WAVE_PARSE_STATE_STATE_COLORS:
        /* Only return to the old state on </trace-state-colors>, not </named-color> */
        if (xmlStrcmp (name, (const xmlChar *) "trace-state-colors") == 0)
            parse_state = old_parse_state;
        break;
    case WAVE_PARSE_STATE_PANE_COLORS:
        /* Only return to the old state on </pane-colors>, not </named-color> */
        if (xmlStrcmp (name, (const xmlChar *) "pane-colors") == 0)
            parse_state = WAVE_PARSE_STATE_CONFIG;
        break;
    case WAVE_PARSE_STATE_ERROR:
        return;
        break;
    }

#ifdef DEBUG_CONFIG_PARSE
    config_log (TRUE, "%d\n", parse_state);
#endif
}

void init_xml_sax_handler (xmlSAXHandler * handler)
{
    handler->internalSubset = NULL;
    handler->isStandalone = NULL;
    handler->hasInternalSubset = NULL;
    handler->hasExternalSubset = NULL;
    handler->resolveEntity = NULL;
    handler->getEntity = NULL;
    handler->entityDecl = NULL;
    handler->notationDecl = NULL;
    handler->attributeDecl = NULL;
    handler->elementDecl = NULL;
    handler->unparsedEntityDecl = NULL;
    handler->setDocumentLocator = NULL;
    handler->startDocument = NULL;
    handler->endDocument = NULL;
    handler->startElement = parse_start_of_element;
    handler->endElement = parse_end_of_element;
    handler->reference = NULL;
    handler->characters = NULL;
    handler->ignorableWhitespace = NULL;
    handler->processingInstruction = NULL;
    handler->comment = NULL;
    handler->warning = warning_sax;
    handler->error = error_sax;
    handler->fatalError = error_sax;
    handler->getParameterEntity = NULL;
    handler->cdataBlock = NULL;
}

/* parse_wave_config_file : parse a wave config file and populate the (pre-initialised
	using new_wave_config) structure `config' */
gboolean parse_wave_config_file (const char *filename, WaveConfig * config)
{
    xmlSAXHandler my_sax_handler;

    init_xml_sax_handler (&my_sax_handler);

    parse_state = WAVE_PARSE_STATE_NONE;
    return !xmlSAXUserParseFile (&my_sax_handler, config, filename);
}
