/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 The University of Manchester
    Copyright (C) 1999-2003 Tony Bybell

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `stime.c'
    Simulation Time Handling

    $Id: stime.c,v 1.6 2003/07/13 15:37:16 bardslea Exp $
*/

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "config/stime.h"

/* {new,free}_trace_marker : make/destroy a marker, populate with copy of `name' and `time' */
TraceMarker *new_trace_marker (const char *name, UTimeType time)
{
    TraceMarker *ret = g_new (TraceMarker, 1);

    ret->name = g_strdup (name);
    ret->time = time;

    return ret;
}
void free_trace_marker (TraceMarker *marker)
{
	if (marker->name) g_free (marker->name);
	g_free (marker);
}

/* Prefix characters for time multiplication */
static char *time_prefices = " munpf";

/* wave_format_time : format a TimeType into a string (using the
	given buffer) using the preferred dimension character, or a smaller
	one if necessary */
void wave_format_time (char *buf, TimeType time, char dimension)
{
    int actual_divs_by_1000, requested_divs_by_1000;
    char *si_prefix = strchr (time_prefices, (int) dimension);

    if (si_prefix)
        requested_divs_by_1000 = si_prefix - time_prefices;
    else                        /* Assume seconds */
        requested_divs_by_1000 = 0;

    for (actual_divs_by_1000 = requested_divs_by_1000;
      actual_divs_by_1000 > 0 && (time % 1000) == 0; actual_divs_by_1000--)
    {
        time /= 1000;
    }

    if (actual_divs_by_1000)
        sprintf (buf, LL_PRINTF_FORMAT " %cs", time, time_prefices[actual_divs_by_1000]);
    else
        sprintf (buf, LL_PRINTF_FORMAT " s", time);
}

/* wave_str_to_time_type : (hopefully!) 64b integer from string parser.
	Returns a TimeType value and sets *end_ptr (if not NULL) to the first
	non-numeric character at the end of the parsed number in the string.
	Returns 0 on failure. */
TimeType wave_str_to_time_type (const char *str, char **end_ptr)
{
    TimeType ret = 0;
    unsigned char ch;
    gboolean negative = FALSE;
    const char *set_end_ptr = NULL;

    /* Skip leading whitespace, g_strchug isn't const char *, suspicious */
    while (isspace (*str))
        str++;

    /* Find sign */
    switch (*str)
    {
    case '-':
        negative = TRUE;
        str++;
        break;
    case '+':
        negative = FALSE;
        str++;
        break;
    default:
        break;
    }

    while (*str)
    {
        ch = *str;

        if (isdigit (ch))
            ret = ret * 10 + (ch - '0');
        else if (ch == '_')     /* Allow underscores in numbers to be ignored */
        {                       /* skip */
        } else
        {
            set_end_ptr = str;
            break;
        }
        str++;
    }

    if (end_ptr)
        *end_ptr = (char *) set_end_ptr;

    return (negative ? -ret : ret);
}

/* wave_parse_time : parse a string with a time value and a dimension
	(ns/ms/s ...) into a TimeType value.  Dimension is the base dimension
	of the returned value.  If no dimension is specified in the string,
	assume that the units are `dimension' */
TimeType wave_parse_time (const char *buf, char dimension)
{
    TimeType ret;
    int delta;
    char *after_number_ptr;
    char *si_prefix = NULL;
    char *target_si_prefix;

    ret = wave_str_to_time_type (buf, &after_number_ptr);

    if (after_number_ptr)
    {
        char ch;

        after_number_ptr = g_strchug (after_number_ptr);

        ch = tolower (*after_number_ptr);
        if (ch == 's')
            ch = ' ';           /* as in plain vanilla seconds */

        si_prefix = strchr (time_prefices, ch);
    }

    if (!si_prefix)
        return ret;

    target_si_prefix = strchr (time_prefices, dimension);
    if (!target_si_prefix)
        return ret;             /* should *never* happen */

    /* +ve delta means that the target type is later in
       the series " munpf" and so we must multiply */
    delta = target_si_prefix - si_prefix;
    if (delta < 0)
    {
        int i;

        for (i = delta; i < 0; i++)
            ret /= 1000;
    } else
    {
        int i;

        for (i = 0; i < delta; i++)
            ret *= 1000;
    }

    return ret;
}

/* match_marker_name : function for g_list_find_custom to match a char * against a marker's name */
static gint match_marker_name (TraceMarker * marker, char *str)
{
    return strcmp ((marker->name ? marker->name : ""), str);
}

/* find_trace_marker_in_list : find a marker in a list of markers */
TraceMarker *find_trace_marker_in_list (GList * markers, const char *name)
{
    GList *found = g_list_find_custom (markers, (char *) name, (GCompareFunc) match_marker_name);

    return (found ? (TraceMarker *) found->data : NULL);
}
