/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 Esterel Technologies [http://www.esterel-technologies.com]
    Copyright (C) 2002-2003 The University of Manchester

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `trace.c'
    Trace configuration handling

    $Id: trace.c,v 1.30 2003/03/24 17:31:27 bardslea Exp $
*/

#include <string.h>
#include <assert.h>
#include <stdio.h>
#include "config/global.h"

static const char *const trace_format_names[TRACE_FORMAT_NOFORMAT + 1] = {
    "bin", "oct", "dec", "sig", "hex", "ascii", "default"
};

/* string_to_trace_state : returns the TraceState associated with the symbolic name `str' */
TraceState string_to_trace_state (const char *str)
{
    int i;

    for (i = 0; wave_trace_state_feature_info[i].name; i++)
    {
        if (strcmp (str, wave_trace_state_feature_info[i].short_name) == 0)
            return i;
    }
    return -1;
}

/* string_to_trace_pane_feature : returns the TracePaneFeature associated with the symbolic name `str' */
TracePaneFeature string_to_trace_pane_feature (const char *str)
{
    int i;

    for (i = 0; wave_trace_pane_feature_info[i].name; i++)
    {
        if (strcmp (str, wave_trace_pane_feature_info[i].short_name) == 0)
            return i;
    }
    return -1;
}

/* string_to_trace_state : returns the StateFormat associated with the symbolic name `str' */
TraceFormat string_to_trace_format (const char *str)
{
    TraceFormat format;

    for (format = 0; format < TRACE_FORMAT_NOFORMAT; format++)
    {
        if (strcmp (str, trace_format_names[format]) == 0)
            return format;
    }
    return TRACE_FORMAT_NOFORMAT;
}

/* trace_state_to_string : returns a (constant) symbolic string for the given trace state */
const char *trace_state_to_string (TraceState state)
{
    return wave_trace_state_feature_info[state].short_name;
}

/* trace_pane_feature_to_string : returns a (constant) symbolic string for the given trace pane feature */
const char *trace_pane_feature_to_string (TracePaneFeature feature)
{
    return wave_trace_pane_feature_info[feature].short_name;
}

/* trace_format_to_string : returns a (constant) symbolic string for the given trace format */
const char *trace_format_to_string (TraceFormat format)
{
    return trace_format_names[format];
}

/* new_trace_config : allocate a TraceConfig */
TraceConfig *new_trace_config (void)
{
    TraceConfig *trace_config = g_new (TraceConfig, 1);

    trace_config->decor = NULL;
    trace_config->group = NULL;
    trace_config->state_colors = NULL;
    trace_config->name = g_strdup ("");
    trace_config->alias_name = NULL;
    trace_config->format = TRACE_FORMAT_NOFORMAT;
    trace_config->height = 1;
    trace_config->drawer = NULL;

    trace_config->flags = 0;
    trace_config_set_flags (trace_config, TRACE_FLAGS_RJUSTIFIED, TRUE);

    trace_config->signals = NULL;

    return trace_config;
}

/* trace_config_set_flags : set/clear a named flag in a TraceConfig. eg.
	trace_config_set_flags (config, TRACE_FLAGS_HIDDEN | TRACE_FLAGS_SELECTED, TRUE) */
void trace_config_set_flags (TraceConfig * config, TraceFlags flags, gboolean state)
{
    if (state)
        config->flags |= flags;
    else
        config->flags &= ~flags;
}

/* trace_config_toggle_flags : invert the named flags */
void trace_config_toggle_flags (TraceConfig * config, TraceFlags flags)
{
    config->flags ^= flags;
}

/* trace_config_get_state_colors : get the state colour table
	for a given config, either from the trace itself, its
	group/decor, or default colours */
TraceStateColors *trace_config_get_state_colors (const TraceConfig * config, const WaveConfig * wave_config)
{
    TraceStateColors *ret = NULL;

    assert (config);

    /* This trace has its own unique colours */
    if (config->state_colors)
        ret = config->state_colors;
    /* Use the colours of the specified decor, decors should
     *always* have well defined colours */
    else if (config->decor)
        ret = config->decor->state_colors;
    /* OK, lets try the group colours */
    else if (config->group)
    {
        TraceGroup *group = config->group;

        /* Search through state colours then decor colours */
        if (group->state_colors)
            ret = group->state_colors;
        /* Then the decor */
        else if (group->decor)
            ret = group->decor->state_colors;
    }

    /* Still no colour assigned */
    if (!ret)
    {
        /* Try the default decor */
        if (wave_config->default_decor)
            ret = wave_config->default_decor->state_colors;
    }

    return ret;
}

/* {new,free}_trace_node_bit : constructor/destructor for trace node bits */
TraceNodeBit *new_trace_node_bit (const char *name, unsigned int lrindex)
{
    TraceNodeBit *ret = g_new (TraceNodeBit, 1);

    ret->node_name = g_strdup (name);
    ret->lrindex = lrindex;

    return ret;
}

void free_trace_node_bit (TraceNodeBit * node_bit)
{
    if (node_bit && node_bit->node_name)
        g_free (node_bit->node_name);
    g_free (node_bit);
}
