/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 Esterel Technologies [http://www.esterel-technologies.com]
    Copyright (C) 2002-2003 The University of Manchester

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `trace.h'
    Trace configuration handling

    $Id: trace.h,v 1.28 2003/03/25 17:26:19 bardslea Exp $
*/

#ifndef WAVE_CONFIG_TRACE_H
#define WAVE_CONFIG_TRACE_H

#include "config/stime.h"
#include "config/color.h"
#include "config/group.h"

/* TraceState : Signal states which merit different colors */
typedef enum
{
    TRACE_STATE_FONT,
    TRACE_STATE_LOW,
    TRACE_STATE_HIGH,
    TRACE_STATE_X,
    TRACE_STATE_XFILL,
    TRACE_STATE_TRANS,
    TRACE_STATE_MID,
    TRACE_STATE_VTRANS,
    TRACE_STATE_VBOX,
    TRACE_STATE_UNLOADED,
    TRACE_STATE_ANALOG,
    TRACE_STATE_CLIP,
	TRACE_STATE_REQ,
	TRACE_STATE_ACK,
	TRACE_STATE_HBOX,
    NB_TRACE_STATES             /* Keep this as the last element */
}
TraceState;

/* TracePaneFeature : separably colourable features in the trace pane */
typedef enum
{
    TRACE_PANE_BACKGROUND,
    TRACE_PANE_GRID,
    TRACE_PANE_NAMED_MARKERS,
    TRACE_PANE_PRIMARY_MARKER,
    TRACE_PANE_FONT,
    NB_TRACE_PANE_FEATURES      /* Keep this as the last element */
}
TracePaneFeature;

/* Trace value display formats */
typedef enum
{
    TRACE_FORMAT_BIN,
    TRACE_FORMAT_OCT,
    TRACE_FORMAT_DEC,
    TRACE_FORMAT_SIGNED_DEC,
    TRACE_FORMAT_HEX,
    TRACE_FORMAT_ASCII,
    TRACE_FORMAT_NOFORMAT       /* Keep this as the last element */
}
TraceFormat;

/* Other modifiers for trace appearance */
/* FIXME, need some descriptions */
typedef enum
{
    TRACE_FLAGS_ALIAS = (1 << 0),
    TRACE_FLAGS_SEPARATOR = (1 << 1),
    TRACE_FLAGS_SELECTED = (1 << 2),
    TRACE_FLAGS_HIDDEN = (1 << 3),
    TRACE_FLAGS_INVERTED = (1 << 4),
    TRACE_FLAGS_REVERSED = (1 << 5),
    TRACE_FLAGS_RJUSTIFIED = (1 << 6),
    /* This trace is just a filler trace for a height != 1 trace,
       FIXME, find a better way to do this */
    TRACE_FLAGS_EXTENSION = (1 << 7)
}
TraceFlags;

/* Forward declaration of global config type */
struct WaveConfig_;

/* Bits of a node known to GTKWave used to build a vector
	this will probably get slicker (and change name!) later */
typedef struct
{
    char *node_name;            /* Full node name known to GTKWave in a Node * */
    int lrindex;                /* index into node counting from 0 on the left-most bit,
                                   -1 for no index */
}
TraceNodeBit;

typedef struct TraceConfig_
{
    char *name, *alias_name;
    TraceFormat format;
    TraceGroup *group;
    TraceDecor *decor;
    TraceStateColors *state_colors;
    TraceFlags flags;
    char *drawer;				/* Drawer to use by name or NULL for default drawer */
    unsigned int height;        /* Multiples of font height to draw trace, mustn't be 0! */

    /* These entries are temporary until I sort out a better way to represent
       signal names to be dumped and look at the interaction of Bits/VectorBits... */
    GList *signals;             /* List of TraceNodeBits */
}
TraceConfig;

/* Predicates on flags from TraceConfig */
#define TRACE_IS_ALIAS(trace_config)		((trace_config)->flags & TRACE_FLAGS_ALIAS)
#define TRACE_IS_SEPARATOR(trace_config)	((trace_config)->flags & TRACE_FLAGS_SEPARATOR)
#define TRACE_IS_SELECTED(trace_config)		((trace_config)->flags & TRACE_FLAGS_SELECTED)
#define TRACE_IS_HIDDEN(trace_config)		((trace_config)->flags & TRACE_FLAGS_HIDDEN)
#define TRACE_IS_INVERTED(trace_config)		((trace_config)->flags & TRACE_FLAGS_INVERTED)
#define TRACE_IS_REVERSED(trace_config)		((trace_config)->flags & TRACE_FLAGS_REVERSED)
#define TRACE_IS_RJUSTIFIED(trace_config)	((trace_config)->flags & TRACE_FLAGS_RJUSTIFIED)
#define TRACE_IS_EXTENSION(trace_config)	((trace_config)->flags & TRACE_FLAGS_EXTENSION)

/* string_to_trace_state : returns the TraceState associated with the symbolic name `str' */
extern TraceState string_to_trace_state (const char *str);

/* string_to_trace_pane_feature : returns the TracePaneFeature associated with the symbolic name `str' */
extern TracePaneFeature string_to_trace_pane_feature (const char *str);

/* string_to_trace_state : returns the StateFormat associated with the symbolic name `str' */
extern TraceFormat string_to_trace_format (const char *str);

/* trace_state_to_string : returns a (constant) symbolic string for the given trace state */
extern const char *trace_state_to_string (TraceState state);

/* trace_pane_feature_to_string : returns a (constant) symbolic string for the given trace pane feature */
extern const char *trace_pane_feature_to_string (TracePaneFeature feature);

/* trace_format_to_string : returns a (constant) symbolic string for the given trace format */
extern const char *trace_format_to_string (TraceFormat format);

/* new_trace_config : allocate a TraceConfig */
extern TraceConfig *new_trace_config (void);

/* trace_config_set_flags : set/clear a named flag in a TraceConfig. eg.
	trace_config_set_flags (config, TRACE_FLAGS_HIDDEN | TRACE_FLAGS_SELECTED, TRUE) */
extern void trace_config_set_flags (TraceConfig * config, TraceFlags flags, gboolean state);

/* trace_config_toggle_flags : invert the named flags */
extern void trace_config_toggle_flags (TraceConfig * config, TraceFlags flags);

/* trace_config_get_state_colors : get the state colour table for a given config, either
	from the trace itself, its group/decor, or default colours */
extern TraceStateColors *trace_config_get_state_colors (const TraceConfig * config,
  const struct WaveConfig_ *wave_config);

/* TraceNodeBit handling */

/* {new,free}_trace_node_bit : constructor/destructor for trace node bits */
extern TraceNodeBit *new_trace_node_bit (const char *name, unsigned int lrindex);
extern void free_trace_node_bit (TraceNodeBit * node_bit);

#endif
