/*
    GTKWave Trace Configuration Library - libwaveconfig
    Copyright (C) 2002-2003 Esterel Technologies [http://www.esterel-technologies.com]
    Copyright (C) 2002-2003 The University of Manchester

    The authors of this library can be contacted through the GTKWave maintainer.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    `write.c'
    Saved traces XML file writer

    $Id: write.c,v 1.28 2003/03/04 17:58:59 bardslea Exp $
*/

#include <stdio.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <time.h>
#include "config/global.h"

/* File scope vars. to reduce need to pass args */
static FILE *saved_traces_stream = NULL;
static WaveConfig *saved_traces_config = NULL;

#define TAB " "
#define TABS(depth) indent (tabs + (depth))
#define NL putc ('\n', saved_traces_stream)
#define OPEN_TAG(depth,tag) { NL; TABS (depth); fprintf (saved_traces_stream, "<%s>", (tag)); }
#define CLOSE_TAG(depth,tag) { NL; TABS (depth); fprintf (saved_traces_stream, "</%s>", (tag)); }
#define ATTR_OPEN_TAG(depth,tag) { NL; TABS (depth); fprintf (saved_traces_stream, "<%s", (tag)); }
#define ATTR_CLOSE_TAG fprintf (saved_traces_stream, ">")
#define ATTR_CLOSE_EMPTY_TAG fprintf (saved_traces_stream, "/>")
#define ATTR_STR(name,value) { fprintf (saved_traces_stream, " %s=\"", (name)); \
	write_escaped_string (saved_traces_stream, (value)); fprintf (saved_traces_stream, "\""); }
#define ATTR_INT(name,value) fprintf (saved_traces_stream, " %s=\"%d\"", (name), (value))
#define ATTR_BOOL(name,value) fprintf (saved_traces_stream, " %s=\"%s\"", (name), ((value) ? "yes" : "no"))
#define ATTR_TRUE(name,value) if (value) fprintf (saved_traces_stream, " %s=\"yes\"", (name))
#define STR(str) fprintf (saved_traces_stream, "%s", (str))

/* AB 2003-01-26, I didn't have time to learn libxml extensively before writing this,
	much of this code will (hopefully!) be replaced by DOC code as that looks like a
	more sensible way to go.  The following escaping function is just because I can't
	for the life of me figure out how to do this in libxml. */

static void write_escaped_string (FILE * stream, const char *string)
{
    while (*string)
    {
        char chr = *string;

        switch (chr)
        {
        case '&':
            fprintf (stream, "&amp;");
            break;
        case '<':
            fprintf (stream, "&lt;");
            break;
        case '>':
            fprintf (stream, "&gt;");
            break;
        case '"':
            fprintf (stream, "&quot;");
            break;
        case '\'':
            fprintf (stream, "&apos;");
            break;
        default:
            if (chr < ' ' || chr > '~')
                fprintf (stream, "&#x%02X;", (int) chr);
            else
                putc (chr, stream);
            break;
        }
        string++;
    }
}

/* indent : indent to a given level */
static void indent (int tabs)
{
    while (tabs)
    {
        fprintf (saved_traces_stream, TAB);
        tabs--;
    }
}

static void write_header (int tabs, const char *banner)
{
    STR ("<?xml version=\"1.0\"?>");
    NL;
    if (banner)
    {
        time_t now = time (NULL);
        char *time_str = g_strchomp (g_strdup (ctime (&now)));

        STR ("<!-- ");
        STR (banner);
        STR (" -->");
        NL;
        STR ("<!-- at ");
        STR (time_str);
        STR (" -->");
        NL;

        g_free (time_str);
    }
    OPEN_TAG (0, "config");
}

static void write_footer (int tabs)
{
    CLOSE_TAG (0, "config");
    NL;
}

static void write_pane_colors (int tabs)
{
    TracePaneFeature i;

    OPEN_TAG (0, "pane-colors");
    for (i = 0; i < NB_TRACE_PANE_FEATURES; i++)
    {
        char *str = wave_color_to_string (saved_traces_config->pane_colors->colors[i]);

        ATTR_OPEN_TAG (1, "named-color");
        ATTR_STR ("name", trace_pane_feature_to_string (i));
        ATTR_STR ("color", str);
        ATTR_CLOSE_EMPTY_TAG;
        g_free (str);
    }
    CLOSE_TAG (0, "pane-colors");
}

static void write_markers (int tabs)
{
    GList *markers = saved_traces_config->markers;

    OPEN_TAG (0, "markers");
    while (markers)
    {
        TraceMarker *marker = (TraceMarker *) markers->data;
        char time_str[MAX_FORMATTED_TIME_LENGTH];

        wave_format_time (time_str, marker->time, saved_traces_config->time_dimension);

        ATTR_OPEN_TAG (1, "marker");
        ATTR_STR ("name", marker->name);
        ATTR_STR ("time", time_str);
        ATTR_CLOSE_EMPTY_TAG;
        markers = g_list_next (markers);
    }
    CLOSE_TAG (0, "markers");
}

static void write_state_colors (TraceStateColors * state_colors, int tabs)
{
    TraceState i;

    OPEN_TAG (0, "trace-state-colors");
    for (i = 0; i < NB_TRACE_STATES; i++)
    {
        char *str = wave_color_to_string (state_colors->colors[i]);

        ATTR_OPEN_TAG (1, "named-color");
        ATTR_STR ("name", trace_state_to_string (i));
        ATTR_STR ("color", str);
        ATTR_CLOSE_EMPTY_TAG;
        g_free (str);
    }
    CLOSE_TAG (0, "trace-state-colors");
}

static void write_decor (TraceDecor * decor, int tabs)
{
    ATTR_OPEN_TAG (0, "decor");
    ATTR_STR ("name", decor->name);
    ATTR_CLOSE_TAG;
    write_state_colors (decor->state_colors, tabs + 1);
    CLOSE_TAG (0, "decor");
}

static void write_decors (int tabs)
{
    OPEN_TAG (0, "decors");
    g_list_foreach (saved_traces_config->decors, (GFunc) write_decor, GINT_TO_POINTER (tabs + 1));
    CLOSE_TAG (0, "decors");
}

static void write_trace_group (TraceGroup * group, int tabs)
{
    ATTR_OPEN_TAG (0, "trace-group");
    ATTR_STR ("name", group->name);
    if (group->decor)
        ATTR_STR ("decor", group->decor->name);
    ATTR_CLOSE_TAG;
    if (group->state_colors)
        write_state_colors (group->state_colors, tabs + 1);

    CLOSE_TAG (0, "trace-group");
}

static void write_trace_groups (int tabs)
{
    OPEN_TAG (0, "trace-groups");
    g_list_foreach (saved_traces_config->groups, (GFunc) write_trace_group, GINT_TO_POINTER (tabs + 1));
    CLOSE_TAG (0, "trace-groups");
}

static void write_signal (TraceNodeBit * node_bit, int tabs)
{
    ATTR_OPEN_TAG (0, "signal");
    ATTR_STR ("name", node_bit->node_name);
    if (node_bit->lrindex != -1)
        ATTR_INT ("lrindex", node_bit->lrindex);
    ATTR_CLOSE_EMPTY_TAG;
}

static void write_trace_config (TraceConfig * trace_config, int tabs)
{
    if (TRACE_IS_EXTENSION (trace_config))
        return;

    if (TRACE_IS_SEPARATOR (trace_config))
    {
        ATTR_OPEN_TAG (0, "separator");
        ATTR_STR ("name", trace_config->name);
    } else
    {
        ATTR_OPEN_TAG (0, "trace");
        ATTR_STR ("name", trace_config->name);

        if (trace_config->group && strcmp (trace_config->group->name, "default") != 0)
            ATTR_STR ("group", trace_config->group->name);
        if (trace_config->decor && strcmp (trace_config->decor->name, "default") != 0)
            ATTR_STR ("decor", trace_config->decor->name);

        if (trace_config->format != TRACE_FORMAT_NOFORMAT)
            ATTR_STR ("mode", trace_format_to_string (trace_config->format));

        ATTR_TRUE ("hidden", TRACE_IS_HIDDEN (trace_config));
        ATTR_TRUE ("inverted", TRACE_IS_INVERTED (trace_config));
        ATTR_TRUE ("reversed", TRACE_IS_REVERSED (trace_config));
        ATTR_TRUE ("rjustified", TRACE_IS_RJUSTIFIED (trace_config));
    }
    ATTR_TRUE ("selected", TRACE_IS_SELECTED (trace_config));
    if (trace_config->height != 1)
        ATTR_INT ("height", trace_config->height);

    ATTR_CLOSE_TAG;

    if (trace_config->signals)  /* Some named signals */
    {
        GList *signals = trace_config->signals;

        if (g_list_length (signals) == 1)
            write_signal ((TraceNodeBit *) signals->data, tabs + 2);
        else
        {
            OPEN_TAG (1, "vector");
            g_list_foreach (signals, (GFunc) write_signal, GINT_TO_POINTER (tabs + 2));
            CLOSE_TAG (1, "vector");
        }
    }

    if (trace_config->state_colors)
        write_state_colors (trace_config->state_colors, tabs + 2);

    CLOSE_TAG (0, (TRACE_IS_SEPARATOR (trace_config) ? "separator" : "trace"));
}

static void write_traces (int tabs)
{
    OPEN_TAG (0, "traces");
    g_list_foreach (saved_traces_config->traces, (GFunc) write_trace_config, GINT_TO_POINTER (tabs + 1));
    CLOSE_TAG (0, "traces");
}

/* write_wave_config_file : write a wave config file from a WaveConfig, returns TRUE
	on success and moans (returning FALSE as well) on failure */
gboolean write_wave_config_file (const char *filename, WaveConfig * config, const char *banner)
{
    saved_traces_stream = fopen (filename, "wt");
    saved_traces_config = config;

    if (!saved_traces_stream)
    {
        config_log (TRUE, "Can't open file `%s' to write configuration\n", filename);
        return FALSE;
    }

    /* Markup document header */
    write_header (0, banner);
    /* Markup global values for display */
    /* write_wave_globals (saved_traces_stream, config); */
    /* Markup traces */
    write_decors (1);
    NL;
    write_trace_groups (1);
    NL;
    write_pane_colors (1);
    NL;
    write_markers (1);
    NL;
    write_traces (1);
    /* Markup Document end */
    write_footer (0);

    fclose (saved_traces_stream);
    saved_traces_stream = NULL;
    saved_traces_config = NULL;
    return TRUE;
}
