/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `draw.c'
    Drawing functions

    $Id: draw.c,v 1.53 2003/11/09 22:21:33 bardslea Exp $
*/

#include "draw.h"
#include "rc.h"
#include "digital.h"
#include "currenttime.h"
#include "bsearch.h"
#include "tracepane.h"
#include "signalpane.h"

/* wave_get_feature_gc : get/make GC for specified feature ,
	values less than NB_TRACE_STATES are state colors, those >= are
	pane colors + NB_TRACE_STATES */
static GdkGC *wave_get_feature_gc (WaveDrawingContext *context, int feature)
{
	GdkGC *ret = NULL;

	if (feature < NB_TRACE_STATES)
		ret = wave_get_state_color_gc (context->state_colors, feature, context->window);
	else ret = wave_get_pane_color_gc (wave_config->pane_colors, feature - NB_TRACE_STATES, context->window);

	return ret;
}

void wave_draw_gdk_line (WaveDrawingContext *context, int feature,
	int x1, int y1, int x2, int y2)
{
    GdkGC *gc = wave_get_feature_gc (context, feature);
	gdk_draw_line (*(context->data.gdk.wave_pixmap), gc, x1, y1, x2, y2);
}

void wave_draw_gdk_filled_rectangle (WaveDrawingContext *context, int feature,
	int x1, int y1, int x2, int y2)
{
    GdkGC *gc = wave_get_feature_gc (context, feature);
	gdk_draw_rectangle (*(context->data.gdk.wave_pixmap), gc, TRUE, x1, y1, 1 + x2 - x1, 1 + y2 - y1);
}

void wave_draw_gdk_filled_polygon (WaveDrawingContext *context, int feature,
	GdkPoint *points, int num_points)
{
    GdkGC *gc = wave_get_feature_gc (context, feature);
    gdk_draw_polygon (*(context->data.gdk.wave_pixmap), gc, TRUE, points, num_points);
}

void wave_draw_gdk_string (WaveDrawingContext *context, int feature,
	WaveDrawingFont font, int x, int y, char *string)
{
	GdkPixmap *pixmap;
    GdkGC *gc;

	/* Use special GC selection for SIGNAL font, these GC choices are a bit odd */
    switch (font)
    {
    case WAVE_FONT_SIGNAL:
		pixmap = *(context->data.gdk.signal_pixmap);
        gc = context->data.gdk.signal_style->fg_gc[GTK_STATE_NORMAL];
		break;
    case WAVE_FONT_SIGNAL_SELECTED:
		pixmap = *(context->data.gdk.signal_pixmap);
        gc = context->data.gdk.signal_style->white_gc;
		break;
    case WAVE_FONT_SIGNAL_SEPARATOR:
		pixmap = *(context->data.gdk.signal_pixmap);
        gc = context->data.gdk.signal_style->bg_gc[GTK_STATE_SELECTED];
		break;
    default:
		pixmap = *(context->data.gdk.wave_pixmap);
    	gc = wave_get_feature_gc (context, feature);
		break;
	}
	gdk_draw_string (pixmap, context->data.gdk.fonts[font], gc, x, y, string);
}

/* wave_set_ps_color : set the color of printing to that specified by feature,
	values less than NB_TRACE_STATES are state colors, those >= are
	pane colors + NB_TRACE_STATES */
static void wave_set_ps_color (WaveDrawingContext *context, int feature)
{
	WaveColor color;

	if (feature < NB_TRACE_STATES)
		color = context->state_colors->colors[feature];
	else color = wave_config->pane_colors->colors[feature - NB_TRACE_STATES];

	/* Change a colour to black if it doesn't contrast with white enough,
		FIXME, have options here? */
	if (color.red > 0xE0 && color.green > 0xE0 && color.blue > 0xE0)
		color.red = color.green = color.blue = 0;

	fprintf (context->data.ps.file, "%0.3g %0.3g %0.3g rgb\n",
		color.red / 255.0, color.green / 255.0, color.blue / 255.0);
}

void wave_draw_ps_line (WaveDrawingContext *context, int feature,
	int x1, int y1, int x2, int y2)
{
	wave_set_ps_color (context, feature);
	fprintf (context->data.ps.file, "%d %d %d %d l\n", x1, y1, x2, y2);
}

void wave_draw_ps_filled_rectangle (WaveDrawingContext *context, int feature,
	int x1, int y1, int x2, int y2)
{
	wave_set_ps_color (context, feature);
	fprintf (context->data.ps.file, "%d %d %d %d fr\n", x1, y1, x2, y2);
}

void wave_draw_ps_filled_polygon (WaveDrawingContext *context, int feature,
	GdkPoint *points, int num_points)
{
	int i;

	wave_set_ps_color (context, feature);
	for (i = 0; i < num_points; i ++)
		fprintf (context->data.ps.file, "%d %d ", points[i].x, points[i].y);
	fprintf (context->data.ps.file, "%d fp\n", num_points);
}

#undef SHOW_STRING_ORIGIN

void wave_draw_ps_string (WaveDrawingContext *context, int feature,
	WaveDrawingFont font, int x, int y, char *string)
{
	wave_set_ps_color (context, feature);
#ifdef SHOW_STRING_ORIGIN
	fprintf (context->data.ps.file, "%d %d %d %d l\n", x - 1, y, x + 1, y);
	fprintf (context->data.ps.file, "%d %d %d %d l\n", x, y - 1, x, y + 1);
#endif
	fprintf (context->data.ps.file, "%d %d (%s) s\n", x, y, string);
}

WaveDrawingFunctions wave_drawing_functions[] =
{	
	{wave_draw_gdk_line, wave_draw_gdk_filled_rectangle,
		wave_draw_gdk_filled_polygon, wave_draw_gdk_string},
	{wave_draw_ps_line, wave_draw_ps_filled_rectangle,
		wave_draw_ps_filled_polygon, wave_draw_ps_string},
};

/* wave_drawing_context_set_trace : set the transient, individual trace related,
	info in the drawing context.  Remember to do this between each TraceEnt */
void wave_drawing_context_set_trace (WaveDrawingContext *context, TraceEnt *trace_ent)
{
	context->trace = trace_ent;

	if (trace_ent)
		context->state_colors = trace_config_get_state_colors (trace_ent->config, wave_config);
	else context->state_colors = NULL;
}

/* wave_new_gdk_drawing_context : make a new drawing context onto the given pixmaps,
	the pixmaps are passed by reference so that the original locations' pixmaps
	can be reallocated without invalidating this drawing context. */
WaveDrawingContext *wave_new_gdk_drawing_context (GdkPixmap **wave_pixmap,
	GdkPixmap **signal_pixmap, GtkStyle *signal_style, WaveWindow *window)
{
	WaveDrawingContext *ret = g_new (WaveDrawingContext, 1);

	ret->window = window;
	ret->destination = WAVE_DRAWING_GDK;
	ret->data.gdk.wave_pixmap = wave_pixmap;
	ret->data.gdk.signal_pixmap = signal_pixmap;
	ret->data.gdk.signal_style = signal_style;
	ret->data.gdk.fonts[WAVE_FONT_WAVE] = wavefont;
	ret->data.gdk.fonts[WAVE_FONT_SMALL] = wavefont_smaller;
	ret->data.gdk.fonts[WAVE_FONT_SIGNAL] = signalfont;
	ret->data.gdk.fonts[WAVE_FONT_SIGNAL_SELECTED] = signalfont;
	ret->data.gdk.fonts[WAVE_FONT_SIGNAL_SEPARATOR] = signalfont;

	wave_drawing_context_set_trace (ret, NULL);

	return ret;
}

/* wave_new_ps_drawing_context : make a new drawing context for PostScript printing */
WaveDrawingContext *wave_new_ps_drawing_context (FILE *file, WaveWindow *window)
{
	WaveDrawingContext *ret = g_new (WaveDrawingContext, 1);

	ret->window = window;
	ret->destination = WAVE_DRAWING_PS;
	ret->data.ps.file = file;

	wave_drawing_context_set_trace (ret, NULL);

	return ret;

}

/* new_trace_drawing_info: make an empty (referring to the given drawer) TraceDrawingInfo structure */
TraceDrawingInfo *new_trace_drawing_info (TraceDrawer *drawer)
{
   	TraceDrawingInfo *ret = g_new (TraceDrawingInfo, 1);

   	ret->drawer = drawer;
   	ret->context = NULL;
   	ret->free_context = NULL;
   	ret->recommended_height = 0;

	return ret;
}

/* wave_draw_truncated_string : draw a STATE_FONT string truncated to the given width */
extern void wave_draw_truncated_string (WaveDrawingContext *context, int x, int y, int edge, int width, char *string)
{
    if (edge >= context->window->traces.scale.wave_pane_width ||
    	gdk_string_measure (wavefont, string) + wave_flags.vector_padding <= width)
		WAVE_DRAW_STATE_STRING (context, TRACE_STATE_FONT, WAVE_FONT_WAVE, x, y, string);
    else {
        char *mod = bsearch_trunc (string, width - wave_flags.vector_padding);

        if (mod)
        {
            mod[0] = '+';
            mod[1] = '\0';
			WAVE_DRAW_STATE_STRING (context, TRACE_STATE_FONT, WAVE_FONT_WAVE, x, y, string);
        }
    }
}

void unset_trace_drawer (TraceEnt * t)
{
    if (t->drawing_info->context)
    {
        if (t->drawing_info->free_context)
            t->drawing_info->free_context (t->drawing_info->context);
        else g_free (t->drawing_info->context);

        t->drawing_info->drawer = NULL;
    }
    t->drawing_info->drawer = NULL;
    t->drawing_info->free_context = NULL;
}

void draw_bit_trace_start (WaveDrawingContext *context, Node *node, WaveEvent *ev, int y0, int y1)
{
    TraceDrawer *d;

    if (context->trace->drawing_info && context->trace->drawing_info->drawer->compatible (context->trace))
        d = context->trace->drawing_info->drawer;
    else
        d = &digital_trace_drawer;
    if (d->bit_trace_start) d->bit_trace_start (context, node, ev, y0, y1);
}

void draw_bit_trace (WaveDrawingContext *context, Node *node, WaveEvent *ev, WaveEvent *ev2, int whichgc,
  gboolean first_at_this_time, int x0, int y0, int x1, int y1, int x, int ytext)
{
    TraceDrawer *d;

    if (context->trace->drawing_info && context->trace->drawing_info->drawer->compatible (context->trace))
        d = context->trace->drawing_info->drawer;
    else
        d = &digital_trace_drawer;
    if (d->bit_trace) d->bit_trace (context, node, ev, ev2, whichgc, first_at_this_time, x0, y0, x1, y1, x, ytext);
}

void draw_bit_spike (WaveDrawingContext *context, Node *node, int x0, int y0, int x1, int y1)
{
    TraceDrawer *d;

    if (context->trace->drawing_info && context->trace->drawing_info->drawer->compatible (context->trace))
        d = context->trace->drawing_info->drawer;
    else
        d = &digital_trace_drawer;
    if (d->bit_spike) d->bit_spike (context, node, x0, y0, x1, y1);
}

void draw_value_trace1 (WaveDrawingContext *context, Node *node, int *trans, int type,
	int last, int x0, int y0, int y1, int yu)
{
    TraceDrawer *d;

    if (context->trace->drawing_info && context->trace->drawing_info->drawer->compatible (context->trace))
        d = context->trace->drawing_info->drawer;
    else
        d = &digital_trace_drawer;
    if (d->value_trace1) d->value_trace1 (context, node, trans, type, last, x0, y0, y1, yu);
}

void draw_value_trace2 (WaveDrawingContext *context, Node *node, WaveEvent *ev, int *trans, int type,
	int x, int x0, int y0, int x1, int y1, int yu, int ytext)
{
    TraceDrawer *d;

    if (context->trace->drawing_info && context->trace->drawing_info->drawer->compatible (context->trace))
        d = context->trace->drawing_info->drawer;
    else
        d = &digital_trace_drawer;
    if (d->value_trace2) d->value_trace2 (context, node, ev, trans, type, x, x0, y0, x1, y1, yu, ytext);
}

void save_drawing_context (FILE * f, TraceEnt * t)
{
    if (t->drawing_info && t->drawing_info->drawer->save_context)
        t->drawing_info->drawer->save_context (f, t->drawing_info->context);
}
