/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `draw.c'
    Drawing functions

    $Id: draw.h,v 1.29 2003/11/09 22:21:33 bardslea Exp $
*/

#ifndef WAVE_DRAW_H
#define WAVE_DRAW_H

#include <gdk/gdk.h>
#include "analyzer.h"
#include "event.h"
#include "view.h"

extern GdkPixmap *wavepixmap;

/* Destination device for drawing.  Used as indices into wave_drawing_functions */
typedef enum
{
	WAVE_DRAWING_GDK,
	WAVE_DRAWING_PS
}
WaveDrawingDestination;

/* Font selection for WAVE_DRAW_STRING */
typedef enum
{
	WAVE_FONT_WAVE,
	WAVE_FONT_SMALL,
	WAVE_FONT_SIGNAL,
	WAVE_FONT_SIGNAL_SEPARATOR,
	WAVE_FONT_SIGNAL_SELECTED,
	NB_WAVE_FONTS
}
WaveDrawingFont;

/* Gathering of context information for printing */
typedef struct
{
	WaveWindow *window;
	WaveDrawingDestination destination;
	union
	{
		struct {
			GdkPixmap **wave_pixmap;
			GdkPixmap **signal_pixmap;
			GtkStyle *signal_style;
			GdkFont *fonts[NB_WAVE_FONTS];
		} gdk;
		struct {
			FILE *file;
		} ps;
	} data;
	TraceEnt *trace; /* transient entry for currently-being-drawn trace
		also provides link to TraceDrawingInfo */
	TraceStateColors *state_colors; /* transient entry " " */
}
WaveDrawingContext;

typedef struct
{
	void (*line) (WaveDrawingContext *, int, int, int, int, int);
	void (*filled_rectangle) (WaveDrawingContext *, int, int, int, int, int);
	void (*filled_polygon) (WaveDrawingContext *, int, GdkPoint *, int);
	void (*string) (WaveDrawingContext *, int, WaveDrawingFont, int, int, char *);
}
WaveDrawingFunctions;

#define WAVE_DRAW_STATE_LINE(context,state,x1,y1,x2,y2) \
	(wave_drawing_functions[(context)->destination].line \
		((context),(state),(x1),(y1),(x2),(y2)))
#define WAVE_DRAW_PANE_LINE(context,feature,x1,y1,x2,y2) \
	(wave_drawing_functions[(context)->destination].line \
		((context),(feature) + NB_TRACE_STATES,(x1),(y1),(x2),(y2)))
#define WAVE_DRAW_STATE_FILLED_RECTANGLE(context,state,x1,y1,x2,y2) \
	(wave_drawing_functions[(context)->destination].filled_rectangle \
		((context),(state),(x1),(y1),(x2),(y2)))
#define WAVE_DRAW_PANE_FILLED_RECTANGLE(context,feature,x1,y1,x2,y2) \
	(wave_drawing_functions[(context)->destination].filled_rectangle \
		((context),(feature) + NB_TRACE_STATES,(x1),(y1),(x2),(y2)))
#define WAVE_DRAW_STATE_FILLED_POLYGON(context,state,points,num_points) \
	(wave_drawing_functions[(context)->destination].filled_polygon \
		((context),(state),(points),(num_points)))
#define WAVE_DRAW_STATE_STRING(context,state,font,x,y,str) \
	(wave_drawing_functions[(context)->destination].string \
		((context),(state),(font),(x),(y),(str)))
#define WAVE_DRAW_PANE_STRING(context,feature,font,x,y,str) \
	(wave_drawing_functions[(context)->destination].string \
		((context),(feature) + NB_TRACE_STATES,(font),(x),(y),(str)))

typedef struct TraceDrawer_
{
    const char *name;

    /* This function should return whether this trace drawer can correctly render the given trace.
		It is the only field which MUST NOT be NULL.  If the trace drawer can't handle the trace,
     	the drawing engine falls back to the built-in default... */
     gboolean (*compatible) (TraceEnt * t);

    /* These functions are used to draw traces of single bits (which are, of course, four-state values.) */
    void (*bit_trace_start) (WaveDrawingContext *context, Node *node, WaveEvent *ev, int y0, int y1);
    void (*bit_trace) (WaveDrawingContext *context, Node *node, WaveEvent *ev, WaveEvent *ev2, int whichgc,
      gboolean first_at_this_time, int x0, int y0, int x1, int y1, int x, int ytext);
    void (*bit_spike) (WaveDrawingContext *context, Node *node, int x0, int y0, int x1, int y1);

    /* These functions are used to draw traces of single values which are more complex than a
		single bit (e.g. a string or a floating-point number.) */
    void (*value_trace1) (WaveDrawingContext *context, Node *node, int *trans, int type,
    	int lasttype, int x0, int y0, int y1, int yu);
    void (*value_trace2) (WaveDrawingContext *context, Node *node, WaveEvent *ev, int *trans, int type, int x,
      int x0, int y0, int x1, int y1, int yu, int ytext);

    void (*save_context) (FILE * f, void *context);
}
TraceDrawer;

/* Per-trace drawing information, including which TraceDrawer to use.
	Note that this info. is usually set by the reader */
typedef struct TraceDrawingInfo_
{
    TraceDrawer *drawer; /* drawing engine to use */
    void *context;              /* local data for this drawer */
    void (*free_context) (void *context); /* function to free local data */

    unsigned recommended_height; /* recommended starting height for a signal,
    								you need not respect this if a saved trace file
    								tells you otherwise. 0 means `don't care' */
}
TraceDrawingInfo;

/* new_trace_drawing_info: make an empty (referring to the given drawer) TraceDrawingInfo structure */
extern TraceDrawingInfo *new_trace_drawing_info (struct TraceDrawer_ *drawer);

extern WaveDrawingFunctions wave_drawing_functions[];

/* wave_new_gdk_drawing_context : make a new drawing context onto the given pixmaps,
	the pixmaps are passed by reference so that the original locations' pixmaps
	can be reallocated without invalidating this drawing context. */
extern WaveDrawingContext *wave_new_gdk_drawing_context (GdkPixmap **wave_pixmap,
	GdkPixmap **signal_pixmap, GtkStyle *signal_style, WaveWindow *window);
/* wave_new_ps_drawing_context : make a new drawing context for PostScript printing */
extern WaveDrawingContext *wave_new_ps_drawing_context (FILE *file, WaveWindow *window);

/* wave_drawing_context_set_trace : set the transient, individual trace related,
	info in the drawing context.  Remember to do this between each TraceEnt */
extern void wave_drawing_context_set_trace (WaveDrawingContext *context, TraceEnt *trace_ent);

/* wave_draw_truncated_string : draw a STATE_FONT string truncated to the given width */
extern void wave_draw_truncated_string (WaveDrawingContext *context, int x, int y, int edge, int width, char *string);

extern void draw_bit_trace_start (WaveDrawingContext *context, Node *node, WaveEvent *ev, int y0, int y1);
extern void draw_bit_trace (WaveDrawingContext *context, Node *node, WaveEvent *ev, WaveEvent *ev2, int whichgc,
  gboolean first_at_this_time, int x0, int y0, int x1, int y1, int x, int ytext);
extern void draw_bit_spike (WaveDrawingContext *context, Node *node, int x0, int y0, int x1, int y1);

extern void draw_value_trace1 (WaveDrawingContext *context, Node *node, int *trans, int type,
	int lasttype, int x0, int y0, int y1, int yu);
extern void draw_value_trace2 (WaveDrawingContext *context, Node *node, WaveEvent *ev, int *trans, int type, int x,
  int x0, int y0, int x1, int y1, int yu, int ytext);

extern void unset_trace_drawer (TraceEnt * t);
extern void save_drawing_context (FILE * f, TraceEnt * t);

/* Need to find somewhere better to put these once tracewindow.c has been integrated with
	signalwindow.c FIXME AB */
extern void render_signal_names (WaveDrawingContext *context, int trtarget);

#endif
