/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `event.c'
	Event trace list and list iterator handling

    $Id: event.c,v 1.9 2003/11/07 09:56:33 bardslea Exp $
*/

#include "event.h"
#include "log.h"
#include "value.h"

/* FIXME, focus some optimisation effort on the size of HistEnt. Bitfields? */
typedef struct HistEnt_
{
    struct HistEnt_ *next;      /* next transition in history */
    TimeType time;              /* time of transition */

    union
    {
        char val;      			/* value: "0XZ1"[val] */
        char *vector;           /* pointer to a whole vector */
    }
    v;
}
HistEnt;

typedef struct WaveEventPositionPrivate_
{
	HistEnt *previousHistent;
}
WaveEventPositionPrivate;

/* WaveEventList : event lists are traces of events associated with a
	particular Node in which events are held in sequence of their time
	stamps but can be searched for in better than O(n) time.
	Note that the WaveEventList structure is private to event.c */
typedef struct WaveEventList_
{
	/* FIXME, implemented with histents for the moment */
	Node *node;
	HistEnt *head;
	HistEnt *curr;
	TimeType last_time;
}
WaveEventList;

/* WaveEventIterator : iterator for WaveEventLists */
typedef struct WaveEventIteratorPrivate_
{
	WaveEvent events[3]; /* Event structures for previous, curr, next */
	/* FIXME, implemented with histents for the moment */
	HistEnt *previousHistent;
}
WaveEventIteratorPrivate;

/* wave_new_event_list : make a new (empty) event list */
WaveEventList *wave_new_event_list (Node *node)
{
	WaveEventList *ret = g_new (WaveEventList, 1);

	ret->node = node;
	ret->head = NULL;
	ret->curr = NULL;
	ret->last_time = 0;

	return ret;
}

/* wave_copy_histent_to_event : simple copy of time/value, returns a 
	pointer to the source event or NULL if the hist_ent is NULL */
static WaveEvent *wave_copy_histent_to_event (WaveEvent *event, HistEnt *hist_ent)
{
	WaveEvent *ret = NULL;

	if (hist_ent)
	{
		event->time = hist_ent->time;
		event->value.vector = hist_ent->v.vector;
		ret = event;
	}

	return ret;
}

/* wave_event_iterator_init_events : update event structure contents for
	event iterator */
static void wave_event_iterator_init_events (WaveEventIterator *iterator)
{
	HistEnt *previous = iterator->priv->previousHistent;
	WaveEvent *events = iterator->priv->events;

	if (previous) /* Not at start of trace */
	{
		iterator->previous = wave_copy_histent_to_event (events + 0, previous);
		iterator->current = wave_copy_histent_to_event (events + 1, previous->next);
		iterator->next = wave_copy_histent_to_event (events + 2,
			(previous->next ? previous->next->next : NULL));
	} else {
		iterator->previous = NULL;
		iterator->current = wave_copy_histent_to_event (events + 1, iterator->event_list->head);
		if (iterator->current)
			iterator->next = wave_copy_histent_to_event (events + 2, iterator->event_list->head->next);
		else iterator->next = NULL;
	}	
}

/* wave_new_event_iterator : make an iterator for the given event list
	(starts with `curr' being the first event */
WaveEventIterator *wave_new_event_iterator (WaveEventList *event_list)
{
	WaveEventIterator *ret = g_new (WaveEventIterator, 1);

	ret->priv = g_new (WaveEventIteratorPrivate, 1);
	ret->priv->previousHistent = NULL;
	ret->event_list = event_list;

	wave_event_iterator_init_events (ret);

	return ret;
}

/* wave_free_event_iterator : frees an iterator and it's previous/curr/next event elements */
void wave_free_event_iterator (WaveEventIterator *iterator)
{
	g_free (iterator);
}

/* wave_event_iterator_step_{forward,backward} : step to next/previous element of event_list.
	Returns FALSE and doesn't update current/previous/next pointers if this operation
	isn't possible (ie. we're already at the end/start of the event trace) */
gboolean wave_event_iterator_step_forward (WaveEventIterator *iterator)
{
	if (! iterator->next)
		return FALSE;
		
	/* Move current and next events to previous and current */
	iterator->priv->events[0] = iterator->priv->events[1];
	iterator->previous = iterator->priv->events + 0; /* May not have been a previous before */
	iterator->priv->events[1] = iterator->priv->events[2];

	/* Find the new next event */
	{
		HistEnt *previous = iterator->priv->previousHistent;

		if (previous)
			previous = previous->next;
		else previous = iterator->event_list->head;

		iterator->next = wave_copy_histent_to_event (iterator->priv->events + 2,
			(previous->next ? previous->next->next : NULL));

		iterator->priv->previousHistent = previous;
	}

	return TRUE;
}
gboolean wave_event_iterator_step_backward (WaveEventIterator *iterator)
{
	if (! iterator->previous)
		return FALSE;

	/* Move previous and current events on to current and next */
	iterator->previous = iterator->priv->events + 2; /* May not have been a next before */
	iterator->priv->events[1] = iterator->priv->events[0];
	iterator->priv->events[2] = iterator->priv->events[1];

	/* Find the new next event */
	{
		HistEnt *previous = iterator->priv->previousHistent;

		fprintf (stderr, "FIXME, can't step backwards yet\n");
	}

	return TRUE;
}

/* wave_event_iterator_seek : set the current event in the given iterator to be
	an event matching a search for time `time' under the direction of match conditions
	`whence'.  Returns TRUE on success and sets previous/current/next pointers of
	the iterator.  On failure, returns FALSE and makes no change to iterator */
gboolean wave_event_iterator_seek (WaveEventIterator *iterator, TimeType time,
	WaveEventIteratorWhence whence)
{
	/* Just search from the beginning for now */
	iterator->priv->previousHistent = NULL;
	wave_event_iterator_init_events (iterator);

	if (! iterator->current)
		return FALSE;

	do
	{
	 /* FIXME, need to support the whence */
	 	if (iterator->current->time <= time &&
	 		(! iterator->next || iterator->next->time > time))
	 		return TRUE;
	} while (wave_event_iterator_step_forward (iterator));

	return FALSE;
}

/* wave_event_iterator_get_position : make a position describing structure for this iterator's
	current position */
WaveEventPosition wave_event_iterator_get_position (WaveEventIterator *iterator)
{
	WaveEventPosition ret;
	ret.priv = g_new (WaveEventPositionPrivate, 1);

	ret.priv->previousHistent = iterator->priv->previousHistent;

	return ret;
}
/* wave_event_iterator_set_position : move to a position returned by wave_event_iterator_get_position */
void wave_event_iterator_set_position (WaveEventIterator *iterator,
	WaveEventPosition position)
{
	iterator->priv->previousHistent = position.priv->previousHistent;
	wave_event_iterator_init_events (iterator);
}

/* wave_event_position_free : destructor for allocated structure returned by
	wave_event_iterator_get_position */
void wave_event_position_free (WaveEventPosition position)
{
	g_free (position.priv);
}

/*
 * Specialised allocator for histent structures.  These are NEVER
 * freed so this is OK...
 */
#ifndef HISTENT_GRANULARITY
#define HISTENT_GRANULARITY 100
#endif

HistEnt *wave_new_hist_ent_raw (void)
{
    static HistEnt *he_curr = NULL, *he_fini = NULL;

    if (he_curr == he_fini)
    {
        he_curr = g_new (HistEnt, HISTENT_GRANULARITY);
        he_fini = he_curr + HISTENT_GRANULARITY;
    }

    return he_curr ++;
}

/* wave_new_hist_ent : make a new HistEnt structure with the provided entries.
	The use of `val' or `vector' for the HistEnt value is chosen by `valNvector' */
static HistEnt *wave_new_hist_ent (TimeType time, gboolean valNvector, char val, void *vector)
{
	HistEnt *ret = wave_new_hist_ent_raw ();

	ret->next = NULL;
	ret->time = time;

	if (valNvector)
		ret->v.val = val;
	else ret->v.vector = vector;

	return ret;
}

/* find_hist_ent_at_time (HistEnt *hist_ents, TimeType time) : find the first
	(if firstNlast is true) or last (if firstNlast is false)
	HistEnt in the list which is at a time equal or less than time.  Returns
	NULL if `hist_ents' is NULL or the last element in `hist_ents' if no entry matches.
*/
static HistEnt *find_hist_ent_at_time (HistEnt *hist_ents, TimeType time, gboolean firstNlast)
{
	HistEnt *ret = hist_ents;

	if (firstNlast)
	{
		while (hist_ents && hist_ents->time < time)
		{
			ret = hist_ents;
			hist_ents = hist_ents->next;
		}
	} else {
		while (hist_ents && hist_ents->time <= time)
		{
			ret = hist_ents;
			hist_ents = hist_ents->next;
		}
	}

	return ret;
}

/* hist_ent_insert_after_first : insert the HistEnt h2 after the first element of h1 and return h2 */
static HistEnt *hist_ent_insert_after_first (HistEnt *h1, HistEnt *h2)
{
	if (h1)
	{
		if (h2) h2->next = h1->next;
		h1->next = h2;
	}

	return h2;
}

/* node_fix_curr_after_insertion : fix the event_list->curr end of trace pointer after an insertion
	into the Node trace history ahead of event_list->curr */
static void node_fix_curr_after_insertion (WaveEventList *event_list)
{
	while (event_list->curr && event_list->curr->next)
		event_list->curr = event_list->curr->next;
}

/* node_append_hist_ent : append the given HistEnt to a node's trace history.  Assumes that
	event_list->curr is not NULL */
static void node_append_hist_ent (WaveEventList *event_list, HistEnt *hist_ent)
{
	event_list->curr->next = hist_ent;
	event_list->curr = hist_ent;
}

/* hist_ent_flags_is_vector_valued : do the given HistEnt flags imply that this HistEnt
	has a ``vector'' value. */
static gboolean hist_ent_flags_is_vector_valued (NodeFlags flags)
{
	return (flags & NODE_FLAGS_VECTOR_MASK) != 0;
}

/* wave_insert_initial_dummy_event : insert an initial dummy HistEnt into a signal's history.
	This must be done for each signal as they are declared. */
static HistEnt *wave_insert_initial_dummy_event (Node *n, NodeFlags flag)
{
	char val = 'X';
	void *vector = NULL;
	gboolean valNvector = FALSE;

	switch (flag & NODE_FLAGS_TYPE_MASK)
	{
	case NODE_FLAGS_DIGITAL:
		valNvector = TRUE;
		break;
	case NODE_FLAGS_VECTOR:
		vector = g_strnfill ((n->ext ? ABS (n->ext->msi - n->ext->lsi) + 1 : 0), 'X');
		break;
	case NODE_FLAGS_STRING:
		vector = g_strdup ("UNDEF");
		break;
	case NODE_FLAGS_REAL:
		vector = NULL;
		break;
    case NODE_FLAGS_SYNC_HANDSHAKE:
    	val = HANDSHAKE_STATE_QUIESCENT;
    	valNvector = TRUE;
    	break;
    case NODE_FLAGS_DATA_HANDSHAKE:
    	{
			HandshakeValue *handshake_value = g_new (HandshakeValue, 1);
			handshake_value->val = HANDSHAKE_STATE_QUIESCENT;
			handshake_value->data = NewFormatDataFromUnsignedInt (1, 0);

			vector = handshake_value;
		}
    	break;
    default:
    	wave_log (TRUE, "Invalid flag passed to insert_initial_histent (%d) for node %s\n", flag, n->nname);
    	exit (EXIT_FAILURE);
    	break;
	}

    return wave_new_hist_ent (-1, valNvector, val, vector);
}


/* wave_node_insert_event : insert an event into the Node `n' trace history.
	`event_type' should be set to the appropriate flag for the Node's data type and `val' or `vector'
	should be the required value.  Note that event_type is present for error checking
	against the node type flag.  Setting event_type to 0 will turn off this feature.
*/
void wave_node_insert_event (Node *node, TimeType time, NodeFlags event_type, char val, void *vector)
{
    HistEnt *he = NULL;
    WaveEventList *event_list = node->iterator->event_list;
	NodeFlags node_event_type = node->flags & NODE_FLAGS_TYPE_MASK;
	gboolean is_vector = hist_ent_flags_is_vector_valued (node_event_type);

	if (event_type != 0 && node_event_type != (event_type & NODE_FLAGS_TYPE_MASK))
	{
		wave_log (TRUE, "Event type mismatch in trace `%s', expecting %s, got %s\n", node->nname,
			wave_node_event_type_name (node_event_type), wave_node_event_type_name (event_type));
		return;
	}

    if (! event_list->curr) /* Make the first element */
		event_list->head = event_list->curr = wave_insert_initial_dummy_event (node, node_event_type);

	he = wave_new_hist_ent (time, ! is_vector, val, vector);

	if (time > event_list->last_time)
	{
		event_list->last_time = time;
		node_append_hist_ent (event_list, he);
	} else {
		HistEnt *insert_point = find_hist_ent_at_time (event_list->head, time, FALSE);

		hist_ent_insert_after_first (insert_point, he);
		node_fix_curr_after_insertion (event_list);
	}
}
