/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `event.h'
	Event trace list and list iterator handling

    $Id: event.h,v 1.8 2003/11/07 09:56:33 bardslea Exp $
*/

#ifndef WAVE_EVENT_H
#define WAVE_EVENT_H

#include "analyzer.h"

/* WaveEventList : event lists are traces of events associated with a
	particular Node in which events are held in sequence of their time
	stamps but can be searched for in better than O(n) time.
	Note that the WaveEventList structure is private to event.c */
struct WaveEventList_;

/* WaveValue : values of events are either characters ('x', 'z' ...) or
	pointers to some other structure.  What type of structure can be
	determined by the flags in this Event's EventList's Node */ 
typedef union
{
	char val;
	void *vector;	/* vector char *, double *, ... */
}
WaveEventValue;

/* WaveEvent : a single event reporting structure */
typedef struct
{
	TimeType time;
	WaveEventValue value;
}
WaveEvent;

/* WaveEventIterator : iterator for WaveEventLists */
struct WaveEventIteratorPrivate_;
typedef struct WaveEventIterator_
{
	struct WaveEventList_ *event_list;
	WaveEvent *previous, *current, *next; /* Previous, current and next events, should always
		be kept upto date after by iterator operations */
	struct WaveEventIteratorPrivate_ *priv;
}
WaveEventIterator;

/* WaveEventPosition : position in event list which can be
	noted and moved (back) to */
struct WaveEventPositionPrivate_;
typedef struct 
{
	struct WaveEventPositionPrivate_ *priv;
} WaveEventPosition;

/* WaveEventIteratorWhence : which event should be searched upto by wave_event_iterator_seek */
typedef enum
{
	WAVE_EVENT_LAST_EQ,		/* last event at time equal to search time */
	WAVE_EVENT_FIRST_EQ,	/* first event " */
	WAVE_EVENT_LAST_LE,		/* last event at time equal or before search time */
	WAVE_EVENT_FIRST		/* first of all events */
}
WaveEventIteratorWhence;

/* wave_new_event_list : make a new (empty) event list */
extern struct WaveEventList_ *wave_new_event_list (Node *node);

/* wave_new_event_iterator : make an iterator for the given event list
	(starts with `curr' being the first event */
extern WaveEventIterator *wave_new_event_iterator (struct WaveEventList_ *event_list);

/* wave_free_event_iterator : frees an iterator and it's prev/curr/next event elements */
extern void wave_free_event_iterator (WaveEventIterator *iterator);

/* wave_event_iterator_step_{forward,backward} : step to next/previous element of event_list.
	Returns FALSE and doesn't update current/previous/next pointers if this operation
	isn't possible (ie. we're already at the end/start of the event trace) */
extern gboolean wave_event_iterator_step_forward (WaveEventIterator *iterator);
extern gboolean wave_event_iterator_step_backward (WaveEventIterator *iterator);

/* wave_event_iterator_get_position : make a position describing structure for this iterator's
	current position */
extern WaveEventPosition wave_event_iterator_get_position (WaveEventIterator *iterator);
/* wave_event_iterator_set_position : move to a position returned by wave_event_iterator_get_position */
extern void wave_event_iterator_set_position (WaveEventIterator *iterator,
	WaveEventPosition position);

/* wave_event_position_free : destructor for allocated structure returned by
	wave_event_iterator_get_position */
extern void wave_event_position_free (WaveEventPosition position);

/* wave_event_iterator_seek : set the current event in the given iterator to be
	an event matching a search for time `time' under the direction of match conditions
	`whence'.  Returns TRUE on success and sets previous/current/next pointers of
	the iterator.  On failure, returns FALSE and makes no change to iterator */
extern gboolean wave_event_iterator_seek (WaveEventIterator *iterator, TimeType time,
	WaveEventIteratorWhence whence);

/* wave_node_insert_event : insert an event into the Node `n' trace history.
	`event_type' should be set to the appropriate flag for the Node's data type and `val' or `vector'
	should be the required value.  Note that event_type is present for error checking
	against the node type flag.  Setting event_type to 0 will turn off this feature.
*/
extern void wave_node_insert_event (Node *node, TimeType time, NodeFlags event_type, char val, void *vector);

#endif
