/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2002-3, The University of Manchester
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `fileio.c'
    Trace file I/O (and compression) handling

    $Id: fileio.c,v 1.23 2003/11/12 23:06:22 bardslea Exp $
*/

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include "fileio.h"
#include "str.h"
#include "log.h"

/* get_file_size : get the size in bytes of the requested file */
off_t get_file_size (const char *filename)
{
    struct stat buf;
    int result = stat (filename, &buf);

    return (result == 0 ? buf.st_size : 0);
}

/* file_desc_close : close a file (if not a stream) so it can be reopened later */
void file_desc_close (FileDesc * file_desc)
{
    if (file_desc->name)
    {
        FILE_DESC_CLOSE (file_desc);
        file_desc->handle = NULL;
    }
}

/* file_desc_reopen : reopen a closed file with the same mode it was file_desc_open'ed with */
void file_desc_reopen (FileDesc * file_desc)
{
    if (file_desc->name)
    {
        FILE_DESC_OPEN (file_desc, file_desc->name, file_desc->mode);

        if (!file_desc->handle)
            wave_log (TRUE, "Unable to reopen file %s\n", file_desc->name);
    }
}

/* eof and getc are often macros */
static int file_desc_none_eof (FILE * handle)
{
    return feof (handle);
}

static int file_desc_none_getc (FILE * handle)
{
    return getc (handle);
}

static int file_desc_posix_read (void *ptr, size_t size, size_t nelem, void *handle)
{
    return read ((int) handle, ptr, size * nelem);
}

static int file_desc_posix_write (void *ptr, size_t size, size_t nelem, void *handle)
{
    return write ((int) handle, ptr, size * nelem);
}

static int file_desc_posix_always_eof (void *ptr)
{
    return 1;
}

static int do_nothing (void)
{
	return 0;
}

/* Posix functions using read/write rather than fread and fwrite are
	provided for stream use using file_desc_fdopen */
static FileFormatData file_format_data_posix = {
    "posix",
    (int (*)(const char *, const char *)) NULL,
    (void (*)(void *)) close,
    (int (*)(void *)) file_desc_posix_always_eof,
    (int (*)(void *)) NULL, 
    (int (*)(void *, size_t, size_t, void *)) file_desc_posix_read,
    (int (*)(const void *, size_t, size_t, void *)) file_desc_posix_write,
    (int (*)(void *, long, int)) NULL,
    (long (*)(void *)) NULL,
	(void (*)(void *)) do_nothing
};

static FileFormatData file_format_data_none = {
    "no",
    (int (*)(const char *, const char *)) fopen,
    (void (*)(void *)) fclose,
    (int (*)(void *)) file_desc_none_eof,
    (int (*)(void *)) file_desc_none_getc,
    (int (*)(void *, size_t, size_t, void *)) fread,
    (int (*)(const void *, size_t, size_t, void *)) fwrite,
    (int (*)(void *, long, int)) fseek,
    (long (*)(void *)) ftell,
	(void (*)(void *)) clearerr
};

#ifdef HAVE_LIBZ

static int file_desc_zlib_read (void *ptr, size_t size, size_t nelem, void *handle)
{
    return gzread ((gzFile) handle, ptr, size * nelem);
}

static int file_desc_zlib_write (void *ptr, size_t size, size_t nelem, void *handle)
{
    return gzwrite ((gzFile) handle, ptr, size * nelem);
}

static void clear_no_errors (void *ptr)
{
    wave_log (TRUE, "This file reader cannot recover from an error/EOF\n");
}

static FileFormatData file_format_data_zlib = {
    "zlib",
    (int (*)(const char *, const char *)) gzopen,
    (void (*)(void *)) gzclose,
    (int (*)(void *)) gzeof,
    (int (*)(void *)) gzgetc,
    (int (*)(void *, size_t, size_t, void *)) file_desc_zlib_read,
    (int (*)(const void *, size_t, size_t, void *)) file_desc_zlib_write,
    (int (*)(void *, long, int)) gzseek,
    (long (*)(void *)) gztell,
	(void (*)(void *)) clear_no_errors
};
#endif

#ifdef HAVE_LIBBZ2
static FileFormatData file_format_data_bzip2 = {
    "bzip2",
    /* FIXME, libbz2 doesn't have seek */
    (int (*)(const char *, const char *)) gzopen,
    (void (*)(void *)) gzclose,
    (int (*)(void *)) gzeof,
    (int (*)(void *)) gzgetc,
    (int (*)(void *, size_t, size_t, void *)) gzread,
    (int (*)(const void *, size_t, size_t, void *)) gzwrite,
    (int (*)(void *, long, int)) gzseek,
    (long (*)(void *)) gztell
};
#endif

FileDesc *file_desc_new (void)
{
    FileDesc *ret = g_new (FileDesc, 1);

    ret->handle = NULL;
    ret->compression_type = FILE_COMP_NONE;
    ret->format_data = &file_format_data_none;
    ret->name = NULL;
    ret->mode = "r";

    return ret;
}

/* file_desc_fdopen: fdopen a posix stream on a file descriptor
	and return a FileDesc for that stream.  Returns NULL on failure */
FileDesc *file_desc_fdopen (int fildes, const char *mode)
{
	FileDesc *ret = file_desc_new ();

    ret->format_data = &file_format_data_posix;
	ret->name = NULL;
	ret->handle = (void *) fildes;
	ret->mode = g_strdup (mode);

	return ret;
}

/* file_desc_open : open a file, handling compression extensions, in the specified
	fopen mode */
FileDesc *file_desc_open (const char *filename, const char *mode)
{
    FileDesc *file_desc = file_desc_new ();
    gboolean is_stdin = strcmp (filename, "-") == 0;
    char *extension;

    /* Check for stdin */
    if (is_stdin)
    {
        file_desc->name = NULL; /* Stream! */
        file_desc->handle = (void *) stdin;
        return file_desc;
    }
    file_desc->name = g_strdup (filename);
    file_desc->mode = g_strdup (mode);

    /* Set name and get extension */
    extension = wave_get_filename_extension (filename);

    /* Check available compressions : */
    if (extension)
    {
#ifdef HAVE_LIBZ
		if (g_strcasecmp (extension, "gz") == 0)
		{
			file_desc->format_data = &file_format_data_zlib;
			file_desc->compression_type = FILE_COMP_ZLIB;
		}
#endif
	}

    /* Open the file */
    file_desc->handle = (void *) FILE_DESC_OPEN (file_desc, filename, mode);

    if (!file_desc->handle)
    {
        wave_log (TRUE, "Unable to open file %s\n", filename);
        return NULL;
    }

    wave_log (FALSE, "Opening file %s with %s compression\n", filename, file_desc->format_data->name);

    return file_desc;
}

#define INITIAL_FILE_DESC_LINE_PARSER_BUFFER_SIZE (4096)
#define DFI(arg)

/* file_desc_line_parser_read : read more of the file.  The rules are:
	Always try to read to fill the buffer, ie. (buffer + allocated_buffer_size) - buffer_end_ptr
	bytes).  Update buffer_end_ptr after read.  Returns the return value from the `read' but also
	moans (badly!) on read return < 0 */
static int file_desc_line_parser_read (FileDescLineParser *parser)
{
	int read_ret = FILE_DESC_READ (parser->buffer_end_ptr, 1,
		(parser->buffer + parser->allocated_buffer_size) - parser->buffer_end_ptr, parser->file);

	/* Assume that 0 or < 0 return values don't bode well, check for EOF */
	if (read_ret <= 0)
	{
		if (! FILE_DESC_EOF (parser->file))
		{
			wave_log (TRUE, "Error reading file in line parser\n");
			wave_exit (EXIT_FAILURE);
		} else {
			FILE_DESC_CLEAR_ERROR (parser->file);
		}
	} else
		parser->buffer_end_ptr += read_ret;

	DFI (fprintf (stderr, "Read %d bytes\n", read_ret));
	return read_ret;
}

/* file_desc_line_parser_move_line_to_start_of_buffer : move everything in the buffer
	from this_line onwards to the start of the buffer */
static void file_desc_line_parser_move_line_to_start_of_buffer (FileDescLineParser *parser)
{
	int buffer_shift_distance = parser->this_line - parser->buffer;

	memcpy (parser->buffer, parser->this_line, parser->buffer_end_ptr - parser->this_line);
	parser->buffer_end_ptr -= buffer_shift_distance;
	parser->working_eol -= buffer_shift_distance;
	parser->this_line = parser->buffer;
	DFI (fprintf (stderr, "M\n"));
}

/* file_desc_line_parser_double_buffer_size : double the buffer size while preserving contents */
static void file_desc_line_parser_double_buffer_size (FileDescLineParser *parser)
{
	int old_buffer_read_length = parser->buffer_end_ptr - parser->buffer;
	int old_buffer_working_eol_offset = parser->working_eol - parser->buffer;

	parser->allocated_buffer_size *= 2;
	parser->buffer = g_renew (char, parser->buffer, parser->allocated_buffer_size);
	parser->this_line = parser->buffer;
	parser->buffer_end_ptr = parser->buffer + old_buffer_read_length;
	parser->working_eol = parser->buffer + old_buffer_working_eol_offset;

	DFI (fprintf (stderr, "Resizing file desc line parser buffer to %d bytes\n", parser->allocated_buffer_size));
}

/* file_desc_line_parser_read_line : read a line (or just point to the next one in the
	buffer).  Returns a pointer to the line in the parser buffer if a line was successfully
	read (an EOL char was found) or NULL if the file could not yet provide a complete line.
	After execution, parser->this_line will be invalid.  The read line's line termination
	will be replaced by `\0' (think strtok).  If dont_read is TRUE then don't try and
	perform a read from the file (i.e. don't block). */
char *file_desc_line_parser_read_line (FileDescLineParser *parser, gboolean dont_read)
{
	char *ret = NULL;

	while (! ret)
	{
		gboolean found_eol = FALSE;

		/* Now try and find the end of the line */
		while (! found_eol && parser->working_eol < parser->buffer_end_ptr)
		{
			if (parser->working_eol[0] == '\r' || parser->working_eol[0] == '\n')
				found_eol = TRUE;
			else parser->working_eol ++;
		}

		if (found_eol)
		{
			DFI (fprintf (stderr, "G %d\n", (int) parser->working_eol[0]));

			if (parser->working_eol[0] == '\r') /* CR */
			{
				/* Yep, we have the LF */
				if ((parser->working_eol + 1) < parser->buffer_end_ptr && parser->working_eol[1] == '\n')
				{
					ret = parser->this_line;
					parser->working_eol[0] = '\0';
					parser->working_eol += 2;
					parser->this_line = parser->working_eol;
					parser->line_num ++;
				}
				/* Otherwise spin around again */
			} else { /* LF */
				ret = parser->this_line;
				parser->working_eol[0] = '\0';
				parser->working_eol += 1;
				parser->this_line = parser->working_eol;
				parser->line_num ++;
			}
		} else
		/* No EOL found in buffer, reposition in buffer then reread */
		if (parser->working_eol == parser->buffer_end_ptr)
		{
			/* Don't block if asked not to */
			if (dont_read) return NULL;

			DFI (fprintf (stderr, "E\n"));
			/* Try shifting the line start to the start of the buffer first */
			if (parser->this_line != parser->buffer)
				file_desc_line_parser_move_line_to_start_of_buffer (parser);
			/* If the buffer really is full, resize */
			else if (parser->buffer_end_ptr == (parser->buffer + parser->allocated_buffer_size))
				file_desc_line_parser_double_buffer_size (parser);

			/* Otherwise (and also), just try reading more of the file */
			/* Return if we can't actually read anything */
			if (file_desc_line_parser_read (parser) == 0)
				break;
		}
	}

	return ret;
}

/* file_desc_line_parser_new : start a new line parser, assume that the file_desc is open */
FileDescLineParser *file_desc_line_parser_new (FileDesc *file_desc)
{
	FileDescLineParser *ret = g_new (FileDescLineParser, 1);

	ret->file = file_desc;
	ret->allocated_buffer_size = INITIAL_FILE_DESC_LINE_PARSER_BUFFER_SIZE;
	ret->buffer = g_new (char, ret->allocated_buffer_size);
	ret->buffer_end_ptr = ret->buffer;
	ret->this_line = ret->buffer;
	ret->working_eol = ret->buffer;
	ret->line_num = 0;

	/* Do some speculative reading, no, don't, let's not block
	file_desc_line_parser_read (ret);
	*/

	return ret;
}

/* simple_parse_token : parse a token from the given line, returns an
	enum value indicating what sort of token was read, also updates
	`*line' and `*token' to point past the end of the current token
	and to the start of the parsed token respectively */
SimpleTokenType simple_parse_token (char **line, char **token, WaveErrorContext *context)
{
	char *token_start = wave_skip_ws (*line);
	SimpleTokenType ret = SIMPLE_NO_TOKEN;

	if (*token_start != '\0')
	{ /* We may have a valid token */
		char *end_of_token_ptr;

		if (*token_start == '"')
		{
			gboolean in_escape = FALSE;
			ret = SIMPLE_TOKEN_STRING;	
			token_start ++;
			end_of_token_ptr = token_start;

			while (*end_of_token_ptr && (in_escape || *end_of_token_ptr != '"'))
			{
				in_escape = (*end_of_token_ptr == '\\');
				end_of_token_ptr ++;
			}
			
			if (*end_of_token_ptr == '\0')
			{
				wave_log_context (context);
				wave_log (TRUE, "invalid end of line in string token, using whole line\n");
			}
		} else {
			ret = SIMPLE_TOKEN_SYMBOL;
			end_of_token_ptr = token_start;	

			while (*end_of_token_ptr && ! isspace (*end_of_token_ptr))
				end_of_token_ptr ++;
		}

		/* Update line and token return values */
		if (token)
			*token = token_start;
		*line = end_of_token_ptr + (*end_of_token_ptr ? 1 : 0);
		*end_of_token_ptr = '\0';
	}

    return ret;
}
