/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2002-3, The University of Manchester
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `fileio.h'
    Trace file I/O (and compression) handling

    $Id: fileio.h,v 1.17 2003/06/23 12:37:40 bardslea Exp $
*/

#ifndef WAVE_FILEIO_H
#define WAVE_FILEIO_H

#include <stdio.h>
#include <glib.h>
#include <sys/types.h>
#include "error.h"
#include "../config.h"

#ifdef HAVE_LIBZ
# include <zlib.h>
#endif

#ifdef HAVE_LIBBZ2
# include <bzlib.h>
#endif

typedef enum
{
    FILE_COMP_NONE,             /* no file compression, fopen/fclose ... */
    FILE_COMP_ZLIB,             /* zlib compression */
    FILE_COMP_BZIP2             /* bzip2 compression */
}
FileCompressionType;

typedef struct
{
    char *name;                 /* name of compression type */
    int (*open) (const char *, const char *); /* filename mode */
    void (*close) (void *);     /* handle */
    int (*eof) (void *);        /* handle */
    int (*getc) (void *);       /* handle */
    int (*read) (void *, size_t, size_t, void *); /* ptr size nelem handle */
    int (*write) (const void *, size_t, size_t, void *); /* ptr size nelem handle */
    int (*seek) (void *, long, int); /* handle offset whence */
    long (*tell) (void *);      /* handle */
	void (*clear_error) (void *);	/* handle */
}
FileFormatData;

typedef struct
{
    FileCompressionType compression_type;
    void *handle;
    FileFormatData *format_data;
    char *name;                 /* filename and mode for reopen */
    char *mode;
}
FileDesc;

#define FILE_DESC_OPEN(stream, filename, mode) \
	(((stream)->format_data->open) ((filename), (mode)))
#define FILE_DESC_CLOSE(stream) \
	(((stream)->format_data->close) ((stream)->handle))
#define FILE_DESC_EOF(stream) \
	(((stream)->format_data->eof) ((stream)->handle))
#define FILE_DESC_GETC(stream) \
	(((stream)->format_data->getc) ((stream)->handle))
#define FILE_DESC_READ(ptr, size, nelem, stream) \
	(((stream)->format_data->read) ((ptr), (size), (nelem), (stream)->handle))
#define FILE_DESC_WRITE(ptr, size, nelem, stream) \
	(((stream)->format_data->write) ((ptr), (size), (nelem), (stream)->handle))
#define FILE_DESC_SEEK(stream, offset, whence) \
	(((stream)->format_data->seek) ((stream)->handle, (offset), (whence)))
#define FILE_DESC_TELL(stream) \
	(((stream)->format_data->tell) ((stream)->handle))
#define FILE_DESC_CLEAR_ERROR(stream) \
	(((stream)->format_data->clear_error) ((stream)->handle))

/* get_file_size : get the size in bytes of the requested file */
extern off_t get_file_size (const char *filename);

extern FileDesc *file_desc_new (void);

/* file_desc_fdopen: fdopen a posix stream on a file descriptor
	and return a FileDesc for that stream.  Returns NULL on failure */
extern FileDesc *file_desc_fdopen (int fildes, const char *mode);

/* file_desc_open : open a file, handling compression extensions, in the specified
	fopen mode */
extern FileDesc *file_desc_open (const char *filename, const char *mode);

/* file_desc_close : close a file (if not a stream) so it can be reopened later */
extern void file_desc_close (FileDesc * file_desc);

/* file_desc_reopen : reopen a closed file with the same mode it was file_desc_open'ed with */
extern void file_desc_reopen (FileDesc * file_desc);

/* Line parsing types/functions for FileDesc objects */

typedef struct
{
	FileDesc *file;		/* File to parse from */
	char *buffer;		/* Buffered read block */
	char *buffer_end_ptr;	/* End of valid buffer entries + 1 (so that buffer_end_ptr == buffer
		denotes an empty, but allocated, buffer) */
	char *this_line;	/* Pointer to present line in buffer, invalid after failed calls to
		file_desc_line_parser_read_line */
	char *working_eol;	/* Pointer to end of currently-being-constructed line, never directly read this */
	int allocated_buffer_size; /* Buffer grows if line length > previous buffer size */
	int line_num;		/* Line number of last read line, set to 0 if no lines have been read */
}
FileDescLineParser;

/* file_desc_line_parser_new : start a new line parser, assume that the file_desc is open */
extern FileDescLineParser *file_desc_line_parser_new (FileDesc *file_desc);

/* file_desc_line_parser_read_line : read a line (or just point to the next one in the
	buffer).  Returns a pointer to the line in the parser buffer if a line was successfully
	read (an EOL char was found) or NULL if the file could not yet provide a complete line.
	After execution, parser->this_line will be invalid.  The read line's line termination
	will be replaced by `\0' (think strtok).  If dont_read is TRUE then don't try and
	perform a read from the file (i.e. don't block). */
extern char *file_desc_line_parser_read_line (FileDescLineParser *parser, gboolean dont_read);

/* SimpleTokenType : token type for simple line tokeniser */
typedef enum
{
	SIMPLE_TOKEN_SYMBOL,	/* [a-zA-Z_][a-zA-Z0-9_]* */
	SIMPLE_TOKEN_STRING,	/* "[^"]*" */
	SIMPLE_NO_TOKEN
}
SimpleTokenType;

/* simple_parse_token : parse a token from the given line, returns an
	enum value indicating what sort of token was read, also updates
	`*line' and `*token' to point past the end of the current token
	and to the start of the parsed token respectively */
extern SimpleTokenType simple_parse_token (char **line, char **token, WaveErrorContext *context);

#endif
