/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `help.c'
    Help box handling

    $Id: help.c,v 1.19 2003/05/05 20:27:14 bardslea Exp $
*/

#include <gtk/gtk.h>
#include <assert.h>
#include "utils.h"
#include "help.h"
#include "gui_if.h"
#include "support.h"
#include "main.h"

static gboolean helpbox_is_active = FALSE;

/* Add some text to our text widget - this is a callback that is invoked
when our window is realized. We could also force our window to be
realized with gtk_widget_realize, but it would have to be part of
a hierarchy first */

static GtkWidget *text;

static void help_text (const char *str)
{
    gtk_text_insert (GTK_TEXT (text), NULL, &text->style->black, NULL, str, -1);
}

static void help_text_bold (const char *str)
{
    gtk_text_insert (GTK_TEXT (text), NULL,
      &text->style->fg[GTK_STATE_SELECTED], &text->style->bg[GTK_STATE_SELECTED], str, -1);
}

static void help_realize_text (GtkWidget * text, gpointer data)
{
    help_text ("Click on any menu item or button that corresponds to a "
      "menu item for its full description. Pressing a hotkey for a " "menu item is also allowed.");
}

/* Create a scrolled text area that displays a "message" */
static GtkWidget *create_help_text (void)
{

/* Create a table to hold the text widget and scrollbars */
    GtkWidget *table = gtk_table_new (1, 16, FALSE);
	GtkWidget *vscrollbar;

/* Put a text widget in the upper left hand corner. Note the use of
    GTK_SHRINK in the y direction */
    text = gtk_text_new (NULL, NULL);
    gtk_table_attach (GTK_TABLE (table), text, 0, 14, 0, 1,
      GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);
    gtk_widget_set_usize (GTK_WIDGET (text), 100, 50);
    gtk_widget_show (text);

/* And a VScrollbar in the upper right */
    vscrollbar = gtk_vscrollbar_new (GTK_TEXT (text)->vadj);
    gtk_table_attach (GTK_TABLE (table), vscrollbar, 15, 16, 0, 1, GTK_FILL, GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);
    gtk_widget_show (vscrollbar);

/* Add a handler to put a message in the text widget when it is realized */
    gtk_signal_connect (GTK_OBJECT (text), "realize", GTK_SIGNAL_FUNC (help_realize_text), NULL);

    gtk_text_set_word_wrap (GTK_TEXT (text), TRUE);
    return table;
}

static GtkWidget *window = NULL;

static void ok_callback (GtkWidget * widget, GtkWidget * nothing)
{
    helpbox_is_active = FALSE;
    gtk_widget_destroy (window);
}

/* wave_create_help_window : create the help window making menu/button functions report usage
	rather than working normally */
void wave_create_help_window (const char *title, int width, const char *default_text)
{
    GtkWidget *vbox, *hbox, *button1, *label, *separator, *ctext;

    if (helpbox_is_active)
        return;
    helpbox_is_active = TRUE;

    /* create a new nonmodal window */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_usize (GTK_WIDGET (window), width, 400);
    gtk_window_set_title (GTK_WINDOW (window), title);
    gtk_signal_connect (GTK_OBJECT (window), "delete_event", (GtkSignalFunc) ok_callback, NULL);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox);
    gtk_widget_show (vbox);

    label = gtk_label_new (default_text);
    gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);

    ctext = create_help_text ();
    gtk_box_pack_start (GTK_BOX (vbox), ctext, TRUE, TRUE, 0);
    gtk_widget_show (ctext);

    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);

    hbox = gtk_hbox_new (FALSE, 1);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show (hbox);

    button1 = gtk_button_new_with_label ("Close Help");
    gtk_widget_set_usize (button1, 100, -1);
    gtk_signal_connect (GTK_OBJECT (button1), "clicked", GTK_SIGNAL_FUNC (ok_callback), NULL);
    gtk_widget_show (button1);
    gtk_container_add (GTK_CONTAINER (hbox), button1);
    GTK_WIDGET_SET_FLAGS (button1, GTK_CAN_DEFAULT);
    gtk_signal_connect_object (GTK_OBJECT (button1), "realize",
      GTK_SIGNAL_FUNC (gtk_widget_grab_default), GTK_OBJECT (button1));

    gtk_widget_show (window);
}

/* wave_help : report help information if the help window is showing.
	`bold_string' should be the help title and `normal_string' (if not
	empty/NULL should be the explanation, returns TRUE if we're in
	help mode */
gboolean wave_help (const char *bold_string, const char *normal_string)
{
	if (helpbox_is_active)
	{
		help_text_bold ("\n\n");
		help_text_bold (bold_string);
		if (normal_string && *normal_string)
		{
			help_text ("\n");
			help_text (normal_string);
		}
	}
	return helpbox_is_active;
}

/* Callbacks and GUI stuff */
void on_help_menu_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    /* Does nothing for now */
}

void on_help_about_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
	GtkWidget *about_dialog;

    if (wave_help ("Wave Version", "pops up a dialog box describing the current version of this program"))
        return;

	about_dialog = create_about_dialog ();
	gtk_label_set_text (GTK_LABEL (lookup_widget (about_dialog, "about_dialog_title")),
		wave_about_title);
	gtk_label_set_text (GTK_LABEL (lookup_widget (about_dialog, "about_dialog_copyright")),
		wave_copyright_info);
	gtk_widget_show (about_dialog);
}

void on_help_wave_help_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_help ("Wave Help", "pops up this dialog box"))
        return;

    wave_create_help_window ("Wave Help", 300, "Select any main window menu item");
}

