/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2003, The University of Manchester
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `hhh.c'
    HHH file reader

    $Id: hhh.c,v 1.68 2004/05/17 01:32:29 janinl Exp $
*/

#include <ctype.h>
#include "hhh.h"

#include "draw.h"
#include "gui_cb.h"
#include "log.h"
#include "utils.h"
#include "str.h"
#include "value.h"
#include "format/external.h"
#include "tracepane.h"
#include "signalpane.h"

/* FIXME, Only need main.h for wave_polling, remove when this goes away */
#include "main.h"

typedef struct
{
	FileDescLineParser *parser;
	TimeType now;
	int nodes_allocated_size;
	Node **nodes;
	WaveErrorContext *error_context;

	/* Strings stored for use with hhh_string_get */
	char **strings;
	int strings_allocated_size;
	int strings_count;
}
HHHPrivateData;

#define HHH_PRIVATE(reader,var) (((HHHPrivateData *) reader->instance_data)->var)

#define HHH_INIT_TIME (0)

/* fwd decl. */
static TraceDrawer hhhd_trace_drawer;

/* hhh_register_string : remember a new string (makes a copy of the argument's first
	`length' characters (or the whole argument if length == -1)) and returns the
	index of that new string in the HHH_PRIVATE (strings) array */
static int hhh_register_string (WaveReader *reader, const char *string, int length)
{
	char **strings = HHH_PRIVATE (reader, strings);
	int index = HHH_PRIVATE (reader, strings_count);

	if (HHH_PRIVATE (reader, strings_allocated_size) == HHH_PRIVATE (reader, strings_count))
	{
		HHH_PRIVATE (reader, strings_allocated_size) *= 2;
		strings = g_renew (char *, strings, HHH_PRIVATE (reader, strings_allocated_size));
	}

	strings[index] = (length == -1 ? g_strdup (string) : g_strndup (string, length));
	FormatDeescapeString (strings[index]);

	HHH_PRIVATE (reader, strings_count) = index + 1;
	HHH_PRIVATE (reader, strings) = strings;

	return index;
}

/* hhh_string_get : string extraction function, uses HHH_PRIVATE (strings) */
static char *hhh_string_get (void *index_ptr, WaveReader *reader)
{
	int index = (int) index_ptr;
	
	if (index < HHH_PRIVATE (reader, strings_count))
		return HHH_PRIVATE (reader, strings[index]);
	else return "Invalid string";
}

static void hhh_cleanup (WaveReader *reader)
{
	g_free (reader->instance_data);
}

static void hhh_sort_symbols (void)
{
    int i;
    Symbol *curnode = firstnode;

    wave_symbols = g_new (Symbol *, wave_num_symbols);

    for (i = 0; i < wave_num_symbols; i ++)
    {
        char *subst, ch;
        int len;

        wave_symbols[i] = curnode;
        subst = curnode->name;
        len = strlen (subst);
        if (len > longestname)
            longestname = len;
        curnode = curnode->next_in_file;
        do
        {
            ch = *subst;

            if (ch == hier_delimiter)
            {
                *subst = 0x01;  /* not NULL, but lowest valued other char */
            }                   /* forces sort at hier boundaries */
            subst ++;
        }
        while (ch);
    }

    wave_heapsort (wave_symbols, wave_num_symbols);

    for (i = 0; i < wave_num_symbols; i ++)
    {
        char *subst, ch;

        subst = wave_symbols[i]->name;
        while ((ch = (*subst)))
        {
            if (ch == 0x01)
            {
                *subst = hier_delimiter;
            }                   /* restore back to normal */
            subst ++;
        }
    }

    wave_symbols_are_sorted = TRUE;

    init_tree ();
    for (i = 0; i < wave_num_symbols; i ++)
    {
        build_tree_from_name (wave_symbols[i]->name, i);
    }
    treeprune (treeroot);
}

/* hhh_new_node : make a new Node with the given name (the given name
	becomes the property of the Node) */
static Node *hhh_new_node (char *name, WaveReader *reader, int width)
{
    Node *ret = wave_new_node (name, reader,
    	(width == 0 ? NODE_FLAGS_SYNC_HANDSHAKE : NODE_FLAGS_DATA_HANDSHAKE));

    ret->nname = (char *) name;
	ret->drawing_info = new_trace_drawing_info (&hhhd_trace_drawer);

    return ret;
}

/* hhh_null_array : NULL the elements of the given array, note
	the very fine distinction between this and clearing the array
	with zeroes */
static void hhh_null_array (void **ptr_array, int size)
{
    int i;

    for (i = 0; i < size; i ++)
    {
        *ptr_array = NULL;
        ptr_array ++;
    }
}

/* hhh_load_channels : load channel definitions from hhh file. */
static void hhh_load_channels (WaveReader *reader, FileDescLineParser *parser)
{
    Symbol *curnode = NULL;
	char *line, *decl_type_token;
	SimpleTokenType token_type;
	gboolean end_of_decls = FALSE;
	WaveErrorContext *error_context = HHH_PRIVATE (reader, error_context);

	do {
		line = file_desc_line_parser_read_line (parser, FALSE);
		error_context->line_num = parser->line_num;
	
		if (! line)
		{
			wave_log_context (error_context);
			wave_log (TRUE, "invalid end of file in declarations\n");
			exit (EXIT_FAILURE);
		}

		token_type = simple_parse_token (&line, &decl_type_token, error_context);
		if (token_type != SIMPLE_TOKEN_SYMBOL)
		{
			if (token_type != SIMPLE_NO_TOKEN) /* Allow blank lines */
			{
				wave_log_context (error_context);
				wave_log (TRUE, "expecting declaration symbol, skipping line\n");
			}
			continue;
		}

		if (strcmp (decl_type_token, "init-channel") == 0)
		{
			char *channel_num_token, *channel_width_token;
			char *channel_name_token, *channel_format_token;
			int channel_num, channel_width;
			char *channel_name;
			Format channel_format = NULL;
			Node *new_node;
			Symbol *new_symbol;

			if (simple_parse_token (&line, &channel_num_token, error_context) != SIMPLE_TOKEN_SYMBOL ||
				simple_parse_token (&line, &channel_width_token, error_context) != SIMPLE_TOKEN_SYMBOL)
			{
				wave_log_context (error_context);
				wave_log (TRUE, "invalid channel declaration\n");
				continue;
			}
			channel_num = atoi (channel_num_token);
			channel_width = atoi (channel_width_token);

			/* Channel name, formt are optional */
			if (simple_parse_token (&line, &channel_name_token, error_context) == SIMPLE_TOKEN_STRING &&
				*channel_name_token /* Empty string means no name */ )
			{
				char *format_ptr = wave_skip_ws (line);
				channel_name = g_strdup (channel_name_token);

				/* Look for format */
				switch (*format_ptr)
				{
				case '\"':
					if (simple_parse_token (&format_ptr, &channel_format_token, error_context) ==
						SIMPLE_TOKEN_STRING && *channel_format_token /* Empty string means no format */)
						channel_format = FormatExternalToInternal (format_ptr);
					break;
				case '(':
					channel_format = FormatListExternalToInternal (format_ptr, (FormatGetFunction)
						hhh_string_get, NULL, '\"');
					{ /* Set the userData fields in each STRING format element to the `reader' */
						FormatElement *elements = channel_format;

						while (elements->nature != FORMAT_END)
						{
							if (elements->nature == FORMAT_STRING)
								elements->info.string.userData = reader;
							elements ++;
						}
					}
					break;
				}
			} else {
				channel_name = g_strdup_printf ("Unnamed channel %d", channel_num);
			}

			/* Default format is just an unsigned int across the whole word */
			if (! channel_format)
			{
				FormatPosition format_pos;
				channel_format = NewFormat (2);
				format_pos.offset = 0;
				format_pos.bitCount = channel_width;
				NewFormatNumber (channel_format, format_pos);
			}

			new_node = hhh_new_node (channel_name, reader, channel_width);
			new_symbol = symadd (channel_name, hash (channel_name));
			new_node->format = channel_format;
			new_node->trace_format = TRACE_FORMAT_DEC;

			if (channel_width != 0) /* Not a sync channel */
				new_node->drawing_info->recommended_height = 2;

			{ /* Check that we've enough space */
				int allocated_size = HHH_PRIVATE (reader, nodes_allocated_size);
				Node **nodes = HHH_PRIVATE (reader, nodes);

				if (allocated_size < channel_num)
				{
					int new_allocated_size = allocated_size * 2;

					nodes = g_renew (Node *, nodes, new_allocated_size);
					hhh_null_array ((void **) (nodes + allocated_size), allocated_size);
					HHH_PRIVATE (reader, nodes) = nodes;
					HHH_PRIVATE (reader, nodes_allocated_size) = new_allocated_size;
				}
			}

			new_symbol->n = new_node;

			if (!firstnode)
				firstnode = curnode = new_symbol;
			else
			{
				curnode->next_in_file = new_symbol;
				curnode = new_symbol;
			}
			wave_num_symbols ++;
			HHH_PRIVATE (reader, nodes)[channel_num - 1] = new_node;
			/* hhh_max_channel_num = MAX (hhh_max_channel_num, channel_num); */
		} else if (strcmp (decl_type_token, "time") == 0)
        { /* First `event' line must be time = 0 */
			char *time_token;

			if (simple_parse_token (&line, &time_token, error_context) != SIMPLE_TOKEN_SYMBOL)
			{
				wave_log_context (error_context);
				wave_log (TRUE, "invalid time line, skipping\n");
			} else {
				HHH_PRIVATE (reader, now) = wave_str_to_time_type (time_token, NULL);
				end_of_decls = TRUE;
			}
        } else {
			wave_log_context (error_context);
			wave_log (TRUE, "invalid declaration line, skipping\n");
		}
	} while (! end_of_decls);
}

/* hhh_parse_data_strings : parse (string ...) parts of a `dataon' line, create internal strings
	and put the relevant indices into the given FormatData */
static void hhh_parse_data_strings (WaveReader *reader, char *line, FormatData **value)
{
	SimpleTokenType token_type;
	WaveErrorContext *error_context = HHH_PRIVATE (reader, error_context);
	char *token;

	line = wave_skip_ws (line);

	/* Look for "(string <offset> <string>) ..." */
	while (*line == '(')
	{
		line ++;
		token_type = simple_parse_token (&line, &token, error_context);

		if (token_type == SIMPLE_TOKEN_SYMBOL && strcmp (token, "string") == 0)
		{
			/* Parse offset and string */
			char *offset_str_ptr, *string_ptr;

			if (simple_parse_token (&line, &offset_str_ptr, error_context) == SIMPLE_TOKEN_SYMBOL &&
				simple_parse_token (&line, &string_ptr, error_context) == SIMPLE_TOKEN_STRING)
			{
				FormatData *index_format_data = NewFormatData (FormatBitCountToWordCount (64));
				int string_index = hhh_register_string (reader, string_ptr, -1);
				int bit_offset = atoi (offset_str_ptr);

				/* Need to resize value if string pokes out of the end */
				if ((bit_offset + 64) > (FORMAT_BITS_PER_INT * (*value)->wordCount))
				{
					FormatData *new_value = NewFormatData (FormatBitCountToWordCount (bit_offset + 64));

					CopyFormatDataWords (new_value, *value);
					DeleteFormatData (*value);
					*value = new_value;
				}
				index_format_data->words[0] = string_index;
				FormatDataInsertBitField (*value, index_format_data, bit_offset, 64);
				DeleteFormatData (index_format_data);
			}
		} else {
			wave_log_context (error_context);
			wave_log (TRUE, "unrecognised object `%s'\n", token);
			return;
		}

		if (*line != ')')
		{
			wave_log_context (error_context);
			wave_log (TRUE, "invalid string definitions\n");
			return;
		}

		line = wave_skip_ws (line + 1);
	}
}

/* hhh_load_events_body : load events from the HHH file. */
static void hhh_load_events_body (WaveReader *reader, FileDescLineParser *parser)
{
	char *line, *event_type_token;
	SimpleTokenType token_type;
	WaveErrorContext *error_context = HHH_PRIVATE (reader, error_context);

    do
    {
		line = file_desc_line_parser_read_line (parser, FALSE);
		error_context->line_num = parser->line_num;
		if (! line)
			return;

		token_type = simple_parse_token (&line, &event_type_token, error_context);
		if (token_type != SIMPLE_TOKEN_SYMBOL)
		{
			if (token_type != SIMPLE_NO_TOKEN) /* Allow blank lines */
			{
				wave_log_context (error_context);
				wave_log (TRUE, "expecting event symbol, skipping line\n");
			}
			continue;
		}

        if (strcmp (event_type_token, "signal") == 0)
        {
            int channel_num;
			Node *node;
			char *channel_num_token, *event_token;
			char event_char;

			if (simple_parse_token (&line, &channel_num_token, error_context) != SIMPLE_TOKEN_SYMBOL ||
				simple_parse_token (&line, &event_token, error_context) != SIMPLE_TOKEN_SYMBOL)
			{
				wave_log_context (error_context);
				wave_log (TRUE, "invalid signal event, skipping\n");
				continue;
			}
			channel_num = atoi (channel_num_token);
			node = HHH_PRIVATE (reader, nodes)[channel_num - 1]; /* FIXME, check number? */
			event_char = HANDSHAKE_STATE_QUIESCENT;

			if (strcmp (event_token, "requp") == 0)
				event_char = HANDSHAKE_STATE_REQUP;
			else if (strcmp (event_token, "reqdown") == 0)
				event_char = HANDSHAKE_STATE_REQDOWN;
			else if (strcmp (event_token, "ackup") == 0)
				event_char = HANDSHAKE_STATE_ACKUP;
			else if (strcmp (event_token, "ackdown") == 0)
				event_char = HANDSHAKE_STATE_QUIESCENT;
			else {
				wave_log_context (error_context);
				wave_log (FALSE, "unknown event: %s\n", event_token);
				continue;
			}

			/* Is this a data channel */
			if (node->flags & NODE_FLAGS_DATA_HANDSHAKE)
			{
				WaveEventIterator *iterator = node->iterator;
				HandshakeValue *data = g_new (HandshakeValue, 1);
				data->val = event_char;

				/* Find data value from prev. events */
                if (wave_event_iterator_seek (iterator, HHH_PRIVATE (reader, now), WAVE_EVENT_LAST_LE))
				{
					data->data = HANDSHAKE_VALUE (iterator->current->value.vector)->data;
				} else {
					/* FIXME, correct data width */
					data->data = NewFormatDataFromUnsignedInt (1, 0);
				}

				wave_node_insert_event (node, HHH_PRIVATE (reader, now), NODE_FLAGS_DATA_HANDSHAKE, '\0', data);
			} else {
				wave_node_insert_event (node, HHH_PRIVATE (reader, now), NODE_FLAGS_SYNC_HANDSHAKE, event_char, NULL);
			}
        } else if (strcmp (event_type_token, "dataon") == 0 ||
        	strcmp (event_type_token, "data") == 0)
        {
            int channel_num;
			Node *node;
			char *channel_num_token, *event_data_token;
            HandshakeValue *data;
            gboolean hex;
            gboolean isSigned;

			if (simple_parse_token (&line, &channel_num_token, error_context) != SIMPLE_TOKEN_SYMBOL ||
				simple_parse_token (&line, &event_data_token, error_context) != SIMPLE_TOKEN_SYMBOL)
			{
				wave_log_context (error_context);
				wave_log (TRUE, "invalid data event, skipping\n");
				continue;
			}
			channel_num = atoi (channel_num_token);

			node = HHH_PRIVATE (reader, nodes)[channel_num - 1];
            data = g_new (HandshakeValue, 1);

			{
				WaveEventIterator *iterator = node->iterator;

				/* Find state value from prev. events */
                if (wave_event_iterator_seek (iterator, HHH_PRIVATE (reader, now), WAVE_EVENT_LAST_LE))
					data->val = HANDSHAKE_VALUE (iterator->current->value.vector)->val;
				else data->val = HANDSHAKE_STATE_QUIESCENT;

				/* Make upper case to indicate new data arrival */
				data->val = toupper (data->val);
			}

			hex = g_strncasecmp (event_data_token, "0x", 2) == 0;
			if (hex) event_data_token += 2;

			isSigned = *event_data_token == '-';
			if (isSigned) event_data_token ++;

			data->data = FormatDataParseUInt (event_data_token, (hex ? 16 : 10));
			if (isSigned)
			{
				/* FIXME, need better way to retrieve width of channel */
				FormatData *negated = FormatDataNegate (data->data,
					node->format->position.bitCount);
				DeleteFormatData (data->data);
				data->data = negated;
			}

			hhh_parse_data_strings (reader, line, &(data->data));

			wave_node_insert_event (node, HHH_PRIVATE (reader, now), NODE_FLAGS_DATA_HANDSHAKE, '\0', data);
        } else if (strcmp (event_type_token, "time") == 0)
        {
			char *time_token;

			if (simple_parse_token (&line, &time_token, error_context) != SIMPLE_TOKEN_SYMBOL)
			{
				wave_log_context (error_context);
				wave_log (TRUE, "invalid time line, skipping\n");
			} else
				HHH_PRIVATE (reader, now) = wave_str_to_time_type (time_token, NULL);
		} else if (strcmp (event_type_token, "dataoff") == 0)
		{
			/* Ignore me */
		} else if (strcmp (event_type_token, "new-thread") == 0)
		{
			/* Ignore me */
		} else if (strcmp (event_type_token, "message-output") == 0)
		{
			/* Ignore me */
		} else if (strcmp (event_type_token, "end") == 0)
		{
			reader->flags &= ~WAVE_READER_DYNAMIC;

			/* Unpoll the file (if we are polling). */
			if (wave_polling)
			{
				wave_log_context (error_context);
				wave_log (FALSE, "end of file found, unpolling");
				wave_unpoll_file (wave_polling);	
				wave_set_periodically_reread_active (FALSE);
			}
			return;
		} else {
			wave_log_context (error_context);
			wave_log (TRUE, "invalid event line, skipping\n");
        }
    } while (TRUE);
}

static void hhh_seek_end (WaveReader *reader)
{
	TimeType old_end_time = tims.end;
	TimeType old_last_time = tims.last;
	TimeType now = HHH_PRIVATE (reader, now);

	hhh_load_events_body (reader, HHH_PRIVATE (reader, parser));

    tims.last = now;

	if (old_last_time != now)
	{
		GtkAdjustment *hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);

		if (old_end_time >= old_last_time)
		{
			TimeType ntinc;

			ntinc = wave_pane_width_in_time_units (wave_window);

			tims.timecache = tims.last - ntinc + 1;
			if (tims.timecache < tims.first)
				tims.timecache = tims.first;

			hadj->value = tims.timecache;
		}

		fix_wavehadj (wave_window);
		gtk_adjustment_changed (hadj);
		gtk_adjustment_value_changed (hadj);
	}
}

static TimeType hhh_load_declarations (WaveReader *reader)
{
	const char *fname = reader->filename;
    FileDesc *file_desc;
    TimeType max_time;
	FileDescLineParser *parser;

    if (! fname)
    {
        wave_log (TRUE, "HHH Reader: cannot read HHH file from stdin\n");
        exit (EXIT_FAILURE);
    }

    file_desc = file_desc_open (fname, "rt");

    if (! file_desc)
    {
        wave_log (TRUE, "HHH Reader: error opening HHH file '%s'.\n", fname);
        exit (EXIT_FAILURE);
    }

    wave_config->time_dimension = 'n';
	HHH_PRIVATE (reader, error_context)->filename = fname;

    parser = file_desc_line_parser_new (file_desc);
    HHH_PRIVATE (reader, parser) = parser;

    hhh_load_channels (reader, parser);

	if (wave_num_symbols == 0)
	{
        wave_log (TRUE, "HHH Reader: no channels in HHH file\n");
        exit (EXIT_FAILURE);
	}
    hhh_sort_symbols ();
	hhh_load_events_body (reader, parser);
	max_time = HHH_PRIVATE (reader, now);

    return max_time;

    wave_log (TRUE, "HHH Reader: maximum time: " LL_PRINTF_FORMAT " %cs\n", max_time, wave_config->time_dimension);

    /* Compute the starting and ending times for the dump */
    /* min_time = start_time * wave_config->time_scale; */
    /* max_time = end_time * wave_config->time_scale; */
    /* max_time = 1000; */

    wave_set_end_time (max_time);
    wave_set_start_time (0);

    return max_time;
}

Node *hhh_signal_key_find (WaveReader *reader ,const char *key)
{
  int channel_num = atoi (key);

  if (channel_num <= 0 || channel_num > HHH_PRIVATE (reader, nodes_allocated_size))
    return NULL;

  Node *node = HHH_PRIVATE (reader, nodes)[channel_num - 1];

  return node;
}

WaveReader *hhh_init (void)
{
	HHHPrivateData *reader_data = g_new (HHHPrivateData, 1); 
	WaveReader *ret = wave_new_reader ("hhh",
		"Balsa handshake activity trace", WAVE_READER_DYNAMIC, reader_data);
	int allocated_size = 1024;

	reader_data->parser = NULL;
	reader_data->now = 0;
	reader_data->nodes_allocated_size = allocated_size;
    reader_data->nodes = g_new (Node *, allocated_size);
    hhh_null_array ((void **) reader_data->nodes, allocated_size);
	reader_data->error_context = wave_new_error_context ();
	reader_data->error_context->operation_name = "HHH Reader";
	reader_data->error_context->line_num = 1;

	reader_data->strings_allocated_size = 128;
	reader_data->strings = g_new (char *, reader_data->strings_allocated_size);
	reader_data->strings_count = 0;

	wave_reader_add_function (ret, "load-declarations", hhh_load_declarations);
	wave_reader_add_function (ret, "seek-end", hhh_seek_end);
	wave_reader_add_function (ret, "cleanup", hhh_cleanup);
	wave_reader_add_function (ret, "signal_key_find", hhh_signal_key_find);

	hhh_register_string (ret, "NULL", -1);

	return ret;
}

/*** HHHD HHH Drawer routines */

static void hhhd_bit_trace (WaveDrawingContext *context, Node *node, WaveEvent * event, WaveEvent * event2, int whichgc,
  gboolean first_at_this_time, int x0, int y0, int x1, int y1, int x, int ytext)
{
	int height = y0 - y1;
	int req_bottom, req_top, ack_bottom, ack_top, data_top, data_bottom = y0;
    char event_value, event2_value;
	gboolean data_handshake;
	gboolean new_data = FALSE;

	if (!event) return;

	data_handshake = node->flags & NODE_FLAGS_DATA_HANDSHAKE;

	if (data_handshake)
	{
		event_value = HANDSHAKE_VALUE (event->value.vector)->val;
		new_data = isupper (event_value) != 0;
		event_value = tolower (event_value);
	} else {
		event_value = event->value.val;
	}

	if (event2)
	{
		event2_value = tolower ((data_handshake ?
			HANDSHAKE_VALUE (event2->value.vector)->val : event2->value.val));
	} else event2_value = HANDSHAKE_STATE_QUIESCENT;

	if (! data_handshake || height < fontheight)
	{
		int half_height = (height / 2) - 1;	
		req_top = y1;
		req_bottom = y1 + half_height;
		ack_top = y0 - half_height;
		ack_bottom = y0;
		data_top = 0;
	} else {
		int trace_height = fontheight - 6;
		int half_height = (trace_height / 2) - 1;
		req_top = y1;
		req_bottom = y1 + half_height;
		ack_top = y1 + half_height + 2;
		ack_bottom = y1 + trace_height;
		data_top = ack_bottom + 2;
	}

	switch (event_value)
	{
	case HANDSHAKE_STATE_QUIESCENT:
		break;
	case HANDSHAKE_STATE_REQUP:
		WAVE_DRAW_STATE_FILLED_RECTANGLE (context, TRACE_STATE_REQ, x0, req_top, x1, req_bottom);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_top, x1, req_top);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_bottom, x1, req_bottom);
		break;
	case HANDSHAKE_STATE_ACKUP:
		WAVE_DRAW_STATE_FILLED_RECTANGLE (context, TRACE_STATE_ACK, x0, ack_top, x1, ack_bottom);
		WAVE_DRAW_STATE_FILLED_RECTANGLE (context, TRACE_STATE_REQ, x0, req_top, x1, req_bottom);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, ack_top, x1, ack_top);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, ack_bottom, x1, ack_bottom);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_top, x1, req_top);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_bottom, x1, req_bottom);
		break;
	case HANDSHAKE_STATE_REQDOWN:
		WAVE_DRAW_STATE_FILLED_RECTANGLE (context, TRACE_STATE_ACK, x0, ack_top, x1, ack_bottom);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, ack_top, x1, ack_top);
		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, ack_bottom, x1, ack_bottom);
		break;
	default:
		break;
	}

	if (! (first_at_this_time && new_data))
	{
		switch (event_value)
		{
		case HANDSHAKE_STATE_QUIESCENT:
			break;
		case HANDSHAKE_STATE_REQUP:
			WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_top, x0, req_bottom);
			break;
		case HANDSHAKE_STATE_ACKUP:
			WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, ack_bottom, x0, ack_top);
			break;
		case HANDSHAKE_STATE_REQDOWN:
			WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, req_top, x0, req_bottom);
			WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x1, ack_bottom, x1, ack_top);
			break;
		default:
			break;
		}
	}

	/* Draw data */
	if (new_data && data_top)
	{
		char *data_string;
		int mid_data_y = data_top + (data_bottom - data_top) / 2;
		int width = 10000; /* Arbitrarily wide */
		int string_y = mid_data_y + (wavefont->ascent / 2);

		{ /* Must be currently using this node's iterator */
			WaveEventIterator *iterator = node->iterator;
			WaveEventPosition old_position = wave_event_iterator_get_position (iterator);
			TimeType event_time = event->time; /* Note that event will change until we wind back to
				old_position */

			while (wave_event_iterator_step_forward (iterator)) 
			{
				if (isupper (HANDSHAKE_VALUE (iterator->current->value.vector)->val))
				{
					TimeType time_difference = iterator->current->time - event_time;

					width = time_difference * wave_window->traces.scale.pixels_per_time_unit - 2;
					break;
				}
			}
			wave_event_iterator_set_position (iterator, old_position);
			wave_event_position_free (old_position);
		}

		data_string = wave_event_handshake_string (context->trace, event, TRUE);

		WAVE_DRAW_STATE_LINE (context, TRACE_STATE_HBOX, x0, data_top + 2, x0, data_bottom - 2);
        // WAVE_DRAW_STATE_STRING (context, TRACE_STATE_FONT, WAVE_FONT_WAVE, x0 + 4, string_y, data_string);

		wave_draw_truncated_string (context, x0 + 4, string_y, 0, width, data_string);

        g_free (data_string);
	}
}

static gboolean hhhd_compat (TraceEnt * t)
{
    /* Need something sensible here */
    return TRUE;
}

static TraceDrawer hhhd_trace_drawer = {
    "default",
    hhhd_compat,
    NULL,
    hhhd_bit_trace,
    NULL, /* spike */
    NULL, NULL, /* value */
    NULL
};
