/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `log.c'
    Log window control

    $Id: log.c,v 1.4 2003/07/16 11:24:04 bardslea Exp $
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "log.h"
#include "gui_if.h"
#include "support.h"
#include "main.h"
#include "help.h"

WaveLogLevel wave_log_level = WAVE_LOG_ERROR;

static GtkWidget *log_window;
static int log_location = 0;
static char *log_accum = NULL;
static int wave_exit_value = 0;

/* wave_init_log_window : create the log window (if necessary) and raise/show it
	iff `show' is TRUE */
void wave_init_log_window (gboolean show)
{
    if (log_window == NULL)
    {
        log_window = create_log_window ();
        gtk_text_set_word_wrap (GTK_TEXT (lookup_widget (log_window, "log_text")), TRUE);
        log_location = GTK_EDITABLE (lookup_widget (log_window, "log_text"))->current_pos;
    }
    /* Deal with any saved log messages */
    if (log_accum)
    {
        gtk_editable_insert_text (GTK_EDITABLE (lookup_widget (log_window, "log_text")), log_accum,
        	strlen (log_accum), &log_location);
        g_free (log_accum);
        log_accum = NULL;
    }
    if (show)
    {
        gtk_widget_show (log_window);
        /* Gtk 1.2 has no other way to do this.  :^( */
        gdk_window_raise (log_window->window);
    }
}

/* wave_log_context : print an error context header for a subsequent wave_log call */
void wave_log_context (WaveErrorContext *context)
{
	if (context)
	{
		wave_log (FALSE, "%s:%s:%d: ", (context->operation_name ? context->operation_name : ""),
			(context->filename ? context->filename : ""), context->line_num);
	}
}

/* wave_log : error logging to window or stderr */
void G_GNUC_PRINTF (2, 3) wave_log (gboolean is_error, const char *fmt, ...)
{
    va_list args;
    char *str;
    GtkWidget *log_text;

    /*
     * No main window yet; send to stdout/stderr instead.
     */
    if (! wave_have_main_window && is_error)
    {
        va_start (args, fmt);
        vfprintf (stderr, fmt, args);
        va_end (args);
        return;
    }

    va_start (args, fmt);
    str = g_strdup_vprintf (fmt, args);
    va_end (args);

    if (! wave_have_main_window)
    {
        /* No main window yet, so save for later */
        if (log_accum)
        {
            int len1 = strlen (log_accum);
            int len2 = strlen (str);

            log_accum = g_realloc (log_accum, len1 + len2 + 1);
            memcpy (log_accum + len1, str, len2 + 1);
            g_free (str);
        } else
        {
            log_accum = str;
        }
        return;
    }

    wave_init_log_window (is_error || wave_log_level == WAVE_LOG_ALWAYS);
    log_text = GTK_WIDGET (lookup_widget (log_window, "log_text"));
    gtk_editable_insert_text (GTK_EDITABLE (log_text), str, strlen (str), &log_location);
    g_free (str);

    if (is_error)
    {
        gdk_beep ();
    }
}

/* wave_exit : exit GTKWave on a fatal error giving a handy exit-delaying
	dialogue box if we're graphical yet */
void wave_exit (int return_val)
{
	wave_exit_value = return_val;

    if (wave_have_main_window)
	{
		GtkWidget *exit_dialog = create_exit_dialog ();

		wave_init_log_window (TRUE);
		gtk_widget_show (exit_dialog);
	} else
		on_exit_gtkwave_exit_clicked (NULL, NULL);
}

void on_open_log_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_help ("Open Log", "opens the log window (which normally only happens when an error occurs)"))
        return;

    wave_init_log_window (TRUE);
}

gboolean on_log_window_delete_event (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
    gtk_widget_hide (log_window);
    return TRUE;
}

void on_close_log_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_hide (log_window);
}

gboolean on_log_window_destroyed (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
    log_window = NULL;
    return FALSE;
}

void on_exit_gtkwave_exit_clicked (GtkButton *button, gpointer user_data)
{
	exit (wave_exit_value);
}

