/*
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `lxt.c' LXT-format reader
    
    $Id: lxt.c,v 1.70 2003/11/09 22:21:34 bardslea Exp $
 */

#include "lxt.h"

#include <stdio.h>

#ifndef _MSC_VER
#include <unistd.h>
#include <sys/mman.h>
#else
#include <windows.h>
#include <io.h>
#define dup _dup;
#endif

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "utils.h"
#include "bsearch.h"
#include "value.h"
#include "log.h"

#define LT_SECTION_END                          (0)
#define LT_SECTION_CHG                          (1)
#define LT_SECTION_SYNC_TABLE                   (2)
#define LT_SECTION_FACNAME                      (3)
#define LT_SECTION_FACNAME_GEOMETRY             (4)
#define LT_SECTION_TIMESCALE                    (5)
#define LT_SECTION_TIME_TABLE                   (6)
#define LT_SECTION_INITIAL_VALUE                (7)
#define LT_SECTION_DOUBLE_TEST                  (8)
#define LT_SECTION_TIME_TABLE64                 (9)
#define LT_SECTION_ZFACNAME_PREDEC_SIZE         (10)
#define LT_SECTION_ZFACNAME_SIZE                (11)
#define LT_SECTION_ZFACNAME_GEOMETRY_SIZE       (12)
#define LT_SECTION_ZSYNC_SIZE                   (13)
#define LT_SECTION_ZTIME_TABLE_SIZE             (14)
#define LT_SECTION_ZCHG_PREDEC_SIZE             (15)
#define LT_SECTION_ZCHG_SIZE                    (16)
#define LT_SECTION_ZDICTIONARY                  (17)
#define LT_SECTION_ZDICTIONARY_SIZE             (18)
#define LT_SECTION_EXCLUDE_TABLE                (19)

#define LT_SYM_F_BITS           (0)
#define LT_SYM_F_INTEGER        (1<<0)
#define LT_SYM_F_DOUBLE         (1<<1)
#define LT_SYM_F_STRING         (1<<2)
#define LT_SYM_F_ALIAS          (1<<3)

#define LT_HDRID (0x0138)
#define LT_VERSION (0x0004)
#define LT_TRLID (0xB4)

/****************************************************************************/

#define LXTHDR "LXTLOAD | "
#define DBP(params) WAVE_DEBUG(printf params)
#define LT_SYM_CONCRETE (LT_SYM_F_INTEGER|LT_SYM_F_DOUBLE|LT_SYM_F_STRING)

/*
 * globals 
 */

struct fac
{
    int array_height, msb, lsb, len;
    char *name;
    unsigned int lastchange, flags;
    Node *resolve_lxt_alias_to;
    char bus;
};

static void *mm, *mmcache;
static int version;

static struct fac *mvlfacs = NULL;
static int first_cycle, last_cycle, total_cycles;
static int maxchange = 0, maxindex = 0;
static TimeType lxt_scaled_start_time = -1;
static TimeType lxt_scaled_end_time = -1;

static off_t f_len;

static int *positional_information = NULL;
static TimeType *time_information = NULL;

static unsigned int change_field_offset = 0;
static unsigned int facname_offset = 0;
static unsigned int facgeometry_offset = 0;
static unsigned int time_table_offset = 0;
static unsigned int time_table_offset64 = 0;
static unsigned int sync_table_offset = 0;
static unsigned int initial_value_offset = 0;
static unsigned int timescale_offset = 0;
static unsigned int double_test_offset = 0;
static unsigned int zdictionary_offset = 0;

static unsigned int zfacname_predec_size = 0;
static unsigned int zfacname_size = 0;
static unsigned int zfacgeometry_size = 0;
static unsigned int zsync_table_size = 0;
static unsigned int ztime_table_size = 0;
static unsigned int zchg_predec_size = 0;
static unsigned int zchg_size = 0;
static unsigned int zdictionary_predec_size = 0;

static unsigned char initial_value = '1';

static unsigned int dict_num_entries;
static unsigned int dict_string_mem_required;
static unsigned int dict_16_offset;
static unsigned int dict_24_offset;
static unsigned int dict_32_offset;
static unsigned int dict_width;
static char **dict_string_mem_array = NULL;

static unsigned int exclude_offset = 0;

static char *lt_buf = NULL;     /* max memory ever needed for parse_offset() */
static int lt_len = 0;

static gboolean lxt_compress_clock = FALSE;

#ifndef _MSC_VER
static int fd;                  /* XXX : win32 requires decompression tempfiles since */

                                /*
                                 * its mmap() variant doesn't use file descriptors 
                                 */
#else
static HANDLE hIn, hInMap;
#endif

/****************************************************************************/

/*
 * functions which emit various big endian
 * data to a file
 */
static int fpos;

static int lt_emit_u8 (FILE * handle, int value)
{
    unsigned char buf[1];
    int nmemb;

    buf[0] = value & 0xff;
    nmemb = fwrite (buf, sizeof (char), 1, handle);
    fpos += nmemb;
    return (nmemb);
}

static int lt_emit_u16 (FILE * handle, int value)
{
    unsigned char buf[2];
    int nmemb;

    buf[0] = (value >> 8) & 0xff;
    buf[1] = value & 0xff;
    nmemb = fwrite (buf, sizeof (char), 2, handle);
    fpos += nmemb;
    return (nmemb);
}

static int lt_emit_u24 (FILE * handle, int value)
{
    unsigned char buf[3];
    int nmemb;

    buf[0] = (value >> 16) & 0xff;
    buf[1] = (value >> 8) & 0xff;
    buf[2] = value & 0xff;
    nmemb = fwrite (buf, sizeof (char), 3, handle);
    fpos += nmemb;
    return (nmemb);
}

static int lt_emit_u32 (FILE * handle, int value)
{
    unsigned char buf[4];
    int nmemb;

    buf[0] = (value >> 24) & 0xff;
    buf[1] = (value >> 16) & 0xff;
    buf[2] = (value >> 8) & 0xff;
    buf[3] = value & 0xff;
    nmemb = fwrite (buf, sizeof (char), 4, handle);
    fpos += nmemb;
    return (nmemb);
}

/****************************************************************************/

#if 0
/* WAVE_BE32_UNALIGNED_OK */
/*
 * reconstruct 8/16/24/32 bits out of the aet's representation
 * of a big-endian integer.  this is for 32-bit PPC so no byte
 * swizzling needs to be done at all.  for 24-bit ints, we have no danger of
 * running off the end of memory so just grab an int and rsh it by 8...
 */

#define get_byte(offset) 	((unsigned int)(*((unsigned char *)mm+(offset))))
#define get_16(offset) 		((unsigned int)(*((unsigned short *)(((unsigned char *)mm)+(offset)))))
#define get_24(offset)		((*(unsigned int *)(((unsigned char *)mm)+(offset)))>>8)
#define get_32(offset)		(*(unsigned int *)(((unsigned char *)mm)+(offset)))
#define get_64(offset)		((((UTimeType)get_32(offset))<<32)|((UTimeType)get_32((offset)+4)))

#else

/*
 * reconstruct 8/16/24/32 bits out of the aet's representation
 * of a big-endian integer.  this should work on all architectures.
 */
static unsigned int get_byte (int offset)
{
    return *((unsigned char *) mm + offset);
}

static unsigned int get_16 (int offset)
{
    unsigned char *nn = (unsigned char *) mm + offset;
    unsigned int m1 = *((unsigned char *) (nn++));
    unsigned int m2 = *((unsigned char *) nn);

    return ((m1 << 8) | m2);
}

static unsigned int get_24 (int offset)
{
    unsigned char *nn = (unsigned char *) mm + offset;
    unsigned int m1 = *((unsigned char *) (nn++));
    unsigned int m2 = *((unsigned char *) (nn++));
    unsigned int m3 = *((unsigned char *) nn);

    return ((m1 << 16) | (m2 << 8) | m3);
}

static unsigned int get_32 (int offset)
{
    unsigned char *nn = (unsigned char *) mm + offset;
    unsigned int m1 = *((unsigned char *) (nn++));
    unsigned int m2 = *((unsigned char *) (nn++));
    unsigned int m3 = *((unsigned char *) (nn++));
    unsigned int m4 = *((unsigned char *) nn);

    return ((m1 << 24) | (m2 << 16) | (m3 << 8) | m4);
}

static UTimeType get_64 (int offset)
{
    return ((((UTimeType) get_32 (offset)) << 32) | ((UTimeType) get_32 (offset + 4)));
}

#endif

/****************************************************************************/

static char double_mask[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
static char double_is_native = 0;

static void create_double_endian_mask (int offset)
{
    static double p = 3.14159;
    double d;
    int i, j;

    d = *((double *) ((unsigned char *) mm + offset));
    if (p == d)
    {
        double_is_native = 1;
    } else
    {
        char *remote, *here;

        remote = (char *) &d;
        here = (char *) &p;
        for (i = 0; i < 8; i++)
        {
            for (j = 0; j < 8; j++)
            {
                if (here[i] == remote[j])
                {
                    double_mask[i] = j;
                    break;
                }
            }
        }
    }
}

static char *swab_double_via_mask (int offset)
{
    char swapbuf[8];
    char *pnt = g_new (char, 8);
    int i;

    memcpy (swapbuf, (unsigned char *) (mm + offset), 8);
    for (i = 0; i < 8; i++)
    {
        pnt[i] = swapbuf[(unsigned char) double_mask[i]];
    }

    return pnt;
}

/****************************************************************************/

/*
 * convert 0123 to an mvl character representation 
 */
static unsigned char convert_mvl (int value)
{
    return "01zx1xx0xxxxxxxx"[value & 15];
}

/*
 * given an offset into the aet, calculate the "time" of that offset in simulation.  this func similar to how gtkwave
 * can find the most recent valuechange that corresponds with an arbitrary timevalue. 
 */
static int max_compare_time_tc, max_compare_pos_tc;
static int compar_mvl_timechain (const void *s1, const void *s2)
{
    int key = *((int *) s1), obj = *((int *) s2), delta;

    if ((obj <= key) && (obj > max_compare_time_tc))
    {
        max_compare_time_tc = obj;
        max_compare_pos_tc = (int *) s2 - positional_information;
    }

    delta = key - obj;
    if (delta < 0)
        return -1;
    else if (delta > 0)
        return 1;
    else
        return 0;
}

static TimeType bsearch_mvl_timechain (int key)
{
    max_compare_time_tc = -1;
    max_compare_pos_tc = -1;

    bsearch ((void *) &key, (void *) positional_information, total_cycles, sizeof (int), compar_mvl_timechain);
    if ((max_compare_pos_tc <= 0) || (max_compare_time_tc < 0))
    {
        max_compare_pos_tc = 0; /* aix bsearch fix */
    }

    return (time_information[max_compare_pos_tc]);
}

/**************************************************************************************************/

/*
 * build up decompression dictionary for MVL2 wave_symbols >16 bits ...
 */
static void build_dict (void)
{
#ifdef _MSC_VER
    FILE *tmp;
    unsigned char *t;
#endif
    gzFile zhandle;
    int offs = zdictionary_offset + 24;
    int total_mem, rc, i;
    char *decmem = NULL;
    char *pnt;

    dict_num_entries = get_32 (zdictionary_offset + 0);
    dict_string_mem_required = get_32 (zdictionary_offset + 4);
    dict_16_offset = get_32 (zdictionary_offset + 8);
    dict_24_offset = get_32 (zdictionary_offset + 12);
    dict_32_offset = get_32 (zdictionary_offset + 16);
    dict_width = get_32 (zdictionary_offset + 20);

    DBP ((LXTHDR "zdictionary_offset = %08x\n", zdictionary_offset));
    DBP ((LXTHDR "zdictionary_predec_size = %08x\n\n", zdictionary_predec_size));
    DBP ((LXTHDR "dict_num_entries = %d\n", dict_num_entries));
    DBP ((LXTHDR "dict_string_mem_required = %d\n", dict_string_mem_required));
    DBP ((LXTHDR "dict_16_offset = %d\n", dict_16_offset));
    DBP ((LXTHDR "dict_24_offset = %d\n", dict_24_offset));
    DBP ((LXTHDR "dict_32_offset = %d\n", dict_32_offset));

    printf (LXTHDR "Dictionary compressed MVL2 change records detected...\n");

#ifdef _MSC_VER
    t = ((char *) mm) + offs;
    tmp = tmpfile ();
    if (!tmp)
    {
        fprintf (stderr, LXTHDR "could not open decompression tempfile!\n");
        exit (255);
    }

    fwrite (t, zdictionary_predec_size, 1, tmp);
    fseek (tmp, 0, SEEK_SET);

    zhandle = gzdopen (dup (fileno (tmp)), "rb");
#else
    if (offs != lseek (fd, offs, SEEK_SET))
    {
        fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
        exit (255);
    }
    zhandle = gzdopen (dup (fd), "rb");
#endif
    total_mem = dict_string_mem_required;
    decmem = g_new (char, total_mem);

    rc = gzread (zhandle, decmem, total_mem);
    DBP ((LXTHDR "section offs for name decompression = %08x of len %d\n", offs, dict_string_mem_required));
    DBP (("Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
    if (rc != total_mem)
    {
        fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
        exit (255);
    }

    dict_string_mem_array = g_new (char *, dict_num_entries);
    pnt = decmem;
    for (i = 0; i < dict_num_entries; i++)
    {
        dict_string_mem_array[i] = pnt;
        pnt += (strlen (pnt) + 1);
        DBP ((LXTHDR "Dict %d: '1%s'\n", i, dict_string_mem_array[i]));
    }

    gzclose (zhandle);
#ifdef _MSC_VER
    fclose (tmp);
#endif

    printf (LXTHDR "...expanded %d entries from %08x into %08x bytes.\n", dict_num_entries, zdictionary_predec_size,
      dict_string_mem_required);
}

/*
 * decompress facility names and build up fac geometry
 */
static void build_facs (const char *fname)
{
    char *buf, *bufprev, *bufcurr;
    int offs = facname_offset + 8;
    int i, j, clone;
    char *pnt;
    int total_mem = get_32 (facname_offset + 4);
    gzFile zhandle = NULL;
    char *decmem = NULL;

#ifdef _MSC_VER
    FILE *tmp;
#endif

    buf = g_new (char, total_mem);
    pnt = bufprev = buf;

    if (zfacname_size)
    {
        int rc;

#ifdef _MSC_VER
        unsigned char *t = ((char *) mm) + offs;

        tmp = tmpfile ();
        if (!tmp)
        {
            fprintf (stderr, LXTHDR "could not open decompression tempfile!\n");
            exit (255);
        }

        fwrite (t, zfacname_size, 1, tmp);
        fseek (tmp, 0, SEEK_SET);

        zhandle = gzdopen (dup (fileno (tmp)), "rb");
#else
        if (offs != lseek (fd, offs, SEEK_SET))
        {
            fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
            exit (255);
        }
        zhandle = gzdopen (dup (fd), "rb");
#endif
        mmcache = mm;
        total_mem = zfacname_predec_size;
        decmem = g_new (char, total_mem);
        mm = decmem;

        rc = gzread (zhandle, decmem, total_mem);
        DBP ((LXTHDR "section offs for name decompression = %08x of len %d\n", offs, zfacname_size));
        DBP ((LXTHDR "Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
        if (rc != total_mem)
        {
            fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
            exit (255);
        }

        offs = 0;               /* we're in our new memory region now.. */
    }

    printf (LXTHDR "Building %d facilities.\n", wave_num_symbols);
    for (i = 0; i < wave_num_symbols; i++)
    {
        clone = get_16 (offs);
        offs += 2;
        bufcurr = pnt;
        for (j = 0; j < clone; j++)
        {
            *(pnt++) = *(bufprev++);
        }
        while ((*(pnt++) = get_byte (offs++)));
        mvlfacs[i].name = bufcurr;
        DBP ((LXTHDR "Encountered facility %d: '%s'\n", i, bufcurr));
        bufprev = bufcurr;
    }

    if (zfacname_size)
    {
        mm = mmcache;
        g_free (decmem);
        decmem = NULL;
        gzclose (zhandle);
#ifdef _MSC_VER
        fclose (tmp);
#endif
    }

    if (!facgeometry_offset)
    {
        fprintf (stderr, "LXT '%s' is missing a facility geometry section, exiting.\n", fname);
        exit (255);
    }

    offs = facgeometry_offset;

    if (zfacgeometry_size)
    {
        int rc;

#ifdef _MSC_VER
        unsigned char *t = ((char *) mm) + offs;

        tmp = tmpfile ();
        if (!tmp)
        {
            fprintf (stderr, LXTHDR "could not open decompression tempfile!\n");
            exit (255);
        }

        fwrite (t, zfacgeometry_size, 1, tmp);
        fseek (tmp, 0, SEEK_SET);

        zhandle = gzdopen (dup (fileno (tmp)), "rb");
#else
        if (offs != lseek (fd, offs, SEEK_SET))
        {
            fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
            exit (255);
        }
        zhandle = gzdopen (dup (fd), "rb");
#endif
        mmcache = mm;
        total_mem = wave_num_symbols * 16;
        decmem = g_new (char, total_mem);
        mm = decmem;

        rc = gzread (zhandle, decmem, total_mem);
        DBP ((LXTHDR "section offs for facgeometry decompression = %08x of len %d\n", offs, zfacgeometry_size));
        DBP ((LXTHDR "Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
        if (rc != total_mem)
        {
            fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
            exit (255);
        }

        offs = 0;               /* we're in our new memory region now.. */
    }

    for (i = 0; i < wave_num_symbols; i++)
    {
        mvlfacs[i].array_height = get_32 (offs);
        mvlfacs[i].msb = get_32 (offs + 4);
        mvlfacs[i].lsb = get_32 (offs + 8);
        mvlfacs[i].flags = get_32 (offs + 12);
        mvlfacs[i].len =
          (mvlfacs[i].lsb >
          mvlfacs[i].msb) ? (mvlfacs[i].lsb - mvlfacs[i].msb + 1) : (mvlfacs[i].msb - mvlfacs[i].lsb + 1);

        if (mvlfacs[i].len > lt_len)
            lt_len = mvlfacs[i].len;
        DBP ((LXTHDR "%s[%d:%d]\n", mvlfacs[i].name, mvlfacs[i].msb, mvlfacs[i].lsb));

        offs += 0x10;
    }

    lt_buf = g_new (char, (lt_len ? lt_len : 1));

    if (zfacgeometry_size)
    {
        mm = mmcache;
        g_free (decmem);
        decmem = NULL;
        gzclose (zhandle);
#ifdef _MSC_VER
        fclose (tmp);
#endif
    }
}

/*
 * build up the lastchange entries so we can start to walk
 * through the aet..
 */
static void build_facs2 (const char *fname)
{
    int i;
    int offs;
    int chg;
    int maxchg = 0, maxindx = 0;
    int last_position;
    TimeType last_time;
    char *decmem = NULL;
    int total_mem;
    gzFile zhandle = NULL;

#ifdef _MSC_VER
    FILE *tmp;
#endif

    if ((!time_table_offset) && (!time_table_offset64))
    {
        fprintf (stderr, "LXT '%s' is missing a time table section, exiting.\n", fname);
        exit (255);
    }

    if ((time_table_offset) && (time_table_offset64))
    {
        fprintf (stderr, "LXT '%s' has both 32 and 64-bit time table sections, exiting.\n", fname);
        exit (255);
    }

    if (time_table_offset)
    {
        offs = time_table_offset;

        DBP ((LXTHDR "Time table position: %08x\n", time_table_offset + 12));
        total_cycles = get_32 (offs + 0);
        DBP ((LXTHDR "Total cycles: %d\n", total_cycles));

        if (ztime_table_size)
        {
            int rc;

#ifdef _MSC_VER
            unsigned char *t = ((char *) mm) + offs + 4;

            tmp = tmpfile ();
            if (!tmp)
            {
                fprintf (stderr, LXTHDR "could not open decompression tempfile!\n");
                exit (255);
            }

            fwrite (t, ztime_table_size, 1, tmp);
            fseek (tmp, 0, SEEK_SET);

            zhandle = gzdopen (dup (fileno (tmp)), "rb");
#else
            if ((offs + 4) != lseek (fd, offs + 4, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
                exit (255);
            }
            zhandle = gzdopen (dup (fd), "rb");
#endif
            mmcache = mm;
            total_mem = 4 + 4 + (total_cycles * 4) + (total_cycles * 4);
            decmem = g_new (char, total_mem);
            mm = decmem;

            rc = gzread (zhandle, decmem, total_mem);
            DBP ((LXTHDR "section offs for timetable decompression = %08x of len %d\n", offs, ztime_table_size));
            DBP ((LXTHDR "Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
            if (rc != total_mem)
            {
                fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
                exit (255);
            }

            offs = 0;           /* we're in our new memory region now.. */
        } else
        {
            offs += 4;          /* skip past count to make consistent view between compressed/uncompressed */
        }

        first_cycle = get_32 (offs);
        DBP ((LXTHDR "First cycle: %d\n", first_cycle));
        last_cycle = get_32 (offs + 4);
        DBP ((LXTHDR "Last cycle: %d\n", last_cycle));
        DBP ((LXTHDR "Total cycles (actual): %d\n", last_cycle - first_cycle + 1));

        /*
         * rebuild time table from its deltas... 
         */

        positional_information = g_new (int, total_cycles);
        last_position = 0;
        offs += 8;
        for (i = 0; i < total_cycles; i++)
        {
            last_position = positional_information[i] = get_32 (offs) + last_position;
            offs += 4;
        }
        time_information = g_new (TimeType, total_cycles);
        last_time = LL_LITERAL (0);
        for (i = 0; i < total_cycles; i++)
        {
            last_time = time_information[i] = ((TimeType) get_32 (offs)) + last_time;
            time_information[i] *= wave_config->time_scale;
            offs += 4;
        }

        if (ztime_table_size)
        {
            mm = mmcache;
            g_free (decmem);
            decmem = NULL;
            gzclose (zhandle);
#ifdef _MSC_VER
            fclose (tmp);
#endif
        }
    } else
    {                           /* 64-bit read */

        offs = time_table_offset64;

        DBP ((LXTHDR "Time table position: %08x\n", time_table_offset64 + 20));

        total_cycles = (TimeType) ((unsigned int) get_32 (offs + 0));
        DBP ((LXTHDR "Total cycles: %d\n", total_cycles));

        if (ztime_table_size)
        {
            int rc;

#ifdef _MSC_VER
            unsigned char *t = ((char *) mm) + offs + 4;

            tmp = tmpfile ();
            if (!tmp)
            {
                fprintf (stderr, LXTHDR "could not open decompression tempfile!\n");
                exit (255);
            }

            fwrite (t, ztime_table_size, 1, tmp);
            fseek (tmp, 0, SEEK_SET);

            zhandle = gzdopen (dup (fileno (tmp)), "rb");
#else
            if ((offs + 4) != lseek (fd, offs + 4, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
                exit (255);
            }
            zhandle = gzdopen (dup (fd), "rb");
#endif
            mmcache = mm;
            total_mem = 8 + 8 + (total_cycles * 4) + (total_cycles * 8);
            decmem = g_new (char, total_mem);
            mm = decmem;

            rc = gzread (zhandle, decmem, total_mem);
            DBP ((LXTHDR "section offs for timetable decompression = %08x of len %d\n", offs, ztime_table_size));
            DBP ((LXTHDR "Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
            if (rc != total_mem)
            {
                fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
                exit (255);
            }

            offs = 0;           /* we're in our new memory region now.. */
        } else
        {
            offs += 4;          /* skip past count to make consistent view between compressed/uncompressed */
        }

        first_cycle = get_64 (offs);
        DBP ((LXTHDR "First cycle: %d\n", first_cycle));
        last_cycle = get_64 (offs + 8);
        DBP ((LXTHDR "Last cycle: %d\n", last_cycle));
        DBP ((LXTHDR "Total cycles (actual): %Ld\n", last_cycle - first_cycle + 1));

        /*
         * rebuild time table from its deltas... 
         */

        positional_information = g_new (int, total_cycles);
        last_position = 0;
        offs += 16;
        for (i = 0; i < total_cycles; i++)
        {
            last_position = positional_information[i] = get_32 (offs) + last_position;
            offs += 4;
        }
        time_information = g_new (TimeType, total_cycles);
        last_time = LL_LITERAL (0);
        for (i = 0; i < total_cycles; i++)
        {
            last_time = time_information[i] = ((TimeType) get_64 (offs)) + last_time;
            time_information[i] *= wave_config->time_scale;
            offs += 8;
        }

        if (ztime_table_size)
        {
            mm = mmcache;
            g_free (decmem);
            decmem = NULL;
            gzclose (zhandle);
#ifdef _MSC_VER
            fclose (tmp);
#endif
        }
    }

    if (sync_table_offset)
    {
        offs = sync_table_offset;

        if (zsync_table_size)
        {
            int rc;

            if (offs != lseek (fd, offs, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", offs);
                exit (255);
            }
            zhandle = gzdopen (dup (fd), "rb");

            mmcache = mm;
            total_mem = wave_num_symbols * 4;
            decmem = g_new (char, total_mem);
            mm = decmem;

            rc = gzread (zhandle, decmem, total_mem);
            DBP ((LXTHDR "section offs for synctable decompression = %08x of len %d\n", offs, zsync_table_size));
            DBP ((LXTHDR "Decompressed size is %d bytes (vs %d)\n", rc, total_mem));
            if (rc != total_mem)
            {
                fprintf (stderr, LXTHDR "decompression size disparity  %d bytes (vs %d)\n", rc, total_mem);
                exit (255);
            }

            offs = 0;           /* we're in our new memory region now.. */
        }

        for (i = 0; i < wave_num_symbols; i++)
        {
            chg = get_32 (offs);
            offs += 4;
            if (chg > maxchg)
            {
                maxchg = chg;
                maxindx = i;
            }
            mvlfacs[i].lastchange = chg;
            DBP ((LXTHDR "Changes: %d '%s' %08x\n", i, mvlfacs[i].name, chg));
        }

        if (zsync_table_size)
        {
            mm = mmcache;
            g_free (decmem);
            decmem = NULL;
            gzclose (zhandle);
        }

        DBP ((LXTHDR "Maxchange at: %08x for symbol '%s' of len %d\n", maxchg, mvlfacs[maxindx].name,
            mvlfacs[maxindx].len));

        maxchange = maxchg;
        maxindex = maxindx;
    }

    if (zchg_size)
    {
        if (zchg_predec_size > 128 * 1024 * 1024)
        {
            char *nam = tmpnam (NULL);
            FILE *tmp = fopen (nam, "wb");
            unsigned int len = zchg_predec_size;
            int rc;
            char buf[32768];
            int fd2 = open (nam, O_RDONLY);
            char testbyte[2] = { 0, 0 };
            char is_bz2;

            unlink (nam);

            printf (LXTHDR "Compressed change records detected, making tempfile...\n");
            if (change_field_offset != lseek (fd, change_field_offset, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", change_field_offset);
                exit (255);
            }

            read (fd, &testbyte, 2);
            is_bz2 = (testbyte[0] == 'B') && (testbyte[1] == 'Z');

            if (change_field_offset != lseek (fd, change_field_offset, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", change_field_offset);
                exit (255);
            }

            if (is_bz2)
            {
#ifndef HAVE_LIBBZ2
                fprintf (stderr, LXTHDR "bzip2 compression not available");
                exit (255);
#else
                zhandle = BZ2_bzdopen (dup (fd), "rb");

                while (len)
                {
                    int siz = (len > 32768) ? 32768 : len;

                    rc = BZ2_bzread (zhandle, buf, siz);
                    if (rc != siz)
                    {
                        fprintf (stderr, LXTHDR "gzread error to tempfile %d (act) vs %d (exp), exiting.\n", rc, siz);
                        exit (255);
                    }
                    if (1 != fwrite (buf, siz, 1, tmp))
                    {
                        fprintf (stderr, LXTHDR "fwrite error to tempfile, exiting.\n");
                        exit (255);
                    };
                    len -= siz;
                }

                printf (LXTHDR "...expanded %08x into %08x bytes.\n", zchg_size, zchg_predec_size);
                BZ2_bzclose (zhandle);
#endif
            } else
            {
                zhandle = gzdopen (dup (fd), "rb");

                while (len)
                {
                    int siz = (len > 32768) ? 32768 : len;

                    rc = gzread (zhandle, buf, siz);
                    if (rc != siz)
                    {
                        fprintf (stderr, LXTHDR "gzread error to tempfile %d (act) vs %d (exp), exiting.\n", rc, siz);
                        exit (255);
                    }
                    if (1 != fwrite (buf, siz, 1, tmp))
                    {
                        fprintf (stderr, LXTHDR "fwrite error to tempfile, exiting.\n");
                        exit (255);
                    };
                    len -= siz;
                }

                printf (LXTHDR "...expanded %08x into %08x bytes.\n", zchg_size, zchg_predec_size);
                gzclose (zhandle);
            }
            munmap (mm, f_len);
            close (fd);
            fflush (tmp);
            fseek (tmp, 0, SEEK_SET);
            fclose (tmp);

            fd = fd2;
            mm = mmap (NULL, zchg_predec_size, PROT_READ, MAP_SHARED, fd, 0);
            mm -= 4;            /* because header and version don't exist in packed change records */
        } else
        {
            unsigned int len = zchg_predec_size;
            int rc;
            char *buf = g_new (char, zchg_predec_size);
            char *pnt = buf;
            char testbyte[2] = { 0, 0 };
            char is_bz2;

            printf (LXTHDR "Compressed change records detected...\n");
            if (change_field_offset != lseek (fd, change_field_offset, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", change_field_offset);
                exit (255);
            }

            read (fd, &testbyte, 2);
            is_bz2 = (testbyte[0] == 'B') && (testbyte[1] == 'Z');

            if (change_field_offset != lseek (fd, change_field_offset, SEEK_SET))
            {
                fprintf (stderr, LXTHDR "lseek error at offset %08x\n", change_field_offset);
                exit (255);
            }

            if (is_bz2)
            {
#ifndef HAVE_LIBBZ2
                fprintf (stderr, LXTHDR "bzip2 compression not available");
                exit (EXIT_FAILURE);
#else
                zhandle = BZ2_bzdopen (dup (fd), "rb");

                while (len)
                {
                    int siz = (len > 32768) ? 32768 : len;

                    rc = BZ2_bzread (zhandle, pnt, siz);
                    if (rc != siz)
                    {
                        fprintf (stderr, LXTHDR "BZ2_bzread error to buffer %d (act) vs %d (exp), exiting.\n", rc, siz);
                        exit (255);
                    }
                    pnt += siz;
                    len -= siz;
                }

                printf (LXTHDR "...expanded %08x into %08x bytes.\n", zchg_size, zchg_predec_size);
                BZ2_bzclose (zhandle);
#endif
            } else
            {
                zhandle = gzdopen (dup (fd), "rb");

                while (len)
                {
                    int siz = (len > 32768) ? 32768 : len;

                    rc = gzread (zhandle, pnt, siz);
                    if (rc != siz)
                    {
                        fprintf (stderr, LXTHDR "gzread error to buffer %d (act) vs %d (exp), exiting.\n", rc, siz);
                        exit (255);
                    }
                    pnt += siz;
                    len -= siz;
                }

                printf (LXTHDR "...expanded %08x into %08x bytes.\n", zchg_size, zchg_predec_size);
                gzclose (zhandle);
            }

            munmap (mm, f_len);
            close (fd);

            fd = -1;
            mm = buf - 4;       /* because header and version don't exist in packed change records */
        }
    }

    if (!sync_table_offset)
    {
        unsigned int vlen = zchg_predec_size ? zchg_predec_size + 4 : 0;
        unsigned int offs = zchg_predec_size ? 4 : 0;
        unsigned int wave_num_symbols_bytes;
        unsigned int num_records = 0;
        unsigned int last_change_delta, numbytes;
        int *positional_compar = positional_information;

        unsigned int dict_16_offset_new = 0;
        unsigned int dict_24_offset_new = 0;
        unsigned int dict_32_offset_new = 0;

        char *nam;
        FILE *tmp;
        int recfd;

        fprintf (stderr, LXTHDR "Linear LXT encountered...\n");

        if (!zchg_predec_size)
        {
            fprintf (stderr, LXTHDR "Uncompressed linear LXT not supported, exiting.\n");
            exit (255);
        }

        if (wave_num_symbols >= 256 * 65536)
        {
            wave_num_symbols_bytes = 3;
        } else if (wave_num_symbols >= 65536)
        {
            wave_num_symbols_bytes = 2;
        } else if (wave_num_symbols >= 256)
        {
            wave_num_symbols_bytes = 1;
        } else
        {
            wave_num_symbols_bytes = 0;
        }

        nam = tmpnam (NULL);
        tmp = fopen (nam, "wb");
        fpos = 4;               /* fake 4 bytes padding */
        recfd = open (nam, O_RDONLY);

        unlink (nam);

        while (offs < vlen)
        {
            unsigned int facidx;
            unsigned char cmd;
            unsigned int offscache2, offscache3;
            unsigned int height;
            unsigned char cmdkill;

            num_records++;

            /* remake time vs position table on the fly */
            if (offs == *positional_compar)
            {
                *positional_compar = fpos;
                positional_compar++;
            }

            switch (wave_num_symbols_bytes & 3)
            {
            case 0:
                facidx = get_byte (offs);
                break;
            case 1:
                facidx = get_16 (offs);
                break;
            case 2:
                facidx = get_24 (offs);
                break;
            case 3:
                facidx = get_32 (offs);
                break;
            }

            if (facidx > wave_num_symbols)
            {
                fprintf (stderr, LXTHDR "Facidx %d out of range (vs %d) at offset %08x, exiting.\n", facidx, wave_num_symbols,
                  offs);
                exit (255);
            }

            offs += (wave_num_symbols_bytes + 1);

            cmdkill = mvlfacs[facidx].flags & (LT_SYM_F_DOUBLE | LT_SYM_F_STRING);
            if (!cmdkill)
            {
                cmd = get_byte (offs);

                if (cmd > 0xf)
                {
                    fprintf (stderr, LXTHDR "Command byte %02x invalid at offset %08x, exiting.\n", cmd, offs);
                    exit (0);
                }

                offs++;
            } else
            {
                cmd = 0;
            }

            offscache2 = offs;

            height = mvlfacs[facidx].array_height;
            if (height)
            {
                if (height >= 256 * 65536)
                {
                    offs += 4;
                } else if (height >= 65536)
                {
                    offs += 3;
                } else if (height >= 256)
                {
                    offs += 2;
                } else
                {
                    offs += 1;
                }
            }

            offscache3 = offs;
            if (!dict_16_offset_new)
            {
                if (offs == dict_16_offset)
                {
                    dict_16_offset_new = fpos;
                }
            } else if (!dict_24_offset_new)
            {
                if (offs == dict_24_offset)
                {
                    dict_24_offset_new = fpos;
                }
            } else if (!dict_32_offset_new)
            {
                if (offs == dict_32_offset)
                {
                    dict_32_offset_new = fpos;
                }
            }

            /* printf("%08x : %04x %02x (%d) %s[%d:%d]\n", offscache, facidx, cmd, mvlfacs[facidx].len, mvlfacs[facidx].name, mvlfacs[facidx].msb, mvlfacs[facidx].lsb); */

            if (!cmdkill)
                switch (cmd)
                {
                    int modlen;

                case 0x0:
                    modlen = (!(mvlfacs[facidx].flags & LT_SYM_F_INTEGER)) ? mvlfacs[facidx].len : 32;
                    if ((dict_string_mem_array) && (modlen > dict_width))
                    {
                        if ((!dict_16_offset) || (offscache3 < dict_16_offset))
                        {
                            offs += 1;
                        } else if ((!dict_24_offset) || (offscache3 < dict_24_offset))
                        {
                            offs += 2;
                        } else if ((!dict_32_offset) || (offscache3 < dict_32_offset))
                        {
                            offs += 3;
                        } else
                        {
                            offs += 4;
                        }
                    } else
                    {
                        offs += (mvlfacs[facidx].len + 7) / 8;
                    }

                    break;

                case 0x1:
                    offs += (mvlfacs[facidx].len + 3) / 4;
                    break;

                case 0x2:
                    offs += (mvlfacs[facidx].len + 1) / 2;
                    break;

                case 0x3:
                case 0x4:
                case 0x5:
                case 0x6:
                case 0x7:
                case 0x8:
                case 0x9:
                case 0xa:
                case 0xb:
                    break;      /* single byte, no extra "skip" */

                case 0xc:
                case 0xd:
                case 0xe:
                case 0xf:
                    offs += ((cmd & 3) + 1); /* skip past numbytes_trans */
                    break;
            } else
            {
                /* cmdkill = 1 for strings + reals skip bytes */
                if (mvlfacs[facidx].flags & LT_SYM_F_DOUBLE)
                {
                    offs += 8;
                } else          /* strings */
                {
                    while (get_byte (offs))
                        offs++;
                    offs++;
                }
            }

            last_change_delta = fpos - mvlfacs[facidx].lastchange - 2;
            mvlfacs[facidx].lastchange = fpos;

            maxchange = fpos;
            maxindex = facidx;

            if (last_change_delta >= 256 * 65536)
            {
                numbytes = 3;
            } else if (last_change_delta >= 65536)
            {
                numbytes = 2;
            } else if (last_change_delta >= 256)
            {
                numbytes = 1;
            } else
            {
                numbytes = 0;
            }

            lt_emit_u8 (tmp, (numbytes << 4) | cmd);
            switch (numbytes & 3)
            {
            case 0:
                lt_emit_u8 (tmp, last_change_delta);
                break;
            case 1:
                lt_emit_u16 (tmp, last_change_delta);
                break;
            case 2:
                lt_emit_u24 (tmp, last_change_delta);
                break;
            case 3:
                lt_emit_u32 (tmp, last_change_delta);
                break;
            }

            if (offs - offscache2)
            {
                fpos += fwrite (mm + offscache2, 1, offs - offscache2, tmp); /* copy rest of relevant info */
            }
        }

        dict_16_offset = dict_16_offset_new;
        dict_24_offset = dict_24_offset_new;
        dict_32_offset = dict_32_offset_new;

        fflush (tmp);
        fseek (tmp, 0, SEEK_SET);
        fclose (tmp);
        fprintf (stderr, LXTHDR "%d linear records converted into %08x bytes.\n", num_records, fpos - 4);

        if (zchg_predec_size > 128 * 1024 * 1024)
        {
            munmap (mm + 4, zchg_predec_size);
            close (fd);
        } else
        {
            free (mm + 4);
        }

        fd = recfd;
        mm = mmap (NULL, fpos - 4, PROT_READ, MAP_SHARED, recfd, 0);
        mm -= 4;                /* because header and version don't exist in packed change records */
    }
}

/*
 * given a fac+offset, return the binary data for it
 */
static char *parse_offset (struct fac *which, unsigned int offs)
{
    int v, v2, j;
    int k, l;
    char *pnt;
    char repeat;

    l = which->len;
    pnt = lt_buf;
    v = get_byte (offs);
    v2 = v & 0x0f;

    switch (v2)
    {
    case 0x00:                 /* MVL2 */
        {
            unsigned int msk;
            int bitcnt = 0;
            int ch;

            offs += ((v >> 4) & 3) + 2; /* skip value */

            if ((dict_string_mem_array) && (l > dict_width))
            {
                unsigned int dictpos;
                unsigned int ld;

                if ((!dict_16_offset) || (offs < dict_16_offset))
                {
                    dictpos = get_byte (offs);
                } else if ((!dict_24_offset) || (offs < dict_24_offset))
                {
                    dictpos = get_16 (offs);
                } else if ((!dict_32_offset) || (offs < dict_32_offset))
                {
                    dictpos = get_24 (offs);
                } else
                {
                    dictpos = get_32 (offs);
                }

                if (dictpos <= dict_num_entries)
                {
                    ld = strlen (dict_string_mem_array[dictpos]);
                    for (j = 0; j < (l - (ld + 1)); j++)
                    {
                        *(pnt++) = '0';
                    }
                    *(pnt++) = '1';
                    memcpy (pnt, dict_string_mem_array[dictpos], ld);
                } else
                {
                    fprintf (stderr, LXTHDR "dict entry at offset %08x [%d] out of range, ignoring!\n", dictpos, offs);
                    for (j = 0; j < l; j++)
                    {
                        *(pnt++) = '0';
                    }
                }
            } else
                for (j = 0;; j++)
                {
                    ch = get_byte (offs + j);
                    msk = 0x80;
                    for (k = 0; k < 8; k++)
                    {
                        *(pnt++) = (ch & msk) ? '1' : '0';
                        msk >>= 1;
                        bitcnt++;
                        if (bitcnt == l)
                            goto bail;
                    }
                }
        }
        break;

    case 0x01:                 /* MVL4 */
        {
            int bitcnt = 0;
            int ch;
            int rsh;

            offs += ((v >> 4) & 3) + 2; /* skip value */
            for (j = 0;; j++)
            {
                ch = get_byte (offs + j);
                rsh = 6;
                for (k = 0; k < 4; k++)
                {
                    *(pnt++) = convert_mvl ((ch >> rsh) & 0x3);
                    rsh -= 2;
                    bitcnt++;
                    if (bitcnt == l)
                        goto bail;
                }
            }
        }
        break;

    case 0x02:                 /* MVL9 */
        {
            int bitcnt = 0;
            int ch;
            int rsh;

            offs += ((v >> 4) & 3) + 2; /* skip value */
            for (j = 0;; j++)
            {
                ch = get_byte (offs + j);
                rsh = 4;
                for (k = 0; k < 2; k++)
                {
                    *(pnt++) = convert_mvl (ch >> rsh);
                    rsh -= 4;
                    bitcnt++;
                    if (bitcnt == l)
                        goto bail;
                }
            }
        }
        break;

    case 0x03:                 /* mvl repeat expansions */
    case 0x04:
    case 0x05:
    case 0x06:
    case 0x07:
    case 0x08:
    case 0x09:
    case 0x0a:
    case 0x0b:
        repeat = convert_mvl (v2 - 3);
        for (j = 0; j < l; j++)
        {
            *(pnt++) = repeat;
        }
        break;

    default:
        fprintf (stderr, LXTHDR "Unknown %02x at offset: %08x\n", v, offs);
        exit (255);
    }

  bail:
    return (lt_buf);
}

/*
 * XXX : this needs to get re-integrated as most of the function calls for mmap/etc
 * are cut and paste straight out of the 1.3 version!
 */
static void XXX_PLEASE_REUSE_ME__lowlevel_openAndMap (char *filename, void **mappingPtr, off_t * lengthPtr)
{
    void *map;
    off_t len;

#ifdef _MSC_VER
    HANDLE hIn = CreateFile (filename, GENERIC_READ, 0, NULL,
      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

    if (hIn != INVALID_HANDLE_VALUE)
    {
        HANDLE hInMap = CreateFileMapping (hIn, NULL, PAGE_READONLY, 0, 0, NULL);

        map = MapViewOfFile (hInMap, FILE_MAP_READ, 0, 0, 0);
        len = GetFileSize (hIn, NULL);
        *mappingPtr = map;
        *lengthPtr = len;
        return;
    }
#else /* !_MSC_VER */
    int fd = open (filename, O_RDONLY);

    if (fd >= 0)
    {
        len = lseek (fd, 0, SEEK_END);
        map = mmap (NULL, len, PROT_READ, MAP_SHARED, fd, 0);
        if (map != MAP_FAILED)
        {
            *mappingPtr = map;
            *lengthPtr = len;
            return;
        }
    }
#endif /* _MSC_VER */
    perror ("could not open and map file");
    exit (1);
}

/*
 * mainline
 */
TimeType lxt_main (WaveReader *reader)
{
	const char *fname = reader->filename;
    int i;
    Node *n;
    Symbol *curnode;
    Symbol *s, *prevsymroot = NULL, *prevsym = NULL;
    unsigned int tagpnt;
    int tag;

#ifndef _MSC_VER

    fd = open (fname, O_RDONLY);
    if (fd < 0)
    {
        fprintf (stderr, "Could not open '%s', exiting.\n", fname);
        exit (0);
    }

    f_len = lseek (fd, 0, SEEK_END);
    mm = mmap (NULL, f_len, PROT_READ, MAP_SHARED, fd, 0);
#else
    hIn = CreateFile (fname, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hIn == INVALID_HANDLE_VALUE)
    {
        fprintf (stderr, "Could not open '%s', exiting.\n", fname);
        exit (0);
    }
    hInMap = CreateFileMapping (hIn, NULL, PAGE_READONLY, 0, 0, NULL);
    mm = MapViewOfFile (hInMap, FILE_MAP_READ, 0, 0, 0);
    f_len = GetFileSize (hIn, NULL);
#endif

    if ((i = get_16 (0)) != LT_HDRID)
    {
        fprintf (stderr, "Not an LXT format AET, exiting.\n");
        exit (255);
    }

    if ((version = get_16 (2)) > LT_VERSION)
    {
        fprintf (stderr, "Version %d of LXT format AETs not supported, exiting.\n", version);
        exit (255);
    }

    if (get_byte (f_len - 1) != LT_TRLID)
    {
        fprintf (stderr, "LXT '%s' is truncated, exiting.\n", fname);
        exit (255);
    }

    DBP ((LXTHDR "Loading LXT '%s'...\n", fname));
    DBP ((LXTHDR "Len: %d\n", (unsigned int) f_len));

    tagpnt = f_len - 2;
    while ((tag = get_byte (tagpnt)) != LT_SECTION_END)
    {
        int offset = get_32 (tagpnt - 4);

        tagpnt -= 5;

        switch (tag)
        {
        case LT_SECTION_CHG:
            change_field_offset = offset;
            DBP ((LXTHDR "LT_SECTION_CHG at %08x\n", offset));
            break;
        case LT_SECTION_SYNC_TABLE:
            sync_table_offset = offset;
            DBP ((LXTHDR "LT_SECTION_SYNC_TABLE at %08x\n", offset));
            break;
        case LT_SECTION_FACNAME:
            facname_offset = offset;
            DBP ((LXTHDR "LT_SECTION_FACNAME at %08x\n", offset));
            break;
        case LT_SECTION_FACNAME_GEOMETRY:
            facgeometry_offset = offset;
            DBP ((LXTHDR "LT_SECTION_FACNAME_GEOMETRY at %08x\n", offset));
            break;
        case LT_SECTION_TIMESCALE:
            timescale_offset = offset;
            DBP ((LXTHDR "LT_SECTION_TIMESCALE at %08x\n", offset));
            break;
        case LT_SECTION_TIME_TABLE:
            time_table_offset = offset;
            DBP ((LXTHDR "LT_SECTION_TIME_TABLE at %08x\n", offset));
            break;
        case LT_SECTION_TIME_TABLE64:
            time_table_offset64 = offset;
            DBP ((LXTHDR "LT_SECTION_TIME_TABLE64 at %08x\n", offset));
            break;
        case LT_SECTION_INITIAL_VALUE:
            initial_value_offset = offset;
            DBP ((LXTHDR "LT_SECTION_INITIAL_VALUE at %08x\n", offset));
            break;
        case LT_SECTION_DOUBLE_TEST:
            double_test_offset = offset;
            DBP ((LXTHDR "LT_SECTION_DOUBLE_TEST at %08x\n", offset));
            break;

        case LT_SECTION_ZFACNAME_PREDEC_SIZE:
            zfacname_predec_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZFACNAME_PREDEC_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZFACNAME_SIZE:
            zfacname_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZFACNAME_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZFACNAME_GEOMETRY_SIZE:
            zfacgeometry_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZFACNAME_GEOMETRY_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZSYNC_SIZE:
            zsync_table_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZSYNC_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZTIME_TABLE_SIZE:
            ztime_table_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZTIME_TABLE_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZCHG_PREDEC_SIZE:
            zchg_predec_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZCHG_PREDEC_SIZE = %08x\n", offset));
            break;
        case LT_SECTION_ZCHG_SIZE:
            zchg_size = offset;
            DBP ((LXTHDR "LT_SECTION_ZCHG_SIZE = %08x\n", offset));
            break;

        case LT_SECTION_ZDICTIONARY:
            zdictionary_offset = offset;
            DBP ((LXTHDR "LT_SECTION_ZDICTIONARY = %08x\n", offset));
            break;
        case LT_SECTION_ZDICTIONARY_SIZE:
            zdictionary_predec_size = offset;
            break;

        case LT_SECTION_EXCLUDE_TABLE:
            exclude_offset = offset;
            DBP ((LXTHDR "LT_SECTION_EXCLUDE_TABLE = %08x\n", offset));
            break;

        default:
            fprintf (stderr, "Skipping unknown section tag %02x.\n", tag);
            break;
        }
    }

    if (exclude_offset)
    {
        unsigned int offset = exclude_offset;
        int i, num_blackouts = get_32 (offset);
        TimeType bs, be;
        BlackoutRegion *bt;

        offset += 4;

        for (i = 0; i < num_blackouts; i++)
        {
            bs = get_64 (offset);
            offset += 8;
            be = get_64 (offset);
            offset += 8;

            bt = g_new (BlackoutRegion, 1);
            bt->bstart = bs;
            bt->bend = be;
            bt->next = blackout_regions;

            blackout_regions = bt;
        }
    }

    if (double_test_offset)
    {
        create_double_endian_mask (double_test_offset);
    }

    if (timescale_offset)
    {
        signed char scale;

        scale = (signed char) get_byte (timescale_offset);
        switch (scale)
        {
        case 0:
            wave_config->time_dimension = 's';
            break;
        case -1:
            wave_config->time_scale = LL_LITERAL (100);
            wave_config->time_dimension = 'm';
            break;
        case -2:
            wave_config->time_scale = LL_LITERAL (10);
        case -3:
            wave_config->time_dimension = 'm';
            break;

        case -4:
            wave_config->time_scale = LL_LITERAL (100);
            wave_config->time_dimension = 'u';
            break;
        case -5:
            wave_config->time_scale = LL_LITERAL (10);
        case -6:
            wave_config->time_dimension = 'u';
            break;

        case -10:
            wave_config->time_scale = LL_LITERAL (100);
            wave_config->time_dimension = 'p';
            break;
        case -11:
            wave_config->time_scale = LL_LITERAL (10);
        case -12:
            wave_config->time_dimension = 'p';
            break;

        case -13:
            wave_config->time_scale = LL_LITERAL (100);
            wave_config->time_dimension = 'f';
            break;
        case -14:
            wave_config->time_scale = LL_LITERAL (10);
        case -15:
            wave_config->time_dimension = 'f';
            break;

        case -7:
            wave_config->time_scale = LL_LITERAL (100);
            wave_config->time_dimension = 'n';
            break;
        case -8:
            wave_config->time_scale = LL_LITERAL (10);
        case -9:
        default:
            wave_config->time_dimension = 'n';
            break;
        }
    } else
    {
        wave_config->time_dimension = 'n';
    }

    if (!facname_offset)
    {
        fprintf (stderr, "LXT '%s' is missing a facility name section, exiting.\n", fname);
        exit (255);
    }

    wave_num_symbols = get_32 (facname_offset);
    DBP ((LXTHDR "Number of facs: %d\n", wave_num_symbols));
    /* FIXME, check out elements */
    mvlfacs = g_new0 (struct fac, wave_num_symbols);

    if (initial_value_offset)
    {
        switch (get_byte (initial_value_offset))
        {
        case 0:
        case 7:
            initial_value = '0';
            break;
        case 1:
        case 4:
            initial_value = '1';
            break;
        case 2:
            initial_value = 'Z';
            break;
        default:
            initial_value = 'X';
            break;
        }
    } else
    {
        initial_value = '1';
    }

    if (zdictionary_offset)
    {
        if (zdictionary_predec_size)
        {
            build_dict ();
        } else
        {
            fprintf (stderr, "LXT '%s' is missing a zdictionary_predec_size chunk, exiting.\n", fname);
            exit (255);
        }
    }

    build_facs (fname);
    build_facs2 (fname);

    /*
     * do your stuff here..all useful info has been initialized by now 
     */

    if (!hier_was_explicitly_set)
    {                           /* set default hierarchy split char */
        hier_delimiter = '.';
    }

    for (i = 0; i < wave_num_symbols; i++)
    {
        char buf[4096];
        char *str;
        struct fac *f;

        if (mvlfacs[i].flags & LT_SYM_F_ALIAS)
        {
            int alias = mvlfacs[i].array_height;

            f = mvlfacs + alias;

            while (f->flags & LT_SYM_F_ALIAS)
            {
                f = mvlfacs + f->array_height;
            }
        } else
        {
            f = mvlfacs + i;
        }

        if ((f->len > 1) && (!(f->flags & (LT_SYM_F_INTEGER | LT_SYM_F_DOUBLE | LT_SYM_F_STRING))))
        {
            sprintf (buf, "%s[%d:%d]", mvlfacs[i].name, mvlfacs[i].msb, mvlfacs[i].lsb);
            str = g_new (char, strlen (buf) + 1);
            copy_change_delim (str, buf);
            s = symadd (str, hash (str));
            prevsymroot = prevsym = NULL;
        } else if ((f->len == 1) && (!(f->flags & (LT_SYM_F_INTEGER | LT_SYM_F_DOUBLE | LT_SYM_F_STRING))) &&
          ((i != wave_num_symbols - 1) && (!strcmp (mvlfacs[i].name, mvlfacs[i + 1].name))) ||
          ((i != 0) && (!strcmp (mvlfacs[i].name, mvlfacs[i - 1].name))) &&
          (mvlfacs[i].msb != -1) && (mvlfacs[i].lsb != -1))
        {
            sprintf (buf, "%s[%d]", mvlfacs[i].name, mvlfacs[i].msb);
            str = g_new (char, strlen (buf) + 1);
            copy_change_delim (str, buf);
            s = symadd (str, hash (str));
            if ((prevsym) && (i > 0) && (!strcmp (mvlfacs[i].name, mvlfacs[i - 1].name)))
            {                   /* allow chaining for
                                 * search functions.. */
                prevsym->vec_root = prevsymroot;
                prevsym->vec_chain = s;
                s->vec_root = prevsymroot;
                prevsym = s;
            } else
            {
                prevsymroot = prevsym = s;
            }
        } else
        {
            str = g_new (char, strlen (mvlfacs[i].name) + 1);
            copy_change_delim (str, mvlfacs[i].name);
            s = symadd (str, hash (str));
            prevsymroot = prevsym = NULL;

            if (f->flags & LT_SYM_F_INTEGER)
            {
                mvlfacs[i].msb = 31;
                mvlfacs[i].lsb = 0;
                mvlfacs[i].len = 32;
            }
        }

        if (!firstnode)
        {
            firstnode = curnode = s;
        } else
        {
            curnode->next_in_file = s;
            curnode = s;
        }

		{ /* Make the Nodes and choose their type */
			NodeFlags node_type = 0;

			if (f->flags & LT_SYM_F_DOUBLE)
				node_type = NODE_FLAGS_REAL;
			else if (f->flags & LT_SYM_F_STRING)
				node_type = NODE_FLAGS_STRING;
			else node_type = (f->len == 1 ? NODE_FLAGS_DIGITAL : NODE_FLAGS_VECTOR);

			n = wave_new_node (s->name, reader, node_type);
		}
        n->import_info = (void *) (mvlfacs + i);

        if ((f->len > 1) || (f->flags & (LT_SYM_F_DOUBLE | LT_SYM_F_STRING)))
        {
            ExtNode *ext = g_new (ExtNode, 1);

            ext->msi = mvlfacs[i].msb;
            ext->lsi = mvlfacs[i].lsb;
            n->ext = ext;
        }

        s->n = n;
    }

    wave_symbols = g_new (Symbol *, wave_num_symbols);
    curnode = firstnode;
    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;
        int len;

        wave_symbols[i] = curnode;
        if ((len = strlen (subst = curnode->name)) > longestname)
            longestname = len;
        curnode = curnode->next_in_file;
        while ((ch = (*subst)))
        {
            if (ch == hier_delimiter)
            {
                *subst = 0x01;
            }                   /* forces sort at hier boundaries */
            subst++;
        }
    }

    printf (LXTHDR "Sorting facilities at hierarchy boundaries...");
    fflush (stdout);
    wave_heapsort (wave_symbols, wave_num_symbols);
    printf ("sorted.\n");

    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;

        subst = wave_symbols[i]->name;
        while ((ch = (*subst)))
        {
            if (ch == 0x01)
            {
                *subst = hier_delimiter;
            }                   /* restore back to normal */
            subst++;
        }

#ifdef WAVE_DEBUG_FACILITIES
        printf ("%-4d %s\n", i, wave_symbols[i]->name);
#endif
    }

    wave_symbols_are_sorted = TRUE;

    printf (LXTHDR "Building facility hierarchy tree...");
    fflush (stdout);
    init_tree ();
    for (i = 0; i < wave_num_symbols; i++)
    {
        build_tree_from_name (wave_symbols[i]->name, i);
    }
    treeprune (treeroot);
    printf ("built.\n\n");

#ifdef WAVE_DEBUG_FACILITIES
    treedebug (treeroot, "");
#endif

    lxt_scaled_start_time = first_cycle * wave_config->time_scale;
    lxt_scaled_end_time = last_cycle * wave_config->time_scale;
    printf ("[" LL_PRINTF_FORMAT "] start time.\n[" LL_PRINTF_FORMAT "] end time.\n", lxt_scaled_start_time, lxt_scaled_end_time);
    return (lxt_scaled_end_time);
}

/*
 * this is the black magic that handles aliased signals...
 */
static void lxt_resolver (Node * np, Node * resolve)
{
    np->ext = resolve->ext;
/* FIXME FIXME FIXME NOHIST
    np->head = resolve->head;
    np->curr = resolve->curr;
    */
    np->import_info = NULL;
}

/*
 * actually import an lxt trace but don't do it if
 * 1) it's already been imported
 * 2) an alias of this trace has been imported--instead
 *    copy over the relevant info and be done with it.
 */
void import_lxt_trace (Node *np)
{
    unsigned int offs, offsdelta;
    int v, w;
    TimeType tmval;
    TimeType prevtmval;
    char *parsed;
    int len;
    struct fac *f;
    struct fac *np_mvlfac = (np ? (struct fac *) np->import_info : NULL);

    f = (struct fac *) np->import_info;
    if (!f)
        return;                 /* already imported */

    if (((struct fac *) np->import_info)->flags & LT_SYM_F_ALIAS)
    {
        int alias = np_mvlfac->array_height;

        f = mvlfacs + alias;

        if (f->resolve_lxt_alias_to)
        {
            if (!np_mvlfac->resolve_lxt_alias_to)
                np_mvlfac->resolve_lxt_alias_to = f->resolve_lxt_alias_to;
        } else
        {
            f->resolve_lxt_alias_to = np;
        }

        while (f->flags & LT_SYM_F_ALIAS)
        {
            f = mvlfacs + f->array_height;

            if (f->resolve_lxt_alias_to)
            {
                if (!np_mvlfac->resolve_lxt_alias_to)
                    np_mvlfac->resolve_lxt_alias_to = f->resolve_lxt_alias_to;
            } else
            {
                f->resolve_lxt_alias_to = np;
            }
        }
    }

    /*
     * f is the head minus any aliases, np->mvlfac is us... 
     */
    if (np_mvlfac->resolve_lxt_alias_to)
    {                           /* in case we're an alias head for later.. */
        lxt_resolver (np, np_mvlfac->resolve_lxt_alias_to);
        return;
    }

    np_mvlfac->resolve_lxt_alias_to = np; /* in case we're an alias head for later.. */
    offs = f->lastchange;

    tmval = LL_LITERAL (-1);
    prevtmval = LL_LITERAL (-1);
    len = np_mvlfac->len;

    if (f->array_height < 1)    /* sorry, arrays not supported */
    {
        while (offs)
        {
            unsigned char val = 0;

            if ((w = ((v = get_byte (offs)) & 0xF)) > 0xb)
            {
                unsigned int offsminus1, offsminus2, offsminus3;
                TimeType delta, time_offsminus1;
                int skip;

                switch (v & 0xF0)
                {
                case 0x00:
                    skip = 2;
                    offsdelta = get_byte (offs + 1);
                    break;
                case 0x10:
                    skip = 3;
                    offsdelta = get_16 (offs + 1);
                    break;
                case 0x20:
                    skip = 4;
                    offsdelta = get_24 (offs + 1);
                    break;
                case 0x30:
                    skip = 5;
                    offsdelta = get_32 (offs + 1);
                    break;
                default:
                    fprintf (stderr, "Unknown %02x at offset: %08x\n", v, offs);
                    exit (0);
                }
                offsminus1 = offs - offsdelta - 2;

                switch (get_byte (offsminus1) & 0xF0)
                {
                case 0x00:
                    offsdelta = get_byte (offsminus1 + 1);
                    break;
                case 0x10:
                    offsdelta = get_16 (offsminus1 + 1);
                    break;
                case 0x20:
                    offsdelta = get_24 (offsminus1 + 1);
                    break;
                case 0x30:
                    offsdelta = get_32 (offsminus1 + 1);
                    break;
                default:
                    fprintf (stderr, "Unknown %02x at offset: %08x\n", get_byte (offsminus1), offsminus1);
                    exit (0);
                }
                offsminus2 = offsminus1 - offsdelta - 2;

                delta = (time_offsminus1 = bsearch_mvl_timechain (offsminus1)) - bsearch_mvl_timechain (offsminus2);

                if (len > 1)
                {
                    WAVE_DEBUG (fprintf (stderr, "!!! DELTA = %Ld\n", delta));
                    WAVE_DEBUG (fprintf (stderr, "!!! offsminus1 = %08x\n", offsminus1));

                    if (!lxt_compress_clock)
                    {
                        int val = get_byte (offsminus1) & 0xF;
                        int reps = 0;
                        int rcnt;
                        unsigned int reconstructm1 = 0;
                        unsigned int reconstructm2 = 0;
                        unsigned int reconstructm3 = 0;
                        unsigned int rle_delta[2];
                        int i;

                        if ((val != 0) && (val != 3) && (val != 4))
                        {
                            fprintf (stderr, "Unexpected clk compress byte %02x at offset: %08x\n",
                              get_byte (offsminus1), offsminus1);
                            exit (0);
                        }

                        switch (w & 3)
                        {
                        case 0:
                            reps = get_byte (offs + skip);
                            break;
                        case 1:
                            reps = get_16 (offs + skip);
                            break;
                        case 2:
                            reps = get_24 (offs + skip);
                            break;
                        case 3:
                            reps = get_32 (offs + skip);
                            break;
                        }

                        reps++;

                        WAVE_DEBUG (fprintf (stderr, "!!! reps = %d\n", reps));

                        parsed = parse_offset (f, offsminus1);
                        for (i = 0; i < len; i++)
                        {
                            reconstructm1 <<= 1;
                            reconstructm1 |= (parsed[i] & 1);
                        }

                        WAVE_DEBUG (fprintf (stderr, "!!! M1 = '%08x'\n", reconstructm1));

                        parsed = parse_offset (f, offsminus2);
                        for (i = 0; i < len; i++)
                        {
                            reconstructm2 <<= 1;
                            reconstructm2 |= (parsed[i] & 1);
                        }

                        WAVE_DEBUG (fprintf (stderr, "!!! M2 = '%08x'\n", reconstructm2));

                        switch (get_byte (offsminus2) & 0xF0)
                        {
                        case 0x00:
                            offsdelta = get_byte (offsminus2 + 1);
                            break;
                        case 0x10:
                            offsdelta = get_16 (offsminus2 + 1);
                            break;
                        case 0x20:
                            offsdelta = get_24 (offsminus2 + 1);
                            break;
                        case 0x30:
                            offsdelta = get_32 (offsminus2 + 1);
                            break;
                        default:
                            fprintf (stderr, "Unknown %02x at offset: %08x\n", get_byte (offsminus2), offsminus2);
                            exit (0);
                        }
                        offsminus3 = offsminus2 - offsdelta - 2;

                        parsed = parse_offset (f, offsminus3);
                        for (i = 0; i < len; i++)
                        {
                            reconstructm3 <<= 1;
                            reconstructm3 |= (parsed[i] & 1);
                        }

                        WAVE_DEBUG (fprintf (stderr, "!!! M3 = '%08x'\n", reconstructm3));

                        rle_delta[0] = reconstructm2 - reconstructm3;
                        rle_delta[1] = reconstructm1 - reconstructm2;

                        WAVE_DEBUG (fprintf (stderr, "!!! RLE0 = '%08x'\n", rle_delta[0]));
                        WAVE_DEBUG (fprintf (stderr, "!!! RLE1 = '%08x'\n", rle_delta[1]));

                        WAVE_DEBUG (fprintf (stderr, "!!! time_offsminus1 = %Ld\n", time_offsminus1));

                        tmval = time_offsminus1 + (delta * reps);
                        for (rcnt = 0; rcnt < reps; rcnt++)
                        {
                            int k;
                            int j = (reps - rcnt);
                            unsigned int res = reconstructm1 +
                              ((j / 2) + (j & 0)) * rle_delta[1] + ((j / 2) + (j & 1)) * rle_delta[0];

                            WAVE_DEBUG (fprintf (stderr, "!!! %Ld -> '%08x'\n", tmval, res));

                            for (k = 0; k < len; k++)
                            {
                                parsed[k] = ((1 << (len - k - 1)) & res) ? '1' : '0';
                            }

							{
								char *vector = g_new (char, len + 1);
								memcpy (vector, parsed, len);
								vector[len] = '\0';

								wave_node_insert_event (np, tmval, NODE_FLAGS_VECTOR, '\0', vector);
							}

                            tmval -= delta;
                        }
                    } else
                    {           /* compress to z on multibit */
                        int i;
						char *vector = g_new (char, len + 1);

                        for (i = 0; i < len; i++)
                            vector[i] = 'Z';
                        vector[len] = '\0';

                        tmval = time_offsminus1 + delta;

						wave_node_insert_event (np, tmval, NODE_FLAGS_VECTOR, '\0', vector);
                    }

                    offs = offsminus1; /* no need to recalc it again! */
                    continue;
                } else
                {
                    if (!lxt_compress_clock)
                    {
                        int val = get_byte (offsminus1) & 0xF;
                        int reps = 0;
                        int rcnt;

                        if ((val < 3) || (val > 4))
                        {
                            fprintf (stderr, "Unexpected clk compress byte %02x at offset: %08x\n",
                              get_byte (offsminus1), offsminus1);
                            exit (0);
                        }

                        switch (w & 3)
                        {
                        case 0:
                            reps = get_byte (offs + skip);
                            break;
                        case 1:
                            reps = get_16 (offs + skip);
                            break;
                        case 2:
                            reps = get_24 (offs + skip);
                            break;
                        case 3:
                            reps = get_32 (offs + skip);
                            break;
                        }

                        reps++;
                        val = (reps & 1) ^ (val == 4); /* because x3='0', x4='1' */
						val = (val ? '1' : '0');

                        tmval = time_offsminus1 + (delta * reps);
                        for (rcnt = 0; rcnt < reps; rcnt++)
                        {
                        	wave_node_insert_event (np, tmval, NODE_FLAGS_DIGITAL, val, NULL);

                            tmval -= delta;
                            val = wave_invert_digital_value (val);
                        }
                    } else
                    {
                        tmval = time_offsminus1 + delta;

                       	wave_node_insert_event (np, tmval, NODE_FLAGS_DIGITAL, 'Z', NULL);
                    }
                }

                offs = offsminus1; /* no need to recalc it again! */
                continue;
            } else if ((tmval = bsearch_mvl_timechain (offs)) != prevtmval)
            {                   /* get rid of glitches (if even possible) 
                                 */
                DBP ((LXTHDR "offs: %08x is time %08x\n", offs, tmval));
                if (!(f->flags & (LT_SYM_F_DOUBLE | LT_SYM_F_STRING)))
                {
                    parsed = parse_offset (f, offs);

                    if (len == 1)
                    {
						val = toupper (parsed[0]);

                       	wave_node_insert_event (np, tmval, NODE_FLAGS_DIGITAL, val, NULL);
                    } else
                    {
                    	char *vector = g_new (char, len + 1);
                        memcpy (vector, parsed, len);
                        vector[len] = '\0';

                       	wave_node_insert_event (np, tmval, NODE_FLAGS_VECTOR, '\0', vector);
                    }
                } else if (f->flags & LT_SYM_F_DOUBLE)
                {
                	char *vector;
                    int offs_dbl = offs + ((get_byte (offs) >> 4) & 3) + 2; /* skip value */

                    if (double_is_native)
                    {
                        vector = ((unsigned char *) mm + offs_dbl);
                        DBP ((LXTHDR "Added double '%.16g'\n", *((double *) (mm + offs_dbl))));
                    } else
                    {
                        vector = swab_double_via_mask (offs_dbl);
                        DBP ((LXTHDR "Added bytefixed double '%.16g'\n", *((double *) (vector))));
                    }
					wave_node_insert_event (np, tmval, NODE_FLAGS_REAL, '\0', vector);
                } else
                {               /* defaults to if(f->flags&LT_SYM_F_STRING) */
                	char *vector;
                    int offs_str = offs + ((get_byte (offs) >> 4) & 3) + 2; /* skip value */
                    vector = ((unsigned char *) mm + offs_str);

                    DBP ((LXTHDR "Added string '%s'\n", (unsigned char *) mm + offs_str));
					wave_node_insert_event (np, tmval, NODE_FLAGS_STRING, '\0', vector);
                }
            }

            prevtmval = tmval;

            /*
             * v=get_byte(offs); 
             */
            switch (v & 0xF0)
            {
            case 0x00:
                offsdelta = get_byte (offs + 1);
                break;

            case 0x10:
                offsdelta = get_16 (offs + 1);
                break;

            case 0x20:
                offsdelta = get_24 (offs + 1);
                break;

            case 0x30:
                offsdelta = get_32 (offs + 1);
                break;

            default:
                fprintf (stderr, "Unknown %02x at offset: %08x\n", v, offs);
                exit (0);
            }

            offs = offs - offsdelta - 2;
        }
    }

    np->import_info = NULL;     /* it's imported and cached so we can forget it's an mvlfac now */
}

/***********************************************************************************************/

static int f_lxt_clock_compress_to_z (char *str, int offset)
{
    WAVE_DEBUG (wave_log (FALSE, "WAVE_DEBUG: f_lxt_clock_compress_to_z(\"%s\")\n", str));
    lxt_compress_clock = wave_parse_boolean_rc_entry (str);
    return 0;
}
static void clock_compress_toggled (GtkToggleButton * togglebutton, gpointer unused)
{
    lxt_compress_clock = !lxt_compress_clock;
}
static void add_clock_compress_checkbutton (GtkNotebook * n)
{
    GtkBox *box = GTK_BOX (gtk_notebook_get_nth_page (n, 0));
    GtkWidget *w = gtk_check_button_new_with_label ("Compress LXT Clock");

    gtk_box_pack_start (box, w, FALSE, FALSE, 0);
    gtk_widget_show (w);
    gtk_signal_connect (GTK_OBJECT (w), "toggled", GTK_SIGNAL_FUNC (clock_compress_toggled), NULL);
}

static WaveRcEntry lxt_entries[] = {
    {"lxt_clock_compress_to_z", f_lxt_clock_compress_to_z, NULL}
};

WaveReader *lxt_init (void)
{
	WaveReader *ret = wave_new_reader ("lxt", "LXT", 0, NULL);

	wave_reader_add_function (ret, "load-declarations", lxt_main);
	wave_reader_add_function (ret, "import-trace", import_lxt_trace);

	wave_reader_add_rc_entries (ret, 1, lxt_entries);

	return ret;
}
