/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `main.c'
    GTKWave top level

    $Id: main.c,v 1.101 2004/04/25 19:23:42 bardslea Exp $
*/

#include "../config.h"

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "utils.h"
#include "main.h"
#include "currenttime.h"
#include "rc.h"
#include "log.h"
#include "bsearch.h"
#include "reader.h"
#include "gui_cb.h"
#include "support.h"
#include "color.h"
#include "tracefile.h"
#include "polling.h"
#include "fileio.h"
#include "command.h"
#include "value.h"
#include "tracepane.h"
#include "signalpane.h"
#include "marker.h"

/* Global configration for the tool */
WaveConfig *wave_config = NULL;

gboolean wave_debug_ab = FALSE;

WaveFilePollingInfo *wave_polling = NULL;

/* Show all signals as traces immediately the tool is started */
static gboolean show_all_traces = FALSE;
/* Check streams 3 & 4 for a command interface to another process */
static gboolean check_for_command = FALSE;

gboolean wave_have_main_window = FALSE;

const char *wave_about_title = "GTKWave Waveform Viewer " VERSION;
const char *wave_copyright_info =
	"Licensed under the GNU GPL\n"
	"(C) 2004 Andrew Bardsley\n"
	"(C) 1999-2003 Tony Bybell\n"
	"(C) 2000-2003 The University of Manchester\n"
	"    and other contributors";

/* Filescope vars. to span command line processing and main */
static gboolean read_from_stdin = FALSE;
static char *forced_reader_name = NULL;
static gboolean dynamic_polling_enabled = FALSE;
static int dynamic_polling_period = 0;

/* Command line switch types, don't bother with long prefices 'cause these
	never go outside this file */
typedef enum
{
    POSIX_LAST_SWITCH,          /* -- */
    BARE_HYPHEN_SWITCH,         /* - */
    FORMAT_SWITCH,              /* reader selection switch */
    DYNAMIC_SWITCH,				/* dynamic end-of-file following */
    SHOW_ALL_SWITCH,			/* show all traces straight away */
	COMMAND_IF_SWITCH,			/* check for command interface on streams 3 & 4 */
    DEBUG_SWITCH,
    IGNORED_SWITCH,
    IGNORED_ARG_SWITCH,
    NO_SWITCH
}
WaveSwitchTypes;

typedef struct
{
    char short_name;
    char *long_name;
    WaveSwitchTypes type;
}
WaveSwitchData;

const WaveSwitchData wave_switches[] = {
    {'-', "", POSIX_LAST_SWITCH},
    {'f', "format", FORMAT_SWITCH},
    {'g', "gdk-debug", IGNORED_ARG_SWITCH},
    {'_', "debug-ab", DEBUG_SWITCH},
    {'d', "dynamic", DYNAMIC_SWITCH},
    {'a', "show-all", SHOW_ALL_SWITCH},
    {'c', "command-interface", COMMAND_IF_SWITCH},
    {'\0', NULL, NO_SWITCH}
};

/* wave_usage : moan about usage and exit */
static void wave_usage (void)
{
    fprintf (stderr, "%s\n%s\n\nusage: %s <switch>* [<dump-file> [<saved-trace-file>]]\n\n",
    	wave_about_title, wave_copyright_info, g_get_prgname ());

    fprintf (stderr,
      "switches:     --           Accept no more command line switches\n"
      "              -f <format>  Assume the dump file is of (preferred extension)\n"
      "                             format <format> (--format)\n"
      "              -d <period>  Poll the dump file every <period> seconds and add\n"
      "                             new trace data to the end of the display (--dynamic)\n"
      "              -a           Show all traces in view at startup (--show-all)\n"
      "              -c           Use streams 3 and 4 as a command line interface to\n"
	  "                             GTKWave (--command-interface)\n"
      "\n" "When <dump-file> is `-', GTKWave reads waveform dump data from stdin\n\n");

#ifdef INIT_BUILTIN_READERS
#define REG(format) ", " #format
	fprintf (stderr,
	  "This GTKWave binary was built with the following modules built in (other\n"
	  "modules may be dynamically loaded): %s\n\n", (INIT_BUILTIN_READERS) + 2);
#undef REG
#endif

    exit (EXIT_FAILURE);
}

/* parse_switch : identify which switch is meant by swtch */
static WaveSwitchTypes parse_switch (char *swtch)
{
    const WaveSwitchData *patterns = wave_switches;

    if (strcmp (swtch, "-") == 0)
        return BARE_HYPHEN_SWITCH;
    if (swtch[0] != '-')
        return NO_SWITCH;       /* Not -... */
    if (swtch[1] == '-')        /* Long Switch */
    {
        while (patterns->type != NO_SWITCH)
        {
            if (patterns->long_name && strcmp (swtch + 2, patterns->long_name) == 0)
                return patterns->type;
            patterns++;
        }
    } else if (swtch[1] && swtch[2] == '\0') /* Short option */
    {
        while (patterns->type != NO_SWITCH)
        {
            if (patterns->short_name != '_' && swtch[1] == patterns->short_name)
                return patterns->type;
            patterns++;
        }
    }
    return NO_SWITCH;
}

/* process_command_line_switches : process command line switches and leave the trailing,
	non-switch arguments on the argv list.  Modifies passed in argv/argv */
static void process_command_line_switches (int *argc, char ***argv)
{
    gboolean bad_exit = FALSE;
    gboolean last_switch = FALSE;

    while (!last_switch && *argc && ***argv == '-')
    {
        char *swtch = (*argv)[0];

        (*argv)++;
        (*argc)--;

        switch (parse_switch (swtch))
        {
        case FORMAT_SWITCH:
            if (*argc == 0)     /* Need an arg. */
                bad_exit = TRUE;
            else
            {
                if (**argv)
                {
                    forced_reader_name = **argv;
                    (*argv)++;
                    (*argc)--;
                } else
                {
                    wave_log (TRUE, "Format switch requires an argument\n");
                    wave_usage ();
                }
            }
            break;
        case DYNAMIC_SWITCH:
            if (*argc == 0)     /* Need an arg. */
                bad_exit = TRUE;
            else
            {
                if (**argv)
                {
                    int time = atoi (**argv);

                    if (time <= 0 || time > 30)
                        wave_log (TRUE, "Dynamic poll period must be >= 0s and <= 30s\n");
                    else
                    {
                        dynamic_polling_period = time;
                        dynamic_polling_enabled = TRUE;
                    }
                    (*argv)++;
                    (*argc)--;
                } else
                {
                    wave_log (TRUE, "Dynamic switch requires an argument\n");
                    wave_usage ();
                }
            }
            break;
        case SHOW_ALL_SWITCH:
            show_all_traces = TRUE;
            break;
		case COMMAND_IF_SWITCH:
			check_for_command = TRUE;
			break;
        case BARE_HYPHEN_SWITCH:
            read_from_stdin = TRUE;
            last_switch = TRUE;
            break;
        case POSIX_LAST_SWITCH: /* Posix style -- switch meaning take no more arguments */
            last_switch = TRUE;
            break;
        case IGNORED_SWITCH:   /* Ignore switch */
            break;
        case IGNORED_ARG_SWITCH: /* Ignore switch and one argument */
            (*argv)++;
            (*argc)--;
            break;
        case DEBUG_SWITCH:
            wave_debug_ab = TRUE;
            break;
        default:
            wave_log (TRUE, "Unrecognised command line switch `%s'\n", swtch);
            wave_usage ();
            break;
        }
    }
    if (bad_exit)
        wave_usage ();
}

void wave_update_times (void)
{
    wave_set_end_time (tims.last);   /* until the configure_event of wave_drawing_area */
    tims.first = tims.laststart;
    wave_set_start_time (tims.first);
    tims.zoom = tims.prevzoom = 0; /* 1 pixel/ns default */
    tims.marker = tims.lmbcache = -1; /* uninitialized at first */

    if (tims.last - tims.first <= 400)
        wave_flags.do_initial_zoom_fit = TRUE; /* force zoom on small traces */

    wave_zoom (tims.zoom, wave_window);
}

void wave_reset_markers (void)
{
	int i;

	for (i = 0; i < 26; i++)
	{
		named_markers[i] = -1;
	}
}

void dump_2 (Node *node)
{
	NodeFlags node_type = node->flags & NODE_FLAGS_TYPE_MASK;
	WaveEventIterator *event_iter = wave_new_event_iterator (/* event_list */ NULL);

	if (! event_iter->current)
		return;

	do {
		char type_str[7];
		char *type_str_ptr = type_str;
		char *value;
		fprintf (stderr, LL_PRINTF_FORMAT, event_iter->current->time);

		switch (node_type)
		{
		case NODE_FLAGS_DIGITAL:
			*type_str_ptr = 'd';
			type_str_ptr ++;
			break;
		case NODE_FLAGS_VECTOR:
			*type_str_ptr = 'v';
			type_str_ptr ++;
			break;
		case NODE_FLAGS_STRING:
			*type_str_ptr = 's';
			type_str_ptr ++;
			break;
		case NODE_FLAGS_REAL:
			*type_str_ptr = 'r';
			type_str_ptr ++;
			break;
		case NODE_FLAGS_SYNC_HANDSHAKE:
			*type_str_ptr = 'h';
			type_str_ptr ++;
			break;
		case NODE_FLAGS_DATA_HANDSHAKE:
			*type_str_ptr = 'H';
			type_str_ptr ++;
			break;
		default:
			break;
		}
		*type_str_ptr = '\0';

		value = wave_event_value_string (NULL, event_iter->current);
		fprintf (stderr, "%s(%s) ", type_str, value);
		g_free (value);
	} while (wave_event_iterator_step_forward (event_iter));

	wave_free_event_iterator (event_iter);
}

void dump_1 (void)
{
	int i;

	for (i = 0; i < wave_num_symbols; i ++)
	{
		Node *node = wave_symbols[i]->n;

		if (node)
		{
			fprintf (stderr, "%s: ", node->nname);
			dump_2 (node);
			fprintf (stderr, "\n");
		}
	}
}

/* wave_load_trace_file : load a trace file from `filename'.  If `reader_name'
	is not NULL, then use that as the name of the reader format to use. */
WaveReader *wave_load_trace_file (char *filename, char *reader_name)
{
	WaveReader *reader;

    symnew ();

	if (! filename)
		return NULL;

	reader = wave_find_init_reader (filename, reader_name, read_from_stdin);
	/* Load format reader */
	if (! reader)
	{
		wave_log (TRUE, "Can't find reader for file `%s'\n", filename);

		return NULL;
	}

	/* Set flags of reader */
	if (dynamic_polling_enabled)
		reader->flags |= WAVE_READER_DYNAMIC;

	tims.last = wave_reader_load_declarations (reader);

	wave_reset_markers ();
	wave_update_times ();
	wave_set_step_size (wave_config->time_scale, wave_config->time_dimension, wave_window);

    if (show_all_traces)
    {
		wave_add_all_signals_to_trace ();
		wave_max_signal_length (wave_window);
		wave_signal_pane_adjust_after_insertion (wave_window);
		wave_redraw (wave_window);
	}

    if (reader->flags & WAVE_READER_DYNAMIC)
	{
		/* NB. might need to change this if the sensitivity changes in
			wave_set_periodically_reread_active */
        wave_polling = wave_install_file_polling (reader, dynamic_polling_period);
		wave_set_periodically_reread_active (TRUE);
	}
    update_markertime (tims.marker);

	return reader;
}

int main (int argc, char *argv[])
{
    char *dump_filename = NULL;
    char *saved_traces_filename = NULL;

    /* Configuration and command line option stuff */
    config_log = wave_log;

    /* Initialise graphics stuff */
    gtk_init (&argc, &argv);

    if (argv[0])
        g_set_prgname (argv[0]);
    argv++;
    argc--;
    process_command_line_switches (&argc, &argv);

    /* Choose file names for waveform file and save traces file */
    if (read_from_stdin)
    {
        dump_filename = "-";

        if (argc == 1)          /* Saved traces file */
            saved_traces_filename = argv[0];
        if (argc > 1)
            wave_usage ();
    } else
    {
        if (argc == 2)
            saved_traces_filename = argv[1];

        if (argc >= 1)
            dump_filename = argv[0];
        /* AB
        else
            wave_usage ();
        */
    }

    { /* Handle pixmap startup */
        GtkWidget *sample_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

        wave_register_pixmaps (sample_window);
        gtk_widget_destroy (sample_window);
    }

    /* Banner! */
    wave_log (FALSE, "%s\n%s\n\n", wave_about_title, wave_copyright_info);

    wave_config = new_wave_config ();
    wave_init_flags ();
    wave_read_rc_file ();
    wave_register_builtin_readers ();

	if (check_for_command)
	{
		if (! wave_command_try_to_open (3, 4))
			wave_log (FALSE, "Failed to open command interface on streams 3 and 4\n");
		else wave_log (FALSE, "Opened command interface on streams 3 and 4\n");
	}

    build_main_window ();
    wave_have_main_window = TRUE;

	if (dump_filename)
	{
		if (! wave_load_trace_file (dump_filename, forced_reader_name))
		{
			wave_exit (EXIT_FAILURE);
		}
		// symdump (stderr, wave_symbols, wave_num_symbols);
	}
	if (saved_traces_filename)
	{
		if (! parse_wave_config_file (saved_traces_filename, wave_config))
		{
			wave_log (TRUE, "Couldn't successfully parse saved traces file `%s'\n", saved_traces_filename);
		} else {
			add_traces_from_trace_config_list (wave_config->traces);
			wave_redraw (wave_window);
		}
	}

    update_markertime (tims.marker);

    gtk_main ();

    return EXIT_SUCCESS;
}
