/*
	The GTKWave electronic waveform viewer
	Copyright (C) 2000-2003 Department of Computer Science
	The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
	Copyright (C) 1999-2003 Tony Bybell

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

	`navigate.c'
	Navigation GUI elements and functions

	$Id: navigate.c,v 1.13 2004/04/25 19:23:42 bardslea Exp $
*/

#include "navigate.h"
#include "currenttime.h"
#include "rc.h"
#include "help.h"
#include "reader.h"
#include "main.h"
#include "log.h"
#include "support.h"
#include "tracepane.h"
#include "signalpane.h"
#include "gui_if.h"

/* Maximum string length to hold a formatted time */
#define WAVE_MAX_TIME_LENGTH (40)

static TimeType step_length = 0;

static gdouble page_divisor = (gdouble) 1.0;

/* wave_calculate_zoom_for_span : calculate the zoom factor to show the span of
	time from `from' to `to'. */
double wave_calculate_zoom_for_span (TimeType from, TimeType to)
{
    int fixedwidth;
    double estimated;

    if (wave_window->traces.scale.wave_pane_width > 4)
        fixedwidth = wave_window->traces.scale.wave_pane_width - 4;
    else fixedwidth = wave_window->traces.scale.wave_pane_width;

    estimated = log ((to - from + 1) / (double) fixedwidth * (double) 200.0) / log (wave_flags.zoom_base);

    if (estimated < (double) 0.0)
        estimated = (double) 0.0;

    return -estimated;
}

/* wave_zoom : zoom the display to zoom level wave_flags.zoom_base^z0, adjust this zoom factor
	to make it sensible */
void wave_zoom (gdouble z0, WaveWindow *window)
{
    gdouble ppf = (gdouble) 200, frame;

    window->traces.scale.pixels_per_frame = 200;
    frame = pow (wave_flags.zoom_base, -z0);

    if (frame > (MAX_HISTENT_TIME / (gdouble) 4.0))
    {
        window->traces.scale.time_units_per_frame = MAX_HISTENT_TIME / (gdouble) 4.0;
    } else if (frame < (gdouble) 1.0)
    {
        window->traces.scale.time_units_per_frame = 1.0;
    } else
    {
        window->traces.scale.time_units_per_frame = frame;
    }

    if (wave_flags.pow10snap_flag && window->traces.scale.time_units_per_frame > 100.0)
    {
        gdouble p = 10.0;
        int l = (int) (log (window->traces.scale.time_units_per_frame) / log (p) + 0.5);

        window->traces.scale.time_units_per_frame = pow (p, (gdouble) l);
    }

    window->traces.scale.time_units_per_pixel = window->traces.scale.time_units_per_frame / ppf;
    window->traces.scale.pixels_per_time_unit = ppf / window->traces.scale.time_units_per_frame;
    window->traces.scale.tickstep = 10;

    WAVE_DEBUG (printf ("Zoom: %e Pixelsperframe: %lld, window->traces.scale.time_units_per_frame: %e\n", z0, window->traces.scale.pixels_per_frame, (float) window->traces.scale.time_units_per_frame));
}

static void adjust_zoom (void)
{
    GtkAdjustment *hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);

    fix_wavehadj (wave_window);
    /* force zoom update */
    gtk_adjustment_changed (hadj);
    gtk_adjustment_value_changed (hadj);
}

static void update_time (void)
{
    WAVE_DEBUG (printf ("Timeentry Configure Event\n"));

    wave_zoom (tims.zoom, wave_window);
    adjust_zoom ();
}

void fix_wavehadj (WaveWindow *window)
{
    gfloat pageinc;
    GtkAdjustment *hadj = GTK_ADJUSTMENT (window->traces.hslider);

    hadj->lower = tims.first;
    hadj->upper = tims.last + 2.0;

    pageinc = wave_pane_width_in_time_units (window);
    hadj->page_size = hadj->page_increment = (pageinc >= 1.0) ? pageinc : 1.0;
    hadj->step_increment = (window->traces.scale.time_units_per_pixel >= 1.0)
    	? window->traces.scale.time_units_per_pixel : 1.0;

    if (hadj->page_size >= (hadj->upper - hadj->lower))
        hadj->value = hadj->lower;
    if (hadj->value + hadj->page_size > hadj->upper)
    {
        hadj->value = hadj->upper - hadj->page_size;
        if (hadj->value < hadj->lower)
            hadj->value = hadj->lower;
    }
}


void on_time_menu_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    /* Does nothing for now */
}

void on_time_move_to_time_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    char *new_time;

    if (wave_help ("Move To Time", "scrolls the waveform display such that the left border is the time entered in the dialog"))
        return;

	gtk_widget_show (create_move_dialog ());
}

static void move_dialog_ok (GtkWidget *widget)
{
	char *time_str = gtk_entry_get_text (GTK_ENTRY (lookup_widget (widget, "move_entry")));
	TimeType new_time = wave_parse_time (time_str, wave_config->time_dimension);
    TimeType page_increment;
	char canonical_time_value[WAVE_MAX_TIME_LENGTH];

	new_time = CLAMP (new_time, tims.first, tims.last);

	GTK_ADJUSTMENT (wave_window->traces.hslider)->value = new_time;

	page_increment = wave_pane_width_in_time_units (wave_window);
	if (new_time < tims.last - page_increment + 1)
	{
		tims.timecache = new_time;
	} else
	{
		tims.timecache = tims.last - page_increment + 1;
		if (tims.timecache < tims.first)
			tims.timecache = tims.first;
	}

	wave_format_time (canonical_time_value, tims.timecache, wave_config->time_dimension);
	wave_log (FALSE, "Moved to time: %s\n", canonical_time_value);

	update_time ();
}

void on_move_dialog_cancel_clicked (GtkButton *button, gpointer user_data)
{
	on_move_dialog_delete_event (widget_top_parent (GTK_WIDGET (button)), NULL, NULL);
}

void on_move_dialog_ok_clicked (GtkButton *button, gpointer user_data)
{
	move_dialog_ok (GTK_WIDGET (button));
	on_move_dialog_delete_event (widget_top_parent (GTK_WIDGET (button)), NULL, NULL);
}

void on_move_entry_activate (GtkEditable *editable, gpointer user_data)
{
	move_dialog_ok (GTK_WIDGET (editable));
	on_move_dialog_delete_event (widget_top_parent (GTK_WIDGET (editable)), NULL, NULL);
}

gboolean on_move_dialog_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	gtk_widget_destroy (widget);
	return TRUE;
}

void on_time_zoom_zoom_in_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_in (NULL, user_data);
}

void on_zoom_out_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_out (NULL, user_data);
}

void on_time_zoom_best_fit_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_fit (NULL, user_data);
}

void on_time_zoom_start_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_start (NULL, user_data);
}

void on_time_zoom_end_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_end (NULL, user_data);
}

void on_time_zoom_undo_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_undo (NULL, user_data);
}

void on_time_zoom_redo_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    clicked_zoom_redo (NULL, user_data);
}

void on_time_step_step_left_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    on_step_left_clicked (NULL, user_data);
}

void on_time_step_step_right_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    on_step_right_clicked (NULL, user_data);
}

void on_time_page_page_left_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    on_page_left_clicked (NULL, user_data);
}

void on_time_page_page_right_menuitem_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    on_page_right_clicked (NULL, user_data);
}

void on_page_left_clicked (GtkButton * button, gpointer user_data)
{
    TimeType ntinc, ntfrac;

    if (wave_help ("Page Left", "scrolls the display window left one page worth of data. The net action is that the data scrolls right a page"))
        return;

    ntinc = wave_pane_width_in_time_units (wave_window);

    /* really don't need this var but speed of ui code is human dependent */
    ntfrac = ntinc * page_divisor;
    if (ntfrac < 1 || ntinc < 1)
        ntfrac = ntinc = 1;

    if (tims.start - ntfrac > tims.first)
        tims.timecache = tims.start - ntfrac;
    else
        tims.timecache = tims.first;
    wave_reader_load_times (wave_reader, tims.timecache, tims.start);

    GTK_ADJUSTMENT (wave_window->traces.hslider)->value = tims.timecache;
    update_time ();
}

void on_page_right_clicked (GtkButton * button, gpointer user_data)
{
    TimeType ntinc, ntfrac;

    if (wave_help ("Page Right", "scrolls the display window right one page worth of data. The net action is that the data scrolls left a page"))
        return;

    ntinc = wave_pane_width_in_time_units (wave_window);

    ntfrac = ntinc * page_divisor;
    if (ntfrac < 1 || ntinc < 1)
        ntfrac = ntinc = 1;

    if (tims.start + ntfrac < tims.last - ntinc + 1)
    {
        tims.timecache = tims.start + ntfrac;
    } else
    {
        tims.timecache = tims.last - ntinc + 1;
        if (tims.timecache < tims.first)
            tims.timecache = tims.first;
    }
    wave_reader_load_times (wave_reader, tims.start, tims.timecache);

    GTK_ADJUSTMENT (wave_window->traces.hslider)->value = tims.timecache;
    update_time ();
}

static gboolean in_start_set = FALSE;
static gboolean in_end_set = FALSE;
static gboolean in_step_size_set = FALSE;

/* wave_set_{start,end}_time : set start/end time entry text to the given time */
void wave_set_start_time (TimeType start_time)
{
    GtkWidget *w;
    char buf[WAVE_MAX_TIME_LENGTH];

    tims.start = start_time;
    if (! wave_have_main_window)
        return;
    w = lookup_widget (wave_window->window, "from_entry");
    wave_format_time (buf, start_time, wave_config->time_dimension);
    in_start_set = TRUE;
    gtk_entry_set_text (GTK_ENTRY (w), buf);
    in_start_set = FALSE;
}
void wave_set_end_time (TimeType end_time)
{
    GtkWidget *w;
    char buf[WAVE_MAX_TIME_LENGTH];

    tims.end = end_time;
    if (! wave_have_main_window)
        return;
    w = lookup_widget (wave_window->window, "to_entry");
    wave_format_time (buf, end_time, wave_config->time_dimension);
    in_end_set = TRUE;
    gtk_entry_set_text (GTK_ENTRY (w), buf);
    in_end_set = FALSE;
}

/* wave_set_step_size : set the step size to the given value/dimension */
void wave_set_step_size (TimeType time, char time_dimension, WaveWindow *window)
{
	char buf[WAVE_MAX_TIME_LENGTH];

	wave_format_time (buf, time, time_dimension);
	gtk_entry_set_text (GTK_ENTRY (lookup_widget (window->window,
		"step_size_entry")), buf);
	step_length = time;
}

void on_from_entry_changed (GtkEditable * editable, gpointer user_data)
{
    TimeType t, pageinc;
    GtkEntry *e = GTK_ENTRY (editable);
    char buf[WAVE_MAX_TIME_LENGTH];

    if (in_start_set)
        return;

    t = wave_parse_time (gtk_entry_get_text (e), wave_config->time_dimension);
    if (t < tims.first || t > tims.last)
    {
        wave_set_start_time (tims.start);
    } else
    {
        GTK_ADJUSTMENT (wave_window->traces.hslider)->value = t;

        pageinc = wave_pane_width_in_time_units (wave_window);
        if (t < tims.last - pageinc + 1)
        {
            tims.timecache = t;
        } else
        {
            tims.timecache = tims.last - pageinc + 1;
            if (tims.timecache < tims.first)
                tims.timecache = tims.first;
        }

        wave_format_time (buf, tims.timecache, wave_config->time_dimension);
        wave_log (FALSE, "Moved to time: %s\n", buf);

        update_time ();
    }
}

void on_to_entry_changed (GtkEditable * editable, gpointer user_data)
{
    TimeType t, pageinc;
    GtkEntry *e = GTK_ENTRY (editable);
    char buf[WAVE_MAX_TIME_LENGTH];

    if (in_start_set)
        return;

    t = wave_parse_time (gtk_entry_get_text (e), wave_config->time_dimension);

    if (t < tims.first || t > tims.last)
    {
        wave_set_end_time (tims.end);
    } else
    {
        GTK_ADJUSTMENT (wave_window->traces.hslider)->value = t;

        pageinc = wave_pane_width_in_time_units (wave_window);
        t -= pageinc;           /* As though we're setting the from time */

        if (t < tims.last - pageinc + 1)
        {
            tims.timecache = t;
        } else
        {
            tims.timecache = tims.last - pageinc + 1;
            if (tims.timecache < tims.first)
                tims.timecache = tims.first;
        }

        wave_format_time (buf, tims.timecache, wave_config->time_dimension);
        wave_log (FALSE, "Moved to time: %s\n", buf);

        update_time ();
    }
}

void on_step_size_entry_changed (GtkEditable * editable, gpointer user_data)
{
    GtkEntry *e = GTK_ENTRY (editable);
    TimeType t;
    char buf[WAVE_MAX_TIME_LENGTH];

    if (in_step_size_set)
        return;

    t = wave_parse_time (gtk_entry_get_text (e), wave_config->time_dimension);

    if (!(t > (tims.last - tims.first)))
    {
        wave_format_time (buf, t, wave_config->time_dimension);
        in_step_size_set = TRUE;
        gtk_entry_set_text (e, buf);
        in_step_size_set = FALSE;
        step_length = t;
    }
}

void on_step_left_clicked (GtkButton * button, gpointer user_data)
{
    GtkAdjustment *hadj;
    gfloat inc;
    TimeType ntinc;

    if (wave_help ("Shift Left", "scrolls the display window left one tick worth of data. The net action is that the data scrolls right a tick"))
        return;

    inc = step_length;

    hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);
    if (hadj->value - inc > tims.first)
        hadj->value = hadj->value - inc;
    else
        hadj->value = tims.first;

    ntinc = (TimeType) inc;
    if (tims.start - ntinc > tims.first)
        tims.timecache = tims.start - ntinc;
    else
        tims.timecache = tims.first;

    update_time ();
}

void on_step_right_clicked (GtkButton * button, gpointer user_data)
{
    GtkAdjustment *hadj;
    gfloat inc;
    TimeType ntinc, pageinc;

    if (wave_help ("Shift Right", "scrolls the display window right one tick worth of data. The net action is that the data scrolls left a tick"))
        return;

    inc = step_length;
    ntinc = (TimeType) inc;

    hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);
    if ((hadj->value + inc) < tims.last)
        hadj->value = hadj->value + inc;
    else
        hadj->value = tims.last - inc;

	pageinc = wave_pane_width_in_time_units (wave_window);

    if (tims.start + ntinc < tims.last - pageinc + 1)
        tims.timecache = tims.start + ntinc;
    else
    {
        tims.timecache = tims.last - pageinc + 1;
        if (tims.timecache < tims.first)
            tims.timecache = tims.first;
    }

    update_time ();
}

void service_dragzoom (TimeType time1, TimeType time2)
{
    gdouble estimated;
    TimeType temp;
    GtkAdjustment *hadj;
    TraceEnt *t;

    if (time2 < time1)
    {
        temp = time1;
        time1 = time2;
        time2 = temp;
    }
    wave_reader_load_times (wave_reader, time1, time2);
    if (time2 > time1)
    {                           /* ensure at least 1 tick */
        estimated = wave_calculate_zoom_for_span (time1, time2);

        tims.prevzoom = tims.zoom;
        tims.timecache = tims.laststart = time1;
        wave_set_start_time (tims.timecache);

        for (t = traces.first; t; t = t->next)
        {                       /* have to nuke string refs so printout is ok! */
            if (t->asciivalue)
            {
                g_free (t->asciivalue);
                t->asciivalue = NULL;
            }
        }
        for (t = traces.buffer; t; t = t->next)
        {
            if (t->asciivalue)
            {
                g_free (t->asciivalue);
                t->asciivalue = NULL;
            }
        }

        update_markertime (tims.marker = -1);
        wave_window->signals.window_width_dirty = TRUE;
        wave_max_signal_length (wave_window);

        hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);
        hadj->value = time1;

        wave_zoom (estimated, wave_window);
        tims.zoom = estimated;

        adjust_zoom ();
        WAVE_DEBUG (wave_log (FALSE, "WAVE_DEBUG: Drag Zoom\n"));
    }
}

void clicked_zoom_out (GtkButton * button, gpointer user_data)
{
    TimeType middle = 0, width;

    if (wave_help ("Zoom Out", "is used to decrease the zoom factor"))
        return;

    if (wave_flags.center_zooms_flag)
    {
        if (tims.marker < 0 || tims.marker < tims.first || tims.marker > tims.last)
        {
            if (tims.end > tims.last)
                wave_set_end_time (tims.last);
            middle = (tims.start / 2) + (tims.end / 2);
            if (tims.start & 1 && tims.end & 1)
                middle++;
        } else
        {
            middle = tims.marker;
        }
    }

    tims.prevzoom = tims.zoom;

    tims.zoom--;
    wave_zoom (tims.zoom, wave_window);

    if (wave_flags.center_zooms_flag)
    {
        width = wave_pane_width_in_time_units (wave_window);
        wave_set_start_time (middle - width / 2);
        if (tims.start + width > tims.last)
            wave_set_start_time (tims.last - width);
        if (tims.start < tims.first)
            wave_set_start_time (tims.first);
        GTK_ADJUSTMENT (wave_window->traces.hslider)->value = tims.timecache = tims.start;
    } else
    {
        tims.timecache = 0;
    }

    adjust_zoom ();
}

void clicked_zoom_in (GtkButton * button, gpointer user_data)
{
    TimeType middle = 0, width;

    if (wave_help ("Zoom In", "is used to increase the zoom factor"))
        return;

    if (tims.zoom >= 0)
    {                           /* otherwise it's ridiculous and can */
        return;                 /* cause overflow problems in the scope. */
    }

    if (wave_flags.center_zooms_flag)
    {
        if (tims.marker < 0 || tims.marker < tims.first || tims.marker > tims.last)
        {
            if (tims.end > tims.last)
                wave_set_end_time (tims.last);
            middle = (tims.start / 2) + (tims.end / 2);
            if (tims.start & 1 && tims.end & 1)
                middle++;
        } else
        {
            middle = tims.marker;
        }
    }

    tims.prevzoom = tims.zoom;

    tims.zoom++;
    wave_zoom (tims.zoom, wave_window);

    if (wave_flags.center_zooms_flag)
    {
		width = wave_pane_width_in_time_units (wave_window);
        wave_set_start_time (middle - width / 2);
        if (tims.start + width > tims.last)
            wave_set_start_time (tims.last - width);
        if (tims.start < tims.first)
            wave_set_start_time (tims.first);
        GTK_ADJUSTMENT (wave_window->traces.hslider)->value = tims.timecache = tims.start;
    } else
    {
        tims.timecache = 0;
    }

    adjust_zoom ();
}

void clicked_zoom_fit (GtkButton * button, gpointer user_data)
{
    gdouble estimated;

    if (wave_help ("Zoom Fit", "tries to fit the whole trace on screen. Note that the trace may be more or less than a whole screen since this isn't a \"perfect fit\""))
        return;

    estimated = wave_calculate_zoom_for_span (tims.first, tims.last);

    tims.prevzoom = tims.zoom;
    tims.timecache = 0;

    wave_zoom (estimated, wave_window);
    tims.zoom = estimated;

    adjust_zoom ();
}

void clicked_zoom_start (GtkButton * button, gpointer user_data)
{
    GtkAdjustment *hadj;

    if (wave_help ("Zoom To Start", "is used to jump scroll to the trace's beginning"))
        return;

    hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);
    hadj->value = tims.timecache = tims.first;
    update_time ();
}

void clicked_zoom_end (GtkButton * button, gpointer user_data)
{
    GtkAdjustment *hadj;
    TimeType ntinc;

    if (wave_help ("Zoom To End", "is used to jump scroll to the trace's end"))
        return;

	ntinc = wave_pane_width_in_time_units (wave_window);

    tims.timecache = tims.last - ntinc + 1;
    if (tims.timecache < tims.first)
        tims.timecache = tims.first;

    hadj = GTK_ADJUSTMENT (wave_window->traces.hslider);
    hadj->value = tims.timecache;
    update_time ();
}

void clicked_zoom_undo (GtkButton * button, gpointer user_data)
{
    gdouble temp = tims.zoom;

    if (wave_help ("Zoom Undo", "is used to revert to the previous zoom value used"))
        return;

    tims.zoom = tims.prevzoom;
    tims.prevzoom = temp;
    tims.timecache = 0;
    wave_zoom (tims.zoom, wave_window);

    adjust_zoom ();
}

void clicked_zoom_redo (GtkButton * button, gpointer user_data)
{
    wave_log (TRUE, "That function is not implemented!\n");
}
