/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `out.c'
    Synopsis (EPIC) TimeMill .out file parser

    $Id: out.c,v 1.65 2003/11/09 22:21:34 bardslea Exp $
*/

#include "out.h"

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "utils.h"
#include "alloc.h"
#include "fileio.h"
#include "log.h"
#include "gui_cb.h"
#include "draw.h"
#include "analog.h"
#include "tracepane.h"

#undef VCD_BSEARCH_IS_PERFECT

#define OUT_BSIZ 32768          /* size of getch() emulation buffer--this val should be ok */

typedef struct OutSymbol_
{
    struct OutSymbol_ *root, *chain;

    struct OutSymbol_ *next;
    char *name;
    char *id;
    unsigned char vartype;
    int msi, lsi;
    int size;
    char *value;
    char branch_index;
    char one_of_set;
    Node **narray;
}
OutSymbol;

typedef struct
{
    char *name;
    double vdd;
    double time_resolution;
    double current_resolution;
    double voltage_resolution;
    double math_resolution;
    double high_threshold;
    double low_threshold;
    long num_of_nodes;
    long num_of_elems;
    long num_of_extras;
    char bus_notation[3];
    char hier_separator;
    char cse;
}
OutFileInfo;

static char make_out_save_file = 0;
static FILE *out_save_handle = NULL;
const static char *out_autosave_name = "out_autosave.sav";

static FileDesc *out_handle = NULL;
static char out_is_compressed = 0;
static OutFileInfo *this_file;

static void out_build_symbols (void);

static int outbyteno = 0;       /* really should be size_t, but this is only used for debugging mangled traces */
static int header_over = 0;

/*  static int dumping_off = 0; */
static TimeType out_start_time = -1;
static TimeType out_end_time = -1;
static TimeType out_scaled_start_time = -1;
static TimeType out_scaled_end_time = -1;
static TimeType out_current_time = -1;
static char out_hier_delimiter[2] = { 0, 0 }; /* fill in after rc reading code */

static OutSymbol *pv = NULL, *rooto = NULL;

enum Tokens
{
    T_VDD, T_TIMERESOLUTION, T_CURRESOLUTION, T_VOLRESOLUTION,
    T_MATHRESOLUTION, T_HIGHTHRESH, T_LOWTHRESH, T_NNODES, T_NELEMS,
    T_EXTRA, T_BUSN, T_HIERSEP, T_CASE, T_INDEX, T_BIO, T_OCT, T_HEX, T_BIG,
    T_HEAD,
    T_VALUE, T_TIME, T_UNKNOWN_KEY, T_EOF
};

static char *out_tokens[] = {
    "vdd", "time_resolution", "current_resolution", "voltage_resolution",
    "math_resolution", "high_threshold", "low_threshold", "nnodes", "nelems",
    "extra_nodes", "bus_notation", "hier_separator", "case", "index", "bio",
    "oct", "hex", "big", "", "", "", "", ""
};

#define NUM_TOKENS 18

static int t_max_str = 1024;    /* was originally a const..now it reallocs */
static char *yytext = NULL;
static int yylen = 0;

static OutSymbol *outsymroot = NULL, *outsymcurr = NULL;
static OutSymbol **sorted = NULL;

enum Indices
{ N_ANALOGUE, N_INSTANTANEOUS, N_AVECURR, N_RMS, N_DIDT,
    N_MATH, N_AVEMATH, N_RMSMATH, N_DMDT, N_LOGIC
};

/*
static char *indices[] = { "v", "i", "I", "I_rms", "di/dt", "m", "M", "M_rms", "dm/dt", "l" };

#define NUM_INDICES 10
*/

static int numsyms = 0;

/*
    bsearch compare
*/
static int outsymbsearchcompare (const void *s1, const void *s2)
{
    char *v1;
    OutSymbol *v2;

    v1 = (char *) s1;
    v2 = *((OutSymbol **) s2);

    return (strcmp (v1, v2->id));
}

/*
    actual bsearch
*/
static OutSymbol *bsearch_out (char *key)
{
    OutSymbol **v;
    OutSymbol *t;

    v = (OutSymbol **) bsearch (key, sorted, numsyms, sizeof (OutSymbol *), outsymbsearchcompare);

    if (v)
    {
#ifndef VCD_BSEARCH_IS_PERFECT
        for (;;)
        {
            t = *v;

            if ((v == sorted) || (strcmp ((*(--v))->id, key)))
            {
                return (t);
            }
        }
#else
        return (*v);
#endif
    } else
    {
        return (NULL);
    }
}

/*
    sort on outsymbol pointers
*/
static int outsymcompare (const void *s1, const void *s2)
{
    OutSymbol *v1, *v2;

    v1 = *((OutSymbol **) s1);
    v2 = *((OutSymbol **) s2);

    return (strcmp (v1->id, v2->id));
}

/*
    create sorted (by id) table
*/
static void create_sorted_table (void)
{
    OutSymbol *v;
    OutSymbol **pnt;

    fprintf (stderr, "Sorted table!\n");

    if (sorted)
    {
        g_free (sorted);        /* this means we saw a 2nd enddefinition chunk! */
    }

    if (numsyms)
    {
        pnt = sorted = g_new (OutSymbol *, numsyms);
        for (v = outsymroot; v; v = v->next)
        {
            *(pnt++) = v;
        }

        qsort (sorted, numsyms, sizeof (OutSymbol *), outsymcompare);
    }
}

static char *outbuf = NULL, *ost = NULL, *oend = NULL;

static void getch_alloc (void)
{
    oend = ost = outbuf = g_new (char, OUT_BSIZ);
}

static void getch_free (void)
{
    if (outbuf)
        g_free (outbuf);
    outbuf = ost = oend = NULL;
}

static long lastpos;

static int getch_fetch (void)
{
    size_t rd;

    if (FILE_DESC_EOF (out_handle) || errno)
    {
        lastpos += (oend - outbuf);
        return (-1);
    }
    outbyteno += (oend - outbuf);
    lastpos = FILE_DESC_TELL (out_handle);

    rd = FILE_DESC_READ (outbuf, sizeof (char), OUT_BSIZ, out_handle);

    oend = (ost = outbuf) + rd;

    if (!rd)
        return (-1);

    return ((int) (*(ost++)));
}

#define getch() ((ost!=oend)?((int)(*(ost++))):(getch_fetch()))

static char out_val_convert (int val)
{
    switch (val)
    {
    case 0:
    case 3:
    case 6:
        return '0';
        break;

    case 1:
    case 4:
    case 7:
        return '1';
        break;

    case 5:
        return 'Z';
        break;

    case 2:
    case 8:
    case 9:
    case 10:
        return 'X';
        break;

    default:
        return 0;
        break;
    }
}

static int get_str (void)
{
    char ch;

    for (;;)
    {
        ch = getch ();
        if (ch < 0)
            return -1;
        if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
            continue;
        break;
    }                           /* Remove White space *//* FIXME */

    for (yylen = 0;;)
    {
        if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
            break;
        yytext[yylen] = ch;
        yylen++;
        if (yylen == t_max_str)
        {
            t_max_str *= 2;
            yytext = (char *) g_realloc (yytext, t_max_str + 1);
        }
        ch = getch ();
    }                           /* Get String */

    yytext[yylen] = '\0';       /* Terminate and adjust length */
    return 0;
}

static int get_token (void)     /* Returns Token and 1 argument */
{
    char *cmnd;
    int ch;
    int i = 0;

    yylen = 0;

    for (;;)
    {
        ch = getch ();
        if (ch < 0)
            return (T_EOF);
        if (ch != ' ' && ch != '\t' && ch != '\n' && ch != '\r')
            break;
    }

    switch (ch)
    {
    case ';':
        do
        {
            ch = getch ();
        }
        while (ch != EOF && ch != '\n');

        return T_HEAD;          /* Ignore Header */
        break;

    case '.':
        ch = getch ();
        yylen = 0;
        for (;;)
        {
            if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
                break;
            yytext[yylen] = ch;
            yylen++;

            if (yylen == t_max_str)
            {
                yytext = (char *) g_realloc (yytext, (t_max_str = t_max_str * 2) + 1);
            }

            ch = getch ();
        }                       /* Get In value */

        cmnd = g_new (char, yylen + 1);

        strncpy (cmnd, yytext, yylen);
        cmnd[yylen] = '\0';     /* Terminator */

        for (;;)
        {
            ch = getch ();
            if (ch < 0)
                return (T_EOF);
            if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
                continue;
            break;
        }                       /* SKIP WHITE SPACE */

        for (yylen = 0;;)
        {
            if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
                break;
            yytext[yylen] = ch;
            yylen++;

            if (yylen == t_max_str)
            {
                yytext = (char *) g_realloc (yytext, (t_max_str = t_max_str * 2) + 1);
            }

            ch = getch ();
        }                       /* Get 1st Arg */

        yytext[yylen] = 0;      /* Terminate and adjust length */

        for (i = 0; i < NUM_TOKENS; i++)
        {
            if (!strcmp (cmnd, out_tokens[i]))
            {
                return i;
            }
        }                       /* Calculate Token Value */

        return (T_UNKNOWN_KEY);

        break;

    default:
        for (;;)
        {
            if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
                break;
            yytext[yylen] = ch;
            yylen++;

            if (yylen == t_max_str)
            {
                yytext = (char *) g_realloc (yytext, (t_max_str = t_max_str * 2) + 1);
            }

            ch = getch ();
        }                       /* Get First? integer */

        yytext[yylen] = 0;      /* Terminate and adjust length */
        if (ch == '\n' || ch == '\r')
        {
            return T_TIME;
        }
        return T_VALUE;         /* Need to set to end of line when handling T_VALUE */
    }
}

static void get_bus (OutSymbol * o, char *name, int nmesze)
{

    int nme, i;
    char *tmp, ch;
    tmp = g_new (char, 5);

    for (nme = 0;; nme++)
    {
        ch = name[nme];
        if (ch == this_file->bus_notation[0])
            break;
    }
    o->name = g_new (char, nme + 1);

    strncpy (o->name, name, nme);
    o->name[nme] = '\0';

    nme++;
    ch = name[nme];
    i = 0;
    while (ch != this_file->bus_notation[2])
    {
        tmp[i] = ch;
        i++;
        nme++;
        ch = name[nme];
    }
    tmp[i] = '\0';
    o->lsi = o->msi = atoi (tmp);
    o->size = 1;
    o->root = o->chain = NULL;
    o->one_of_set = 1;
}

static void parse_valuechange (int dir)
{
    OutSymbol *o;
    int v;
    double *va;

    if (!sorted)
    {
        fprintf (stderr, "Symbols should be sorted\n");
        exit (1);
    }
    o = bsearch_out (yytext);
    if (!o)
    {
        fprintf (stderr, "Unknown Out file Identifier : '%s'\n", yytext);
    }

    switch (o->vartype)
    {
    case N_LOGIC:
        get_str ();
        v = atoi (yytext);
        o->value[0] = out_val_convert (v);
        wave_node_insert_event (o->narray[0], out_current_time * wave_config->time_scale, NODE_FLAGS_DIGITAL, o->value[0], NULL);
        break;
    default:
        get_str ();
        va = g_new (double, 1);

        *va = atof (yytext);
        wave_node_insert_event (o->narray[0], out_current_time * wave_config->time_scale, NODE_FLAGS_REAL, '\0', (char *) va);
        break;
    }
}

/*  static int decs = 0; */

static void parse_outdec (WaveReader *reader)
{
    OutSymbol *o;
    int flag = 0;

    o = g_new (OutSymbol, 1);

    /* Lines look like: .index signal_name signal_id signal_type */
    /* eg. .index irq_or.za 105 l */

    /* signal name */
    if (yytext[yylen - 1] == this_file->bus_notation[2])
    {
        get_bus (o, yytext, yylen);
        WAVE_DEBUG (wave_log (FALSE, "Found Bussed Signal %s\n", o->name));
    } else
    {
        o->name = g_new (char, yylen + 1);

        strcpy (o->name, yytext);
        o->name[yylen] = '\0';

        o->msi = o->lsi = -1;
        o->size = 1;
        o->root = NULL;
        o->chain = NULL;
        o->one_of_set = 0;
        WAVE_DEBUG (wave_log (FALSE, "Found Signal %s\n", o->name));
    }

    /* identifier */
    get_str ();
    o->id = g_new (char, yylen + 1);

    strcpy (o->id, yytext);     /* Get Identifier */
    WAVE_DEBUG (wave_log (FALSE, "With identifier %s\n", o->id));

    /* signal type */
    get_str ();

    switch (yytext[0])
    {                           /* Remove Brackets etc from around sig name */
    case 'v':
        /*  yytext[n] tests here should be tests on o->name */
        if (0)                  /* AB if (yytext[1] != '(') */
        {
            goto norm_var;
        }
        o->vartype = N_ANALOGUE;

        break;

    case 'i':
        if (yytext[1] >= '0' && yytext[1] <= '5')
        {
            if (yytext[2] != '(')
                goto norm_var;
        } else if (0)           /* AB if (yytext[1] != '(') */
        {
            goto norm_var;
        }
        o->vartype = N_INSTANTANEOUS;
        break;

    case 'I':
        if (yytext[1] == '_')
        {
            if (yytext[5] >= '0' && yytext[5] <= '5')
            {
                if (yytext[6] != '(')
                    goto norm_var;
            } else if (0)       /* AB if (yytext[5] != '(') */
            {
                goto norm_var;
            }
            o->vartype = N_RMS;
        } else
        {
            if (yytext[1] >= '0' && yytext[1] <= '5')
            {
                if (yytext[2] != '(')
                    goto norm_var;
            } else if (0)       /* AB if (yytext[1] != '(') */
            {
                goto norm_var;
            }
            o->vartype = N_AVECURR;
        }
        break;

    case 'd':
        if (yytext[5] >= '0' && yytext[5] <= '5')
        {
            if (yytext[6] != '(')
                goto norm_var;
        } else if (yytext[5] != '(')
        {
            goto norm_var;
        }
        if (yytext[1] == 'm')
        {                       /* dm/dt... */
            o->vartype = N_DMDT;
        } else
        {                       /* di/dt... */
            o->vartype = N_DIDT;
        }
        break;

    case 'm':
        if (yytext[1] >= '0' && yytext[1] <= '5')
        {
            if (yytext[2] != '(')
                goto norm_var;
        } else if (yytext[1] != '(')
        {
            goto norm_var;
        }
        o->vartype = N_MATH;
        break;

    case 'M':
        if (yytext[1] == '_')
        {
            if (yytext[5] >= '0' && yytext[5] <= '5')
            {
                if (yytext[6] != '(')
                    goto norm_var;
            } else if (yytext[5] != '(')
            {
                goto norm_var;
            }
            o->vartype = N_RMSMATH;
        } else
        {
            if (yytext[1] >= '0' && yytext[1] <= '5')
            {
                if (yytext[2] != '(')
                    goto norm_var;
            } else if (yytext[1] != '(')
            {
                goto norm_var;
            }
            o->vartype = N_AVEMATH;
        }
        break;

    default:
      norm_var:
        o->vartype = N_LOGIC;
        flag = 1;
        break;
    }

    if (!flag)
    {
        yytext[yylen] = ')';
        yylen++;
        yytext[yylen] = '\0';
    }

    o->value = g_new (char, o->size + 1);
    o->value[o->size] = '\0';

/*
 * if (o->size > 1) {
 */
    o->narray = g_new (Node *, o->size);
/*
 * } else {
 *     o->narray = (Node **) malloc_2(sizeof(Node *));
 * }
 */
    {
    	TraceDrawingInfo *drawing_info = NULL;
    	NodeFlags node_type = (o->size > 1 ? NODE_FLAGS_VECTOR : NODE_FLAGS_DIGITAL);

        int i;
		for (i = 0; i < o->size; i++)
			o->value[i] = 'X';

        /* AB FIXME, assuming o->size == 1 */
        /* Set the drawer appropriately */
        switch (o->vartype)
        {
        case N_ANALOGUE:
            drawing_info = new_trace_drawing_info (&analog_trace_drawer);
            drawing_info->context = make_analog_trace_drawer_context (0.0,
              this_file->vdd / this_file->voltage_resolution);
            node_type = NODE_FLAGS_REAL;
            break;
        case N_INSTANTANEOUS:
        case N_AVECURR:
        case N_RMS:
            drawing_info = new_trace_drawing_info (&analog_trace_drawer);
            drawing_info->context = make_analog_trace_drawer_context (0.0, 100);
            node_type = NODE_FLAGS_REAL;
            break;
        case N_LOGIC:          /* Don't set a drawer, use digital drawer */
            break;
        default:
            wave_log (FALSE, "Not sure how to render trace %s, using analog trace drawer\n", o->name);
            drawing_info = new_trace_drawing_info (&analog_trace_drawer);
            drawing_info->context = make_analog_trace_drawer_context (0.0, 1.0);
            node_type = NODE_FLAGS_REAL;
            break;
        }

        o->narray[0] = wave_new_node (o->name, reader, node_type);
        o->narray[0]->drawing_info = drawing_info;
    }

    if (!outsymroot)
    {
        outsymroot = outsymcurr = o;
    } else
    {
        outsymcurr->next = o;
        outsymcurr = o;
    }

    if (numsyms != 0 && (numsyms % 10000) == 0)
    {
        printf ("%d signals processed.\n", numsyms);
    }

    numsyms++;

    if (out_save_handle)
    {
        if (o->msi == o->lsi)
        {
            if (o->vartype != N_LOGIC)
            {
                fprintf (out_save_handle, "%s\n", o->name);
            } else if (o->msi >= 0)
            {
                fprintf (out_save_handle, "%s[%d]\n", o->name, o->msi);
            } else
            {
                fprintf (out_save_handle, "%s\n", o->name);
            }
        } else
        {
            fprintf (out_save_handle, "%s[%d:%d]\n", o->name, o->msi, o->lsi);
        }
    }
}

static void out_parsedefs (WaveReader *reader)
{
    int tok, j;
    double time_res;

    for (;;)
    {
        tok = get_token ();
        switch (tok)
        {

        case T_VDD:
            this_file->vdd = atof (yytext);
            break;

        case T_TIMERESOLUTION:
            this_file->time_resolution = (time_res = atof (yytext));
            j = 0;

            /* FIXME, make me into a function */
            if (time_res >= 1)
            {
                wave_config->time_dimension = 'n';
                wave_config->time_scale = (long) time_res;
            } else
            {
                while (time_res < 1)
                {
                    time_res *= 10;
                    j++;
                }
                switch (j)
                {
                case 1:
                    wave_config->time_dimension = 'p';
                    wave_config->time_scale = 100 * (long) time_res;
                    break;

                case 2:
                    wave_config->time_dimension = 'p';
                    wave_config->time_scale = 10 * (long) time_res;
                    break;

                case 3:
                    wave_config->time_dimension = 'p';
                    wave_config->time_scale = 1 * (long) time_res;
                    break;

                case 4:
                    wave_config->time_dimension = 'f';
                    wave_config->time_scale = 100 * (long) time_res;
                    break;

                case 5:
                    wave_config->time_dimension = 'f';
                    wave_config->time_scale = 10 * (long) time_res;
                    break;

                case 6:
                    wave_config->time_dimension = 'f';
                    wave_config->time_scale = 1 * (long) time_res;
                    break;

                default:
                    fprintf (stderr, "Time Dimension too small, exiting ..\n");
                    exit (1);
                    break;
                }
                if (wave_config->time_scale < 1)
                    wave_config->time_scale = 1;
            }
            break;

        case T_CURRESOLUTION:
            this_file->current_resolution = atof (yytext);
            break;
        case T_VOLRESOLUTION:
            this_file->voltage_resolution = atof (yytext);
            break;
        case T_MATHRESOLUTION:
            this_file->math_resolution = atof (yytext);
            break;
        case T_HIGHTHRESH:
            this_file->high_threshold = atof (yytext);
            break;
        case T_LOWTHRESH:
            this_file->low_threshold = atof (yytext);
            break;
        case T_INDEX:
            /* WAAH */
            parse_outdec (reader);
            break;
        case T_NNODES:
            this_file->num_of_nodes = atol (yytext);
            break;
        case T_NELEMS:
            this_file->num_of_elems = atol (yytext);
            break;
        case T_EXTRA:
            this_file->num_of_extras = atol (yytext);
            break;
        case T_BUSN:
            this_file->bus_notation[0] = (char) *yytext;
            get_str ();
            this_file->bus_notation[1] = (char) *yytext;
            get_str ();
            this_file->bus_notation[2] = (char) *yytext;
            break;
        case T_HIERSEP:
            this_file->hier_separator = (char) *yytext;
            out_hier_delimiter[0] = (hier_delimiter = (char) *yytext);
            break;
        case T_CASE:
            this_file->cse = (char) *yytext;
            break;
        case T_BIO:
        case T_OCT:
        case T_HEX:
        case T_BIG:
            break;
        case T_HEAD:
            break;
        case T_TIME:
            if (header_over == 0)
            {
                header_over = 1;
                create_sorted_table ();
                if (!sorted)
                {
                    fprintf (stderr, "No OUT symbols to transform. Exiting...\n");
                    exit (1);
                }
            }
            fprintf (stderr, "Returned from header read after seeing first time\n");
            return;
            break;
        case T_VALUE:
        case T_UNKNOWN_KEY:
            break;
        case T_EOF:
            return;
            break;
        default:
            fprintf (stderr, "Invalid file contents `%s'\n", yytext);

            break;
        }
    }
}

static void out_parse (WaveReader *reader)
{
    int tok;
    TimeType time;

    out_parsedefs (reader);           /* Parse Declarations */

    for (;;)
	{
		tok = get_token ();
		switch (tok)
		{
		case T_HEAD:
			break;
		case T_TIME:
			time = wave_str_to_time_type (yytext, NULL);

			if (out_start_time < 0)
				out_start_time = time;

			if (out_end_time < time)
				out_end_time = time;

			/* Set Times */
			out_current_time = time;
			break;
		case T_VALUE:
			parse_valuechange (1);
			break;
		case T_UNKNOWN_KEY:
			break;
		case T_EOF:
			return;
			break;
		default:
			fprintf (stderr, "Invalid file contents `%s'\n", yytext);

			break;
		}
	}
}

static void out_build_symbols (void)
{
    int i, j;
    int max_slen = -1;
    OutSymbol *v, *vprime;
	Symbol *curnode = NULL;
    int msi;
    int delta;
    char *str = NULL;
    int slen;
    int substnode;

    for (i = 0; i < numsyms; i++)
    {
        v = sorted[i];
        msi = v->msi;
        delta = ((v->lsi - v->msi) < 0) ? -1 : 1;
        substnode = 0;

        slen = strlen (v->name);
        if (slen > max_slen)
        {
            max_slen = slen;
            str = wave_alloca (slen + 32);
        }
        strcpy (str, v->name);

        if (v->msi >= 0)
        {
            strcpy (str + slen, out_hier_delimiter);
            slen++;
        }

        vprime = bsearch_out (v->id);
        if (vprime != v)
        {                       /* hash mish means dup net */
            if (v->size != vprime->size)
            {
                wave_log (TRUE, "ERROR: Duplicate IDs with differing width: %s %s\n", v->name, vprime->name);
            } else
            {
                substnode = 1;
            }
        }

        if (v->size == 1 && v->vartype == N_LOGIC)
        {
            Symbol *s;

            for (j = 0; j < v->size; j++)
            {
                if (v->msi >= 0)
                {
                    sprintf (str + slen - 1, "[%d]", msi);
                }
                if (!symfind (str))
                {
                    s = symadd (str, hash (str));

                    s->n = v->narray[j];
                    if (substnode)
                    {
                        Node *n = s->n;
                        Node *n2 = vprime->narray[j];

						wave_alias_node (n, n2);
                    }

                    s->n->nname = s->name;
                    if (!firstnode)
                    {
                        firstnode = curnode = s;
                    } else
                    {
                        curnode->next_in_file = s;
                        curnode = s;
                    }

                    wave_num_symbols++;
                    WAVE_DEBUG (wave_log (FALSE, "Added: %s\n", str));
                } else
                {
                    wave_log (TRUE, "Warning: %s is a duplicate net name id = %s.\n", str, v->id);
                }
                msi += delta;
            }

        } else
        {                       /* atomic vector */
            if (v->vartype == N_LOGIC)
            {
                sprintf (str + slen - 1, "[%d:%d]", v->msi, v->lsi);
            } else
            {
                *(str + slen - 1) = 0;
            }
            if (!symfind (str))
            {
                Symbol *s = symadd (str, hash (str));

                s->n = v->narray[0];
                if (substnode)
                {
                    Node *n = s->n, *n2 = vprime->narray[0];

                    /* nname stays same */
                        /* NOHISTENT FIXME
                    n->head = n2->head;
                    n->curr = n2->curr;
                    */
                    n->ext = n2->ext;
                } else
                {
                    ExtNode *en = g_new (ExtNode, 1);

                    en->msi = v->msi;
                    en->lsi = v->lsi;

                    s->n->ext = en;
                }

                s->n->nname = s->name;
                if (!firstnode)
                {
                    firstnode = curnode = s;
                } else
                {
                    curnode->next_in_file = s;
                    curnode = s;
                }

                wave_num_symbols++;
                WAVE_DEBUG (wave_log (FALSE, "Added: %s\n", str));
            } else
            {
                wave_log (TRUE, "Warning: %s is a duplicate net name id = %s.\n", str, v->id);
            }
        }
    }
}

static void sort_bus (void)
{
    Symbol *s = NULL, *curroot = NULL;
    char c;
    int flen, len;
    int i;

    for (i = 0; i < wave_num_symbols; i++)
    {
        s = wave_symbols[i];
        flen = strlen (s->name);
        if (s->name[flen - 1] == this_file->bus_notation[2])
        {

            len = 0;
            c = s->name[len];
            while (c && c != this_file->bus_notation[0])
            {
                len++;
                c = s->name[len]; /* Get length of Name with out range */
            }
            if ((i == 0) || strncmp (s->name, wave_symbols[i - 1]->name, len))
            {
                s->vec_root = curroot = s;
                WAVE_DEBUG (wave_log (FALSE, "%s is the Root\n", s->name));
            } else
            {
                s->vec_root = curroot;
                WAVE_DEBUG (wave_log (FALSE, "%s is part of a bus\n", s->name));
            }

            if (i < wave_num_symbols - 1 && !strncmp (s->name, wave_symbols[i + 1]->name, len))
            {
                if (wave_symbols[i + 1]->name[len] && wave_symbols[i + 1]->name[len] == '[')
                {
                    s->vec_chain = wave_symbols[i + 1];
                    WAVE_DEBUG (wave_log (FALSE, "%s is chained to %s\n", s->name, wave_symbols[i + 1]->name));
                }
            }
        }
    }
}

static void out_sort_symbols (void)
{
    int i;
    Symbol *curnode = firstnode;

    wave_symbols = g_new (Symbol *, wave_num_symbols);

    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;
        int len;

        wave_symbols[i] = curnode;
        if ((len = strlen (subst = curnode->name)) > longestname)
            longestname = len;
        curnode = curnode->next_in_file;
        while ((ch = (*subst)))
        {
            if (ch == hier_delimiter)
            {
                *subst = 0x01;
            }                   /* forces sort at hier boundaries */
            subst++;
        }
    }
    wave_heapsort (wave_symbols, wave_num_symbols);
    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;

        subst = wave_symbols[i]->name;
        while ((ch = (*subst)))
        {
            if (ch == 0x01)
            {
                *subst = hier_delimiter;
            }                   /* restore back to normal */
            subst++;
        }
    }
    sort_bus ();
    wave_symbols_are_sorted = TRUE;

    init_tree ();
    for (i = 0; i < wave_num_symbols; i++)
    {
        build_tree_from_name (wave_symbols[i]->name, i);
    }
    treeprune (treeroot);
}

static void out_cleanup (WaveReader *reader)
{
    int i;

    if (sorted)
    {
        for (i = 0; i < numsyms; i++)
        {
            OutSymbol *v = sorted[i];

            if (v)
            {
                if (v->name)
                    g_free (v->name);
                if (v->id)
                    g_free (v->id);
                if (v->value)
                    g_free (v->value);
                if (v->narray)
                    g_free (v->narray);
                g_free (v);
            }
        }
        g_free (sorted);
        sorted = NULL;
    }

    if (out_is_compressed)
    {
        pclose ((FILE *) out_handle->handle);
    } else
    {
        FILE_DESC_CLOSE (out_handle);
    }

    if (yytext)
    {
        g_free (yytext);
        yytext = NULL;
    }
}

static void out_seek_end (WaveReader *reader)
{
    int tok, ch; 
    TimeType time;

    ch = FILE_DESC_GETC (out_handle);
    if (ch == -1)
        return;
    else
    {
        while (ch != '\n' && ch != EOF)
        {
            ch = FILE_DESC_GETC (out_handle); /* Get to "word boundary" */
        }
        getch_alloc ();
        for (;;)
		{
			tok = get_token ();
			switch (tok)
			{
			case T_HEAD:
				break;
			case T_TIME:
				time = wave_str_to_time_type (yytext, NULL);

				if (out_start_time < 0)
					out_start_time = time;

				if (out_end_time < time)
					out_end_time = time;

				/* Set Times */
				out_current_time = time;
				break;
			case T_VALUE:
				do
				{
					ch = getch ();
				}
				while (ch != '\n' && ch != EOF);
				break;
			case T_UNKNOWN_KEY:
				break;
			case T_EOF:
				goto end;
				break;
			default:
				WAVE_DEBUG (fprintf (stderr, "UNKNOWN TOKEN: %d\n", tok));
				break;
			}
		}
      end:

        tims.last = out_scaled_end_time;
        wave_set_end_time (tims.last);
        clicked_zoom_fit (NULL, NULL);
        wave_redraw_trace_pane (wave_window);
        return;
    }
}

static TimeType out_main (WaveReader *reader)
{
	const char *fname = reader->filename;

    if (make_out_save_file)
    {
        out_save_handle = fopen (out_autosave_name, "wb");
        errno = 0;
    }

    pv = rooto = NULL;

    errno = 0;                  /* reset in case it's set for some reason */
    yytext = g_new (char, t_max_str + 1);

    out_handle = file_desc_open (fname, "rt");

    if (!out_handle)
    {
        fprintf (stderr, "Error opening %s .out file '%s'.\n", out_is_compressed ? "compressed" : "", fname);
        exit (1);
    }

    getch_alloc ();
    this_file = g_new (OutFileInfo, 1);

    out_parse (reader);

    if (out_save_handle)
        fclose (out_save_handle);

    wave_log (FALSE, "[" LL_PRINTF_FORMAT "] start time.\n[" LL_PRINTF_FORMAT "] end time.\n", out_start_time, out_end_time);

    out_build_symbols ();
    out_sort_symbols ();
/* out_cleanup(); */

    getch_free ();

    out_scaled_start_time = out_start_time * wave_config->time_scale;
    out_scaled_end_time = out_end_time * wave_config->time_scale;
    if (out_scaled_start_time == out_scaled_end_time || out_scaled_end_time == 0)
    {
        fprintf (stderr, "OUT times range is equal to zero.  Exiting.\n");
        exit (1);
    }

    return out_scaled_end_time;
}

static int f_enable_out_autosave (char *str, int offset)
{
    WAVE_DEBUG (wave_log (FALSE, "WAVE_DEBUG: f_enable_out_autosave(\"%s\")\n", str));
    make_out_save_file = wave_parse_boolean_rc_entry (str);
    return (0);
}

static WaveRcEntry out_rc[] = {
    {"out_enable_autosave", f_enable_out_autosave, NULL},
};

WaveReader *out_init (void)
{
	WaveReader *ret = wave_new_reader ("out", "Synopsys (EPIC) TimeMill .out",
		WAVE_READER_COMPRESSED, NULL);

	wave_reader_add_function (ret, "load-declarations", out_main);
	wave_reader_add_function (ret, "seek-end", out_seek_end);
	wave_reader_add_function (ret, "cleanup", out_cleanup);

	wave_reader_add_rc_entries (ret, 1, out_rc);

	return ret;
}
