/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
    Copyright (C) 2000 Udi Finkelstein
    Copyright (C) 1999-2003 Tony Bybell

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `print.c'
    Printing

    $Id: print.c,v 1.49 2003/11/13 11:05:38 bardslea Exp $
*/

#include <errno.h>
#include <stdio.h>
#include "print.h"
#include "printps.h"
#include "currenttime.h"
#include "gui_if.h"
#include "support.h"
#include "log.h"
#include "help.h"
#include "signalpane.h"
#include "tracepane.h"

static WavePrinter *wave_printers[] =
{
	&wave_printer_ps,
	&wave_printer_eps,
	NULL
};

/* Initialize print related constants */
static void pr_wave_init (WavePrintingContext * context, WaveWindow *window)
{
	int wave_width = window->signals.pixmap_width + wave_window->traces.scale.wave_pane_width;
	int wave_height = wave_window->traces.scale.wave_pane_height;
	double page_aspect_ratio = context->page_width / context->page_height;
	double wave_aspect_ratio = (double) wave_width / wave_height;

	fprintf (stderr, "%f %f\n", context->page_width, context->page_height);
	fprintf (stderr, "%d %d %f %f ", wave_width, wave_height, wave_aspect_ratio, page_aspect_ratio);

    context->minimum_x = 0;
    context->minimum_y = 0;

	if (wave_aspect_ratio > page_aspect_ratio) /* Waste y space */
	{
		fprintf (stderr, "Ping\n");
		context->scale = context->page_width / wave_width;
	} else { /* Waste x space */
		fprintf (stderr, "Pong\n");
		context->scale = context->page_height / wave_height;
	}

	fprintf (stderr, "%f\n", context->scale);
}

/* wave_print_image : print using a fully set up print context */
void wave_print_image (WavePrintingContext * context)
{
    int trtarget;

    pr_wave_init (context, wave_window);

    context->printer->header (context);

    trtarget = (int) GTK_ADJUSTMENT (wave_window->traces.vslider)->value;
    wave_render_signal_names (context->drawing_context, trtarget, wave_window);

    context->printer->traces_init (context);

	render_times (context->drawing_context, wave_window);
    render_traces (context->drawing_context, wave_window);

    context->printer->trailer (context);
}

/* GUI elements */

static GtkWidget *print_dialog = NULL;
static GtkWidget *print_file_dialog = NULL;
static char *print_filename = NULL;

/* wave_init_print_dialog : create/show print dialogue box */
void wave_init_print_dialog (void)
{
    if (! print_dialog)
    {
    	WavePrinter **printers = wave_printers;
    	GtkWidget *format_option_menu;
    	GtkWidget *format_new_menu;
        print_dialog = create_print_dialog ();
        print_file_dialog = create_print_file_selector ();

		format_option_menu = lookup_widget (print_dialog, "print_format");
		format_new_menu = gtk_menu_new ();
        gtk_option_menu_remove_menu (GTK_OPTION_MENU (format_option_menu));

		while (*printers)
		{
			GtkWidget *menu_item = gtk_menu_item_new_with_label ((*printers)->format_name);

            gtk_widget_show (menu_item);
            gtk_object_set_data (GTK_OBJECT (menu_item), "printer", *printers);
            gtk_menu_append (GTK_MENU (format_new_menu), menu_item);

			printers ++;
		}
        gtk_option_menu_set_menu (GTK_OPTION_MENU (format_option_menu), format_new_menu);
    }

    if (! print_filename)
		print_filename = g_strdup ("gtkwave.ps");
    gtk_entry_set_text (GTK_ENTRY (lookup_widget (print_dialog, "print_file")), print_filename);

    gtk_widget_show (print_dialog);
}

void on_print_file_do_select_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_show (print_file_dialog);
}

void on_print_cancel_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_hide (print_dialog);
    gtk_widget_hide (print_file_dialog);
}

void on_print_file_selector_ok_clicked (GtkButton * button, gpointer user_data)
{
    char *filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (print_file_dialog));

    gtk_entry_set_text (GTK_ENTRY (lookup_widget (print_dialog, "print_file")), filename);
    gtk_widget_hide (print_file_dialog);
}

void on_print_file_selector_cancel_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_hide (print_file_dialog);
}

void on_print_ok_clicked (GtkButton * button, gpointer user_data)
{
    FILE *print_file;

    /* Paper sizes (in inches) Legal, Letter, A4, A3 */
    static const double px[] = { 215.9, 215.9, 210, 297 };
    static const double py[] = { 355.6, 279.4, 297, 420 };
    int print_size, print_orientation;
    double print_width, print_height;
    gboolean print_rotate;
    WavePrintingContext context;

    WavePrinter *printer = (WavePrinter *) wave_get_option_menu_active_data (GTK_OPTION_MENU
    		(lookup_widget (GTK_WIDGET (print_dialog), "print_format")), "printer");

    if (print_filename)
        g_free (print_filename);
    print_filename = gtk_editable_get_chars (GTK_EDITABLE
    	(lookup_widget (print_dialog, "print_file")), 0, -1);

    /* use the cancel action... */
    on_print_cancel_clicked (NULL, NULL);

    print_file = fopen (print_filename, "wb");
    if (! print_file)
    {
        wave_log (TRUE, "Error opening file for writing: %s\n", g_strerror (errno));
        errno = 0;
        return;
    }

    print_size = wave_get_option_menu_index (GTK_OPTION_MENU (lookup_widget (print_dialog, "print_size")));
    print_orientation =
      wave_get_option_menu_index (GTK_OPTION_MENU (lookup_widget (print_dialog, "print_orientation")));

	fprintf (stderr, "px %f py %f\n", px[print_size], py[print_size]);

#if 0
    if (print_orientation == 1) /* Landscape */
    {
        print_width = py[print_size];
        print_height = px[print_size];
        print_rotate = TRUE;
    } else
#endif
    {                           /* Portrait */
        print_width = px[print_size];
        print_height = py[print_size];
        print_rotate = FALSE;
    }

	context.printer = printer;
    printer->print (&context, print_file, print_width, print_height, print_rotate);

    fclose (print_file);
}

void on_file_print_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_help ("Print", "pops up a dialog box to let you convert the trace window to PostScript or MIF and save it in the specified file"))
        return;

    wave_init_print_dialog ();
}

