/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2004 Andrew Bardsley
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `rc.c'
    Configuration file handling and preferences dialogue

    $Id: rc.c,v 1.72 2004/04/25 19:23:42 bardslea Exp $
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include "analyzer.h"
#include "currenttime.h"
#include "utils.h"
#include "fgetdynamic.h"
#include "color.h"
#include "rc.h"
#include "gui_cb.h"
#include "log.h"
#include "support.h"
#include "gui_if.h"
#include "reader.h"
#include "help.h"
#include "tracepane.h"

#ifndef _MSC_VER
#include <unistd.h>
#include <pwd.h>
static char *rc_file_name = ".gtkwave2rc"; /* name of environment file--POSIX */
#else
static char *rc_file_name = "gtkwave2.ini"; /* name of environment file--WIN32 */
#endif

static int rc_line_num = 0;

WaveFlags wave_flags; /* Initialised by wave_init_flags */

/* wave_find_radix_data : find radix info for the given radix (2, 8, 16, 10).  Returns NULL on failure */
FormatRadixData *wave_find_radix_data (unsigned radix)
{
	FormatRadixData **radix_data_iter;
	int i;

	if (! wave_radix_style)
		return NULL;

	radix_data_iter = wave_radix_style->radices;

	for (i = 0; i < WAVE_RADIX_COUNT; i++)
	{
		if ((* radix_data_iter)->radix == radix)
			return (* radix_data_iter);
		radix_data_iter ++;
	}

	return NULL;
}

static void write_rc_file (void);

static int parse_wave_log_level (char *str, int offset)
{
    if (!strcmp (str, "never"))
    {
        wave_log_level = WAVE_LOG_NEVER;
    } else if (!strcmp (str, "error"))
    {
        wave_log_level = WAVE_LOG_ERROR;
    } else if (!strcmp (str, "always"))
    {
        wave_log_level = WAVE_LOG_ALWAYS;
    } else
    {
        wave_log (TRUE, "unknown logging level \"%s\"\n", str);
        return 1;
    }
    return 0;
}
static void write_wave_log_level (FILE * f, char *rc_name, int offset)
{
    switch (wave_log_level)
    {
    case WAVE_LOG_NEVER:
        fprintf (f, "wave_log_level = never\n");
        return;
    case WAVE_LOG_ALWAYS:
        fprintf (f, "wave_log_level = always\n");
        return;
    default:
        fprintf (f, "wave_log_level = error\n");
    }
}

/*
    functions that set the individual rc variables..
*/
static int parse_alt_hier_delimiter (char *str, int offset)
{
    if (str && strlen (str))
    {
        alt_hier_delimiter = str[0];
    }
    return 0;
}
static void write_alt_hier_delimiter (FILE * f, char *rc_name, int offset)
{
    if (alt_hier_delimiter)
    {
        fprintf (f, "alt_hier_delimiter = %c\n", alt_hier_delimiter);
    }
}

/* wave_parse_boolean_rc_entry : parse a boolean rc file option.  Anything
	starting with [yY] or a number that isn't 0 is true */
gboolean wave_parse_boolean_rc_entry (const char *str)
{
    return ((str && tolower (*str) == 'y') || (atoi (str) != 0));
}

/*
    functions that set the individual rc variables..
*/
static int parse_font_height_override (char *str, int offset)
{
    font_height_override = atoi (str);
    return 0;
}
static void write_font_height_override (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "font_height_override = %d\n", font_height_override);
}

static int parse_fontname_signals (char *str, int offset)
{
    if (fontname_signals)
        g_free (fontname_signals);
    fontname_signals = g_strdup (str);
    return 0;
}
static void write_fontname_signals (FILE * f, char *rc_name, int offset)
{
    if (fontname_signals)
    {
        fprintf (f, "fontname_signals = %s\n", fontname_signals);
    }
}

static int parse_fontname_waves (char *str, int offset)
{
    if (fontname_waves)
        g_free (fontname_waves);
    fontname_waves = g_strdup (str);
    return 0;
}
static void write_fontname_waves (FILE * f, char *rc_name, int offset)
{
    if (fontname_waves)
    {
        fprintf (f, "fontname_waves = %s\n", fontname_waves);
    }
}

static int parse_hier_delimiter (char *str, int offset)
{
    if (str && *str)
    {
        hier_delimiter = str[0];
        hier_was_explicitly_set = TRUE;
    }
    return 0;
}
static void write_hier_delimiter (FILE * f, char *rc_name, int offset)
{
    if (hier_was_explicitly_set)
    {
        fprintf (f, "hier_delimiter = %c\n", hier_delimiter);
    }
}

static int parse_vector_padding (char *str, int offset)
{
    wave_flags.vector_padding = atoi (str);
    wave_flags.vector_padding = CLAMP (wave_flags.vector_padding, 4, 16);

    return 0;
}
static void write_vector_padding (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "vector_padding = %d\n", wave_flags.vector_padding);
}

static int parse_window_width (char *str, int offset)
{
    wave_flags.window_width = atoi (str);
    wave_flags.window_width = CLAMP (wave_flags.window_width, 100, 2000);

    return 0;
}
static void write_window_width (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "window_width = %d\n", wave_flags.window_width);
}

static int parse_window_height (char *str, int offset)
{
    wave_flags.window_height = atoi (str);
    wave_flags.window_height = CLAMP (wave_flags.window_height, 100, 2000);

    return 0;
}
static void write_window_height (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "window_height = %d\n", wave_flags.window_height);
}

static int parse_signal_width (char *str, int offset)
{
    wave_flags.signal_width = atoi (str);
    wave_flags.signal_width = CLAMP (wave_flags.signal_width, 50, 2000);

    return 0;
}
static void write_signal_width (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "signal_width = %d\n", wave_flags.signal_width);
}

static int parse_zoom_base (char *str, int offset)
{
    float f;

    sscanf (str, "%f", &f);
    if (f < 1.5)
        f = 1.5;
    else if (f > 10.0)
        f = 10.0;
    wave_flags.zoom_base = (gdouble) f;
    return 0;
}
static void write_zoom_base (FILE * f, char *rc_name, int offset)
{
    fprintf (f, "zoom_base = %f\n", wave_flags.zoom_base);
}

static void write_color (FILE * rc_file, char *name, WaveColor color)
{
    char *color_str = wave_color_to_string (color);

    fprintf (rc_file, "color_%s = %s\n", name, color_str);
    g_free (color_str);
}

/* wave_rc_string_to_wave_color : parse a colour from the rc file, allow GDK
	to have a shot at recognising named colours */
WaveColor wave_rc_color_to_wave_color (const char *str, int line_no)
{
	static WaveColor gray = { 0x80, 0x80, 0x80 };

    /* Numeric colour #RRGGBB */
    if (*str == '#')
        return string_to_wave_color (str);
    else { /* Let GTK worry about named/other format colours */
		GdkColor color;

		if (gdk_color_parse (str, &color))
			return wave_gdk_color_to_wave_color (&color);
    }

    if (!*str)
        wave_log (TRUE, "Empty color name at rc line %d, falling back on gray\n", line_no);
    else
        wave_log (TRUE, "Unknown color name '%s' at rc line %d, falling back on gray\n", str, line_no);
    return gray;
}

/* make the color functions */
#define PANE_COLOR_MAKE(Z,pane_feature) \
	static int parse_color_##Z (char *str, int offset) { \
	    wave_config->pane_colors->colors[(pane_feature)] = \
	    	wave_rc_color_to_wave_color(str,rc_line_num); \
	    return 0; \
	} \
	static void write_color_##Z (FILE *f, char *rc_name, int offset) { \
		write_color (f, #Z, wave_config->pane_colors->colors[(pane_feature)]); \
	}

#define TRACE_COLOR_MAKE(Z,state) \
	static int parse_color_##Z (char *str, int offset) { \
	    wave_config->default_decor->state_colors->colors[(state)] = \
			wave_rc_color_to_wave_color(str,rc_line_num); \
	    return 0; \
	} \
	static void write_color_##Z (FILE *f, char *rc_name, int offset) { \
		write_color (f, #Z, wave_config->default_decor->state_colors->colors[(state)]); \
	}

PANE_COLOR_MAKE (back, TRACE_PANE_BACKGROUND);
PANE_COLOR_MAKE (grid, TRACE_PANE_GRID);
PANE_COLOR_MAKE (mark, TRACE_PANE_NAMED_MARKERS);
PANE_COLOR_MAKE (umark, TRACE_PANE_PRIMARY_MARKER);
PANE_COLOR_MAKE (pfont, TRACE_PANE_FONT);

TRACE_COLOR_MAKE (font, TRACE_STATE_FONT);
TRACE_COLOR_MAKE (low, TRACE_STATE_LOW);
TRACE_COLOR_MAKE (high, TRACE_STATE_HIGH);
TRACE_COLOR_MAKE (x, TRACE_STATE_X);
TRACE_COLOR_MAKE (xfill, TRACE_STATE_XFILL);
TRACE_COLOR_MAKE (trans, TRACE_STATE_TRANS);
TRACE_COLOR_MAKE (mid, TRACE_STATE_MID);
TRACE_COLOR_MAKE (vtrans, TRACE_STATE_VTRANS);
TRACE_COLOR_MAKE (vbox, TRACE_STATE_VBOX);
TRACE_COLOR_MAKE (unloaded, TRACE_STATE_UNLOADED);
TRACE_COLOR_MAKE (analog, TRACE_STATE_ANALOG);
TRACE_COLOR_MAKE (clip, TRACE_STATE_CLIP);

static int parse_boolean (char *str, int offset)
{
	*((gboolean *) (((char *) &wave_flags) + offset)) = wave_parse_boolean_rc_entry (str);
	return 0;
}
static void write_boolean (FILE *file, char *rc_name, int offset)
{
    fprintf (file, "%s = %d\n", rc_name, *((gboolean *) (((char *) &wave_flags) + offset)));
}
/*
    rc variables...these MUST be in alphabetical order for the bsearch!
*/
#define RC_ENTRY(s) {#s, parse_##s, write_##s, 0}
#define RC_BOOL(rc_name,flag_name) {#rc_name, parse_boolean, write_boolean, (int) &(((WaveFlags *) 0)->flag_name)}
static WaveRcEntry rcitems[] = {
    RC_ENTRY (alt_hier_delimiter),
    RC_BOOL (autocoalesce, autocoal_flag),
    RC_BOOL (autocoalesce_reversal, autocoal_rev_flag),
    RC_BOOL (autoname_bundles, autoname_flag),
    RC_ENTRY (color_analog),
    RC_ENTRY (color_back),
    RC_ENTRY (color_clip),
    RC_ENTRY (color_font),
    RC_ENTRY (color_grid),
    RC_ENTRY (color_high),
    RC_ENTRY (color_low),
    RC_ENTRY (color_mark),
    RC_ENTRY (color_mid),
    RC_ENTRY (color_pfont),
    RC_ENTRY (color_trans),
    RC_ENTRY (color_umark),
    RC_ENTRY (color_unloaded),
    RC_ENTRY (color_vbox),
    RC_ENTRY (color_vtrans),
    RC_ENTRY (color_x),
    RC_ENTRY (color_xfill),
    RC_BOOL (constant_marker_update, const_mark_update_flag),
    RC_BOOL (do_initial_zoom_fit, do_initial_zoom_fit),
    RC_BOOL (enable_ghost_marker, enable_ghost_marker),
    RC_BOOL (enable_horiz_grid, enable_horiz_grid),
    RC_BOOL (enable_vert_grid, enable_vert_grid),
    RC_ENTRY (font_height_override),
    RC_ENTRY (fontname_signals),
    RC_ENTRY (fontname_waves),
    RC_ENTRY (hier_delimiter),
    RC_BOOL (hier_grouping, search_hier_group_flag),
    RC_BOOL (left_justify_sigs, sig_just_flag),
    RC_BOOL (show_base_symbols, show_base_flag),
    RC_BOOL (show_grid, show_grid_flag),
    RC_BOOL (show_menubar, show_menubar),
    RC_BOOL (show_navigation, show_navigation),
    RC_BOOL (show_operations, show_operations),
    RC_BOOL (show_times, show_times),
    RC_ENTRY (signal_width),
    RC_BOOL (use_roundcaps, roundcap_vectors_flag),
    RC_BOOL (values_below, values_below_flag),
    RC_ENTRY (vector_padding),
    RC_ENTRY (wave_log_level),
    RC_ENTRY (window_height),
    RC_ENTRY (window_width),
    RC_ENTRY (zoom_base),
    RC_BOOL (zoom_center, center_zooms_flag),
    RC_BOOL (zoom_pow10_snap, pow10snap_flag)
};

#define NUM_RC_ITEMS (sizeof(rcitems)/sizeof(WaveRcEntry))

typedef struct
{
    char *key, *value;
    int line;
}
WaveRcLineInfo;
static WaveRcLineInfo *saved_rc_lines = NULL;
static int num_saved_rc_lines = 0;
static int reader_rc_entry_count;
static WaveRcEntry *reader_rc_entries;

static int rc_compare (const void *v1, const void *v2)
{
    return g_strcasecmp ((char *) v1, ((WaveRcEntry *) v2)->name);
}

static void wave_init_radix_styles (void);

/* wave_init_flags : initialise flags to preferred values, call this
	before parsing rc file */
void wave_init_flags (void)
{
    wave_flags.autocoal_flag = TRUE;
    wave_flags.autocoal_rev_flag = FALSE;
    wave_flags.autoname_flag = FALSE;
    wave_flags.center_zooms_flag = FALSE;
    wave_flags.const_mark_update_flag = FALSE;
    wave_flags.do_initial_zoom_fit = TRUE;
    wave_flags.enable_ghost_marker = TRUE;
    wave_flags.enable_horiz_grid = TRUE;
    wave_flags.enable_vert_grid = TRUE;
    wave_flags.pow10snap_flag = FALSE;
    wave_flags.roundcap_vectors_flag = FALSE;
    wave_flags.search_hier_group_flag = FALSE;
    wave_flags.show_base_flag = TRUE;
    wave_flags.show_grid_flag = TRUE;
    wave_flags.values_below_flag = FALSE;
    wave_flags.sig_just_flag = FALSE;

    wave_flags.vector_padding = 4;

    wave_flags.window_width = 600;
    wave_flags.window_height = 350;
    wave_flags.signal_width = 100;

    wave_flags.show_menubar = TRUE;
    wave_flags.show_operations = TRUE;
    wave_flags.show_navigation = TRUE;
    wave_flags.show_times = TRUE;

    wave_flags.zoom_base = 2.0;

	wave_init_radix_styles ();
}

void wave_read_rc_file (void)
{
    FILE *handle = fopen (rc_file_name, "rt");

#ifndef _MSC_VER                /* no concept of ~ in win32 */
    if (!handle)
    {
        char *rcpath, *home = getpwuid (geteuid ())->pw_dir;
        int len = strlen (home);

        rcpath = alloca (strlen (home) + strlen (rc_file_name) + 2);
        memcpy (rcpath, home, len);
        rcpath[len] = '/';
        strcpy (rcpath + len + 1, rc_file_name);

        handle = fopen (rcpath, "rt");
    }
#endif
    if (!handle)
    {
        /* no .rc file */
        errno = 0;
        return;
    }

    rc_line_num = 0;
    while (!feof (handle))
    {
        char *str;

        rc_line_num++;
        str = fgetmalloc (handle);
        if (str)
        {
            int i, j;
            char *name, *value;
            WaveRcEntry *r;

            /* skip leading ws */
            for (i = 0; str[i]; i++)
            {
                if (!isspace ((unsigned char) str[i]))
                    break;
            }

            if (!str[i] || str[i] == '#')
            {
                /* blank or comment */
                free (str);
                continue;
            }

            name = str + i;     /* FOUND START OF KEY */

            for (i++; str[i]; i++)
            {
                /* find end of key */
                if (isspace ((unsigned char) str[i]))
                    break;
            }

            if (!str[i])
            {
                /* key without value! */
              no_value:
                wave_log (FALSE, "warning: key '%s' without value skipped at line %d\n", name, rc_line_num);
                free (str);
                continue;
            }

            str[i++] = '\0';    /* FOUND END OF KEY */

            for (; str[i]; i++)
            {
                /* permit both 'key value' and 'key = value' forms */
                if (str[i] == '=')
                    continue;
                /* skip more whitespace */
                if (!isspace ((unsigned char) str[i]))
                    break;
            }
            if (!str[i])
                goto no_value;  /* key without value! */

            value = str + i;    /* FOUND START OF VALUE */

            /* Note that there must be at least one non-ws character */
            for (j = i + strlen (value) - 1; j > i; j--)
            {
                /* strip trailing ws */
                if (isspace ((unsigned char) str[j]))
                {
                    str[j] = '\0';
                } else
                {
                    break;
                }
            }

            r = bsearch (name, rcitems, NUM_RC_ITEMS, sizeof (WaveRcEntry), rc_compare);
            if (r)
            {
                r->parse (value, r->offset);
            } else if (num_saved_rc_lines)
            {
                WaveRcLineInfo *rcline;

                saved_rc_lines = (WaveRcLineInfo *)
                  realloc (saved_rc_lines, sizeof (WaveRcLineInfo) * (num_saved_rc_lines + 1));
                rcline = saved_rc_lines + (num_saved_rc_lines++);
                rcline->key = g_strdup (name);
                rcline->value = g_strdup (value);
                rcline->line = rc_line_num;
            } else
            {
                WaveRcLineInfo *rcline = saved_rc_lines = g_new (WaveRcLineInfo, 1);

                num_saved_rc_lines++;
                rcline->key = g_strdup (name);
                rcline->value = g_strdup (value);
                rcline->line = rc_line_num;
            }
            g_free (str);
        }
    }
    fclose (handle);
    errno = 0;
    return;
}

void wave_parse_reader_rc_entries (int reader_entryc, WaveRcEntry * reader_entryv)
{
    int i;

    reader_rc_entry_count = reader_entryc;
    reader_rc_entries = reader_entryv;
    for (i = 0; i < num_saved_rc_lines; i++)
    {
        WaveRcLineInfo *line = saved_rc_lines + i;
        WaveRcEntry *r = bsearch ((void *) line->key,
          (void *) reader_entryv,
          reader_entryc,
          sizeof (WaveRcEntry),
          rc_compare);

        if (r)
        {
            r->parse (line->value, r->offset);
        } else
        {
            wave_log (FALSE, "warning: unknown key '%s' skipped at line %d\n", line->key, line->line);
            line->line = -1; /* Make line invalid, so it won't get printed */
        }
    }
}

static void write_rc_file (void)
{
    FILE *handle = NULL;
    char *where;
    WaveRcEntry *r;
    int i;
    gboolean *reader_done;
    time_t now;

#ifdef _MSC_VER
    where = rc_file_name;
    handle = fopen (rc_file_name, "wt");
#else /* no concept of ~ in win32 */
    char *rcpath, *home = getpwuid (geteuid ())->pw_dir;
    int len = strlen (home);

    rcpath = alloca (len + strlen (rc_file_name) + 2);
    memcpy (rcpath, home, len);
    rcpath[len] = '/';
    strcpy (rcpath + len + 1, rc_file_name);

    where = rcpath;
    handle = fopen (rcpath, "wt");
#endif

    if (!handle)
    {
        /* can't write .rc file; should really do something like perror() */
        wave_log (TRUE, "could not write '%s': %s\n", where, g_strerror (errno));
        errno = 0;
        return;
    }

    /*
     * Write a header
     */
    now = time (NULL);
    fprintf (handle, "# GTKWave RC file, written by version %s\n" "# %s\n", VERSION, ctime (&now));
    /*
     * Write the main rc items
     */
    for (r = rcitems; r < rcitems + NUM_RC_ITEMS; r++)
    {
        void (*writer) (FILE *, char *, int) = r->write;

        if (writer)
        {
            writer (handle, r->name, r->offset);
        }
    }
    /*
     * Write the unrecognised and reader rc items; this is fairly
     * complicated because we must write an item only once, and we
     * must give readers an opportunity to save their current state.
     */
    reader_done = g_new (gboolean, reader_rc_entry_count);
    for (i = 0; i < reader_rc_entry_count; i ++)
    	reader_done[i] = FALSE;
    for (i = 0; i < num_saved_rc_lines; i++)
    {
        WaveRcLineInfo *line = saved_rc_lines + i;

		/* Don't print lines which have had their line no. set to < 0 as these are
			invalid read lines */
		if (line->line >= 0)
		{
			r = bsearch ((void *) line->key, (void *) reader_rc_entries,
			  reader_rc_entry_count, sizeof (WaveRcEntry), rc_compare);
			if (r && r->write)
			{
				r->write (handle, r->name, r->offset);
				reader_done[r - reader_rc_entries] = TRUE;
			} else
				fprintf (handle, "%s = %s\n", line->key, line->value);
		}
    }
    for (i = 0; i < reader_rc_entry_count; i++)
    {
        r = reader_rc_entries + i;
        if (! reader_done[i] && r && r->write)
        {
            r->write (handle, r->name, r->offset);
        }
    }
    g_free (reader_done);

    fclose (handle);
    errno = 0;
    return;
}

static GtkWidget *wave_preferences_dialog = NULL;

/* setup_check : set the state of a checkbox in the preferences dialogue box */
static void setup_check (GtkWidget * master, const gchar * name, gboolean flag)
{
    GtkWidget *w = lookup_widget (wave_preferences_dialog, name);

    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), flag);
}

gboolean wave_preferences_sensitive = TRUE;

static char *wave_digit_separators[] = {"None", "Underscore: 123_456", "Quote: 123'456",
	"Space: 123 456", "Comma: 123,456", "Dot: 123.456", NULL};
static char *wave_digit_separators_strings[] = {"", "_", "'", " ", ",", ".", NULL};
static char *wave_radix_style_names[] = {"Verilog", "VHDL", "C/Balsa", "GTKWave", NULL};
static WaveRadixStyle *wave_radix_styles[4];

WaveRadixStyle *wave_radix_style = NULL;
static char *wave_digit_separator = "";

/* wave_new_radix_style : new radix style */
WaveRadixStyle *wave_new_radix_style (void)
{
	WaveRadixStyle *ret = g_new (WaveRadixStyle, 1);

	return ret;
}
/* wave_copy_radix_style : deep copy a radix style */
WaveRadixStyle *wave_copy_radix_style (WaveRadixStyle *radix_style)
{
	int i;
	WaveRadixStyle *ret = wave_new_radix_style ();

	for (i = 0; i < WAVE_RADIX_COUNT; i ++)
		ret->radices[i] = CopyFormatRadixData (radix_style->radices[i]);

	return ret;
}

/* wave_init_radix_style : initialise the global radix style with the
	usual GTKWave defaults, also initialise the example wave_radix_styles */
static void wave_init_radix_styles (void)
{
	WaveRadixStyle *style;

	/* Verilog */
	style = wave_new_radix_style ();
	style->radices[0] = NewFormatRadixData (2, "b'", NULL, FALSE, 0);
	style->radices[1] = NewFormatRadixData (8, "o'", NULL, FALSE, 0);
	style->radices[2] = NewFormatRadixData (10, NULL, NULL, FALSE, 0);
	style->radices[3] = NewFormatRadixData (16, "x'", NULL, FALSE, 0);
	wave_radix_styles[0] = style;

	/* VHDL */
	style = wave_new_radix_style ();
	style->radices[0] = NewFormatRadixData (2, "2#", "#", FALSE, 0);
	style->radices[1] = NewFormatRadixData (8, "8#", "#", FALSE, 0);
	style->radices[2] = NewFormatRadixData (10, NULL, NULL, FALSE, 0);
	style->radices[3] = NewFormatRadixData (16, "16#", "#", FALSE, 0);
	wave_radix_styles[1] = style;

	/* C/Balsa */
	style = wave_new_radix_style ();
	style->radices[0] = NewFormatRadixData (2, "0b", NULL, FALSE, 0);
	style->radices[1] = NewFormatRadixData (8, "0", NULL, FALSE, 0);
	style->radices[2] = NewFormatRadixData (10, NULL, NULL, FALSE, 0);
	style->radices[3] = NewFormatRadixData (16, "0x", NULL, FALSE, 0);
	wave_radix_styles[2] = style;

	/* GTKWave */
	style = wave_new_radix_style ();
	style->radices[0] = NewFormatRadixData (2, "%", NULL, FALSE, 0);
	style->radices[1] = NewFormatRadixData (8, "#", NULL, FALSE, 0);
	style->radices[2] = NewFormatRadixData (10, NULL, NULL, FALSE, 0);
	style->radices[3] = NewFormatRadixData (16, "$", NULL, FALSE, 0);
	wave_radix_styles[3] = style;

	wave_radix_style = wave_copy_radix_style (wave_radix_styles[3]);
}

/* wave_radix_list_row_refresh_from_radix_data : refresh a given row from
	the given `radix_data' and make that `radix_data' that row's data, if
	row == -1, then this function adds a row rather than refreshing */
static void wave_radix_list_row_refresh_from_radix_data (GtkWidget *list, int row,
	FormatRadixData *radix_data)
{
	int i;
    char *text[5];

    text[0] = g_strdup_printf ("%d", radix_data->radix);
    text[1] = (radix_data->prefix ? radix_data->prefix : "");
    text[2] = (radix_data->suffix ? radix_data->suffix : "");
    text[3] = (radix_data->showLeadingZeroes ? "Yes" : "No");
    text[4] = g_strdup_printf ("%d", radix_data->separatorSpacing);

    if (row == -1)
    {
        gtk_clist_append (GTK_CLIST (list), text);
		gtk_clist_set_row_data (GTK_CLIST (list), GTK_CLIST (list)->rows - 1, radix_data);
	} else {
		for (i = 0; i < WAVE_RADIX_COUNT; i ++)
			gtk_clist_set_text (GTK_CLIST (list), row, i, text[i]);

		gtk_clist_set_row_data (GTK_CLIST (list), row, radix_data);
	}

    g_free (text[0]);
    g_free (text[4]);
}

static void wave_populate_radix_list_from_style (GtkWidget *list, WaveRadixStyle *style)
{
	int i;

    gtk_clist_clear (GTK_CLIST (list));
    gtk_clist_freeze (GTK_CLIST (list));

	for (i = 0; i < WAVE_RADIX_COUNT; i ++)
	{
		wave_radix_list_row_refresh_from_radix_data (list, -1, style->radices[i]);
	}

    gtk_clist_thaw (GTK_CLIST (list));
}

static void wave_radix_separator_selected (GtkMenuItem *menu_item, char *separator)
{
	wave_digit_separator = separator;
}

static char *wave_radix_entry_widgets[] =
{
	"radix_prefix_entry",
	"radix_suffix_entry",
	"radix_showLeadingZeroes",
	"radix_separator_spacing",
	NULL
};

static FormatRadixData *wave_radix_list_selected_radix_data = NULL;
static int wave_radix_list_selected_row_number = 0;
static gboolean wave_in_radix_list_config = FALSE;

/* wave_populate_radices_page : populate the radices page on the preferences
	notebook */
static void wave_populate_radices_page (GtkWidget *preferences)
{
	wave_in_radix_list_config = TRUE;
    setup_check (preferences, "radix_show_radix_indicators", wave_flags.show_base_flag);

	{ /* Set the separator menu */
		GtkWidget *separator_option_menu = lookup_widget (preferences, "radix_separator");
		wave_option_menu_set_menu_from_string_array (GTK_OPTION_MENU (separator_option_menu),
			wave_digit_separators, (void **) wave_digit_separators_strings, NULL, TRUE,
			(void (*)(GtkMenuItem *, void*)) wave_radix_separator_selected);
	}

	{ /* Set the style menu */
		GtkWidget *style_option_menu = lookup_widget (preferences, "radix_style");
		wave_option_menu_set_menu_from_string_array (GTK_OPTION_MENU (style_option_menu),
			wave_radix_style_names, (void **) wave_radix_styles, "style", TRUE, NULL);
	}

	{ /* Set the style list */
		GtkWidget *radix_list = lookup_widget (preferences, "radix_list");
		wave_populate_radix_list_from_style (radix_list, wave_radix_style);
	}
	wave_in_radix_list_config = FALSE;
}

void on_radix_list_unselect_row (GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer user_data)
{
	wave_set_sensitivity_of_named_widgets (wave_preferences_dialog,
		wave_radix_entry_widgets, FALSE);
	wave_radix_list_selected_radix_data = NULL;
}

void on_radix_list_select_row (GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer user_data)
{
	char *tmp;
	FormatRadixData *radix_data = (FormatRadixData *) gtk_clist_get_row_data (clist, row);

	wave_in_radix_list_config = TRUE;
	wave_set_sensitivity_of_named_widgets (wave_preferences_dialog,
		wave_radix_entry_widgets, TRUE);

	/* Update display from row data */
	tmp = g_strdup_printf ("%d", radix_data->radix);
	gtk_label_set_text (GTK_LABEL (lookup_widget (wave_preferences_dialog,
		"radix_label")), tmp);
	g_free (tmp);

	gtk_entry_set_text (GTK_ENTRY (lookup_widget (wave_preferences_dialog,
		"radix_prefix_entry")), (radix_data->prefix ? radix_data->prefix : ""));
	gtk_entry_set_text (GTK_ENTRY (lookup_widget (wave_preferences_dialog,
		"radix_suffix_entry")), (radix_data->suffix ? radix_data->suffix : ""));

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lookup_widget (wave_preferences_dialog,
		"radix_showLeadingZeroes")), radix_data->showLeadingZeroes);

	gtk_spin_button_set_value (GTK_SPIN_BUTTON (lookup_widget (wave_preferences_dialog,
		"radix_separator_spacing")), radix_data->separatorSpacing);
	wave_in_radix_list_config = FALSE;

	wave_radix_list_selected_radix_data = radix_data;
	wave_radix_list_selected_row_number = row;
}

static guint wave_preferences_redraw_timeout_id = 0;
static gboolean wave_preferences_redraw_timeout_set = FALSE;

#define WAVE_REDRAW_TIMEOUT (5.0)

/* wave_preferences_timeout_redraw : redraw the wave plane */
static void wave_preferences_redraw (void)
{
	wave_redraw (wave_window);
	/* FIXME, redraw *all* panes */
}

static gboolean wave_preferences_redraw_timeout (GtkWidget *timeout_scale); /* timeout function */
static void wave_preferences_timeout_redraw (gfloat timeout_period)
{
    GtkWidget *timeout_scale = lookup_widget (wave_preferences_dialog, "timeout");

	if (wave_in_radix_list_config)
		return;

	if (wave_preferences_redraw_timeout_set)
		gtk_timeout_remove (wave_preferences_redraw_timeout_id);

	gtk_adjustment_set_value (gtk_range_get_adjustment (GTK_RANGE (timeout_scale)), timeout_period);
	gtk_widget_set_sensitive (timeout_scale, TRUE);
	wave_preferences_redraw_timeout_id = gtk_timeout_add (1000,
		(GtkFunction) wave_preferences_redraw_timeout, timeout_scale);
	wave_preferences_redraw_timeout_set = TRUE;
}

static void wave_preferences_redraw_timeout_cancel (void)
{
    GtkWidget *timeout_scale = lookup_widget (wave_preferences_dialog, "timeout");
	GtkAdjustment *adjustment = gtk_range_get_adjustment (GTK_RANGE (timeout_scale));

	gtk_adjustment_set_value (adjustment, 0.0);
	gtk_widget_set_sensitive (timeout_scale, FALSE);
	wave_preferences_redraw_timeout_set = FALSE;
}

static gboolean wave_preferences_redraw_timeout (GtkWidget *timeout_scale) /* timeout function */
{
	GtkAdjustment *adjustment = gtk_range_get_adjustment (GTK_RANGE (timeout_scale));
	gfloat timeout = adjustment->value;

	if (timeout < 1.5)
	{
		wave_preferences_redraw ();
		wave_preferences_redraw_timeout_cancel ();
	} else
		wave_preferences_timeout_redraw (timeout - 1.0);

	return FALSE;
}

void on_radix_prefix_entry_activate (GtkEditable *editable, gpointer user_data)
{
	/* Just in case the desensitisation gets confused */
	if (wave_radix_list_selected_radix_data && ! wave_in_radix_list_config)
	{
		char *entry_text = gtk_entry_get_text (GTK_ENTRY (editable));

		FormatRadixDataSetPrefixSuffix (wave_radix_list_selected_radix_data,
			(*entry_text ? entry_text : NULL), NULL);

		wave_radix_list_row_refresh_from_radix_data (lookup_widget (wave_preferences_dialog,
			"radix_list"), wave_radix_list_selected_row_number, wave_radix_list_selected_radix_data);
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
	}
}

void on_radix_suffix_entry_activate (GtkEditable *editable, gpointer user_data)
{
	/* Just in case the desensitisation gets confused */
	if (wave_radix_list_selected_radix_data && ! wave_in_radix_list_config)
	{
		char *entry_text = gtk_entry_get_text (GTK_ENTRY (editable));

		FormatRadixDataSetPrefixSuffix (wave_radix_list_selected_radix_data,
			NULL, (*entry_text ? entry_text : NULL));

		wave_radix_list_row_refresh_from_radix_data (lookup_widget (wave_preferences_dialog,
			"radix_list"), wave_radix_list_selected_row_number, wave_radix_list_selected_radix_data);
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
	}
}

void on_radix_show_leading_zeroes_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	/* Just in case the desensitisation gets confused */
	if (wave_radix_list_selected_radix_data && ! wave_in_radix_list_config)
	{
		wave_radix_list_selected_radix_data->showLeadingZeroes =
			! wave_radix_list_selected_radix_data->showLeadingZeroes;

		wave_radix_list_row_refresh_from_radix_data (lookup_widget (wave_preferences_dialog,
			"radix_list"), wave_radix_list_selected_row_number, wave_radix_list_selected_radix_data);
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
	}
}

void on_radix_separator_spacing_changed (GtkEditable *editable, gpointer user_data)
{
	/* Just in case the desensitisation gets confused */
	if (wave_radix_list_selected_radix_data && ! wave_in_radix_list_config)
	{
		wave_radix_list_selected_radix_data->separatorSpacing =
			gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (editable));

		wave_radix_list_row_refresh_from_radix_data (lookup_widget (wave_preferences_dialog,
			"radix_list"), wave_radix_list_selected_row_number, wave_radix_list_selected_radix_data);
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
	}
}

void on_radix_copy_style_button_clicked (GtkButton *button, gpointer user_data)
{
	int i;
	WaveRadixStyle *style = wave_get_option_menu_active_data (
		GTK_OPTION_MENU (lookup_widget (wave_preferences_dialog, "radix_style")), "style");

	for (i = 0; i < WAVE_RADIX_COUNT; i ++)
		wave_radix_style->radices[i] = CopyFormatRadixData (style->radices[i]);

	wave_populate_radix_list_from_style (lookup_widget (wave_preferences_dialog,
		"radix_list"), wave_radix_style);
	wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
}

/* wave_populate_preferences_dialog : fill the preferences dialogue from the
	settings in tool at the moment gathered from:
		wave_flags, wave_wave_log_level ... */
extern void wave_populate_preferences_dialog (GtkWidget * preferences)
{
    GtkWidget *w;

    wave_preferences_sensitive = FALSE;

    setup_check (preferences, "autocoal", wave_flags.autocoal_flag);
    setup_check (preferences, "autocoal_rev", wave_flags.autocoal_rev_flag);
    setup_check (preferences, "autoname", wave_flags.autoname_flag);
    setup_check (preferences, "search_hier_group", wave_flags.search_hier_group_flag);
    setup_check (preferences, "show_grid", wave_flags.show_grid_flag);
    setup_check (preferences, "values_below", wave_flags.values_below_flag);
    setup_check (preferences, "const_mark_update", wave_flags.const_mark_update_flag);
    setup_check (preferences, "pow10snap", wave_flags.pow10snap_flag);
    setup_check (preferences, "roundcap_vectors", wave_flags.roundcap_vectors_flag);

    /* Signal Justification section is two buttons... */
    if (wave_flags.sig_just_flag != FALSE && wave_flags.sig_just_flag != TRUE)
    {
        wave_flags.sig_just_flag = FALSE;
    }
    setup_check (preferences, "sig_just_left", wave_flags.sig_just_flag);
    setup_check (preferences, "sig_just_right", !wave_flags.sig_just_flag);

    /* Automatic Log Control section is three buttons... */
    switch (wave_log_level)
    {
    case WAVE_LOG_NEVER:
        setup_check (preferences, "log_never", TRUE);
        setup_check (preferences, "log_error", FALSE);
        setup_check (preferences, "log_always", FALSE);
        break;
    case WAVE_LOG_ALWAYS:
        setup_check (preferences, "log_never", FALSE);
        setup_check (preferences, "log_error", FALSE);
        setup_check (preferences, "log_always", TRUE);
        break;
    default:
        /* mysterious value! Default to WAVE_LOG_ERROR */
        wave_log_level = WAVE_LOG_ERROR;
    case WAVE_LOG_ERROR:
        setup_check (preferences, "log_never", FALSE);
        setup_check (preferences, "log_error", TRUE);
        setup_check (preferences, "log_always", FALSE);
    }

    w = lookup_widget (preferences, "zoom_factor");
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (w), wave_flags.zoom_base);

    wave_preferences_sensitive = TRUE;

    wave_populate_radices_page (preferences);
}

void on_preferences_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_preferences_dialog == NULL)
    {
        wave_preferences_dialog = create_preferences_dialog ();
        wave_reader_install_config_panel_extension (wave_reader,
			lookup_widget (wave_preferences_dialog, "preferences_notebook"));
        wave_populate_preferences_dialog (wave_preferences_dialog);
    }
    gtk_widget_show (wave_preferences_dialog);
}

void on_save_preferences_activate (GtkMenuItem * menuitem, gpointer user_data)
{
    if (wave_help ("File->Save Preferences", "saves the current configuration to your .gtkwave2rc file"))
        return;

	{ /* Set geometry elements in flags */
		GtkWidget *window = widget_top_parent (GTK_WIDGET (menuitem));
		GtkWidget *signal_pane = lookup_widget (window, "signal_frame");

		wave_flags.window_width = window->allocation.width;
		wave_flags.window_height = window->allocation.height;
		wave_flags.signal_width = signal_pane->allocation.width;

		wave_flags.show_menubar = ! GTK_TOGGLE_BUTTON (lookup_widget (window, "menubar_toggle_button"))->active;
		wave_flags.show_operations = ! GTK_TOGGLE_BUTTON (lookup_widget (window, "operations_toggle_button"))->active;
		wave_flags.show_navigation = ! GTK_TOGGLE_BUTTON (lookup_widget (window, "navigation_toggle_button"))->active;
		wave_flags.show_times = ! GTK_TOGGLE_BUTTON (lookup_widget (window, "times_toggle_button"))->active;
	}
    write_rc_file ();
}

void on_autocoal_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.autocoal_flag = !wave_flags.autocoal_flag;
}

void on_autocoal_rev_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.autocoal_rev_flag = !wave_flags.autocoal_rev_flag;
}

void on_autoname_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.autoname_flag = !wave_flags.autoname_flag;
}

void on_search_hier_group_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.search_hier_group_flag = !wave_flags.search_hier_group_flag;
}

void on_show_grid_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.show_grid_flag = !wave_flags.show_grid_flag;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_values_below_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.values_below_flag = !wave_flags.values_below_flag;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_show_base_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
	if (! wave_in_radix_list_config)
    {
        wave_flags.show_base_flag = !wave_flags.show_base_flag;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_const_mark_update_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.const_mark_update_flag = !wave_flags.const_mark_update_flag;
}

void on_pow10snap_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_flags.pow10snap_flag = !wave_flags.pow10snap_flag;
}

void on_zoom_factor_changed (GtkEditable * editable, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.zoom_base = (gdouble) gtk_spin_button_get_value_as_float (GTK_SPIN_BUTTON (editable));
    }
}

void on_log_never_clicked (GtkButton * button, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_log_level = WAVE_LOG_NEVER;
}

void on_log_error_clicked (GtkButton * button, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_log_level = WAVE_LOG_ERROR;
}

void on_log_always_clicked (GtkButton * button, gpointer user_data)
{
    if (wave_preferences_sensitive)
        wave_log_level = WAVE_LOG_ALWAYS;
}

void on_roundcap_vectors_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.roundcap_vectors_flag = !wave_flags.roundcap_vectors_flag;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_sig_just_left_clicked (GtkButton * button, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.sig_just_flag = TRUE;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_sig_just_right_clicked (GtkButton * button, gpointer user_data)
{
    if (wave_preferences_sensitive)
    {
        wave_flags.sig_just_flag = FALSE;
		wave_preferences_timeout_redraw (WAVE_REDRAW_TIMEOUT);
    }
}

void on_preferences_close_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_hide (wave_preferences_dialog);
}

void on_preferences_apply_clicked (GtkButton * button, gpointer user_data)
{
	if (wave_preferences_redraw_timeout_set)
	{
		gtk_timeout_remove (wave_preferences_redraw_timeout_id);
		wave_preferences_redraw_timeout_cancel ();
	}

	wave_preferences_redraw ();
}

gboolean on_preferences_delete_event (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
    gtk_widget_hide (wave_preferences_dialog);

    return TRUE;
}
