/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2004 Andrew Bardsley
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `reader.c'
    Generic interface to trace file reading engines.

    $Id: reader.c,v 1.38 2004/05/17 01:32:29 janinl Exp $
*/

#include <gmodule.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <ctype.h>
#include <unistd.h>
#include "reader.h"
#include "main.h"
#include "log.h"
#include "str.h"
#include "rc.h"
#include "navigate.h"

#define REG(format) extern WaveReader format##_init (void);
#ifdef INIT_BUILTIN_READERS
INIT_BUILTIN_READERS
#endif
#undef REG

struct WaveReaderPrivate_
{
    /* Generic stuff */
    unsigned int version;
    const char *name;
    const char *extension;
    WaveReaderFlags capabilities;

    /* Primary operations */
    TimeType (*load_declarations) (WaveReader *reader);
    void (*seek_end) (WaveReader *reader);
    void (*load_times) (WaveReader *reader, TimeType from, TimeType to);
    void (*import_trace) (Node * np);
    void (*cleanup) (WaveReader *reader);
    Node* (*signal_key_find) (WaveReader *reader, const char *key);

    /* Config file extensions */
    int config_length;
    WaveRcEntry *config_vector;

    /* Configure GUI extensions */
    void (*append_config_panel) (WaveReader *reader, GtkNotebook * notebook);
};

GHashTable *wave_reader_function_table = NULL;

typedef struct
{
	char *name;
	unsigned offset;
}
NameOffsetPair;

#define READER_OFFSET(func) ((unsigned) &((WaveReaderPrivate ) 0)->func)
NameOffsetPair wave_reader_function_offsets[] =
{
	{"load-declarations", READER_OFFSET (load_declarations)},
	{"seek-end", READER_OFFSET (seek_end)},
	{"load-times", READER_OFFSET (load_times)},
	{"import-trace", READER_OFFSET (import_trace)},
	{"cleanup", READER_OFFSET (cleanup)},
	{"signal_key_find", READER_OFFSET (signal_key_find)},
	{"append-config-panel", READER_OFFSET (append_config_panel)},
	{NULL, 0}
};

/* init_function_table : populate the above table with associations
	from function names to offsets in a &WaveReader */
static void init_function_table (void)
{
	GHashTable *table = g_hash_table_new (g_str_hash, g_str_equal);
	NameOffsetPair *offsets = wave_reader_function_offsets;

	while (offsets->name)
	{
		g_hash_table_insert (table, offsets->name, GINT_TO_POINTER (offsets->offset));
		offsets ++;
	}

	wave_reader_function_table = table;
}

/* wave_new_reader : make a new reader, each reader module must make a populate
	one of these in its init function */
WaveReader *wave_new_reader (char *extension, char *name, WaveReaderFlags capabilities,
	void *instance_data)
{
	WaveReader *ret = g_new (WaveReader, 1);
	WaveReaderPrivate priv = g_new (struct WaveReaderPrivate_, 1);
	NameOffsetPair *offsets = wave_reader_function_offsets;

	ret->priv = priv;
	ret->filename = NULL;
	ret->flags = 0;
	ret->instance_data = instance_data;

	priv->version = READER_DRIVER_VERSION;
	priv->name = name;
	priv->extension = extension;
	priv->capabilities = capabilities;

	priv->config_length = 0;
	priv->config_vector = NULL;

	if (! wave_reader_function_table)
		init_function_table ();

	while (offsets->name)
	{
		*((void (**)()) (((char *) priv) + offsets->offset)) = NULL;
		offsets ++;
	}
	return ret;
}

/* wave_reader_add_function : add a function to the slots in a WaveReader */
void wave_reader_add_function (WaveReader *reader, char *func_name, void *func)
{
	WaveReaderPrivate priv = reader->priv;	
	unsigned offset = GPOINTER_TO_INT (g_hash_table_lookup (wave_reader_function_table, func_name));

	if (offset == 0)
		wave_log (TRUE, "Unrecognised function `%s' in reader module\n", func_name);
	else {
		*((void (**)()) ((char *) priv + offset)) = func;
	}
}

/* wave_reader_add_rc_entries : add RC entry processing functions... to a reader.
	This function takes an array (and size of array) of RC entries */
void wave_reader_add_rc_entries (WaveReader *reader, int config_length, WaveRcEntry *config_vector)
{
	WaveReaderPrivate priv = reader->priv;	

	priv->config_length = config_length;
	priv->config_vector = config_vector;
}

/* Built-in readers is a list of WaveBuiltinReaders mapping extensions to
	init functions for readers */

typedef struct
{
	char *extension;
	wave_reader_init_func init;
}
WaveBuiltinReader;

static GList *wave_builtin_readers = NULL;

WaveReader *wave_reader = NULL;

/* make_la_filename : make a path to the libtool .a file for reader `name' in
	directory `dir', returns a newly allocated string */
static char *make_la_filename (char *dir, char *name)
{
    return g_strconcat (dir, G_DIR_SEPARATOR_S, READER_PREFIX, name, ".la", NULL);
}

/* find_dl_filename_from_la_file : find the module file name from a libtool .la file
	returns NULL if the .la file cannot be found or the dl_filename cannot be found in that file */
static char *find_dl_filename_from_la_file (char *la_filename)
{
    char *dl_filename = NULL;
    FILE *la_file;
    char line[1024];

    la_file = fopen (la_filename, "rt");

    if (!la_file)
        return NULL;

    /* Find the dl_filename line */
    while (fgets (line, 1024, la_file))
    {
        if (strncmp (line, "dlname='", 8) == 0)
        {
            dl_filename = g_strdup (line + 8);
            break;
        }
    }

    /* Trim the name to the last ' */
    if (dl_filename)
    {
        char *ptr = strrchr (dl_filename, '\'');

        if (ptr)
            *ptr = '\0';
    }

    return dl_filename;
}

/* load_dynamic_reader_module : attempt to load a dynamically-loadable modules for the format `name' */
static WaveReader *load_dynamic_reader_module (char *name)
{
    GModule *handle;
    WaveReader *reader = NULL;
    char *wave_reader_path_env = g_getenv ("WAVE_READER_PATH");
    char *reader_dir = NULL;

    if (!g_module_supported ())
    {
        wave_log (TRUE, "GModule error: %s\n", g_module_error ());
        return NULL;
    }

    /* Path in environment variable */
    if (wave_reader_path_env)
    {
        char *path_ptr = wave_reader_path_env;
        char *la_filename_try;
        char *separator_position = strchr (path_ptr, G_SEARCHPATH_SEPARATOR);

        /* Handle path components */
        while (separator_position)
        {
            char *dir = g_strndup (path_ptr, separator_position - path_ptr);

            la_filename_try = make_la_filename (dir, name);

            if (access (la_filename_try, R_OK) == 0)
            {
                reader_dir = dir;
                break;
            }

            g_free (dir);
            g_free (la_filename_try);
            path_ptr = separator_position + 1;
            separator_position = strchr (path_ptr, G_SEARCHPATH_SEPARATOR);
        }
        /* Use rest of string */
        if (!reader_dir)
        {
            la_filename_try = make_la_filename (path_ptr, name);

            if (access (la_filename_try, R_OK) == 0)
                reader_dir = path_ptr;

            g_free (la_filename_try);
        }
    }
    /* OK, try the default place */
    if (!reader_dir)
    {
        char *la_filename_try = make_la_filename (READER_DIRECTORY, name);

        if (access (la_filename_try, R_OK) == 0)
            reader_dir = READER_DIRECTORY;

        g_free (la_filename_try);
    }

    if (!reader_dir)
    {
        wave_log (TRUE, "Can't find reader module .la file for format %s\n", name);
        return NULL;
    }

    {                           /* Find the dl file name and try to open it */
        char *module_path;
        char *dl_filename;
        char *la_filename = make_la_filename (reader_dir, name);

        dl_filename = find_dl_filename_from_la_file (la_filename);
        g_free (la_filename);

        if (!dl_filename)
        {
            wave_log (TRUE, "Can't find reader module name from .la file for format %s\n", name);
            return NULL;
        }

        module_path = g_strconcat (reader_dir, G_DIR_SEPARATOR_S, dl_filename, NULL);
        g_free (dl_filename);
        handle = g_module_open (module_path, G_MODULE_BIND_LAZY);

        if (!handle)
        {
            wave_log (TRUE, "Can't find and load module %s\n", module_path);
            g_free (module_path);
            return NULL;
        }
        g_free (module_path);
    }

    { 
    	/* Now try and find the init symbol, run that function and return the resulting struct */
        wave_reader_init_func module_init;
        char *module_init_name;
        gboolean success;

        module_init_name = g_strconcat (name, "_init", NULL);
        success = g_module_symbol (handle, module_init_name, (gpointer *) & module_init);

        if (!module_init || !success)
        {
            /* try BSD-style (?) _symbol */
            module_init_name = g_strconcat ("_", name, "_init", NULL);
            success = g_module_symbol (handle, module_init_name, (gpointer *) & module_init);

            if (!module_init || !success)
            {
                wave_log (TRUE, "Can't find entry point for reader module for format `%s'\n", name);
				g_free (module_init_name);
                return NULL;
            }
        }
        g_free (module_init_name);

        reader = module_init ();

        if (reader)
            wave_log (FALSE, "Installed dynamic reader module for format `%s'\n", reader->priv->extension);
        else
            wave_log (TRUE, "Can't initialise reader module for format `%s'\n", name);
    }

    return reader;
}

/* wave_register_builtin_reader : register a reader by noting its init function */
static void wave_register_builtin_reader (char *extension, wave_reader_init_func init)
{
    WaveBuiltinReader *builtin = g_new (WaveBuiltinReader, 1);

    builtin->init = init;
    builtin->extension = extension;

    wave_log (FALSE, "Registering builtin reader module for format `%s'\n", extension);
    wave_builtin_readers = g_list_append (wave_builtin_readers, (gpointer) builtin);
}

/* wave_register_builtin_readers : register the existence of builtin reader modules */
void wave_register_builtin_readers (void)
{
#define REG(format) wave_register_builtin_reader (#format, (wave_reader_init_func) format##_init);
#ifdef INIT_BUILTIN_READERS
    INIT_BUILTIN_READERS
#endif
#undef REG
}

TimeType wave_reader_load_declarations (WaveReader *reader)
{
    return reader->priv->load_declarations (reader);
}

void wave_reader_seek_end (WaveReader *reader)
{
    if (reader->priv->seek_end)
        reader->priv->seek_end (reader);
    else wave_log (TRUE, "Reader cannot re-read a file after opening\n");
}

void wave_reader_load_times (WaveReader *reader, TimeType from, TimeType to)
{
    if (reader->priv->load_times)
        reader->priv->load_times (reader, from, to);
}

void wave_reader_import_trace (Node * np)
{
    if (np && np->reader->priv->import_trace)
        np->reader->priv->import_trace (np);
}

void wave_reader_cleanup (WaveReader *reader)
{
    if (reader->priv->cleanup)
        reader->priv->cleanup (reader);
    else wave_log (TRUE, "%s.cleanup() trapped\n", reader->priv->name);
}

Node *wave_reader_signal_key_find (WaveReader *reader, const char *key)
{
    if (reader->priv->signal_key_find)
        return reader->priv->signal_key_find (reader, key);
    else wave_log (TRUE, "%s.signal_key_find() trapped\n", reader->priv->name);
    return NULL;
}

void wave_reader_install_config_panel_extension (WaveReader *reader, GtkWidget * tabbook)
{
    if (reader && reader->priv && reader->priv->append_config_panel)
        reader->priv->append_config_panel (reader, GTK_NOTEBOOK (tabbook));
}

const char *wave_reader_get_name (WaveReader *reader)
{
    return reader->priv->name;
}

/* wave_find_init_reader : find (and install) a reader module for the given filename */
WaveReader *wave_find_init_reader (char *filename, char *reader_name, gboolean is_stream)
{
    char *extension;
    WaveReader *reader = NULL;
    GList *readers = wave_builtin_readers;

    if (reader_name)
        extension = g_strdup (reader_name);

    if (is_stream)
    {
        if (!reader_name)
        {
            wave_log (TRUE, "A format must be specified using -f on the command line when reader from a stream\n");
            return NULL;
        }
    } else if (!reader_name)
    {
        /* Find an appropriate extension */
        extension = wave_get_filename_extension (filename);

#ifdef HAVE_LIBZ
        if (extension && (g_strcasecmp (extension, "gz") == 0))
        {
            char *dot_position = extension - 1;

            *dot_position = '\0'; 
            extension = g_strdup (wave_get_filename_extension (filename));
            *dot_position = '.';
        } else
#endif
            extension = g_strdup (extension);

        if (!extension)
        {
            wave_log (TRUE, "No extension found on filename %s, please specify format on the command line\n", filename);
            return NULL;
        }
        /* Downcase extension for comparisons */
        g_strdown (extension);
    }

    /* Search for a builtin reader */
    while (!reader && readers)
    {
    	WaveBuiltinReader *builtin_reader = readers->data;

        if (g_strcasecmp (builtin_reader->extension, extension) == 0)
        {
        	reader = builtin_reader->init ();
            if (! reader)
            {
                wave_log (TRUE, "Can't initialise reader module `%s'\n", extension);
                return NULL;
            }
			wave_set_step_size (wave_config->time_scale, wave_config->time_dimension, wave_window);
        }
        readers = g_list_next (readers);
    }

    /* Try and load a dynamic module if no builtin module matches */
    if (!reader)
        reader = load_dynamic_reader_module (extension);

    if (!reader)
    {
        wave_log (TRUE, "No installed reader module for format %s\n", extension);
        return NULL;
    }

    if (is_stream && !(reader->priv->capabilities & WAVE_READER_COMPRESSED))
    {
        wave_log (TRUE, "Driver for format %s (%s) cannot handle streams\n", extension,
			reader->priv->name);
        return NULL;
    }
    g_free (extension);

    if (reader)                 /* This should be true at this point! */
    {
        if (reader->priv->config_length)
            wave_parse_reader_rc_entries (reader->priv->config_length, reader->priv->config_vector);

        /* Do global assignment here in anticipation of removing this global stuff */
        /* Set data in wave_reader */
        reader->filename = g_strdup (filename);

		/* FIXME */
		wave_reader = reader;
    }

    return reader;
}
