/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `reader.h'
    Generic interface to trace file reading engines.

    $Id: reader.h,v 1.27 2004/05/17 01:32:29 janinl Exp $
*/

#ifndef WAVE_READER_H
#define WAVE_READER_H

#include <glib.h>
#include "analyzer.h"
#include "rc.h"
#include "utils.h"
#include "../config.h"

/* 1: upto and including version pre2 */
/* 2: pre3 */
#define READER_DRIVER_VERSION 4

struct WaveReaderPrivate_;
typedef struct WaveReaderPrivate_ *WaveReaderPrivate;

/* WaveReaderFlags are used to signal availability of
	features and to select features

	For availability, these flags are valid:
		WAVE_READER_COMPRESSED: can handle compressed files
		WAVE_READER_DYNAMIC: can handle dynamic EOF tracking

	For active feaures, only these flags make sense:
		WAVE_READER_DYNAMIC: dynamic file following enabled
*/
typedef enum
{
    WAVE_READER_COMPRESSED = (1 << 0),
    WAVE_READER_DYNAMIC = (1 << 1)
}
WaveReaderFlags;

/* Private data of a reader */
typedef struct WaveReader_
{
	const char *filename;	/* Name of dump file */
    WaveReaderPrivate priv; /* Reader private data */
    WaveReaderFlags flags;  /* Active features */
	void *instance_data;	/* Private data of this reader instance */
}
WaveReader;

/* Function type of initialisation functions for reader modules.
	These functions can set up global variables, perform module-specific
	initialisation and then return a WaveReader structure for that reader */
typedef WaveReader *(*wave_reader_init_func) (void);

/* Current reader data, NB. this should *definately* not be
	global in a right thinking world */
extern WaveReader *wave_reader;

extern TimeType wave_reader_load_declarations (WaveReader *wave_reader);
extern void wave_reader_seek_end (WaveReader *wave_reader);
extern void wave_reader_load_times (WaveReader *wave_reader, TimeType from, TimeType to);
extern void wave_reader_import_trace (Node * np);
extern void wave_reader_cleanup (WaveReader *wave_reader);
extern Node *wave_reader_signal_key_find (WaveReader *wave_reader, const char *key);
extern void wave_reader_install_config_panel_extension (WaveReader *wave_reader, GtkWidget * tabbook);

extern const char *wave_reader_get_name (WaveReader *wave_reader);

/* wave_find_init_reader : find (and install) a reader module for the given filename */
extern WaveReader *wave_find_init_reader (char *filename, char *reader_name, gboolean is_stream);

/* wave_register_builtin_readers : register the existence of builtin reader modules */
extern void wave_register_builtin_readers (void);

/* wave_new_reader : make a new reader, each reader module must make a populate
	one of these in its init function */
extern WaveReader *wave_new_reader (char *extension, char *name, WaveReaderFlags capabilities,
	void *instance_data);

/* wave_reader_add_function : add a function to the slots in a WaveReader */
extern void wave_reader_add_function (WaveReader *reader, char *func_name, void *func);

/* wave_reader_add_rc_entries : add RC entry processing functions... to a reader.
	This function takes an array (and size of array) of RC entries */
extern void wave_reader_add_rc_entries (WaveReader *reader, int config_length, WaveRcEntry *config_vector);

#endif /* GTKWAVE_READER_H */
