/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   `signalpane.c'
   Signal pane setup/callbacks

   $Id: signalpane.c,v 1.12 2004/05/14 22:29:57 janinl Exp $
*/

#include "signalpane.h"
#include "tracepane.h"
#include "log.h"
#include "rc.h"
#include "bsearch.h"
#include "value.h"
#include "support.h"

GdkFont *signalfont = NULL;
int fontheight = 1;
gboolean wave_dnd_state = FALSE;
gboolean wave_show_prefixes_state = TRUE;

static gboolean cached_selected_flag = FALSE;
static int cached_trace_line_num = -1;    /* if -1, it's invalid, else */

static TraceEnt *cached_trace = NULL;

/* wave_dnd_error : complain about certain ops conflict with dnd... */
void wave_dnd_error (void)
{
    wave_log (TRUE, "Can't perform that operation when waveform drag and drop is in " "progress!\n");
}

/* wave_render_signal_names_background : paint the background and ``Time'' message in the signal pane */
void wave_render_signal_names_background (WaveWindow * window)
{
    gdk_draw_rectangle (window->signals.pixmap,
      window->signals.drawing_area->style->bg_gc[GTK_STATE_ACTIVE], TRUE, 0, -1, window->signals.fill_width, fontheight);
    gdk_draw_line (window->signals.pixmap, window->signals.drawing_area->style->white_gc,
      0, fontheight - 1, window->signals.fill_width - 1, fontheight - 1);
}

static void wave_update_signal_value (TraceEnt * t)
{
    if (t && t->config->name && !TRACE_IS_SEPARATOR (t->config))
    {
        WAVE_DEBUG (printf ("update_signal_value: %s\n", t->config->name));

        if (tims.marker != -1 && !TRACE_IS_HIDDEN (t->config))
        {
            WaveEventIterator *event_iter = t->n.nd->iterator;

            if (t->asciivalue)
            {
                g_free (t->asciivalue);
                t->asciivalue = NULL;
            }

            if (wave_event_iterator_seek (event_iter, tims.marker, WAVE_EVENT_LAST_LE))
            {
                t->asciivalue = wave_event_value_string (t, event_iter->current);
            }
        }
    }
}

/* wave_render_signal_name : render a single signal name, note that this function is made
	specific to the GDK drawer in places to make life easier. */
static int wave_render_signal_name_body (WaveDrawingContext * context, TraceEnt * t, int i,
  gboolean paint_background, WaveWindow * window)
{
    int y, text_y, bottom_line_y;
    int block_height;
    int equals_width = 0;

    /* Do some special stuff for GDK drawing (so much for generalisation ;) */
    gboolean use_gdk_drawing = context->destination == WAVE_DRAWING_GDK;

    /* Only allow paint_background if we're using the gdk_drawer */
    if (!use_gdk_drawing)
        paint_background = FALSE;

    wave_update_signal_value (t); /* in case it's stale on nonprop */

    y = fontheight * (i + 1);
    block_height = fontheight * t->config->height;
    bottom_line_y = y + block_height - 1;
    text_y = y + fontheight - (signalfont->descent + 1);

    /* Do no drawing for extensions */
    if (TRACE_IS_EXTENSION (t->config))
        return y;

    /*
       Horizontally, each signal name line extends from 0 to window->signals.fill_width-1.
       Each line looks like:
       0..window->signals.name_width-1: signal name, left or right justified with 3 pixel padding
       window->signals.name_width+6..: '='
       window->signals.name_width+9+equals_width: value
     */

    {
        gboolean is_selected = TRACE_IS_SELECTED (t->config);
        gboolean is_separator = TRACE_IS_SEPARATOR (t->config);
        int font_color;

        /* First paint the appropriate background */
        if (is_selected)
        {
            if (use_gdk_drawing)
            {
                gdk_draw_rectangle (window->signals.pixmap,
                  window->signals.drawing_area->style->bg_gc[GTK_STATE_SELECTED], TRUE,
                  0, y, window->signals.fill_width, block_height);
            }
            font_color = WAVE_FONT_SIGNAL_SELECTED;
        } else
        {
            if (paint_background)
            {                   /* for select routines in signal_window.c */
                int gcidx = (paint_background == 2) ? GTK_STATE_NORMAL : GTK_STATE_PRELIGHT;

                gdk_draw_rectangle (window->signals.pixmap, window->signals.drawing_area->style->bg_gc[gcidx],
                  TRUE, 0, y, window->signals.fill_width, block_height);
            }
            font_color = (is_separator ? WAVE_FONT_SIGNAL_SEPARATOR : WAVE_FONT_SIGNAL);
        }

        /* Then the line separating traces */
        if (use_gdk_drawing)
        {
            gdk_draw_line (window->signals.pixmap, window->signals.drawing_area->style->white_gc,
              0, bottom_line_y, window->signals.fill_width - 1, bottom_line_y);
        }

        /* And then the name, value of the signal */
        if (t->config->name)
        {
			char *name;
			int baselen;

			if (!wave_show_prefixes_state) {
				name = strrchr (t->config->name, '.');
				if (name)
					name++;
				else
					name = t->config->name;
			}
			else
				name = t->config->name;

            baselen = gdk_string_measure (signalfont, name);

            WAVE_DRAW_PANE_STRING (context, TRACE_PANE_FONT, font_color,
              (wave_flags.sig_just_flag ? 3 : 3 + window->signals.name_width - baselen), text_y, name);

            if (!is_separator && t->asciivalue && !TRACE_IS_HIDDEN (t->config))
            {
                if (equals_width == 0)
                    equals_width = gdk_string_measure (signalfont, "=");

                WAVE_DRAW_PANE_STRING (context, TRACE_PANE_FONT, font_color,
                  window->signals.name_width + 6, text_y, "=");

                if (wave_flags.values_below_flag && t->config->height > 1)
                {               /* Print value below signal name */
                    int value_length = gdk_string_measure (signalfont, t->asciivalue);

                    WAVE_DRAW_PANE_STRING (context, TRACE_PANE_FONT, font_color,
                      (wave_flags.sig_just_flag ? 3 : 3 + window->signals.name_width - value_length),
                      text_y + fontheight, t->asciivalue);
                } else
                {
                    /* Print value next to signal name */
                    WAVE_DRAW_PANE_STRING (context, TRACE_PANE_FONT, font_color,
                      window->signals.name_width + 3 * 3 + equals_width, text_y, t->asciivalue);
                }
            }
        }
    }

    return y;
}

/* wave_find_num_traces_displayable : calculates the number of trace lines displayable
	(for traces of height 1) based on the current signal drawing area height */
static int wave_find_num_traces_displayable (WaveWindow * window)
{
    /* -1 for the empty time trace that is always there */
    return (window->signals.drawing_area->allocation.height / fontheight) - 1;
}

/* wave_trace_at_line : returns the trace which occupies the given line number (or the
	last trace if the line number is past the end of the trace list) */
static TraceEnt *wave_trace_at_line (int target_line_num)
{
	int line_num = 0;
	TraceEnt *ret = traces.first;

	while (ret && ret->next && line_num != target_line_num)
	{
		ret = ret->next;
		line_num ++;
	}

	return ret;
}

/* wave_render_signal_names : render all visible signals */
void wave_render_signal_names (WaveDrawingContext * context, int first_trace_line, WaveWindow * window)
{
    int trace_visible_line_num;
    int num_traces_displayable = wave_find_num_traces_displayable (window);
    TraceEnt *t = wave_trace_at_line (first_trace_line);

    window->traces.top_visible_trace = t;

    /* OK, cheat on the generalisation here to allow the display specific stuff */
    for (trace_visible_line_num = 0; t && trace_visible_line_num < num_traces_displayable; trace_visible_line_num ++, t = t->next)
        wave_render_signal_name_body (context, t, trace_visible_line_num, FALSE, window);
}

/* wave_render_signal_name : render a single signal name */
int wave_render_signal_name (TraceEnt * t, int i, gboolean paint_background, WaveWindow * window)
{
    return wave_render_signal_name_body (wave_window_drawing_context, t, i, paint_background, window);
}

static void service_hslider (GtkWidget * text, WaveWindow * window)
{
    gint xsrc;

    if (window->signals.pixmap)
    {
        wave_render_signal_names_background (window);
        xsrc = (gint) GTK_ADJUSTMENT (window->signals.hslider)->value;

        gdk_draw_pixmap (window->signals.drawing_area->window,
          window->signals.drawing_area->style->fg_gc[GTK_WIDGET_STATE (window->signals.drawing_area)],
          window->signals.pixmap, xsrc, 0, 0, 0,
          window->signals.drawing_area->allocation.width, window->signals.drawing_area->allocation.height);
    }
}

/* wave_set_trace_selected : set the selection state of a TraceEnt.  Respect the need to
	select (or unselect) all of an extended trace.  Returns the head TraceEnt in the
	trace and corrects `trace_number' to refer to the head trace (if not NULL) */
static TraceEnt *wave_set_trace_selected (TraceEnt * trace, gboolean selected, int *trace_number)
{
    TraceEnt *head_trace = trace;

    /* First go to the head of the trace */
    while (head_trace && TRACE_IS_EXTENSION (head_trace->config))
    {
        head_trace = head_trace->prev;
        if (trace_number)
            (*trace_number)--;
    }

    /* Now set the selected state of the head and all its extensions */
    if (head_trace)
    {
        TraceEnt *extensions = head_trace->next;

        trace_config_set_flags (head_trace->config, TRACE_FLAGS_SELECTED, selected);

        while (extensions && TRACE_IS_EXTENSION (extensions->config))
        {
            trace_config_set_flags (extensions->config, TRACE_FLAGS_SELECTED, selected);
            extensions = extensions->next;
        }
    }

    return head_trace;
}

/* wave_max_signal_length : calculate maximum signal name length */
void wave_max_signal_length (WaveWindow * window)
{
    TraceEnt *t;
    int max_length = 0;
    int value_max_length = 0;
    int equals_width = gdk_string_measure (signalfont, "=");

    WAVE_DEBUG (printf ("window->signals.window_width_dirty: %d\n", window->signals.window_width_dirty));

    window->signals.window_width_dirty = FALSE;

    for (t = traces.first; t; t = t->next)
    {
        int len = 0;

        if (!t->config->name)
            continue;

        if (TRACE_IS_SEPARATOR (t->config))
        {                       /* for "comment" style blank traces */
            len = gdk_string_measure (signalfont, t->config->name);
            max_length = MAX (max_length, len);
        } else
        {
            len = gdk_string_measure (signalfont, t->config->name);
            max_length = MAX (max_length, len);

            wave_update_signal_value (t);

            if (t->asciivalue)
            {
                int vlen = gdk_string_measure (signalfont, t->asciivalue);

                if (wave_flags.values_below_flag && t->config->height > 1)
                {               /* Value printed on next line, these affect the signal name length */
                    max_length = MAX (max_length, vlen);
                } else
                {
                    /* Value printed to right of signal name */
                    value_max_length = MAX (value_max_length, vlen);
                }
            }
        }
    }

    window->signals.name_width = max_length;
    window->signals.pixmap_width = max_length + 2 * 3; /* 3 pixel pad on either size of name */
    if (tims.marker != -1)
    {
        window->signals.pixmap_width += value_max_length + 3 * 3 + equals_width;
    }

    window->signals.pixmap_width = MAX (window->signals.pixmap_width, WAVE_MIN_SIGNAL_PANE_WIDTH);
}

/* wave_redraw_signal_name : redraw a trace name and its extension traces */
static void wave_redraw_signal_name (TraceEnt * trace, int trace_number, WaveWindow * window)
{
    int max_y = wave_render_signal_name (trace, trace_number, TRUE, window);
    int height = 1;             /* Lets not assume that the height field in trace->config is right */

    trace = trace->next;
    while (trace && TRACE_IS_EXTENSION (trace->config))
    {
        wave_render_signal_name (trace, trace_number + height, TRUE, window);
        height++;
        trace = trace->next;
    }

    gdk_draw_pixmap (window->signals.drawing_area->window,
      window->signals.drawing_area->style->fg_gc[GTK_WIDGET_STATE (window->signals.drawing_area)],
      window->signals.pixmap,
      GTK_ADJUSTMENT (window->signals.hslider)->value,
      max_y, 0, max_y, window->signals.drawing_area->allocation.width, height * fontheight);
}

gboolean on_signal_drawing_area_motion_notify_event (GtkWidget * widget, GdkEventMotion * event, WaveWindow * window)
{
    int num_traces_displayable;
    int which;
    int trtarget;
    int new_cached_trace_line_num;
    GtkAdjustment *wadj;
    TraceEnt *t;
    gdouble x, y;
    GdkModifierType state;
    int doloop;

    if (event->is_hint)
    {
        gdk_input_window_get_pointer (event->window, event->deviceid, &x, &y, NULL, NULL, NULL, &state);
    } else
    {
        x = event->x;
        y = event->y;
        state = event->state;
    }

  top:doloop = 0;

/********************* button 1 *********************/

    if ((traces.total) && (state & (GDK_BUTTON1_MASK | GDK_BUTTON3_MASK)) && (window->signals.pixmap) && (cached_trace_line_num != -1))
    {
        num_traces_displayable = wave_find_num_traces_displayable (window);

        which = (int) (y);
        which = (which / fontheight) - 1;

        if (which >= traces.total)
        {
            which = traces.total - 1;
        }

        wadj = GTK_ADJUSTMENT (window->traces.vslider);

        if (num_traces_displayable < traces.total)
        {
            if (which >= num_traces_displayable)
            {
                int target;

                target = ((int) wadj->value) + 1;
                which = num_traces_displayable - 1;

                if (target + which >= (traces.total - 1))
                    target = traces.total - which - 1;
                wadj->value = target;

                if (cached_trace_line_num == which)
                    cached_trace_line_num = which - 1; /* force update */

                gtk_signal_emit_by_name (GTK_OBJECT (wadj), "changed"); /* force bar update */
                gtk_signal_emit_by_name (GTK_OBJECT (wadj), "value_changed"); /* force text update */

                doloop = 1;
            } else if (which < 0)
            {
                int target;

                target = ((int) wadj->value) - 1;
                if (target < 0)
                    target = 0;
                wadj->value = target;

                which = 0;
                if (cached_trace_line_num == which)
                    cached_trace_line_num = -1; /* force update */

                gtk_signal_emit_by_name (GTK_OBJECT (wadj), "changed"); /* force bar update */
                gtk_signal_emit_by_name (GTK_OBJECT (wadj), "value_changed"); /* force text update */

                doloop = 1;
            }
        } else
        {
            if (which < 0)
                which = 0;
        }

        trtarget = ((int) wadj->value) + which;

		cached_trace = t = wave_trace_at_line (trtarget);

        if (wave_dnd_state && (state & GDK_BUTTON3_MASK))
        {
            GtkAdjustment *hadj;
            gint xsrc;
            int yval;

            hadj = GTK_ADJUSTMENT (window->signals.hslider);
            wadj = GTK_ADJUSTMENT (window->traces.vslider);
            gtk_signal_emit_by_name (GTK_OBJECT (wadj), "changed"); /* force bar update */
            gtk_signal_emit_by_name (GTK_OBJECT (wadj), "value_changed"); /* force text update */

            xsrc = (gint) hadj->value;

            yval = wave_render_signal_name (t, which, TRUE, window);
            gdk_draw_pixmap (window->signals.drawing_area->window,
              window->signals.drawing_area->style->fg_gc[GTK_WIDGET_STATE (window->signals.drawing_area)],
              window->signals.pixmap, xsrc, yval, 0, yval,
              window->signals.drawing_area->allocation.width, fontheight - 1);

        } else if ((state & GDK_BUTTON1_MASK) && !wave_dnd_state)
            if ((t) && (which != cached_trace_line_num))
            {
                GtkAdjustment *hadj;
                gint xsrc;

                hadj = GTK_ADJUSTMENT (window->signals.hslider);
                xsrc = (gint) hadj->value;

                new_cached_trace_line_num = which; /* save so fill ins go from deltas in the future */

                do
                {
                    TraceFlags oldflags = oldflags = t->config->flags;
                    int trace_number = which;
                    TraceEnt *head_trace = wave_set_trace_selected (t, cached_selected_flag, &trace_number);

                    if (oldflags != t->config->flags)
                    {
                        WAVE_DEBUG (printf
                          ("Motion select swap in signal_drawing_area at x: %d, y: %d row: %d\n",
                            (int) x, (int) y, which));

                        wave_redraw_signal_name (head_trace, trace_number, window);
                    }

                    if (which > cached_trace_line_num)
                    {
                        which--;
                        t = t->prev;
                    } else if (which < cached_trace_line_num)
                    {
                        which++;
                        t = t->next;
                    }

                }
                while ((which != cached_trace_line_num) && (t));

                cached_trace_line_num = new_cached_trace_line_num; /* for next time around */
            }
    }

    if (doloop)
    {
        gdk_input_window_get_pointer (event->window, event->deviceid, &x, &y, NULL, NULL, NULL, &state);
        goto top;
    }

    return TRUE;
}

gint on_signal_drawing_area_button_release_event (GtkWidget * widget, GdkEventButton * event, WaveWindow * window)
{
    int trace_line_num;

    if (event->button == 1)
    {
        if (!wave_dnd_state)
            cached_trace_line_num = -1;
        gdk_pointer_ungrab (event->time);
        WAVE_DEBUG (printf ("Button 1 released\n"));
    }

/********************* button 3 *********************/

    if (event->button == 3 && window->signals.pixmap)
    {
        cached_trace_line_num = -1;
        gdk_pointer_ungrab (event->time);
        WAVE_DEBUG (printf ("Button 3 released\n"));

        if (wave_dnd_state)
        {
            if (cached_trace)
                wave_set_trace_selected (cached_trace, TRUE, NULL);

            trace_line_num = (int) event->y;
            trace_line_num = (trace_line_num / fontheight) - 1;

            if ((trace_line_num < 0 && window->traces.top_visible_trace == traces.first && prepend_buffer ()) || paste_buffer (NULL))
            {                   /* short circuit on special trace_line_num < 0 case */
                wave_log (FALSE, "Drop completed.\n");

                wave_max_signal_length (window);
                wave_redraw (wave_window);
                wave_signal_pane_adjust_after_insertion (wave_window);
            }
            wave_dnd_state = FALSE;
        }
    }

/********************* button 3 *********************/

    return TRUE;
}

gint on_signal_drawing_area_button_press_event (GtkWidget * widget, GdkEventButton * event, WaveWindow * window)
{
    int num_traces_displayable;
    int which;
    int trtarget;
    GtkAdjustment *wadj;
    TraceEnt *t;

    if (traces.total && window->signals.pixmap)
    {
        gdk_pointer_grab (widget->window, FALSE,
          GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON1_MOTION_MASK
          | GDK_BUTTON3_MOTION_MASK | GDK_BUTTON_RELEASE_MASK, NULL, NULL, event->time);

        num_traces_displayable = wave_find_num_traces_displayable (window);

        which = (int) event->y;
        which = (which / fontheight) - 1;

        if (which >= traces.total || which >= num_traces_displayable || which < 0)
        {
            if (!wave_dnd_state)
                cached_trace_line_num = -1;
            goto check_button_3; /* off in no man's land, but check 3rd anyways.. */
        }

        cached_trace_line_num = which;    /* cache for later fill in */
        wadj = GTK_ADJUSTMENT (window->traces.vslider);
        trtarget = ((int) wadj->value) + which;

		t = wave_trace_at_line (trtarget);

        cached_trace = t;
        if (!wave_dnd_state && event->button == 1)
            if (t)
            {
                if (window->signals.shift_click_trace && (event->state & GDK_SHIFT_MASK))
                {
                    TraceEnt *t2;
                    gboolean is_selected;

                    for (t2 = window->signals.shift_click_trace; t2; t2 = t2->prev)
                    {
                        if (t2 == t)
                        {
                            t2 = window->signals.shift_click_trace;
                            is_selected = TRACE_IS_SELECTED (t2->config);
                            for (; t2; t2 = t2->prev)
                            {
                                wave_set_trace_selected (t2, is_selected, NULL);
                                if (t2 == t)
                                    break;
                            }
                            goto resync_signal_drawing_area;
                        }
                    }

                    for (t2 = window->signals.shift_click_trace; t2; t2 = t2->next)
                    {
                        if (t2 == t)
                        {
                            t2 = window->signals.shift_click_trace;
                            is_selected = TRACE_IS_SELECTED (t2->config);
                            for (; t2; t2 = t2->next)
                            {
                                wave_set_trace_selected (t2, is_selected, NULL);
                                if (t2 == t)
                                    break;
                            }
                            goto resync_signal_drawing_area;
                        }
                    }

                    goto normal_button1_press; /* couldn't find original so make this original... */

                  resync_signal_drawing_area:
                    wave_max_signal_length (window);
                    wave_redraw_signal_pane (window);
                    WAVE_DEBUG (printf ("Shift-Click in signal_drawing_area!\n"));
                    return TRUE;
                } else
                {
                  normal_button1_press:
                    {
                        gboolean current_selection_state = TRACE_IS_SELECTED (t->config);
                        int trace_number = which;
                        TraceEnt *head_trace = wave_set_trace_selected (t, !current_selection_state, &trace_number);

                        cached_selected_flag = !current_selection_state;
                        window->signals.shift_click_trace = t;

                        WAVE_DEBUG (printf
                          ("Button pressed in signal_drawing_area at x: %d, y: %d row: %d\n",
                            (int) event->x, (int) event->y, which));

                        wave_redraw_signal_name (head_trace, trace_number, window);
                    }
                }
            }

      check_button_3:
        if (event->button == 3)
        {
            if (!wave_dnd_state)
            {
                if (cut_buffer ())
                {
                    wave_log (FALSE, "Dragging %d trace%s.\n", traces.buffercount, traces.buffercount != 1 ? "s" : "");

                    wave_max_signal_length (window);
                    wave_redraw (wave_window);
                    wave_dnd_state = TRUE;
                }
            }
        }
    }

    return TRUE;
}

/* wave_signal_pane_scroll_to_right : scroll the signal pane all the way to the right of the
	signal name portion of the pane */
static void wave_signal_pane_scroll_to_right (WaveWindow * window)
{
    int signal_pane_width = window->signals.drawing_area->allocation.width;
    GtkAdjustment *hadj = GTK_ADJUSTMENT (window->signals.hslider);

    hadj->value = MAX (0, (window->signals.name_width + 2 * 3) - signal_pane_width);
    hadj->upper = window->signals.pixmap_width;

    gtk_signal_emit_by_name (GTK_OBJECT (hadj), "value_changed"); /* force bar update */
}

/* wave_signal_pane_adjust_after_insertion : adjust the signal pane scroll bar after
	a signal insertion */
void wave_signal_pane_adjust_after_insertion (WaveWindow * window)
{
    wave_signal_pane_scroll_to_right (window);
}

gint on_signal_drawing_area_configure_event (GtkWidget * widget, GdkEventConfigure * event, WaveWindow * window)
{
    static int trtarget = 0;
    GtkAdjustment *wadj, *hadj;
    int num_traces_displayable;
    int old_signal_fill_width, old_signal_fill_height;

    num_traces_displayable = wave_find_num_traces_displayable (window);

    WAVE_DEBUG (printf ("SigWin Configure Event h: %d, w: %d\n", widget->allocation.height, widget->allocation.width));

    old_signal_fill_width = window->signals.fill_width;
    window->signals.fill_width = MAX (widget->allocation.width, window->signals.pixmap_width);

    old_signal_fill_height = window->signals.fill_height;
    window->signals.fill_height = MAX (widget->allocation.height, window->signals.pixmap_height);

    if (window->signals.pixmap)
    {
        if (old_signal_fill_width != window->signals.fill_width || old_signal_fill_height != window->signals.fill_height)
        {
            gdk_pixmap_unref (window->signals.pixmap);
            window->signals.pixmap =
              gdk_pixmap_new (widget->window, window->signals.fill_width, window->signals.fill_height, -1);
        }
    } else
    {
        window->signals.pixmap =
          gdk_pixmap_new (widget->window, window->signals.fill_width, window->signals.fill_height, -1);
    }

    gdk_draw_rectangle (window->signals.pixmap, widget->style->bg_gc[GTK_STATE_PRELIGHT],
      TRUE, 0, 0, window->signals.fill_width, window->signals.fill_height);

    hadj = GTK_ADJUSTMENT (window->signals.hslider);
    hadj->page_size = hadj->page_increment = (gfloat) widget->allocation.width;
    hadj->step_increment = (gfloat) 10.0; /* approx 1ch at a time */
    hadj->lower = (gfloat) 0.0;
    hadj->upper = (gfloat) window->signals.pixmap_width;

    /* Scroll back onto the signal pane if the signals get narrower */
    if (((int) hadj->value) + widget->allocation.width > window->signals.fill_width)
    {
        hadj->value = (gfloat) (window->signals.fill_width - widget->allocation.width);
    }

    wadj = GTK_ADJUSTMENT (window->traces.vslider);
    wadj->page_size = wadj->page_increment = (gfloat) num_traces_displayable;
    wadj->step_increment = (gfloat) 1.0;
    wadj->lower = (gfloat) 0.0;
    wadj->upper = (gfloat) traces.total;

    if (num_traces_displayable > traces.total)
    {
        trtarget = 0;
        wadj->value = (gfloat) trtarget;
    } else if (wadj->value + num_traces_displayable > traces.total)
    {
        trtarget = traces.total - num_traces_displayable;
        wadj->value = (gfloat) trtarget;
    }

    gtk_widget_realize (window->traces.drawing_area);

    /* Done the resource allocation, now think about appearance stuff */
    gtk_signal_emit_by_name (GTK_OBJECT (wadj), "changed"); /* force bar update */
    gtk_signal_emit_by_name (GTK_OBJECT (wadj), "value_changed"); /* force text update */
    gtk_signal_emit_by_name (GTK_OBJECT (hadj), "changed"); /* force bar update */

    return TRUE;
}

/* wave_redraw_signal_pane : redraw signal pane of main window */
void wave_redraw_signal_pane (WaveWindow * window)
{
    on_signal_drawing_area_configure_event (window->signals.drawing_area, NULL, window);
}

gboolean on_signal_drawing_area_expose_event (GtkWidget * widget, GdkEventExpose * event, WaveWindow * window)
{
    GtkAdjustment *hadj;
    int xsrc;

    hadj = GTK_ADJUSTMENT (window->signals.hslider);
    xsrc = (gint) hadj->value;

    gdk_draw_pixmap (widget->window,
      widget->style->fg_gc[GTK_WIDGET_STATE (widget)], window->signals.pixmap,
      xsrc + event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);

    return FALSE;
}

/* wave_create_signal_pane : create signal pane widgets */
void wave_create_signal_pane (WaveWindow * window)
{
    window->signals.drawing_area = lookup_widget (window->window, "signal_drawing_area");

	window->signals.hslider = gtk_range_get_adjustment (GTK_RANGE (lookup_widget (window->window,	
		"signal_hscroll")));
    gtk_signal_connect (GTK_OBJECT (window->signals.hslider), "value_changed",
      GTK_SIGNAL_FUNC (service_hslider), window);

    window->signals.window = lookup_widget (window->window, "signal_frame");
}
