/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `strace.c'
    Signal trace dialogue and search functions

    $Id: strace.c,v 1.38 2003/11/11 22:53:28 bardslea Exp $
*/

#include "strace.h"
#include "value.h"
#include "addtrace.h"
#include "log.h"
#include "navigate.h"
#include "signalpane.h"
#include "tracepane.h"

TimeType *timearray = NULL;
int timearray_size = 0;

static GtkWidget *ptr_mark_count_label = NULL; /* pointer to mark count label on pattern search dialog */

STrace *straces = NULL;  /* when active, disables certain ops! */

static GtkWidget *window;
static GtkSignalFunc cleanup;

static char *logical[] = {
    "AND", "OR", "XOR", "NAND", "NOR", "XNOR"
};

/* MUST MATCH enum st_stype WITH stype */
enum st_stype
{
    ST_DC, ST_HIGH, ST_MID, ST_X, ST_LOW, ST_STRING,
    ST_RISE, ST_FALL, ST_ANY
};
static char *stype[] = {
    "Don't Care", "High", "Z (Mid)", "X", "Low", "String",
    "Rising Edge", "Falling Edge", "Any Edge"
};

static char logical_mutex[6] = { 0, 0, 0, 0, 0, 0 };

/*
    complain about certain ops conflict with strace...
*/
void strace_error (void)
{
    wave_log (TRUE, "Can't add/delete traces when waveform display search is active!\n");
}

/*
    free the straces...
*/
static void free_straces (void)
{
    STrace *skill, *s = straces;
    int i;

    while (s)
    {
        for (i = 0; i < 9; i++)
        {
            if (s->back[i])
                g_free (s->back[i]);
        }
        if (s->string)
            g_free (s->string);
        skill = s;
        s = s->next;
        g_free (skill);
    }

    straces = NULL;
}

/*
    button/menu/entry activations..
*/
static void logical_clicked (GtkWidget * widget, gpointer which)
{
    int i;
    char *which_char;

    for (i = 0; i < 6; i++)
        logical_mutex[i] = 0;
    which_char = (char *) which;
    *which_char = 1;            /* mark our choice */

    WAVE_DEBUG (printf ("picked: %s\n", logical[which_char - logical_mutex]));
}

static void stype_clicked (GtkWidget * widget, gpointer back)
{
    STraceBack *b = (STraceBack *) back;
    STrace *s = b->parent;

    s->value = b->which;

    WAVE_DEBUG (printf ("Trace %s Search Type: %s\n", s->trace->name, stype[(int) s->value]));
}

static void enter_callback (GtkWidget * widget, gpointer strace_tmp)
{
    gchar *entry_text;
    STrace *s = (STrace *) strace_tmp;

    if (s->string)
    {
        g_free (s->string);
        s->string = NULL;
    }

    entry_text = gtk_entry_get_text (GTK_ENTRY (widget));
    WAVE_DEBUG (printf ("Trace %s Entry contents: %s\n", s->trace->name, entry_text));

    if (! *entry_text)
        return;

    gtk_entry_select_region (GTK_ENTRY (widget), 0, GTK_ENTRY (widget)->text_length);

    s->string = strdup (entry_text);
    g_strup (s->string);
}

static void forwards_callback (GtkWidget * widget, GtkWidget * nothing)
{
/* no cleanup necessary, but do real search */
    WAVE_DEBUG (printf ("Searching Forward..\n"));
    strace_search (STRACE_FORWARD);
}

static void backwards_callback (GtkWidget * widget, GtkWidget * nothing)
{
/* no cleanup necessary, but do real search */
    WAVE_DEBUG (printf ("Searching Backward..\n"));
    strace_search (STRACE_BACKWARD);
}

static void strace_maketimetrace (int mode);

static void mark_callback (GtkWidget * widget, GtkWidget * nothing)
{
    WAVE_DEBUG (printf ("Marking..\n"));
    strace_maketimetrace (1);

    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
}

static void clear_callback (GtkWidget * widget, GtkWidget * nothing)
{
    WAVE_DEBUG (printf ("Clearing..\n"));
    strace_maketimetrace (0);

    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
}

static void destroy_callback (GtkWidget * widget, GtkWidget * nothing)
{
    free_straces ();
    ptr_mark_count_label = NULL;
    gtk_widget_destroy (window);
}


/* update mark count label on pattern search dialog */

static void update_mark_count_label (void)
{
    if (ptr_mark_count_label)
    {
        char mark_count_buf[64];

        if (timearray_size < 1)
            mark_count_buf[0] = 0;
        else
            sprintf (mark_count_buf, "Mark Count: %d", timearray_size >= 1 ? timearray_size - 1 : 0);
        gtk_label_set_text (GTK_LABEL (ptr_mark_count_label), mark_count_buf);
    }
}

void tracesearchbox (char *title, GtkSignalFunc func)
{
    GSList *group;
    GtkWidget *menu, *menuitem, *optionmenu, *entry;
    GtkWidget *vbox, *hbox, *small_hbox, *count_hbox, *vbox_g, *label;
    GtkWidget *button1, *button1a, *button1b, *button1c, *button2;
    GtkWidget *scrolled_win, *frame, *separator;
    TraceEnt *t;
    int i;
    int numtraces;

    if (straces)
    {                           /* is already active */
        gdk_window_raise (window->window);
        return;
    }

    if (wave_add_traces_dialog_is_active ())
    {
        wave_log (TRUE, "Can't activate waveform display search when signal searches are active\n");
        return;
    }

    cleanup = func;
    numtraces = 0;

    /* create a new window */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (window), title);
    gtk_widget_set_usize (GTK_WIDGET (window), 420, -1);
    gtk_signal_connect (GTK_OBJECT (window), "delete_event", GTK_SIGNAL_FUNC (destroy_callback), NULL);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox);
    gtk_widget_show (vbox);

    vbox_g = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox_g);

    frame = gtk_frame_new (NULL);
    gtk_container_border_width (GTK_CONTAINER (frame), 3);
    gtk_widget_show (frame);

    small_hbox = gtk_hbox_new (TRUE, 0);
    gtk_widget_show (small_hbox);

    label = gtk_label_new ("Logical Operation");
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (small_hbox), label, TRUE, FALSE, 0);

    menu = gtk_menu_new ();
    group = NULL;

    for (i = 0; i < 6; i++)
    {
        menuitem = gtk_radio_menu_item_new_with_label (group, logical[i]);
        group = gtk_radio_menu_item_group (GTK_RADIO_MENU_ITEM (menuitem));
        gtk_menu_append (GTK_MENU (menu), menuitem);
        gtk_widget_show (menuitem);
        gtk_signal_connect (GTK_OBJECT (menuitem), "activate", GTK_SIGNAL_FUNC (logical_clicked), &logical_mutex[i]);
        logical_mutex[i] = 0;
    }

    logical_mutex[0] = 1;       /* "and" */

    optionmenu = gtk_option_menu_new ();
    gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu), menu);
    gtk_box_pack_start (GTK_BOX (small_hbox), optionmenu, TRUE, FALSE, 0);
    gtk_widget_show (optionmenu);

    gtk_box_pack_start (GTK_BOX (vbox), small_hbox, FALSE, FALSE, 0);

    scrolled_win = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_set_usize (GTK_WIDGET (scrolled_win), -1, 300);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show (scrolled_win);
    gtk_container_add (GTK_CONTAINER (frame), scrolled_win);
    gtk_container_add (GTK_CONTAINER (vbox), frame);

    for (t = traces.first; t; t = t->next)
    {
        STrace *s;

        if (TRACE_IS_SEPARATOR (t->config) || !TRACE_IS_SELECTED (t->config) || !t->config->name)
            continue;

        numtraces++;
        if (numtraces == 500)
        {
            wave_log (FALSE, "Limiting waveform display search to 500 traces.\n");
            break;
        }

        s = g_new (STrace, 1);

        s->next = straces;
        straces = s;
        s->trace = t;

        if (t != traces.first)
        {
            separator = gtk_hseparator_new ();
            gtk_widget_show (separator);
            gtk_box_pack_start (GTK_BOX (vbox_g), separator, FALSE, FALSE, 0);
        }

        small_hbox = gtk_hbox_new (TRUE, 0);
        gtk_widget_show (small_hbox);

        label = gtk_label_new (t->config->name);
        gtk_widget_show (label);
        gtk_box_pack_start (GTK_BOX (vbox_g), label, FALSE, FALSE, 0);

        menu = gtk_menu_new ();
        group = NULL;

        for (i = 0; i < 9; i++)
        {
            STraceBack *b = g_new (STraceBack, 1);

            b->parent = s;
            b->which = i;
            s->back[i] = b;
            menuitem = gtk_radio_menu_item_new_with_label (group, stype[i]);
            group = gtk_radio_menu_item_group (GTK_RADIO_MENU_ITEM (menuitem));
            gtk_menu_append (GTK_MENU (menu), menuitem);
            gtk_widget_show (menuitem);
            gtk_signal_connect (GTK_OBJECT (menuitem), "activate", GTK_SIGNAL_FUNC (stype_clicked), b);
        }

        optionmenu = gtk_option_menu_new ();
        gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu), menu);
        gtk_box_pack_start (GTK_BOX (small_hbox), optionmenu, TRUE, FALSE, 0);
        gtk_widget_show (optionmenu);

        entry = gtk_entry_new_with_max_length (257); /* %+256ch */
        gtk_signal_connect (GTK_OBJECT (entry), "activate", GTK_SIGNAL_FUNC (enter_callback), s);

        gtk_box_pack_start (GTK_BOX (small_hbox), entry, TRUE, FALSE, 0);
        gtk_widget_show (entry);
        gtk_box_pack_start (GTK_BOX (vbox_g), small_hbox, FALSE, FALSE, 0);
    }

    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled_win), vbox_g);

    /* add a horizontal box to display the mark count */
    count_hbox = gtk_hbox_new (TRUE, 0);
    gtk_widget_show (count_hbox);
    gtk_box_pack_start (GTK_BOX (vbox), count_hbox, FALSE, FALSE, 0);

    ptr_mark_count_label = gtk_label_new ("");
    gtk_widget_show (ptr_mark_count_label);
    gtk_box_pack_start (GTK_BOX (count_hbox), ptr_mark_count_label, TRUE, FALSE, 0);
    update_mark_count_label ();
    /* end of add a horizontal box to display the mark count */

    hbox = gtk_hbox_new (FALSE, 1);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show (hbox);

    button1 = gtk_button_new_with_label ("Fwd");
    gtk_widget_set_usize (button1, 75, -1);
    gtk_signal_connect (GTK_OBJECT (button1), "clicked", GTK_SIGNAL_FUNC (forwards_callback), NULL);
    gtk_widget_show (button1);
    gtk_container_add (GTK_CONTAINER (hbox), button1);
    GTK_WIDGET_SET_FLAGS (button1, GTK_CAN_DEFAULT);
    gtk_signal_connect_object (GTK_OBJECT (button1), "realize",
      GTK_SIGNAL_FUNC (gtk_widget_grab_default), GTK_OBJECT (button1));

    button1a = gtk_button_new_with_label ("Bkwd");
    gtk_widget_set_usize (button1a, 75, -1);
    gtk_signal_connect (GTK_OBJECT (button1a), "clicked", GTK_SIGNAL_FUNC (backwards_callback), NULL);
    gtk_widget_show (button1a);
    gtk_container_add (GTK_CONTAINER (hbox), button1a);
    GTK_WIDGET_SET_FLAGS (button1a, GTK_CAN_DEFAULT);

    button1b = gtk_button_new_with_label ("Mark");
    gtk_widget_set_usize (button1b, 75, -1);
    gtk_signal_connect (GTK_OBJECT (button1b), "clicked", GTK_SIGNAL_FUNC (mark_callback), NULL);
    gtk_widget_show (button1b);
    gtk_container_add (GTK_CONTAINER (hbox), button1b);
    GTK_WIDGET_SET_FLAGS (button1b, GTK_CAN_DEFAULT);

    button1c = gtk_button_new_with_label ("Clear");
    gtk_widget_set_usize (button1c, 75, -1);
    gtk_signal_connect (GTK_OBJECT (button1c), "clicked", GTK_SIGNAL_FUNC (clear_callback), NULL);
    gtk_widget_show (button1c);
    gtk_container_add (GTK_CONTAINER (hbox), button1c);
    GTK_WIDGET_SET_FLAGS (button1c, GTK_CAN_DEFAULT);

    button2 = gtk_button_new_with_label ("Close");
    gtk_widget_set_usize (button2, 75, -1);
    gtk_signal_connect (GTK_OBJECT (button2), "clicked", GTK_SIGNAL_FUNC (destroy_callback), NULL);
    GTK_WIDGET_SET_FLAGS (button2, GTK_CAN_DEFAULT);
    gtk_widget_show (button2);
    gtk_container_add (GTK_CONTAINER (hbox), button2);

    gtk_widget_show (window);
}

/*
    strace backward or forward..
*/
void strace_search (int direction)
{
#if 0
    STrace *s;
    TimeType basetime, maxbase, sttim, fintim;
    TraceEnt *t;
    int totaltraces, passcount;
    int whichpass;
    TimeType middle = 0, width;

    if (direction == STRACE_BACKWARD)
    {                           /* backwards */
        if (tims.marker < 0)
        {
            basetime = MAX_HISTENT_TIME;
        } else
        {
            basetime = tims.marker;
        }
    } else
    {                           /* go forwards */
        if (tims.marker < 0)
        {
            basetime = tims.first;
        } else
        {
            basetime = tims.marker;
        }
    }

    sttim = tims.first;
    fintim = tims.last;

    for (whichpass = 0;; whichpass++)
    {
        if (direction == STRACE_BACKWARD)
        {                       /* backwards */
            maxbase = -1;
            s = straces;
            while (s)
            {
                    HistEnt *h;
                    UTimeType utt;
                    TimeType tt;
                t = s->trace;

                    h = bsearch_node (t->n.nd, basetime);
                    /* FIXME, hp should be the time before h
                    hp = max_compare_index;
                    hp--;
                    h = *hp;
                    */
                    s->his.h = h;
                    utt = h->time;
                    tt = utt;
                    if (tt > maxbase)
                        maxbase = tt;
                s = s->next;
            }
        } else
        {                       /* go forward */
            maxbase = MAX_HISTENT_TIME;
            s = straces;
            while (s)
            {
                    HistEnt *h;
                    UTimeType utt;
                    TimeType tt;
                t = s->trace;

                    h = bsearch_node (t->n.nd, basetime);
                    while (h && h->next && h->time == h->next->time)
                        h = h->next;
                    if ((whichpass) || (tims.marker >= 0))
                        h = h->next;
                    if (!h)
                        return;
                    s->his.h = h;
                    utt = h->time;
                    tt = utt;
                    if (tt < maxbase)
                        maxbase = tt;
                s = s->next;
            }
        }

        s = straces;
        totaltraces = 0;        /* increment when not don't care */
        while (s)
        {
            char str[2];

            t = s->trace;
            s->search_result = 0; /* explicitly must set this */

            if (!t->n.nd->ext)
            {
                if (s->his.h->time != maxbase)
                {
                    s->his.h = bsearch_node (t->n.nd, maxbase);
                    while (s->his.h->next && s->his.h->time == s->his.h->next->time)
                        s->his.h = s->his.h->next;
                }
                str[0] = wave_hist_ent_digital_value (t->config, s->his.h);
                str[1] = '\0';

                switch (s->value)
                {
                case ST_DC:
                    break;

                case ST_HIGH:
                    totaltraces++;
                    if (str[0] == '1')
                        s->search_result = 1;
                    break;

                case ST_RISE:
                    totaltraces++;
                    if (str[0] == '1' && s->his.h->time == maxbase)
                        s->search_result = 1;
                    break;

                case ST_LOW:
                    totaltraces++;
                    if (str[0] == '0')
                        s->search_result = 1;
                    break;

                case ST_FALL:
                    totaltraces++;
                    if ((str[0] == '0') && (s->his.h->time == maxbase))
                        s->search_result = 1;
                    break;

                case ST_MID:
                    totaltraces++;
                    if (str[0] == 'Z')
                        s->search_result = 1;
                    break;

                case ST_X:
                    totaltraces++;
                    if (str[0] == 'X')
                        s->search_result = 1;
                    break;

                case ST_ANY:
                    totaltraces++;
                    if (s->his.h->time == maxbase)
                        s->search_result = 1;
                    break;

                case ST_STRING:
                    totaltraces++;
                    if (s->string)
                        if (strstr (s->string, str))
                            s->search_result = 1;
                    break;

                default:
                    printf ("Internal error: st_type of %d\n", s->value);
                    exit (0);
                }

            }
            s = s->next;
        }

        if ((maxbase < sttim) || (maxbase > fintim))
            return;

        WAVE_DEBUG (printf ("Maxbase: " LL_PRINTF_FORMAT ", total traces: %d\n", maxbase, totaltraces));
        s = straces;
        passcount = 0;
        while (s)
        {
            WAVE_DEBUG (printf ("\tPass: %d, Name: %s\n", s->search_result, s->trace->name));
            if (s->search_result)
                passcount++;
            s = s->next;
        }

        if (totaltraces)
        {
            if (logical_mutex[0])
            {                   /* and */
                if (totaltraces == passcount)
                    break;
            } else if (logical_mutex[1])
            {                   /* or */
                if (passcount)
                    break;
            } else if (logical_mutex[2])
            {                   /* xor */
                if (passcount & 1)
                    break;
            } else if (logical_mutex[3])
            {                   /* nand */
                if (totaltraces != passcount)
                    break;
            } else if (logical_mutex[4])
            {                   /* nor */
                if (!passcount)
                    break;
            } else if (logical_mutex[5])
            {                   /* xnor */
                if (!(passcount & 1))
                    break;
            }
        }

        basetime = maxbase;
    }

    update_markertime (tims.marker = maxbase);

    width = wave_pane_width_in_time_units ();
    if ((tims.marker < tims.start) || (tims.marker >= tims.start + width))
    {
        if ((tims.marker < 0) || (tims.marker < tims.first) || (tims.marker > tims.last))
        {
            if (tims.end > tims.last)
                wave_set_end_time (tims.last);
            middle = (tims.start / 2) + (tims.end / 2);
            if ((tims.start & 1) && (tims.end & 1))
                middle++;
        } else
        {
            middle = tims.marker;
        }

        wave_set_start_time (middle - (width / 2));
        if (tims.start + width > tims.last)
            wave_set_start_time (tims.last - width);
        if (tims.start < tims.first)
            wave_set_start_time (tims.first);
        GTK_ADJUSTMENT (wave_hslider)->value = tims.timecache = tims.start;
    }

    wave_max_signal_length ();
    wave_redraw (wave_window);
#endif
}

/*
    strace forward to make the timetrace
*/
TimeType strace_timetrace (TimeType basetime, int notfirst)
{
#if 0
    STrace *s;
    TimeType maxbase, fintim;
    TraceEnt *t;
    int totaltraces, passcount;
    int whichpass;

    fintim = tims.last;

    for (whichpass = 0;; whichpass++)
    {
        maxbase = MAX_HISTENT_TIME;
        for (s = straces; s; s = s->next)
        {
                HistEnt *h;
                UTimeType utt;
                TimeType tt;
            t = s->trace;

                h = bsearch_node (t->n.nd, basetime);
                s->his.h = h;
                while (h->next && h->time == h->next->time)
                    h = h->next;
                if (whichpass || notfirst)
                    h = h->next;
                if (!h)
                    return (MAX_HISTENT_TIME);
                utt = h->time;
                tt = utt;
                if (tt < maxbase)
                    maxbase = tt;
        }

        s = straces;
        totaltraces = 0;        /* increment when not don't care */
        while (s)
        {
            char str[2];

            t = s->trace;
            s->search_result = 0; /* explicitly must set this */

            if (!t->n.nd->ext)
            {
                if (s->his.h->time != maxbase)
                {
                    s->his.h = bsearch_node (t->n.nd, maxbase);
                    while (s->his.h->next && s->his.h->time == s->his.h->next->time)
                        s->his.h = s->his.h->next;
                }
                str[0] = wave_hist_ent_digital_value (t->config, s->his.h);
                str[1] = '\0';

                switch (s->value)
                {
                case ST_DC:
                    break;

                case ST_HIGH:
                    totaltraces++;
                    if (str[0] == '1')
                        s->search_result = 1;
                    break;

                case ST_RISE:
                    totaltraces++;
                    if (str[0] == '1' && s->his.h->time == maxbase)
                        s->search_result = 1;
                    break;

                case ST_LOW:
                    totaltraces++;
                    if (str[0] == '0')
                        s->search_result = 1;
                    break;

                case ST_FALL:
                    totaltraces++;
                    if (str[0] == '0' && s->his.h->time == maxbase)
                        s->search_result = 1;
                    break;

                case ST_MID:
                    totaltraces++;
                    if (str[0] == 'Z')
                        s->search_result = 1;
                    break;

                case ST_X:
                    totaltraces++;
                    if (str[0] == 'X')
                        s->search_result = 1;
                    break;

                case ST_ANY:
                    totaltraces++;
                    if (s->his.h->time == maxbase)
                        s->search_result = 1;
                    break;

                case ST_STRING:
                    totaltraces++;
                    if (s->string && strstr (s->string, str))
                        s->search_result = 1;
                    break;

                default:
                    fprintf (stderr, "Internal error: st_type of %d\n", s->value);
                    exit (EXIT_FAILURE);
                }

            }
            s = s->next;
        }

        if (maxbase > fintim)
            return (MAX_HISTENT_TIME);

        WAVE_DEBUG (printf ("Maxbase: " LL_PRINTF_FORMAT ", total traces: %d\n", maxbase, totaltraces));
        s = straces;
        passcount = 0;
        while (s)
        {
            WAVE_DEBUG (printf ("\tPass: %d, Name: %s\n", s->search_result, s->trace->name));
            if (s->search_result)
                passcount++;
            s = s->next;
        }

        if (totaltraces)
        {
            if (logical_mutex[0])
            {                   /* and */
                if (totaltraces == passcount)
                    break;
            } else if (logical_mutex[1])
            {                   /* or */
                if (passcount)
                    break;
            } else if (logical_mutex[2])
            {                   /* xor */
                if (passcount & 1)
                    break;
            } else if (logical_mutex[3])
            {                   /* nand */
                if (totaltraces != passcount)
                    break;
            } else if (logical_mutex[4])
            {                   /* nor */
                if (!passcount)
                    break;
            } else if (logical_mutex[5])
            {                   /* xnor */
                if (!(passcount & 1))
                    break;
            }
        }

        basetime = maxbase;
    }

    return (maxbase);
#endif
	return 0;
}

static void strace_maketimetrace (int mode)
{
    TimeType basetime = tims.first;
    int i, notfirst = 0;
    TimeChain *tchead = NULL, *tc = NULL;
    TimeType *t;

    if (timearray)
    {
        g_free (timearray);
        timearray = NULL;
    }

    timearray_size = 0;

    if (!mode)
    {
        update_mark_count_label ();
        return;                 /* merely free stuff up */
    }

    do
    {
        basetime = strace_timetrace (basetime, notfirst);
        notfirst = 1;

        if (!tc)
        {
            tchead = tc = wave_alloca (sizeof (TimeChain));
        } else
        {
            tc->next = wave_alloca (sizeof (TimeChain));
            tc = tc->next;
        }

        tc->t = basetime;
        tc->next = NULL;
        timearray_size++;
    }
    while (basetime != MAX_HISTENT_TIME);

    timearray = t = g_new (TimeType, timearray_size);
    for (i = 0; i < timearray_size; i++)
    {
        *(t++) = tchead->t;
        tchead = tchead->next;
    }

    update_mark_count_label ();
}
