/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `support.c'
    Support for GTK+, originally from Glade, hand modded.

    $Id: support.c,v 1.13 2003/05/05 20:27:15 bardslea Exp $
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include "support.h"

#ifdef BUILTIN_PIXMAPS
#include "../pixmaps/insert_trace.xpm"
#include "../pixmaps/wave_alert.xpm"
#include "../pixmaps/stock_cut.xpm"
#include "../pixmaps/stock_zoom_out.xpm"
#include "../pixmaps/stock_zoom_in.xpm"
#include "../pixmaps/stock_zoom_fit.xpm"
#include "../pixmaps/stock_first.xpm"
#include "../pixmaps/stock_last.xpm"
#include "../pixmaps/stock_undo.xpm"
#include "../pixmaps/stock_redo.xpm"
#include "../pixmaps/prev_page.xpm"
#include "../pixmaps/next_page.xpm"
#include "../pixmaps/stock_left.xpm"
#include "../pixmaps/stock_right.xpm"
#include "../pixmaps/stock_redo_16.xpm"
#include "../pixmaps/menu_16.xpm"
#include "../pixmaps/stock_open_16.xpm"
#include "../pixmaps/time_16.xpm"
#endif

#ifdef BUILTIN_PIXMAPS
/* Hash of builtin pixmap names to GtkWidgets for those pixmaps */
static GHashTable *builtin_pixmaps = NULL;
#endif

/* This is an internally used function to check if a pixmap file exists. */
static gchar *check_file_exists (const gchar * directory, const gchar * filename);

static GtkWidget *create_pixmap_from_data (GtkWidget * widget, char **data);

/* This is an internally used function to create pixmaps. */
static GtkWidget *create_dummy_pixmap (GtkWidget * widget);

/* widget_top_parent : get the top parent (the parent that has no parent) of the
	given widget */
GtkWidget *widget_top_parent (GtkWidget *widget)
{
    GtkWidget *parent;

    for (;;)
    {
        if (GTK_IS_MENU (widget))
            parent = gtk_menu_get_attach_widget (GTK_MENU (widget));
        else
            parent = widget->parent;
        if (parent == NULL)
            break;
        widget = parent;
    }

    return widget;
}

/* lookup_widget : lookup a widget by finding a data entry in the top
	parent of the given widget */
GtkWidget *lookup_widget (GtkWidget * widget, const gchar * widget_name)
{
    GtkWidget *parent = widget_top_parent (widget);
    GtkWidget *found_widget;

    for (;;)
    {
        if (GTK_IS_MENU (widget))
            parent = gtk_menu_get_attach_widget (GTK_MENU (widget));
        else
            parent = widget->parent;
        if (parent == NULL)
            break;
        widget = parent;
    }

    found_widget = (GtkWidget *) gtk_object_get_data (GTK_OBJECT (widget), widget_name);
    if (!found_widget)
        g_warning ("Widget not found: %s", widget_name);
    return found_widget;
}

/* This is a dummy pixmap we use when a pixmap can't be found. */
static char *dummy_pixmap_xpm[] = {
/* columns rows colors chars-per-pixel */
    "1 1 1 1",
    "  c None",
/* pixels */
    " "
};

static GtkWidget *create_pixmap_from_data (GtkWidget * widget, char **data)
{
    GdkColormap *colormap;
    GdkPixmap *gdkpixmap;
    GdkBitmap *mask;
    GtkWidget *pixmap;

    colormap = gtk_widget_get_colormap (widget);
    gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, colormap, &mask, NULL, data);
    if (gdkpixmap == NULL)
        g_error ("Couldn't create replacement pixmap.");
    pixmap = gtk_pixmap_new (gdkpixmap, mask);
    gdk_pixmap_unref (gdkpixmap);
    gdk_bitmap_unref (mask);
    return pixmap;
}

/* This is an internally used function to create pixmaps. */
static GtkWidget *create_dummy_pixmap (GtkWidget * widget)
{
    return create_pixmap_from_data (widget, dummy_pixmap_xpm);
}

static GList *pixmaps_directories = NULL;

/* Use this function to set the directory containing installed pixmaps. */
static void add_pixmap_directory (const gchar * directory)
{
    pixmaps_directories = g_list_prepend (pixmaps_directories, g_strdup (directory));
}

#ifdef BUILTIN_PIXMAPS
/* add_builtin_pixmap : add a pixmap to the internal pixmap hash for a pixmap
	built in to the binary rather than loaded from a file */
static void add_builtin_pixmap (const char *name, GtkWidget * widget, char **data)
{
    g_hash_table_insert (builtin_pixmaps, (gpointer) name, create_pixmap_from_data (widget, data));
}
#endif

/* wave_register_pixmaps : add the pixmap directory to the pixmap path and register
	builtin pixmaps */
void wave_register_pixmaps (GtkWidget * widget)
{
    char *pixmap_directory = g_strdup (PIXMAP_DIRECTORY);
    char *wave_share_dir = g_getenv ("WAVE_SHARE_DIR");
    char *ptr = strchr (pixmap_directory, '/');

    /* Substitute slashes for OS dependent directory separator */
    while (ptr)
    {
        *ptr = G_DIR_SEPARATOR;
        ptr = strchr (ptr + 1, '/');
    }
    add_pixmap_directory (pixmap_directory);
    g_free (pixmap_directory);

    /* Search in the WAVE_SHARE_DIR/pixmaps directory first */
    if (wave_share_dir)
    {
        char *dir = g_strdup_printf ("%s%cpixmaps", wave_share_dir, G_DIR_SEPARATOR);

        add_pixmap_directory (dir);
        g_free (dir);
    }
#ifdef BUILTIN_PIXMAPS
    builtin_pixmaps = g_hash_table_new (g_str_hash, g_str_equal);

    add_builtin_pixmap ("insert_trace.xpm", widget, insert_trace_xpm);
    add_builtin_pixmap ("wave_alert.xpm", widget, wave_alert);
    add_builtin_pixmap ("stock_cut.xpm", widget, stock_cut_xpm);
    add_builtin_pixmap ("stock_zoom_out.xpm", widget, stock_zoom_out_xpm);
    add_builtin_pixmap ("stock_zoom_in.xpm", widget, stock_zoom_in_xpm);
    add_builtin_pixmap ("stock_zoom_fit.xpm", widget, stock_zoom_fit_xpm);
    add_builtin_pixmap ("stock_first.xpm", widget, stock_first_xpm);
    add_builtin_pixmap ("stock_last.xpm", widget, stock_last_xpm);
    add_builtin_pixmap ("stock_undo.xpm", widget, stock_undo_xpm);
    add_builtin_pixmap ("stock_redo.xpm", widget, stock_redo_xpm);
    add_builtin_pixmap ("prev_page.xpm", widget, prev_page_xpm);
    add_builtin_pixmap ("next_page.xpm", widget, next_page_xpm);
    add_builtin_pixmap ("stock_left.xpm", widget, stock_left_xpm);
    add_builtin_pixmap ("stock_right.xpm", widget, stock_right_xpm);

    add_builtin_pixmap ("stock_redo_16.xpm", widget, stock_redo_16_xpm);
    add_builtin_pixmap ("menu_16.xpm", widget, menu_16_xpm);
    add_builtin_pixmap ("stock_open_16.xpm", widget, stock_open_16_xpm);
    add_builtin_pixmap ("time_16.xpm", widget, time_16_xpm);
#endif
}

/* create_pixmap : create a pixmap widget for the pixmap found in file `filename'
	(or a built-in pixmap identified as `filename') */
GtkWidget *create_pixmap (GtkWidget * widget, const gchar * filename)
{
    gchar *found_filename = NULL;
    GdkColormap *colormap;
    GdkPixmap *gdkpixmap;
    GdkBitmap *mask;
    GtkWidget *pixmap;
    GList *elem;

    if (!filename || !filename[0])
        return create_dummy_pixmap (widget);

#ifdef BUILTIN_PIXMAPS
    pixmap = (GtkWidget *) g_hash_table_lookup (builtin_pixmaps, filename);

    if (pixmap)
        return pixmap;
#endif

    /* We first try any pixmaps directories set by the application. */
    elem = pixmaps_directories;
    while (elem)
    {
        found_filename = check_file_exists ((gchar *) elem->data, filename);
        if (found_filename)
            break;
        elem = elem->next;
    }

    /* If we haven't found the pixmap, try the source directory. */
    if (!found_filename)
    {
        found_filename = check_file_exists ("../pixmaps", filename);
    }

    if (!found_filename)
    {
        g_warning ("Couldn't find pixmap file: %s", filename);
        return create_dummy_pixmap (widget);
    }

    colormap = gtk_widget_get_colormap (widget);
    gdkpixmap = gdk_pixmap_colormap_create_from_xpm (NULL, colormap, &mask, NULL, found_filename);
    if (gdkpixmap == NULL)
    {
        g_warning ("Error loading pixmap file: %s", found_filename);
        g_free (found_filename);
        return create_dummy_pixmap (widget);
    }
    g_free (found_filename);
    pixmap = gtk_pixmap_new (gdkpixmap, mask);
    gdk_pixmap_unref (gdkpixmap);
    gdk_bitmap_unref (mask);
    return pixmap;
}

/* This is an internally used function to check if a pixmap file exists. */
static gchar *check_file_exists (const gchar * directory, const gchar * filename)
{
    gchar *full_filename;
    struct stat s;
    gint status;

    full_filename = (gchar *) g_malloc (strlen (directory) + 1 + strlen (filename) + 1);
    strcpy (full_filename, directory);
    strcat (full_filename, G_DIR_SEPARATOR_S);
    strcat (full_filename, filename);

    status = stat (full_filename, &s);
    if (status == 0 && S_ISREG (s.st_mode))
        return full_filename;
    g_free (full_filename);
    return NULL;
}

/* wave_get_option_menu_active_data : get a named data element from the active element of
	the given option menu, returns NULL on any errors */
gpointer wave_get_option_menu_active_data (GtkOptionMenu * option_menu, const char *data_element_name)
{
    GtkWidget *menu, *selected_item;

    menu = gtk_option_menu_get_menu (option_menu);
    if (!menu)
        return NULL;
    selected_item = gtk_menu_get_active (GTK_MENU (menu));
    if (!selected_item)
        return NULL;

    return gtk_object_get_data (GTK_OBJECT (selected_item), data_element_name);
}

/* wave_get_option_menu_index : get the index of the active element in the given option menu */
guint wave_get_option_menu_index (GtkOptionMenu * option_menu)
{
    GtkWidget *menu = gtk_option_menu_get_menu (option_menu);
    GtkWidget *active_item = gtk_menu_get_active (GTK_MENU (menu));

    return (guint) g_list_index (GTK_MENU_SHELL (menu)->children, active_item);
}

/* wave_option_menu_set_menu_from_string_array : remove the current menu from
	an option menu and replace with a menu composed of the strings provided.
	If `data_entry_name' is not NULL then set data for each entry to be either
	the index of the entry (if `dataNindex' is FALSE) or to the element in
	the array `data' matching the index of this entry (if `dataNindex' is TRUE).
	If `callback' is not NULL then set an activate callback, passed the same data as set
	for each enrty, for each entry. */
void wave_option_menu_set_menu_from_string_array (GtkOptionMenu *option_menu,
	char **strings, void **data, char *data_entry_name, gboolean dataNindex,
	void (* callback) (GtkMenuItem *, void *))
{
	GtkWidget *menu = gtk_menu_new ();
	int i = 0;

	gtk_option_menu_remove_menu (option_menu);

	while (*strings)
	{
        GtkWidget *menu_item = gtk_menu_item_new_with_label (*strings);
        void *data_entry = (dataNindex ? data[i]: GINT_TO_POINTER (i));

        if (data_entry_name)
			gtk_object_set_data (GTK_OBJECT (menu_item), data_entry_name, data_entry);

		if (callback)
			gtk_signal_connect (GTK_OBJECT (menu_item), "activate", GTK_SIGNAL_FUNC (callback), data_entry);

        gtk_menu_append (GTK_MENU (menu), menu_item);
        gtk_widget_show (menu_item);
		strings ++;
		i ++;
	}

    gtk_option_menu_set_menu (option_menu, menu);
}

/* wave_set_sensitivity_of_named_widgets : set each of the named widgets'
	sensitivities.  `widget_names' entries are found by using lookup_widget */
void wave_set_sensitivity_of_named_widgets (GtkWidget *reference_widget,
	char **widget_names, gboolean sensitivity)
{
	while (*widget_names)
	{
		gtk_widget_set_sensitive (lookup_widget (reference_widget, *widget_names),
			sensitivity);
		widget_names ++;
	}
}

/* wave_show_dialog : raise/show/otherwise draw the user's attention to the
	dialogue box `dialog' (which must actually be a valid widget) */
void wave_show_dialog (GtkWidget *dialog)
{
	if (dialog && GTK_WIDGET_REALIZED (dialog))
	{
		gdk_window_show (dialog->window);
		gdk_window_raise (dialog->window);
	}
}
