#include <glib.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include "fileio.h"

void wave_exit (int return_val)
{
	exit (return_val);
}

/* wave_log : error logging to window or stderr */
void wave_log (gboolean is_error, const char *fmt, ...)
{
    va_list args;

    va_start (args, fmt);
    vfprintf (stderr, fmt, args);
    va_end (args);
}

/* wave_log_context : print an error context header for a subsequent wave_log call */
void wave_log_context (WaveErrorContext *context)
{
	if (context)
	{
		wave_log (FALSE, "%s:%s:%d: ", (context->operation_name ? context->operation_name : ""),
			(context->filename ? context->filename : ""), context->line_num);
	}
}

void child_death_handler (int not_used)
{
	int status;
	wait (&status);

	exit (EXIT_SUCCESS);
}

void prompt (void)
{
	fprintf (stderr, "> ");
}

void rr (FileDescLineParser *inp, int inp_fd, FileDescLineParser *stdin_parser, FILE *out)
{
	prompt ();

	while (1)
	{
		fd_set inp_or_stdin;
		char *stdin_line, *inp_line;

		FD_ZERO (&inp_or_stdin);
		FD_SET (inp_fd, &inp_or_stdin);
		FD_SET (0, &inp_or_stdin);

		if (select (inp_fd + 1 /* max with stdin + 1 */, &inp_or_stdin, NULL, NULL, NULL) > 0)
		{
			if (FD_ISSET (inp_fd, &inp_or_stdin))
			{
				inp_line = file_desc_line_parser_read_line (inp, FALSE);
				while (inp_line)
				{
					fprintf (stderr, " %s\n", inp_line);
					inp_line = file_desc_line_parser_read_line (inp, TRUE);
				}
			}

			if (FD_ISSET (0, &inp_or_stdin))
			{
				stdin_line = file_desc_line_parser_read_line (stdin_parser, FALSE);
				while (stdin_line)
				{
					fprintf (out, "%s\n", stdin_line);
					fflush (out);
					stdin_line = file_desc_line_parser_read_line (stdin_parser, TRUE);
				}
			}
		}
		prompt ();
	}
}

int main (int argc, char **argv)
{
	char *program_name;
	pid_t new_pid;
	int from_child[2];
	int to_child[2];
	struct sigaction child_death;

	if (argc < 2)
	{
		fprintf (stderr, "usage: talk <binary> <args>\n");
		return EXIT_FAILURE;
	}
	program_name = argv[1];

	if (access (program_name, X_OK) != 0)
	{
		fprintf (stderr, "Can't execute file `%s'\n", program_name);
		return EXIT_FAILURE;
	}

	if (pipe (from_child) != 0 || pipe (to_child) != 0)
	{
		fprintf (stderr, "Problem opening pipes\n");
		return EXIT_FAILURE;
	}

	child_death.sa_handler = child_death_handler;
	child_death.sa_flags = 0;
	sigemptyset (&child_death.sa_mask);

	sigaction (SIGCHLD, &child_death, NULL);
	new_pid = fork ();

	if (new_pid == 0)
	{
		/* Horrible crap to make sure from_child[1] ends up as fildes 3 and
			to_child[0] as 4 */
		int keep_from_child = dup (from_child[1]);
		int keep_to_child = dup (to_child[0]);
		close (from_child[0]);
		close (from_child[1]);
		close (to_child[0]);
		close (to_child[1]);
		if (dup2 (keep_to_child, 3) < 0 || dup2 (keep_from_child, 4) < 0)
		{
			fprintf (stderr, "Problem making child fildess\n");
			return EXIT_FAILURE;
		}
		close (keep_to_child);
		close (keep_from_child);
		close (0);
		execv (program_name, argv + 1);
	} else {
		FileDesc *inp_file_desc;
		FileDesc *stdin_file_desc;
		FILE *out;
		FileDescLineParser *inp_parser;
		FileDescLineParser *stdin_parser;

		close (from_child[1]);
		close (to_child[0]);

		inp_file_desc = file_desc_fdopen (from_child[0], "rt");
		stdin_file_desc = file_desc_fdopen (0, "rt");
		out = fdopen (to_child[1], "wt");

		if (! inp_file_desc || ! stdin_file_desc || ! out)
		{
			fprintf (stderr, "Problem fd-opening pipes\n");
			return EXIT_FAILURE;
		}

		inp_parser = file_desc_line_parser_new (inp_file_desc);
		stdin_parser = file_desc_line_parser_new (stdin_file_desc);


		rr (inp_parser, from_child[0], stdin_parser, out);
	}

	return EXIT_SUCCESS;
}
