/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    
    `tracefile.h'
    Saved traces file parsing

    $Id: tracefile.c,v 1.26 2003/11/11 22:53:28 bardslea Exp $
*/

#include <string.h>
#include "tracefile.h"
#include "reader.h"
#include "currenttime.h"
#include "analyzer.h"
#include "utils.h"
#include "log.h"
#include "main.h"
#include "help.h"
#include "gui_if.h"
#include "navigate.h"
#include "signalpane.h"
#include "tracepane.h"
#include "support.h"

/* make_trace_node_bit_for_node : make a TraceNodeBit matching the given node */
static TraceNodeBit *make_trace_node_bit_for_node (Node * node)
{
    TraceNodeBit *ret = NULL;

    ret = new_trace_node_bit (node->nname, -1);

    return ret;
}

/* single_bit_node_from_trace_node_bit : find the right node to represent a trace_node_bit */
static Node *make_single_bit_node_from_symbol (TraceNodeBit * node_bit)
{
    Node *ret = NULL;
    Symbol *symbol = symfind (node_bit->node_name);

    if (symbol)
    {
        if (node_bit->lrindex != -1) /* There is an index */
        {
        	/* FIXME
            ret = extract_node_single_bit (symbol->n, node_bit->lrindex);
            */
            ret = NULL;

            if (!ret)
                wave_log (TRUE, "Can't extract index `%d' from node `%s'\n", node_bit->lrindex, node_bit->node_name);
        } else
            ret = symbol->n;
    } else
        wave_log (TRUE, "Can't find symbol `%s'\n", node_bit->node_name);

    return ret;
}

#if 0
/* make_bits_from_trace_node_bits : create a vector from the signal list in
	a TraceConfig read from a saved traces file */
static Bits *make_bits_from_trace_node_bits (const char *vector_name, GList * node_bits)
#endif

/* update_wave_config_from_traces : make the `wave_config->traces'
	element match the current state of the wave_drawing_area as described by the TraceEnts
	linked from `traces' */
void update_wave_config_from_traces (TraceEnt * traces, WaveConfig * wave_config)
{
    GList *trace_configs = NULL;
    TraceEnt *traces_iter = traces;
    TraceConfig *extended_trace = NULL;

    while (traces_iter)
    {
        TraceConfig *config = traces_iter->config;

        if (! TRACE_IS_EXTENSION (config))
        {
            /* Update the height of traces from the observed number of
               extension traces */
            extended_trace = config;
            extended_trace->height = 1;
            trace_configs = g_list_prepend (trace_configs, config);

			if (! TRACE_IS_SEPARATOR (config))
			{
				config->signals = g_list_prepend (NULL,
					make_trace_node_bit_for_node (traces_iter->n.nd));
			}
        } else
        {
            if (extended_trace)
                extended_trace->height++;
        }
        traces_iter = traces_iter->next;
    }
    if (wave_config->traces)
        g_list_free (wave_config->traces);
    wave_config->traces = g_list_reverse (trace_configs);
}

static void add_trace_from_trace_config (TraceConfig * trace_config)
{
    TraceEnt *trace = NULL;

    if (TRACE_IS_SEPARATOR (trace_config))
        trace = wave_insert_separator_trace (NULL, trace_config, NULL);
    else if (trace_config->signals) /* Aliased or vector names */
    {
        int signal_count = g_list_length (trace_config->signals);

        if (signal_count == 1)  /* Single signal, dont' both with vectoring */
        {
            Node *node = make_single_bit_node_from_symbol ((TraceNodeBit *) trace_config->signals->data);

            if (node)
                trace = wave_insert_node_trace (node, trace_config);
        } else {
        	wave_log (TRUE, "FIXME, need to readd vector support\n");
        }
    } else
    {                           /* No named signal, just in case */
        Symbol *symbol = symfind (trace_config->name);

        if (symbol)
            trace = wave_insert_node_trace (symbol->n, trace_config);
        else
            wave_log (TRUE, "Can't find signal `%s' to add to trace view\n", trace_config->name);
    }
}

/* add_traces_from_trace_config_list : add each trace from a parsed WaveConfig
	to the wave_drawing_area trace view pane */
void add_traces_from_trace_config_list (GList * trace_configs)
{
    g_list_foreach (trace_configs, (GFunc) add_trace_from_trace_config, NULL);
	wave_signal_pane_adjust_after_insertion (wave_window);
}

/* Callbacks and GUI stuff */
static char *traces_filename = NULL;

void on_file_open_traces_activate (GtkMenuItem * menuitem, gpointer user_data)
{
	GtkWidget *load_traces;

    if (wave_help ("Open Traces", "lets you load a set of trace definitions for the current file previously saved with Save Traces, including the formats of all the traces and the positions of all the markers"))
        return;

    if (wave_dnd_state)
    {                           /* don't mess with sigs when wave_dnd active */
        wave_dnd_error ();
        return;
    }

    load_traces = create_load_traces ();
    if (traces_filename)
        gtk_file_selection_set_filename (GTK_FILE_SELECTION (load_traces), traces_filename);

    gtk_widget_show (load_traces);
}

void on_load_traces_ok_clicked (GtkButton * button, gpointer user_data)
{
    char *filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (
		widget_top_parent (GTK_WIDGET (button))));

    if (traces_filename)
        g_free (traces_filename);
    filename = g_strdup (filename);
    traces_filename = filename;
    on_load_traces_cancel_clicked (button, NULL);

    if (!*filename)
    {
        wave_log (TRUE, "must specify a filename!\n");
        return;
    }

    /* Need to deallocate wave_config, FIXME */
    /* free_wave_config (wave_config); */
    wave_config = new_wave_config ();
    if (!parse_wave_config_file (filename, wave_config))
    {
        wave_log (TRUE, "Couldn't successfully parse saved traces file `%s'\n", filename);
        return;
    }

    add_traces_from_trace_config_list (wave_config->traces);
    update_markertime (tims.marker);
    wave_max_signal_length (wave_window);
    wave_redraw (wave_window);
}

void on_load_traces_cancel_clicked (GtkButton * button, gpointer user_data)
{
    /*gtk_grab_remove(load_traces); */
    gtk_widget_destroy (widget_top_parent (GTK_WIDGET (button)));
}

void on_file_save_traces_activate (GtkMenuItem * menuitem, gpointer user_data)
{
	GtkWidget *save_traces;

    if (wave_help ("Save Traces", "lets you save the current trace definitions for loading later with Open Traces, including the formats of all the traces and the positions of all the markers"))
        return;

    save_traces = create_save_traces ();
    if (traces_filename)
        gtk_file_selection_set_filename (GTK_FILE_SELECTION (save_traces), traces_filename);

    gtk_widget_show (save_traces);
}

/* debug_write_trace : print some info about the traces */
static void debug_write_traces (TraceEnt * traces)
{
    if (!wave_debug_ab)
        return;

    while (traces)
    {
        Node *node = traces->n.nd;

        fprintf (stderr, "node %s", node->nname);
        fprintf (stderr, "\n");

        traces = traces->next;
    }
}

void on_save_traces_ok_clicked (GtkButton * button, gpointer user_data)
{
    char *filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (
		widget_top_parent (GTK_WIDGET (button))));

    if (traces_filename)
        g_free (traces_filename);
    filename = g_strdup (filename);
    traces_filename = filename;
    on_save_traces_cancel_clicked (button, NULL); /* Use cancel func. to destroy dialogue */

    if (!*filename)
        wave_log (TRUE, "Must specify a saved traces file filename!\n");
    else
    {
        update_wave_config_from_traces (traces.first, wave_config);
        debug_write_traces (traces.first);
        write_wave_config_file (filename, wave_config, "GTKWave saved traces, version " VERSION);
    }
}

void on_save_traces_cancel_clicked (GtkButton * button, gpointer user_data)
{
    gtk_widget_destroy (widget_top_parent (GTK_WIDGET (button)));
}

