/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `tracepane.c'
    Trace display rendering

    $Id: tracepane.c,v 1.10 2004/04/25 19:23:42 bardslea Exp $
*/

#include "tracepane.h"
#include "signalpane.h"
#include "color.h"
#include "navigate.h"
#include "rc.h"
#include "bsearch.h"
#include "value.h"
#include "marker.h"
#include "support.h"

GdkFont *wavefont = NULL;
GdkFont *wavefont_smaller = NULL;
WaveDrawingContext *wave_window_drawing_context = NULL;

/* Default, to window, drawing context */
static gint m1x, m2x;
static int in_button_press = 0;
static gfloat old_wvalue = -1.0;
static TimeType prevtim = -1;
static GdkPixmap *wavepixmap = NULL;

static const GdkModifierType bmask[4] = { 0, GDK_BUTTON1_MASK, 0, GDK_BUTTON3_MASK }; /* button 1, 3 press/rel encodings */
static const GdkModifierType m_bmask[4] = { 0, GDK_BUTTON1_MOTION_MASK, 0, GDK_BUTTON3_MOTION_MASK }; /* button 1, 3 motion encodings */

static void render_named_markers (WaveDrawingContext *context, WaveWindow *window)
{
    int i;
    gdouble pixstep = ((gdouble) window->traces.scale.time_units_per_frame) / window->traces.scale.pixels_per_frame;

    for (i = 0; i < 26; i ++)
    {
        TimeType t = named_markers[i];

        if (t != -1 && t >= tims.start && t <= tims.last && t <= tims.end)
        {
            gint xl = ((gdouble) (t - tims.start)) / pixstep; /* snap to integer */

            if ((xl >= 0) && (xl < window->traces.scale.wave_pane_width))
            {
                char nbuff[2];
                gint y, width;

                nbuff[0] = 'A' + i;
                nbuff[1] = 0x00;

                for (y = fontheight - 1; y <= window->traces.scale.wave_pane_height - 1; y += 8)
					WAVE_DRAW_PANE_LINE (context, TRACE_PANE_NAMED_MARKERS, xl, y, xl, y + 5);

                width = gdk_string_measure (wavefont_smaller, nbuff);
				WAVE_DRAW_PANE_STRING (context, TRACE_PANE_NAMED_MARKERS, WAVE_FONT_SMALL,
					xl - (width >> 1), fontheight - 2, nbuff);
            }
        }
    }
}

static void sync_marker (WaveWindow *window)
{
    if ((tims.prevmarker == -1) && (tims.marker != -1))
    {
        window->signals.window_width_dirty = TRUE;
    } else if ((tims.marker == -1) && (tims.prevmarker != -1))
    {
        window->signals.window_width_dirty = TRUE;
    }
    tims.prevmarker = tims.marker;
}

/* AB */
static char *primary_marker_name = "1";
static int primary_marker_name_width = 0;

/* /AB */

static void draw_marker (WaveWindow *window)
{
    gdouble pixstep;
    gint xl;
    GdkGC *gc_umark = wave_get_pane_color_gc (wave_config->pane_colors, TRACE_PANE_PRIMARY_MARKER, window);

    m1x = m2x = -1;

    if (tims.marker >= 0 && tims.marker >= tims.start && tims.marker <= tims.last && tims.marker <= tims.end)
    {
        pixstep = ((gdouble) window->traces.scale.time_units_per_frame) / ((gdouble) window->traces.scale.pixels_per_frame);
        xl = ((gdouble) (tims.marker - tims.start)) / pixstep; /* snap to integer */
        if ((xl >= 0) && (xl < window->traces.scale.wave_pane_width))
        {
            gdk_draw_line (window->traces.drawing_area->window, gc_umark, xl, fontheight - 1, xl, window->traces.scale.wave_pane_height - 1);
            m1x = xl;

            /* AB */
            if (FALSE && !in_button_press)
            {
                if (!primary_marker_name_width)
                {
                    primary_marker_name_width = gdk_string_measure (wavefont_smaller, primary_marker_name) >> 1;
                }

                gdk_draw_string (wavepixmap, wavefont_smaller, gc_umark,
                  xl - primary_marker_name_width, fontheight - 2, primary_marker_name);
            }
            /* /AB */
        }
    }

    if (wave_flags.enable_ghost_marker && in_button_press && tims.lmbcache >= 0
      && tims.lmbcache >= tims.start && tims.lmbcache <= tims.last && tims.lmbcache <= tims.end)
    {
        pixstep = ((gdouble) window->traces.scale.time_units_per_frame) / ((gdouble) window->traces.scale.pixels_per_frame);
        xl = ((gdouble) (tims.lmbcache - tims.start)) / pixstep; /* snap to integer */
        if (xl >= 0 && xl < window->traces.scale.wave_pane_width)
        {
            gdk_draw_line (window->traces.drawing_area->window, gc_umark, xl, fontheight - 1, xl, window->traces.scale.wave_pane_height - 1);
            m2x = xl;
        }
    }

    if (m1x > m2x)
    {                           /* ensure m1x <= m2x for partitioned refresh */
        gint t = m1x;

        m1x = m2x;
        m2x = t;
    }

    if (m1x == -1)
        m1x = m2x;              /* make both markers same if no ghost marker or v.v. */
}

static void draw_marker_partitions (WaveWindow *window)
{
    GdkGC *gc;

    draw_marker (window);

    gc = window->traces.drawing_area->style->fg_gc[GTK_WIDGET_STATE (window->traces.drawing_area)];
    if (m1x == m2x)
    {
        gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m1x, 0, m1x, 0, 1, fontheight - 2);

        if (m1x < 0)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, 0, 0, 0, 0, window->traces.scale.wave_pane_width, window->traces.scale.wave_pane_height);
        else if (m1x == 0)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, 1, 0, 1, 0, window->traces.scale.wave_pane_width - 1, window->traces.scale.wave_pane_height);
        else if (m1x == window->traces.scale.wave_pane_width - 1)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, 0, 0, 0, 0, window->traces.scale.wave_pane_width - 1, window->traces.scale.wave_pane_height);
        else
        {
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, 0, 0, 0, 0, m1x, window->traces.scale.wave_pane_height);
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m1x + 1, 0, m1x + 1, 0, window->traces.scale.wave_pane_width - m1x - 1,
              window->traces.scale.wave_pane_height);
        }
    } else
    {
        gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m1x, 0, m1x, 0, 1, fontheight - 2);
        gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m2x, 0, m2x, 0, 1, fontheight - 2);

        if (m1x > 0)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, 0, 0, 0, 0, m1x, window->traces.scale.wave_pane_height);

        if (m2x - m1x > 1)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m1x + 1, 0, m1x + 1, 0, m2x - m1x - 1,
              window->traces.scale.wave_pane_height);

        if (m2x != window->traces.scale.wave_pane_width - 1)
            gdk_draw_pixmap (window->traces.drawing_area->window, gc, wavepixmap, m2x + 1, 0, m2x + 1, 0, window->traces.scale.wave_pane_width - m2x - 1,
              window->traces.scale.wave_pane_height);
    }
}

static void render_blackout (WaveWindow *window)
{
    gfloat pageinc;
    TimeType lhs, rhs, lclip, rclip;
    BlackoutRegion *bt = blackout_regions;

    if (bt)
    {
        GdkGC *gc_xfill = wave_get_pane_color_gc (wave_config->pane_colors, TRACE_PANE_GRID, window);

        pageinc = wave_pane_width_in_time_units (window);
        lhs = tims.start;
        rhs = pageinc + lhs;

        while (bt)
        {
            if (((bt->bstart <= lhs) && (bt->bend >= lhs)) || ((bt->bstart >= lhs) && (bt->bstart <= rhs)))
            {
                lclip = bt->bstart;
                rclip = bt->bend;

                if (lclip < lhs)
                    lclip = lhs;
                else if (lclip > rhs)
                    lclip = rhs;

                if (rclip < lhs)
                    rclip = lhs;

                lclip -= lhs;
                rclip -= lhs;
                if (rclip > (window->traces.scale.wave_pane_width + 1))
                    rclip = window->traces.scale.wave_pane_width + 1;

                gdk_draw_rectangle (wavepixmap, gc_xfill, TRUE,
                  (((gdouble) lclip) * window->traces.scale.pixels_per_time_unit), fontheight,
                  (((gdouble) (rclip - lclip)) * window->traces.scale.pixels_per_time_unit), window->traces.scale.wave_pane_height - fontheight);
            }

            bt = bt->next;
        }
    }
}

static void render_timebar (WaveWindow *window)
{
    GdkGC *gc_timeb = wave_get_pane_color_gc (wave_config->pane_colors, TRACE_PANE_BACKGROUND, window);

    gdk_draw_rectangle (wavepixmap, gc_timeb, TRUE, 0, -1, window->traces.scale.wave_pane_width, fontheight);
    render_times (wave_window_drawing_context, window);
    render_traces (wave_window_drawing_context, window);
}

static void service_hslider (GtkWidget * text, WaveWindow *window)
{
    GdkGC *gc_back = wave_get_pane_color_gc (wave_config->pane_colors, TRACE_PANE_BACKGROUND, window);

    WAVE_DEBUG (printf ("Wave HSlider Moved\n"));

    if (wavepixmap && window->traces.scale.wave_pane_width > 1)
    {
        GtkAdjustment *hadj = GTK_ADJUSTMENT (window->traces.hslider);

        if (!tims.timecache)
            wave_set_start_time (hadj->value);
        else {
            wave_set_start_time (tims.timecache);
            tims.timecache = 0; /* reset */
        }

        if (tims.start < tims.first)
            wave_set_start_time (tims.first);
        else if (tims.start > tims.last)
            wave_set_start_time (tims.last);

        tims.laststart = tims.start;
        gdk_draw_rectangle (wavepixmap, gc_back, TRUE, 0, 0, window->traces.scale.wave_pane_width, window->traces.scale.wave_pane_height);
        render_timebar (window);
    }
}

static void service_vslider (GtkWidget * text, WaveWindow *window)
{
    GtkAdjustment *sadj, *hadj;
    int trtarget;
    int xsrc;

    if (window->signals.pixmap)
    {
        hadj = GTK_ADJUSTMENT (window->signals.hslider);
        sadj = GTK_ADJUSTMENT (window->traces.vslider);
        xsrc = (gint) hadj->value;

        trtarget = (int) sadj->value;
        WAVE_DEBUG (printf ("Wave VSlider Moved to %d\n", trtarget));

        gdk_draw_rectangle (window->signals.pixmap,
          window->signals.drawing_area->style->bg_gc[GTK_STATE_PRELIGHT], TRUE, 0, 0, window->signals.fill_width,
          window->signals.drawing_area->allocation.height);

        sync_marker (window);
        wave_render_signal_names_background (window);
        wave_render_signal_names (wave_window_drawing_context, trtarget, window);

        /* Need to redraw traces? */
		if (wavepixmap && old_wvalue != sadj->value)
		{
			GdkGC *gc_back = wave_get_pane_color_gc (wave_config->pane_colors, TRACE_PANE_BACKGROUND, window);

			gdk_draw_rectangle (wavepixmap, gc_back, TRUE, 0, fontheight - 1,
				window->traces.scale.wave_pane_width, window->traces.scale.wave_pane_height - fontheight + 1);

			render_times (wave_window_drawing_context, window);
			render_traces (wave_window_drawing_context, window);
		}

        old_wvalue = sadj->value;

        render_named_markers (wave_window_drawing_context, window);
        gdk_draw_pixmap (window->signals.drawing_area->window,
          window->signals.drawing_area->style->fg_gc[GTK_WIDGET_STATE (window->signals.drawing_area)],
          window->signals.pixmap, xsrc, 0, 0, 0, window->signals.drawing_area->allocation.width, window->signals.drawing_area->allocation.height);
        draw_marker (window);
    }
}

static void button_press_release_common (WaveWindow *window)
{
    /* FIXME, less heavy duty func. to use here? */
    wave_max_signal_length (window);
    wave_redraw_signal_pane (window);
}

static void button_motion_common (gint x, gboolean pressrel, WaveWindow *window)
{
    gdouble offset, pixstep;
    TimeType newcurr;

    x = CLAMP (x, 0, window->traces.scale.wave_pane_width - 1);

    pixstep = ((gdouble) window->traces.scale.time_units_per_frame) / window->traces.scale.pixels_per_frame;
    offset = ((gdouble) tims.start) + x * pixstep;
    newcurr = (TimeType) offset;

    if (offset - newcurr > 0.5) /* round to nearest integer ns */
        newcurr ++;

    newcurr = CLAMP (newcurr, tims.start, MIN (tims.end, tims.last));

    tims.marker = newcurr;
    update_markertime (tims.marker);
    if (tims.lmbcache < 0)
        tims.lmbcache = newcurr;

    draw_marker_partitions (window);

    if (pressrel || wave_flags.const_mark_update_flag)
        button_press_release_common (window);
}

gboolean on_traces_drawing_area_motion_notify_event (GtkWidget * widget, GdkEventMotion * event, WaveWindow *window)
{
    gdouble x, y, pixstep, offset;
    GdkModifierType state;
    TimeType newcurr;
    int scrolled;

    if (event->is_hint)
        gdk_input_window_get_pointer (event->window, event->deviceid, &x, &y, NULL, NULL, NULL, &state);
    else {
        x = event->x;
        y = event->y;
        state = event->state;
    }

    do
    {
        scrolled = 0;
        if (state & bmask[in_button_press])
        {
            if (x < 0)
            {
                if (tims.start > tims.first)
                {
                    if (window->traces.scale.time_units_per_frame < 10)
                    {
                        wave_set_start_time (tims.start - window->traces.scale.time_units_per_frame);
                    } else
                    {
                        wave_set_start_time (tims.start - (window->traces.scale.time_units_per_frame / 10));
                    }
                    if (tims.start < tims.first)
                        wave_set_start_time (tims.first);
                    GTK_ADJUSTMENT (window->traces.hslider)->value = tims.marker = tims.timecache = tims.start;

                    gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "changed");
                    gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "value_changed");
                    scrolled = 1;
                }
                x = 0;
            } else if (x > window->traces.scale.wave_pane_width)
            {
                if (tims.start != tims.last)
                {
                    gfloat pageinc = wave_pane_width_in_time_units (window);

                    if (window->traces.scale.time_units_per_frame < 10)
                        wave_set_start_time (tims.start + window->traces.scale.time_units_per_frame);
                    else wave_set_start_time (tims.start + (window->traces.scale.time_units_per_frame / 10));

                    if (tims.start > tims.last - pageinc + 1)
                        wave_set_start_time (tims.last - pageinc + 1);
                    if (tims.start < tims.first)
                        wave_set_start_time (tims.first);

                    tims.marker = tims.start + pageinc;
                    if (tims.marker > tims.last)
                        tims.marker = tims.last;

                    GTK_ADJUSTMENT (window->traces.hslider)->value = tims.timecache = tims.start;

                    gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "changed");
                    gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "value_changed");
                    scrolled = 1;
                }
                x = window->traces.scale.wave_pane_width - 1;
            }
        }

        pixstep = ((gdouble) window->traces.scale.time_units_per_frame) / ((gdouble) window->traces.scale.pixels_per_frame);
        newcurr = tims.start + (offset = x * pixstep);
        if (offset - ((int) offset) > 0.5) /* round to nearest integer ns */
            newcurr ++;

        if (newcurr > tims.last)
            newcurr = tims.last;

        if (newcurr != prevtim)
        {
            update_currenttime (newcurr);
            prevtim = newcurr;
        }

        if (state & bmask[in_button_press])
            button_motion_common (x, FALSE, window);

        if (scrolled)
        {                       /* make sure counters up top update */
            while (gtk_events_pending ())
            {
                gtk_main_iteration ();
            }
        }

        gdk_input_window_get_pointer (event->window, event->deviceid, &x, &y, NULL, NULL, NULL, &state);

    }
    while (scrolled && (state & bmask[in_button_press]));

    return TRUE;
}

gboolean on_traces_drawing_area_button_press_event (GtkWidget * widget, GdkEventButton * event, WaveWindow *window)
{
    if (event->button == 1 || (event->button == 3 && !in_button_press))
    {
        in_button_press = event->button;

        WAVE_DEBUG (printf ("Button Press Event\n"));
        button_motion_common (event->x, TRUE, window);
        tims.timecache = tims.start;

        gdk_pointer_grab (widget->window, FALSE, m_bmask[in_button_press] |
          /* key up on motion for button pressed ONLY */
          GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_RELEASE_MASK, NULL, NULL, event->time);
    }

    return TRUE;
}

gboolean on_traces_drawing_area_button_release_event (GtkWidget * widget, GdkEventButton * event, WaveWindow *window)
{
    if (event->button && event->button == in_button_press)
    {
        in_button_press = 0;

        WAVE_DEBUG (printf ("Button Release Event\n"));
        button_motion_common (event->x, FALSE, window);
        tims.timecache = tims.start;

        gdk_pointer_ungrab (event->time);

        if (event->button == 3)
        {                       /* oh yeah, dragzoooooooom! */
            service_dragzoom (tims.lmbcache, tims.marker);
        }
        tims.lmbcache = -1;
        update_markertime (tims.marker);
    }
    tims.timecache = 0;
    return TRUE;
}

gint on_traces_drawing_area_configure_event (GtkWidget * widget, GdkEventConfigure * event, WaveWindow *window)
{
    WAVE_DEBUG (printf ("WaveWin Configure Event h: %d, w: %d\n", widget->allocation.height, widget->allocation.width));

    if (wavepixmap)
    {
        if (window->traces.scale.wave_pane_width != widget->allocation.width ||
			window->traces.scale.wave_pane_height != widget->allocation.height)
        {
            gdk_pixmap_unref (wavepixmap);
            window->traces.scale.wave_pane_width = widget->allocation.width;
            window->traces.scale.wave_pane_height = widget->allocation.height;
            wavepixmap = gdk_pixmap_new (widget->window, window->traces.scale.wave_pane_width, window->traces.scale.wave_pane_height, -1);

			if (wave_window_drawing_context) g_free (wave_window_drawing_context);
			wave_window_drawing_context = wave_new_gdk_drawing_context (&wavepixmap, &window->signals.pixmap,
				window->signals.drawing_area->style, window);
        }
        old_wvalue = -1.0;
    } else
    {
        window->traces.scale.wave_pane_width = widget->allocation.width;
        window->traces.scale.wave_pane_height = widget->allocation.height;
        wavepixmap = gdk_pixmap_new (widget->window, window->traces.scale.wave_pane_width, window->traces.scale.wave_pane_height, -1);

		if (wave_window_drawing_context) g_free (wave_window_drawing_context);
		wave_window_drawing_context = wave_new_gdk_drawing_context (&wavepixmap, &window->signals.pixmap,
			window->signals.drawing_area->style, window);
    }

    if (window->traces.scale.wave_pane_width > 1)
    {
        if (!wave_flags.do_initial_zoom_fit)
        {
            wave_zoom (tims.zoom, window);
            fix_wavehadj (window);
            /* force zoom update */
            gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "value_changed");
            gtk_signal_emit_by_name (GTK_OBJECT (window->traces.hslider), "changed");
        } else
        {
            wave_flags.do_initial_zoom_fit = FALSE;
            clicked_zoom_fit (NULL, NULL);
        }
    }

/* tims.timecache=tims.laststart; */
    return TRUE;
}

gboolean on_traces_drawing_area_expose_event (GtkWidget * widget, GdkEventExpose * event, WaveWindow *window)
{
    gdk_draw_pixmap (widget->window,
      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
      wavepixmap, event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
    draw_marker (window);

    return FALSE;
}

/* wave_pane_width_in_time_units : returns the wave pane width in the current base time units */
TimeType wave_pane_width_in_time_units (WaveWindow *window)
{
	return window->traces.scale.wave_pane_width * window->traces.scale.time_units_per_pixel;
}

/* wave_create_trace_pane : create a main window trace pane */
void wave_create_trace_pane (WaveWindow *window)
{
    GtkAdjustment *hadj, *vadj;

    window->traces.drawing_area = lookup_widget (window->window, "traces_drawing_area");

	window->traces.vslider = gtk_range_get_adjustment (GTK_RANGE (lookup_widget (window->window,	
		"traces_vscroll")));

    vadj = GTK_ADJUSTMENT (window->traces.vslider);
    gtk_signal_connect (GTK_OBJECT (window->traces.vslider), "value_changed", GTK_SIGNAL_FUNC (service_vslider), window);

	window->traces.hslider = gtk_range_get_adjustment (GTK_RANGE (lookup_widget (window->window,	
		"traces_hscroll")));
    hadj = GTK_ADJUSTMENT (window->traces.hslider);
    gtk_signal_connect (GTK_OBJECT (window->traces.hslider), "value_changed", GTK_SIGNAL_FUNC (service_hslider),
    	window);

    window->traces.window = lookup_widget (window->window, "traces_frame");
}

void update_signal_value (TraceEnt * t)
{
    if (!t)
        return;
    if (t->config->name && !TRACE_IS_SEPARATOR (t->config))
    {
        WAVE_DEBUG (printf ("update_signal_value: %s\n", t->config->name));

        if (tims.marker != -1 && !TRACE_IS_HIDDEN (t->config))
        {
            if (t->asciivalue)
            {
                g_free (t->asciivalue);
                t->asciivalue = NULL;
            }

            {
                WaveEventIterator *event_iter = t->n.nd->iterator;

                if (wave_event_iterator_seek (event_iter, tims.marker, WAVE_EVENT_LAST_LE))
                {
						char *str = wave_event_value_string (t, event_iter->current);
						if (str)
						{
							char *str2 = g_new (char, strlen (str) + 2);
							*str2 = '=';
							strcpy (str2 + 1, str);

							g_free (str);

							t->asciivalue = str2;
						}
                }
            }
        }
    }
}

/* render_time_grid : render time grid/ruler lines and text between times
	`start_time' and `end_time' in steps of `grid_line_spacing_time'.  Vertical
	lines are drawn between `y_start' and `y_end' with text annotation at
	`y_text' only if `y_text' != 0 */
static void render_time_grid (WaveDrawingContext *context, TimeType start_time, TimeType end_time,
	TimeType grid_line_spacing_time, int y_start, int y_end, int y_text, WaveWindow *window)
{
    char time_buffer[32];

	while (start_time <= end_time)
	{
		int x = (start_time - tims.start) * window->traces.scale.pixels_per_time_unit;

        if (y_text)
        {
        	int len;
			wave_format_time (time_buffer, start_time, wave_config->time_dimension);
			len = gdk_string_measure (wavefont, time_buffer);
			WAVE_DRAW_PANE_STRING (context, TRACE_PANE_FONT, WAVE_FONT_WAVE,
				MAX (x - (len >> 1), 0), y_text, time_buffer);
		}

        WAVE_DRAW_PANE_LINE (context, TRACE_PANE_GRID, x, y_start, x, y_end);

		if (start_time == tims.last)
			break;

        start_time = MIN (start_time + grid_line_spacing_time, tims.last);
	}
}

/* render_times : render the time bar (ruler) and vertical grid lines */
void render_times (WaveDrawingContext *context, WaveWindow *window)
{
    TimeType tim, rem, span;
	int order_of_span;
	TimeType grid_line_spacing_time, last_displayed_time;

    render_blackout (window);
	wave_set_end_time (tims.start + wave_pane_width_in_time_units (window));

	span = tims.end - tims.start;
	order_of_span = log10 (span);
	last_displayed_time = MIN (tims.last, tims.end);
	grid_line_spacing_time = pow (10, order_of_span);
	grid_line_spacing_time = MAX (grid_line_spacing_time, 1);

	/* Round up to next grid line, can't think of an easier way */
    tim = tims.start;
	rem = tim % grid_line_spacing_time;
	if (rem != 0)
		tim += grid_line_spacing_time - rem;

	render_time_grid (context, tim, last_displayed_time, grid_line_spacing_time,
		window->traces.scale.time_line_y, window->traces.scale.wave_pane_height,
		wavefont->ascent - 1, window);

	/* Now print the `ticks' */
	if (order_of_span > 0) /* Don't bother with ticks at this scale */
	{
		grid_line_spacing_time = pow (10, order_of_span - 1);

		/* Round up to next grid line, can't think of an easier way */
		tim = tims.start;
		rem = tim % grid_line_spacing_time;
		if (rem != 0)
			tim += grid_line_spacing_time - rem;

		render_time_grid (context, tim, last_displayed_time, grid_line_spacing_time,
			fontheight - 3, window->traces.scale.time_line_y, 0, window);
	}

	/* Line under times, above ticks */
	WAVE_DRAW_PANE_FILLED_RECTANGLE (context, TRACE_PANE_GRID, 0, window->traces.scale.time_line_y,
		(tims.last > tims.end ? /* Draw line all the way across display */
			window->traces.scale.wave_pane_width : /* Just upto tims.end */
			(tims.last - tims.start) * window->traces.scale.pixels_per_time_unit),
		window->traces.scale.time_line_y + 1);
}

/* add_extensions : count the number of extension traces to this one are in
	the traces list */
static int add_extensions (TraceEnt * t, int which)
{
    t = t->next;
    while (t && TRACE_IS_EXTENSION (t->config))
    {
        which ++;
        t = t->next;
    }

    return which;
}

static void do_horizontal_grid_line (WaveDrawingContext *context, int liney, WaveWindow *window)
{
    if (wave_flags.show_grid_flag && wave_flags.enable_horiz_grid)
    {
        int x1, x2;

        if (tims.start < tims.first)
            x1 = (tims.first - tims.start) * window->traces.scale.pixels_per_time_unit;
        else x1 = 0;

        if (tims.last <= tims.end)
            x2 = (tims.last - tims.start) * window->traces.scale.pixels_per_time_unit;
        else x2 = window->traces.scale.wave_pane_width - 1;

		WAVE_DRAW_PANE_LINE (context, TRACE_PANE_GRID, x1, liney, x2, liney);
    }
}

#define WAVE_OPT_SKIP 1

/*
    draw single traces and use this for rendering the grid lines
    for "excluded" traces
*/
static void draw_event_trace (WaveDrawingContext *context, TraceEnt * t,
	WaveEventIterator * event_iter, int which, WaveWindow *window)
{
    TimeType x0, x1, newtime;
    int y0, y1, yu, ytext;
    TimeType tim;
    int height = 1 + add_extensions (t, which) - which;
	int top, bottom;

	wave_drawing_context_set_trace (context, t);

	/* `which' is the trace number, starting from 0 at the top of the display.
		Note that the time ruler occupies the first trace line from the origin.

		Trace (vertically) looks like this:

		top   +0 == top (space of 2 pixels to trace)
		      +1
		      +2 == y1 (sans inversion) Trace is drawn between y1 and y0
		      ...
		         == yu (half way between y1 and y0)
		      ...
		      -4 == y0 Base of drawn trace
	          -3
              -2 == Bottom line ---------------
			  -1
		bottom-0 Repeating bottom of trace

		bottom - top == fontheight * height_of_trace
	*/
	top = (which + 1) * fontheight;
	bottom = top + height * fontheight;

	y1 = top + 2;
    y0 = bottom - 4;
    yu = (y0 + y1) / 2;
    ytext = yu + (wavefont->ascent / 2);

    do_horizontal_grid_line (context, bottom - 2, window);
	TimeType last_time = -1;

	/* No events? don't bother to show anything */
	if (! event_iter->current)
		return;

	if (tims.start == event_iter->current->time &&
		event_iter->current->value.val != 'Z')
		draw_bit_trace_start (context, t->n.nd, event_iter->current, y0, y1);

	do
	{
		tim = event_iter->current->time;
		if (tim > tims.end || tim > tims.last)
			break;

		x0 = (tim - tims.start) * window->traces.scale.pixels_per_time_unit;
		if (x0 < -1)
			x0 = -1;
		else if (x0 > window->traces.scale.wave_pane_width)
			break;

		if (event_iter->next)
			tim = event_iter->next->time;
		else tim = tims.end + 1;

		if (tim > tims.last)
			tim = tims.last;
		else if (tim > tims.end + 1)
			tim = tims.end + 1;
		x1 = (tim - tims.start) * window->traces.scale.pixels_per_time_unit;
		if (x1 < -1)
			x1 = -1;
		else if (x1 > window->traces.scale.wave_pane_width)
			x1 = window->traces.scale.wave_pane_width;

		/* Force drawing of spikes if no spike drawing routine is defined */
		if (x0 != x1 || (t->drawing_info && ! t->drawing_info->drawer->bit_spike))
		{
			gboolean use_x_gc = event_iter->current->value.val == 'X' ||
				(event_iter->next && event_iter->next->value.val == 'X');

			draw_bit_trace (context, t->n.nd, event_iter->current, event_iter->next,
				use_x_gc, last_time != event_iter->current->time, x0, y0, x1, y1, 0, ytext);
		} else
		{
			draw_bit_spike (context, t->n.nd, x0, y0, x1, y1);
			/* skip to next pixel */
			newtime = ((gdouble) (x1 + WAVE_OPT_SKIP)) * window->traces.scale.time_units_per_pixel + tims.start;
/* FIXME FIXME FIXME NOHISTENT
			h3 = bsearch_node (t->n.nd, newtime);
			if (h3->time > h->time)
			{
				h = h3;
				continue;
			}
			*/
		}

		last_time = event_iter->current->time;
	} while (wave_event_iterator_step_forward (event_iter));
}

int vtype (TraceEnt * t, char *vec)
{
    int i, nbits, res;

    if (!vec)
        return 'X';
    nbits = t->n.nd->ext->msi - t->n.nd->ext->lsi;
    if (nbits < 0)
        nbits = -nbits;
    nbits ++;
    res = '1';
    for (i = 0; i < nbits; i ++)
	{
        switch (*vec)
        {
        case 'X':
            return 'X';
        case 'Z':
            if (res == '0')
                return '1';
            res = 'Z';
            break;
        default:
            if (res == 'Z')
                return 'X';
            res = '0';
            break;
        }
        vec ++;
	}

    return res;
}

static void draw_event_trace_vector (WaveDrawingContext *context, TraceEnt * t,
	WaveEventIterator *event_iter, int which, WaveWindow *window)
{
    TimeType x, x0, x1, newtime;
    int y0, y1, yu, liney, ytext, trans;
    TimeType tim;
    int type;
    int lasttype = -1;
    int which2 = add_extensions (t, which);

	wave_drawing_context_set_trace (context, t);

    liney = (which2 + 2) * fontheight - 2;
    y1 = (which + 1) * fontheight + 2;
    y0 = liney - 2;
    yu = (y0 + y1) / 2;
    ytext = yu - (wavefont->ascent / 2) + wavefont->ascent;

    /*
     * Context value.  With the default drawing engine, this is set
     * true for displays which transition through X, e.g. 0-X-1,
     * 1-X-0, but this is a matter for the drawing engine.
     */
    trans = 0;

    do_horizontal_grid_line (context, liney, window);

	/* No events, don't bother printing anything */
    if (! event_iter->current)
    	return;

    do
    {
        x = 0;

        tim = event_iter->current->time;
        if ((tim > tims.end) || (tim > tims.last))
            break;

        x0 = (tim - tims.start) * window->traces.scale.pixels_per_time_unit;
        if (x0 < -1)
            x0 = -1;
        else if (x0 > window->traces.scale.wave_pane_width)
            break;

		if (event_iter->next)
			tim = event_iter->next->time;
		else tim = tims.end + 1;

        if (tim > tims.last)
            tim = tims.last;
        else if (tim > tims.end + 1)
            tim = tims.end + 1;
        x1 = (tim - tims.start) * window->traces.scale.pixels_per_time_unit;
        if (x1 < -1)
            x1 = -1;
        else if (x1 > window->traces.scale.wave_pane_width)
            x1 = window->traces.scale.wave_pane_width;

        /* draw trans */
        type = (t->n.nd->flags & NODE_FLAGS_VECTOR
        	? vtype (t, event_iter->current->value.vector) : '0');
        if (x0 > -1)
            draw_value_trace1 (context, t->n.nd, &trans, type, lasttype, x0, y0, y1, yu);

        if (x0 != x1)
        {
            draw_value_trace2 (context, t->n.nd, event_iter->current, &trans, type, x, x0, y0, x1, y1, yu, ytext);
        } else
        {
            newtime = ((gdouble) (x1 + WAVE_OPT_SKIP)) * window->traces.scale.time_units_per_pixel + tims.start; /* skip to next pixel */
/* FIXME FIXME FIXME NOHISTENT
            h3 = bsearch_node (t->n.nd, newtime);
            if (h3->time > h->time)
            {
                h = h3;
                lasttype = type;
                continue;
            }
            */
        }

        lasttype = type;
	} while (wave_event_iterator_step_forward (event_iter));
}

void render_traces (WaveDrawingContext *context, WaveWindow *window)
{
    if (!window->traces.top_visible_trace)
    {
        window->traces.top_visible_trace = traces.first;
    }
    if (window->traces.top_visible_trace)
    {
        TraceEnt *t;
        int i, num_traces_displayable;

        num_traces_displayable = window->traces.drawing_area->allocation.height / (fontheight);
        num_traces_displayable --; /* for the time trace that is always there */
        for (i = 0, t = window->traces.top_visible_trace; t && (i < num_traces_displayable); i ++, t = t->next)
        {
            if (TRACE_IS_HIDDEN (t->config) || TRACE_IS_SEPARATOR (t->config))
            {
                continue;
            }
            {
                WaveEventIterator *event_iter = t->n.nd->iterator;

				if (! wave_event_iterator_seek (event_iter, tims.start, WAVE_EVENT_LAST_LE))
					continue;

                WAVE_DEBUG (printf ("Bit Trace: %s, %s\n", t->config->name, t->n.nd->nname));
                WAVE_DEBUG (printf ("Start time: " LL_PRINTF_FORMAT ", "
                    "Histent time: " LL_PRINTF_FORMAT "\n", tims.start, h->time));

                if (t->n.nd->flags & NODE_FLAGS_DIGITAL ||
					t->n.nd->flags & NODE_FLAGS_SYNC_HANDSHAKE ||
					t->n.nd->flags & NODE_FLAGS_DATA_HANDSHAKE)
				{
					/* FIXME, _ent_trace and _ent_trace_vector should be the same function */
                    draw_event_trace (context, t, event_iter, i, window);
                } else {
                	draw_event_trace_vector (context, t, event_iter, i, window);
                }
            }
        }
    }

    render_named_markers (context, window);
    draw_marker_partitions (window);
}

/* wave_redraw_trace_pane : redraw only the trace pane */
extern void wave_redraw_trace_pane (WaveWindow *window)
{
    on_traces_drawing_area_configure_event (window->traces.drawing_area, NULL, window);
}

/* wave_redraw : redraw main window signals and traces */
void wave_redraw (WaveWindow *window)
{
    wave_redraw_signal_pane (window);
	wave_redraw_trace_pane (window);
}

