/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `tree.c'
    Signal search tree dialogue tree handling

    $Id: tree.c,v 1.18 2003/11/11 22:53:28 bardslea Exp $
*/

#include "tree.h"

SignalTree *treeroot = NULL;
static char *name_scratch_space = NULL;

char hier_delimiter = '.';      /* default is dot */
gboolean hier_was_explicitly_set = FALSE;
char alt_hier_delimiter = 0x00; /* for vcds with both [hierarchies or .] and / -- used in vcd only */

enum TreeBuildTypes
{
    MAKETREE_FLATTEN, MAKETREE_LEAF, MAKETREE_NODE
};

/* wave_new_signal_tree : make a new signal tree element */
static SignalTree *wave_new_signal_tree (SignalTree *parent)
{
	SignalTree *ret = g_new (SignalTree, 1);

	ret->parent = parent;
	ret->name = NULL;
	ret->next = ret->child = NULL;
	ret->which = -1;

	return ret;
}

/*
    init pointers needed for n-way tree
*/
void init_tree (void)
{
	treeroot = wave_new_signal_tree (NULL);
    name_scratch_space = g_new (char, longestname + 1);
}

/*
    extract the next part of the name in the flattened
    hierarchy name.  return ptr to next name if it exists
    else NULL
*/
static char *get_next_name_element (char *s)
{
    char ch;
    char *pnt;

    pnt = name_scratch_space;

    for (;;)
    {
        ch = *s;

        if (ch == hier_delimiter)
        {
            *(pnt) = 0;
            s++;
            return (s);
        }
        if (!ch)
        {
            *(pnt) = 0;
            return (NULL);      /* nothing left to extract */
        }

        s++;
        *(pnt++) = ch;
    }
}

/*
    build the tree.  to keep the logic simple, we have
    dead leaves at the ends.
*/
void build_tree_from_name (char *s, int which)
{
    SignalTree *t;

    if (s == NULL)
        return;
    if (!strlen (s))
        return;

    t = treeroot;
    while (s)
    {
        s = get_next_name_element (s);

        if (!t->name)
        {
            char *newname;
            SignalTree *t2;

            newname = g_strdup (name_scratch_space);
            t->name = newname;
            if (!s)
                t->which = which;
            else
                t->which = -1;

			t2 = wave_new_signal_tree (t);

            t->child = t2;
            t = t2;

            continue;
        }

        if (!strcmp (t->name, name_scratch_space))
        {
            t = t->child;
        } else
        {
            SignalTree *t2, *t3;
            char *newname;

            newname = g_strdup (name_scratch_space);
			t2 = wave_new_signal_tree (t->parent);
			*t2 = *t;

            t->name = newname;
            t->next = t2;
            if (!s)
                t->which = which;
            else
                t->which = -1;

			t3 = wave_new_signal_tree (t);

            t->child = t3;
            t = t3;
        }
    }
}

/*
    remove dead leaves from the tree
*/
int treeprune (SignalTree * t)
{
    while (t)
    {
        if (t->name)
        {
            if (t->child && !treeprune (t->child))
            {
                t->child = NULL; /* remove empty leaf */
            }
        } else
        {
            g_free (t);         /* free empty leaf */
            return 0;
        }

        t = t->next;
    }
    return 1;
}

/*
    for debugging purposes only
*/
#ifdef DEBUG_FACILITIES
void treedebug (SignalTree * t, char *s)
{
    while (t)
    {
        char *s2;

        s2 = (char *) g_malloc (strlen (s) + strlen (t->name) + 2);
        strcpy (s2, s);
        strcat (s2, ".");
        strcat (s2, t->name);

        if (t->child)
        {
            treedebug (t->child, s2);
        } else
        {
            printf ("%d) %s\n", t->which, s2);
        }

        g_free (s2);
        t = t->next;
    }
}
#endif /* DEBUG_FACILITIES */
