/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 2004 Andrew Bardsley
    Copyright (C) 1999-2003 Tony Bybell
    Copyright (C) 2000-2003 Department of Computer Science
    The University of Manchester, Oxford Road, Manchester, M13 9PL, UK

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `value.c'
    Trace data value handling and print formatting

    $Id: value.c,v 1.13 2004/04/25 19:23:42 bardslea Exp $
*/

#include <ctype.h>
#include <string.h>
#include "value.h"
#include "rc.h"
#include "str.h"
#include "format/format.h"
#include "format/external.h"

/* wave_invert_digital_value : invert the values '0XZ1' by mapping onto '1XZ0' */
char wave_invert_digital_value (char val)
{
	char ret = val;

	switch (val)
	{
	case '0':
		ret = '1';
		break;
	case '1':
		ret = '0';
		break;
	default:
		break;
	}

	return ret;
}

/* wave_event_real_string : convert real value to string */
static char *wave_event_real_string (char *ptr)
{
    char *rv;
    double *d = (double *) ptr;

    if (d)
        rv = g_strdup_printf ("%.16g", *d);
    else rv = g_strdup ("UNDEF");

    return rv;
}

/* wave_event_digital_vector_string : convert a digital vector string into a formatted string of the
	preferred radix */
static char *wave_event_digital_vector_string (TraceEnt *trace_ent, char *vec)
{
    int nbits;
    char *bits;
    char *os, *pnt, *newbuff;
    int i, j, len;
	gboolean inverted;
	FormatRadixData *radix = NULL;

    nbits = trace_ent->n.nd->ext->msi - trace_ent->n.nd->ext->lsi;
    if (nbits < 0)
        nbits = -nbits;
    nbits++;

    if (vec)
    {
        bits = vec;
    } else
    {
        pnt = bits = wave_alloca (nbits);
        for (i = 0; i < nbits; i++)
        {
            *pnt++ = 'X';
        }
    }

	inverted = TRACE_IS_INVERTED (trace_ent->config);

    newbuff = g_new (char, nbits + 6); /* + 6 for justification */
    if (TRACE_IS_REVERSED (trace_ent->config))
    {
        char *fwdpnt, *revpnt;

        fwdpnt = bits;
        revpnt = newbuff + nbits + 6;
        for (i = 0; i < 3; i++)
            *(--revpnt) = (inverted ? '1' : '0');
        for (i = 0; i < nbits; i++)
        {
            *(--revpnt) = (inverted ? wave_invert_digital_value (*fwdpnt) : *fwdpnt);
			fwdpnt ++;
        }
        for (i = 0; i < 3; i++)
            *(--revpnt) = (inverted ? '1' : '0');
    } else
    {
        char *fwdpnt, *fwdpnt2;

        fwdpnt = bits;
        fwdpnt2 = newbuff;
        for (i = 0; i < 3; i++)
            *(fwdpnt2++) = (inverted ? '1' : '0');
        for (i = 0; i < nbits; i++)
        {
            *(fwdpnt2++) = (inverted ? wave_invert_digital_value (*fwdpnt) : *fwdpnt);
			fwdpnt ++;
        }
        for (i = 0; i < 3; i++)
            *(fwdpnt2++) = (inverted ? '1' : '0');
    }

	/* Choose the correct radix and allocate space */
	switch (trace_ent->config->format)
	{
	case TRACE_FORMAT_HEX:
        radix = wave_find_radix_data (16);
        len = nbits / 4 + 2 + radix->prefixSuffixLength;
        break;
	case TRACE_FORMAT_OCT:
        radix = wave_find_radix_data (8);
        len = nbits / 3 + 2 + radix->prefixSuffixLength;
        break;
	case TRACE_FORMAT_BIN:
        radix = wave_find_radix_data (2);
        len = nbits + 2 + radix->prefixSuffixLength;
        break;
	case TRACE_FORMAT_SIGNED_DEC:
	case TRACE_FORMAT_DEC:
		if (nbits > 64)
		{ /* Treat long nums. as hex */
        	radix = wave_find_radix_data (16);
        	len = nbits / 4 + 2 + radix->prefixSuffixLength;
		} else {
			radix = wave_find_radix_data (10);
			len = 21 + radix->prefixSuffixLength;
		}
        break;
    case TRACE_FORMAT_ASCII:
        len = nbits / 8 + 2 + 2;
        break;
    default: /* Just in case, allocate lots of chars */
        len = nbits + 2;
        break;
	}

	/* Allocate space */
    os = pnt = g_new (char, len);

    if (trace_ent->config->format == TRACE_FORMAT_ASCII)
    {
        char *parse;
        gboolean found = FALSE;

        if (wave_flags.show_base_flag)
            *(pnt++) = '"';

        parse = (TRACE_IS_RJUSTIFIED (trace_ent->config)) ? (newbuff + ((nbits + 3) & 3)) : (newbuff + 3);

        for (i = 0; i < nbits; i += 8)
        {
            unsigned long val;

            val = 0;
            for (j = 0; j < 8; j++)
            {
                val <<= 1;

                if (parse[j] == 'X')
                {
                    val = 1000;
                }
                if (parse[j] == 'Z')
                {
                    val = 1000;
                }

                if (parse[j] == '1')
                {
                    val |= 1;
                }
            }

            if (val)
            {
                if (val > 0x7f || !isprint (val))
                    *pnt++ = '.';
                else
                    *pnt++ = val;
                found = TRUE;
            }

            parse += 8;
        }
        if (!found && !wave_flags.show_base_flag)
        {
            *(pnt++) = '"';
            *(pnt++) = '"';
        }

        if (wave_flags.show_base_flag)
            *(pnt++) = '"';
        *(pnt++) = '\0';
    } else if (radix)
    {
    	if (wave_flags.show_base_flag && radix->prefix)
       		pnt = wave_strcpy (pnt, radix->prefix);

    	if (radix->radix == 16)
		{
			char *parse;

			parse = (TRACE_IS_RJUSTIFIED (trace_ent->config)) ? (newbuff + ((nbits + 3) & 3)) : (newbuff + 3);

			for (i = 0; i < nbits; i += 4)
			{
				unsigned char val;

				val = 0;
				for (j = 0; j < 4; j++)
				{
					val <<= 1;

					if (parse[j] == 'X')
					{
						val = 16;
						break;
					}
					if (parse[j] == 'Z')
					{
						val = 17;
						break;
					}

					if (parse[j] == '1')
					{
						val |= 1;
					}
				}

				*(pnt++) = "0123456789ABCDEFXZ"[val];

				parse += 4;
			}
			*pnt = '\0';
		} else if (radix->radix == 8)
		{
			char *parse;

			parse = (TRACE_IS_RJUSTIFIED (trace_ent->config)) ? (newbuff + ((nbits % 3) ? (nbits % 3) : 3)) : (newbuff + 3);

			for (i = 0; i < nbits; i += 3)
			{
				unsigned char val;

				val = 0;
				for (j = 0; j < 3; j++)
				{
					val <<= 1;

					if (parse[j] == 'X')
					{
						val = 8;
						break;
					}
					if (parse[j] == 'Z')
					{
						val = 9;
						break;
					}

					if (parse[j] == '1')
					{
						val |= 1;
					}
				}

				*(pnt++) = "01234567XZ"[val];

				parse += 3;
			}
			*pnt = '\0';
		} else if (radix->radix == 2)
		{
			char *parse;

			parse = newbuff + 3;

			for (i = 0; i < nbits; i++)
			{
				*(pnt++) = *(parse++);
			}
			*pnt = '\0';
		} else if (trace_ent->config->format == TRACE_FORMAT_SIGNED_DEC)
		{
			char *parse;
			TimeType val;
			gboolean fail = FALSE;

			parse = newbuff + 3;

			val = (parse[0] == '1') ? LL_LITERAL (-1) : LL_LITERAL (0);
			if (parse[0] == 'X' || parse[0] == 'Z')
			{
				fail = TRUE;
			} else
			{
				for (i = 1; i < nbits; i++)
				{
					val <<= 1;

					if (parse[i] == 'X' || parse[i] == 'Z')
					{
						fail = TRUE;
						break;
					}
					if (parse[i] == '1')
					{
						val |= LL_LITERAL (1);
					}
				}
			}

			if (!fail)
			{
				sprintf (pnt, LL_PRINTF_FORMAT, val);
			} else {
				strcpy (pnt, "XXX");
			}
			pnt = wave_end_of_string (pnt);
		} else
		{ /* decimal when all else fails */
			char *parse;
			UTimeType val = 0;
			gboolean fail = FALSE;
			radix = wave_find_radix_data (10);

			parse = newbuff + 3;

			for (i = 0; i < nbits; i++)
			{
				val <<= 1;

				if (parse[i] == 'X' || parse[i] == 'Z')
				{
					fail = TRUE;
					break;
				}
				if (parse[i] == '1')
				{
					val |= ULL_LITERAL (1);
				}
			}

			if (!fail)
			{
				sprintf (pnt, ULL_PRINTF_FORMAT, val);
			} else
			{
				strcpy (pnt, "XXX");
			}
			pnt = wave_end_of_string (pnt);
		}

    	if (wave_flags.show_base_flag && radix->suffix)
       		pnt = wave_strcpy (pnt, radix->suffix);
	}

    g_free (newbuff);
    return os;
}

/* wave_handshake_radix_choice : radix choice function for handshake values.
	Usually respects the choices in the `event' flags */
static FormatRadixData **wave_handshake_radix_choice (int preferredRadix, TraceConfig *trace_config)
{
	/* Ignoring preferredRadix for now */
	TraceFormat radix_flag = trace_config->format;
	FormatRadixData **ret = wave_radix_style->radices + 2;

	/* FIXME, shouldn't assume order and position of radices in wave_radix_style */
	switch (radix_flag)
	{
    case TRACE_FORMAT_BIN:
    	ret = wave_radix_style->radices + 0;
    	break;
    case TRACE_FORMAT_OCT:
    	ret = wave_radix_style->radices + 1;
    	break;
    case TRACE_FORMAT_HEX:
    	ret = wave_radix_style->radices + 3;
    	break;
    case TRACE_FORMAT_DEC:
    case TRACE_FORMAT_SIGNED_DEC:
    case TRACE_FORMAT_ASCII:
    default:
    	ret = wave_radix_style->radices + 2;
    	break;
	}

	return ret;
}

/* wave_event_handshake_string : format a suitable printed representation
	of a handshake signal.  Only returns the data on the channel (or "") if
	`data_value_only' is TRUE. If `trace_ent' is NULL then use an appropriate
	fallback representation */
char *wave_event_handshake_string (TraceEnt *trace_ent, WaveEvent *event,
	gboolean data_value_only)
{
	gboolean has_data = (trace_ent->n.nd->flags & NODE_FLAGS_DATA_HANDSHAKE) != 0;
	char event_char = tolower (has_data ? HANDSHAKE_VALUE (event->value.vector)->val : event->value.val);
	char *event_prefix = "";
	char *ret;

	if (! data_value_only)
	{
		switch (event_char)
		{
		case HANDSHAKE_STATE_QUIESCENT:
			event_prefix = "r-a-";
			break;
		case HANDSHAKE_STATE_REQUP:
			event_prefix = "r+a-";
			break;
		case HANDSHAKE_STATE_ACKUP:
			event_prefix = "r+a+";
			break;
		case HANDSHAKE_STATE_REQDOWN:
			event_prefix = "r-a+";
			break;
		default:
			event_prefix = "?";
			break;
		}
	}

	if (has_data)
	{
		Format format;
		TraceConfig *trace_config;
		FormatRadixChoiceFunction radix_choice;
		int event_prefix_length = strlen (event_prefix);
		int formatted_length;
		FormatData *handshake_value = HANDSHAKE_VALUE (event->value.vector)->data;

		/* Choose a format, assume a plain integer with a ? prefix if we
			don't have a trace_ent to guide us */

		if (trace_ent)
		{
			format = trace_ent->n.nd->format;
			trace_config = trace_ent->config;
			radix_choice = (FormatRadixChoiceFunction) wave_handshake_radix_choice;
		} else {
			/* Just use bottom bit */
			FormatPosition pos = {0, 1};
			format = NewFormat (2);
			NewFormatLiteral (& format[0], "? ", 0);
			NewFormatNumber (& format[1], pos);
			trace_config = NULL;
			radix_choice = FormatBareDecimalRadixChoice;
		}

		formatted_length = FormatLength (format, radix_choice, trace_config, handshake_value);

		/* FIXME, just to force binding */
		if (! format) format = FormatExternalToInternal ("ouch!");

		if (data_value_only)
		{
			ret = g_new (char, formatted_length + 1);
			FormatPrint (ret, format, handshake_value, radix_choice, trace_config);
		} else {
			ret = g_new (char, event_prefix_length + 2 + formatted_length);
			sprintf (ret, "%s ", event_prefix);
			FormatPrint (ret + strlen (event_prefix) + 1, format, handshake_value, radix_choice, trace_config);
		}
	} else {
		ret = g_strdup ((data_value_only ? "" : event_prefix));
	}

	return ret;
}

/* wave_event_value_string : convert a WaveEvent's value into a newly allocated string
	representation of that value.  `trace_ent' is allowed to be NULL */
char *wave_event_value_string (TraceEnt *trace_ent, WaveEvent *event)
{
	char *ret = NULL;
	NodeFlags node_type = (trace_ent->n.nd->flags & NODE_FLAGS_TYPE_MASK);

	switch (node_type)
	{
	case NODE_FLAGS_DIGITAL:
		ret = g_new (char, 2);
		ret[0] = wave_event_digital_value ((trace_ent ? trace_ent->config : NULL), event);
        ret[1] = '\0';
        break;
	case NODE_FLAGS_VECTOR:
		if (trace_ent)
			ret = wave_event_digital_vector_string (trace_ent, event->value.vector);
		else ret = g_strdup_printf ("?");
		break;
	case NODE_FLAGS_REAL:
		ret = wave_event_real_string (event->value.vector);
		break;
	case NODE_FLAGS_STRING:
		ret = g_strdup ((event->value.vector ? event->value.vector : "UNDEF"));
		break;
	case NODE_FLAGS_SYNC_HANDSHAKE:
	case NODE_FLAGS_DATA_HANDSHAKE:
		ret = wave_event_handshake_string (trace_ent, event, FALSE);
		break;
	default:
		ret = g_strdup_printf ("Can't display that value %s:%i", __FILE__, __LINE__);
	}

	return ret;
}

/* wave_event_digital_value : return the value of the digital bit in the
	given WaveEvent, note that the Node for this event *must* be DIGITAL */
char wave_event_digital_value (TraceConfig *trace_config, WaveEvent *event)
{
	char ret = '?';

	if (trace_config && TRACE_IS_INVERTED (trace_config))
		ret = wave_invert_digital_value (event->value.val);
	else ret = event->value.val;

	return ret;
}

/* wave_node_event_type_name : returns a printable form of the type portion of
	the given node flags */
char *wave_node_event_type_name (NodeFlags flag)
{
	char *ret;

	switch (flag & NODE_FLAGS_TYPE_MASK)
	{
	case NODE_FLAGS_DIGITAL:
		ret = "bit";
		break;
	case NODE_FLAGS_VECTOR:
		ret = "bit vector";
		break;
	case NODE_FLAGS_STRING:
		ret = "string";
		break;
	case NODE_FLAGS_REAL:
		ret = "floating point";
		break;
	case NODE_FLAGS_SYNC_HANDSHAKE:
		ret = "sync handshake";
		break;
	case NODE_FLAGS_DATA_HANDSHAKE:
		ret = "data handshake";
		break; default:
		ret = "unknown";
		break;
	}

	return ret;
}
