/*    
    The GTKWave electronic waveform viewer
    Copyright (C) 1999-2003 Tony Bybell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    `vcd.c'
    VCD file parser

    $Id: vcd.c,v 1.75 2003/11/09 22:21:35 bardslea Exp $
*/

#include "vcd.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "gui_cb.h"
#include "log.h"
#include "utils.h"
#include "alloc.h"
#include "tracepane.h"

#define VCD_BSIZ 32768          /* size of getch() emulation buffer--this val should be ok */
#define VCD_INDEXSIZ  (2 * 1024 * 1024) /* number of "indexed" entries to use for fast vcdid lookup */

typedef struct SymbolList_
{
    struct SymbolList_ *next;
    Symbol *val;
} SymbolList;

typedef struct StringList_
{
    struct StringList_ *next;
    char *str;
    int len;
} StringList;

struct VcdQueuedEvent_;

typedef struct VcdSymbol_
{
    struct VcdSymbol_ *root, *chain;
    struct VcdSymbol_ *next;
    Symbol *sym_chain;

    char *id;
    unsigned char vartype;
    Node *node;
    struct VcdQueuedEvent_ *ev; /* only if vartype==V_EVENT */
}
VcdSymbol;

static int vcd_max_value_size = 0;
static char *vcd_value_buffer = NULL;

#define GRAN 65536
void *char_alloc (int count)
{
    static char *char_arena = NULL;
    static char *char_ptr = NULL;
    static int char_remaining = 0;
    char *old_char_ptr;

    if (count > char_remaining)
    {
        char_arena = g_new (char, GRAN);

        char_ptr = char_arena;
        char_remaining = GRAN;
    }

    old_char_ptr = char_ptr;
    char_ptr += count;
    char_remaining -= count;

    return old_char_ptr;
}
char *perm_strdup (char *str)
{
    char *ret = char_alloc (strlen (str) + 1);

    strcpy (ret, str);
    return ret;
}

#define VCD_SYMBOL_POOL_SIZE (1024)
VcdSymbol *vcd_new_symbol (void)
{
    VcdSymbol *ret = g_new (VcdSymbol, 1);

    ret->root = ret->chain = NULL;
    ret->sym_chain = NULL;
    ret->next = NULL;
    ret->id = NULL;
    ret->vartype = 0;
    ret->ev = NULL;
    ret->node = NULL;

    return ret;
}

typedef struct VcdQueuedEvent_
{
    struct VcdQueuedEvent_ *next;
    VcdSymbol *sym;
    TimeType last_event_time;   /* make +1 == 0 if there's not an event there too */
} VcdQueuedEvent;

#undef VCD_BSEARCH_IS_PERFECT   /* bsearch is imperfect under linux, but OK under AIX */

static gboolean convert_to_reals = FALSE;

static FileDesc *vcd_handle = NULL;

static void add_verilog_event_events (void);
static void vcd_build_symbols (void);
static void vcd_cleanup (WaveReader * reader);
static void evcd_strcpy (char *dst, char *src);
static void vcd_parse (WaveReader * reader);

static int vcdbyteno = 0;       /* really should be size_t, but this is only used for debugging mangled traces */
static gboolean header_over = FALSE;
static gboolean dumping_off = FALSE;

static char vcd_hier_delimiter[2] = { 0, 0 }; /* fill in after rc reading code */

static VcdSymbol *pv = NULL, *rootv = NULL;

static StringList *slistroot = NULL, *slistcurr = NULL;
static char *slisthier = NULL;
static int slisthier_len = 0;

static TimeType vcd_end_time = -1;
static TimeType vcd_current_time = -1;
static TimeType vcd_start_time = -1;
static TimeType vcd_scaled_start_time = -1;
static TimeType vcd_scaled_end_time = -1;

enum Tokens
{
    T_VAR, T_END, T_SCOPE, T_UPSCOPE,
    T_COMMENT, T_DATE, T_DUMPALL, T_DUMPOFF, T_DUMPON,
    T_DUMPVARS, T_ENDDEFINITIONS,
    T_DUMPPORTS, T_DUMPPORTSOFF, T_DUMPPORTSON, T_DUMPPORTSALL,
    T_TIMESCALE, T_VERSION, T_VCDCLOSE,
    T_EOF, T_STRING, T_UNKNOWN_KEY
};

static char *vcd_tokens[] = {
    "var", "end", "scope", "upscope",
    "comment", "date", "dumpall", "dumpoff", "dumpon",
    "dumpvars", "enddefinitions",
    "dumpports", "dumpportsoff", "dumpportson", "dumpportsall",
    "timescale", "version", "vcdclose",
    "", "", ""
};

#define NUM_TOKENS 18

static int t_max_str = 1024;    /* was originally a const..now it reallocs */
static char *yytext = NULL;
static int yylen = 0, yylen_cache = 0;

#define T_GET tok=get_token();if((tok==T_END)||(tok==T_EOF))break;

static VcdSymbol *vcdsymroot = NULL, *vcdsymcurr = NULL;
static VcdSymbol **sorted = NULL;
static VcdSymbol **indexed = NULL;

enum VarTypes
{
    V_EVENT, V_PARAMETER, V_INTEGER, V_REAL, V_REG, V_SUPPLY0,
    V_SUPPLY1, V_TIME, V_TRI, V_TRIAND, V_TRIOR,
    V_TRIREG, V_TRI0, V_TRI1, V_WAND, V_WIRE, V_WOR, V_PORT,
    V_END, V_LB, V_COLON, V_RB, V_STRING
};

static char *vcd_vartypes[] = {
    "event", "parameter", "integer", "real", "reg", "supply0",
    "supply1", "time", "tri", "triand", "trior",
    "trireg", "tri0", "tri1", "wand", "wire", "wor", "port",
    "$end", "", "", "", ""
};

#define NUM_VTOKENS 19

static int numsyms = 0;

static VcdQueuedEvent *queuedevents = NULL;

static unsigned int vcd_minid = ~0;
static unsigned int vcd_maxid = 0;

static unsigned int vcdid_hash (char *s, int len)
{
    unsigned int val = 0;
    int i;

    s += (len - 1);

    for (i = 0; i < len; i++)
    {
        val *= 95;              /* was 94 but XL uses '!' as right hand side chars which act as leading zeros */
        val += (((unsigned char) *s) - 32); /* was 33 but XL ... */
        s--;
    }

    return (val);
}

/*
    bsearch compare
*/
static int vcdsymbsearchcompare (const void *s1, const void *s2)
{
    char *v1;
    VcdSymbol *v2;

    v1 = (char *) s1;
    v2 = *((VcdSymbol **) s2);

    return strcmp (v1, v2->id);
}

/*
    actual bsearch
*/
static VcdSymbol *bsearch_vcd (char *key, int len)
{
    VcdSymbol **v;
    VcdSymbol *t;

    if (indexed)
    {
        unsigned int hsh = vcdid_hash (key, len);

        if ((hsh >= vcd_minid) && (hsh <= vcd_maxid))
        {
            return (indexed[hsh - vcd_minid]);
        }

        return (NULL);
    }

    v = (VcdSymbol **) bsearch (key, sorted, numsyms, sizeof (VcdSymbol *), vcdsymbsearchcompare);

    if (v)
    {
#ifndef VCD_BSEARCH_IS_PERFECT
        for (;;)
        {
            t = *v;

            if ((v == sorted) || (strcmp ((*(--v))->id, key) != 0))
            {
                return t;
            }
        }
#else
        return *v;
#endif
    } else
    {
        return NULL;
    }
}

/*
    sort on vcdsymbol pointers
*/
static int vcdsymcompare (const void *s1, const void *s2)
{
    VcdSymbol *v1, *v2;

    v1 = *((VcdSymbol **) s1);
    v2 = *((VcdSymbol **) s2);

    return strcmp (v1->id, v2->id);
}

/*
    create sorted (by id) table
*/
static void create_sorted_table (void)
{
    VcdSymbol *v;
    VcdSymbol **pnt;
    unsigned int vcd_distance;

    if (sorted)
    {
        g_free (sorted);        /* this means we saw a 2nd enddefinition chunk! */
    }

    if (indexed)
    {
        g_free (indexed);
        indexed = NULL;
    }

    if (numsyms)
    {
        vcd_distance = vcd_maxid - vcd_minid + 1;

        if (vcd_distance <= VCD_INDEXSIZ)
        {
            int i;

            indexed = g_new (VcdSymbol *, vcd_distance);

            for (i = 0; i < vcd_distance; i++)
                indexed[i] = NULL;

            /* printf("%d symbols span ID range of %d, using indexing...\n", numsyms, vcd_distance); */

            v = vcdsymroot;
            while (v)
            {
                int nid = vcdid_hash (v->id, strlen (v->id));

                if (!indexed[nid - vcd_minid])
                    indexed[nid - vcd_minid] = v;
                v = v->next;
            }
        } else
        {
            pnt = sorted = g_new (VcdSymbol *, numsyms);
            v = vcdsymroot;
            while (v)
            {
                *(pnt++) = v;
                v = v->next;
            }

            qsort (sorted, numsyms, sizeof (VcdSymbol *), vcdsymcompare);
        }
    }
}

/*
    single char get inlined/optimized
*/
static char *vcdbuf = NULL, *vst = NULL, *vend = NULL;

static void getch_alloc (void)
{
    vend = vst = vcdbuf = g_new (char, VCD_BSIZ);
}

static void getch_free (void)
{
    if (vcdbuf)
        g_free (vcdbuf);
    vcdbuf = vst = vend = NULL;
}

static long lastpos;

static int getch_fetch (void)
{
    size_t rd;

    if (FILE_DESC_EOF (vcd_handle) || errno)
    {
        lastpos += (vend - vcdbuf);
        return -1;
    }
    vcdbyteno += (vend - vcdbuf);
    lastpos = FILE_DESC_TELL (vcd_handle);

    rd = FILE_DESC_READ (vcdbuf, sizeof (char), VCD_BSIZ, vcd_handle);

    vst = vcdbuf;
    vend = vcdbuf + rd;

    if (!rd)
        return -1;

    return (int) (*(vst++));
}

#define getch() ((vst!=vend)?((int)(*(vst++))):(getch_fetch()))

static char *varsplit = NULL, *vsplitcurr = NULL;
static int getch_patched (void)
{
    char ch;

    ch = *vsplitcurr;
    if (!ch)
    {
        return -1;
    }
    vsplitcurr++;
    return (int) ch;
}

/*
    simple tokenizer
*/
static int get_token (void)
{
    int ch;
    int i, len = 0;
    int is_string = 0;
    char *yyshadow;

    for (;;)
    {
        ch = getch ();
        if (ch < 0)
            return T_EOF;
        if (ch <= ' ')
            continue;           /* val<=' ' is a quick whitespace check      */
        break;                  /* (take advantage of fact that vcd is text) */
    }
    if (ch == '$')
    {
        yytext[len++] = ch;
        for (;;)
        {
            ch = getch ();
            if (ch < 0)
                return T_EOF;
            if (ch <= ' ')
                continue;
            break;
        }
    } else
    {
        is_string = 1;
    }

    for (yytext[len++] = ch;; yytext[len++] = ch)
    {
        if (len == t_max_str)
        {
            t_max_str *= 2;
            yytext = (char *) g_realloc (yytext, t_max_str + 1);
        }
        ch = getch ();
        if (ch <= ' ')
            break;
    }
    yytext[len] = 0;            /* terminator */

    if (is_string)
    {
        yylen = len;
        return T_STRING;
    }

    yyshadow = yytext;
    do
    {
        yyshadow++;
        for (i = 0; i < NUM_TOKENS; i++)
        {
            if (strcmp (yyshadow, vcd_tokens[i]) == 0)
            {
                return i;
            }
        }
    }
    while (*yyshadow == '$');   /* fix for RCS ids in version strings */

    return T_UNKNOWN_KEY;
}

static int var_prevch = 0;
static int get_vartoken_patched (void)
{
    int ch;
    int i, len = 0;

    if (!var_prevch)
    {
        for (;;)
        {
            ch = getch_patched ();
            if (ch < 0)
            {
                g_free (varsplit);
                varsplit = NULL;
                return V_END;
            }
            if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r'))
                continue;
            break;
        }
    } else
    {
        ch = var_prevch;
        var_prevch = 0;
    }

    if (ch == '[')
        return V_LB;
    if (ch == ':')
        return V_COLON;
    if (ch == ']')
        return V_RB;

    for (yytext[len++] = ch;; yytext[len++] = ch)
    {
        if (len == t_max_str)
        {
            t_max_str *= 2;
            yytext = (char *) g_realloc (yytext, t_max_str + 1);
        }
        ch = getch_patched ();
        if (ch < 0)
            break;
        if ((ch == ':') || (ch == ']'))
        {
            var_prevch = ch;
            break;
        }
    }
    yytext[len] = 0;            /* terminator */

    for (i = 0; i < NUM_VTOKENS; i++)
    {
        if (strcmp (yytext, vcd_vartypes[i]) == 0)
        {
            if (ch < 0)
            {
                g_free (varsplit);
                varsplit = NULL;
            }
            return i;
        }
    }

    yylen = len;
    if (ch < 0)
    {
        g_free (varsplit);
        varsplit = NULL;
    }
    return V_STRING;
}

static int get_vartoken (void)
{
    int ch;
    int i, len = 0;

    if (varsplit)
    {
        int rc = get_vartoken_patched ();

        if (rc != V_END)
            return rc;
        var_prevch = 0;
    }

    if (!var_prevch)
    {
        for (;;)
        {
            ch = getch ();
            if (ch < 0)
                return V_END;
            if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r'))
                continue;
            break;
        }
    } else
    {
        ch = var_prevch;
        var_prevch = 0;
    }

    if (ch == '[')
        return V_LB;
    if (ch == ':')
        return V_COLON;
    if (ch == ']')
        return V_RB;

    for (yytext[len++] = ch;; yytext[len++] = ch)
    {
        if (len == t_max_str)
        {
            t_max_str *= 2;
            yytext = (char *) g_realloc (yytext, t_max_str + 1);
        }
        ch = getch ();
        if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') || (ch < 0))
            break;
        if ((ch == '[') && (yytext[0] != '\\'))
        {
            varsplit = yytext + len; /* keep looping so we get the *last* one */
        } else if ((ch == ':' || ch == ']') && !varsplit && yytext[0] != '\\')
        {
            var_prevch = ch;
            break;
        }
    }
    yytext[len] = 0;            /* absolute terminator */
    if ((varsplit) && (yytext[len - 1] == ']'))
    {
        char *vst;

        vst = g_strdup (varsplit);

        *varsplit = 0x00;       /* zero out var name at the left bracket */
        len = varsplit - yytext;

        varsplit = vsplitcurr = vst;
        var_prevch = 0;
    } else
    {
        varsplit = NULL;
    }

    for (i = 0; i < NUM_VTOKENS; i++)
    {
        if (strcmp (yytext, vcd_vartypes[i]) == 0)
        {
            return i;
        }
    }

    yylen = len;
    return V_STRING;
}

static int get_strtoken (void)
{
    int ch;
    int len = 0;

    if (!var_prevch)
    {
        for (;;)
        {
            ch = getch ();
            if (ch < 0)
                return V_END;
            if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r'))
                continue;
            break;
        }
    } else
    {
        ch = var_prevch;
        var_prevch = 0;
    }

    for (yytext[len++] = ch;; yytext[len++] = ch)
    {
        if (len == t_max_str)
        {
            yytext = (char *) g_realloc (yytext, (t_max_str = t_max_str * 2) + 1);
        }
        ch = getch ();
        if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') || (ch < 0))
            break;
    }
    yytext[len] = 0;            /* terminator */

    yylen = len;
    return V_STRING;
}

static void sync_end (char *hdr)
{
    int tok;

    for (;;)
    {
        tok = get_token ();
        if ((tok == T_END) || (tok == T_EOF))
            break;
    }
}

static char *build_slisthier (void)
{
    StringList *s;
    int len = 0;

    if (!slistroot)
    {
        if (slisthier)
        {
            g_free (slisthier);
        }

        slisthier_len = 0;
        slisthier = g_new (char, 1);

        *slisthier = 0;
        return slisthier;
    }

    len = 0;
    for (s = slistroot; s; s = s->next)
    {
        len += s->len + (s->next ? 1 : 0);
    }

    slisthier_len = len;
    slisthier = g_new (char, len + 1);

    len = 0;
    for (s = slistroot; s; s = s->next)
    {
        strcpy (slisthier + len, s->str);
        len += s->len;
        if (s->next)
        {
            strcpy (slisthier + len, vcd_hier_delimiter);
            len++;
        }
    }
    return slisthier;
}

static void append_vcd_slisthier (char *str)
{
    StringList *s;

    s = g_new (StringList, 1);

    s->len = strlen (str);
    s->str = g_strdup (str);

    if (slistcurr)
    {
        slistcurr->next = s;
        slistcurr = s;
    } else
    {
        slistcurr = slistroot = s;
    }

    build_slisthier ();
}

static int wave_node_size (Node * n)
{
    int ret = 1;

    if (n->ext)
    {
        ret = n->ext->msi - n->ext->lsi;
        if (ret < 0)
            ret = -ret;
        ret++;
    }

    return ret;
}

static void parse_valuechange (void)
{
    VcdSymbol *v;
    char *vector;
    int vlen;

    switch (yytext[0])
    {
    case '0':
    case '1':
    case 'x':
    case 'X':
    case 'z':
    case 'Z':
        if (yylen > 1)
        {
            v = bsearch_vcd (yytext + 1, yylen - 1);
            if (!v)
            {
                wave_log (TRUE, "Near byte %d, Unknown VCD identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext + 1);
            } else if (v->vartype != V_EVENT)
            {
                char val = toupper (yytext[0]);

                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_DIGITAL, val, NULL);
            } else
            {
                char val = (dumping_off ? 'X' : '1'); /* only '1' is relevant */

                if (vcd_current_time != (v->ev->last_event_time + 1))
                {
                    /* dump degating event */
                    wave_node_insert_event (v->node, (v->ev->last_event_time + 1) * wave_config->time_scale, NODE_FLAGS_DIGITAL,
                      '0', NULL);
                }
                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_DIGITAL, val, NULL);
                v->ev->last_event_time = vcd_current_time;
            }
        } else
        {
            wave_log (TRUE, "Near byte %d, Malformed VCD identifier\n", vcdbyteno + (vst - vcdbuf));
        }
        break;

    case 'b':
    case 'B':
        {
            /* extract binary number then.. */
            yylen_cache = yylen;
            vector = g_strdup (yytext + 1);
            vlen = yylen - 1;
            get_strtoken ();
            v = bsearch_vcd (yytext, yylen);
            if (!v)
            {
                wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
                g_free (vector);
            } else
            {
                int size = wave_node_size (v->node);

                if ((v->vartype == V_REAL) || ((convert_to_reals) && ((v->vartype == V_INTEGER)
                      || (v->vartype == V_PARAMETER))))
                {
                    double *d;
                    char *pnt;
                    char ch;
                    TimeType k = 0;

                    pnt = vector;
                    while ((ch = *(pnt++)))
                    {
                        k = (k << 1) | ((ch == '1') ? 1 : 0);
                    }
                    g_free (vector);

                    d = g_new (double, 1);

                    *d = (double) k;

                    if (!v)
                    {
                        wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
                        g_free (d);
                    } else
                    {
                        wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_REAL, '\0',
                          (char *) d);
                    }
                    break;
                }

                if (vlen < size)
                {               /* fill in left part */
                    char extend;
                    int i, fill;

                    extend = (vector[0] == '1') ? '0' : vector[0];

                    fill = size - vlen;
                    for (i = 0; i < fill; i++)
                    {
                        vcd_value_buffer[i] = extend;
                    }
                    strcpy (vcd_value_buffer + fill, vector);
                } else if (vlen == size)
                {               /* straight copy */
                    strcpy (vcd_value_buffer, vector);
                } else
                {               /* too big, so copy only right half */
                    int skip = vlen - size;

                    strcpy (vcd_value_buffer, vector + skip);
                }

                if (size == 1)
                {
                    wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_DIGITAL,
                      toupper (vcd_value_buffer[0]), NULL);
                    g_free (vector);
                } else
                {
                    if (yylen_cache != (size + 1))
                    {
                        g_free (vector);
                        vector = g_new (char, size + 1);
                    }
                    strcpy (vector, vcd_value_buffer);
                    wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_VECTOR, '\0', vector);
                }
            }
            break;
        }

    case 'p':
        /* extract port dump value.. */
        yylen_cache = yylen;
        vector = g_strdup (yytext + 1);
        vlen = yylen - 1;

        get_strtoken ();        /* throw away 0_strength_component */
        get_strtoken ();        /* throw away 0_strength_component */
        get_strtoken ();        /* this is the id                  */
        v = bsearch_vcd (yytext, yylen);
        if (!v)
        {
            wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
            g_free (vector);
        } else
        {
            int size = wave_node_size (v->node);

            if ((v->vartype == V_REAL) || ((convert_to_reals) && ((v->vartype == V_INTEGER)
                  || (v->vartype == V_PARAMETER))))
                /* should never happen, but just in case.. */
            {
                double *d;
                char *pnt;
                char ch;
                TimeType k = 0;

                pnt = vector;
                while ((ch = *(pnt++)))
                {
                    k = (k << 1) | ((ch == '1') ? 1 : 0);
                }
                g_free (vector);

                d = g_new (double, 1);

                *d = (double) k;

                if (!v)
                {
                    wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
                    g_free (d);
                } else
                {
                    wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_REAL, '\0', (char *) d);
                }
                break;
            }

            if (vlen < size)
            {                   /* fill in left part */
                char extend;
                int i, fill;

                extend = '0';

                fill = size - vlen;
                for (i = 0; i < fill; i++)
                {
                    vcd_value_buffer[i] = extend;
                }
                evcd_strcpy (vcd_value_buffer + fill, vector);
            } else if (vlen == size)
            {                   /* straight copy */
                evcd_strcpy (vcd_value_buffer, vector);
            } else
            {                   /* too big, so copy only right half */
                int skip;

                skip = vlen - size;
                evcd_strcpy (vcd_value_buffer, vector + skip);
            }

            if (size == 1)
            {
                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_DIGITAL,
                  toupper (vcd_value_buffer[0]), NULL);
                g_free (vector);
            } else
            {
                if (yylen_cache < size)
                {
                    g_free (vector);
                    vector = g_new (char, size + 1);
                }
                strcpy (vector, vcd_value_buffer);
                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_VECTOR, '\0', vector);
            }
        }
        break;

    case 'r':
    case 'R':
        {
            double *d = g_new (double, 1);

            sscanf (yytext + 1, "%lg", d);

            get_strtoken ();
            v = bsearch_vcd (yytext, yylen);
            if (!v)
            {
                wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
                g_free (d);
            } else
            {
                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_REAL, '\0', (char *) d);
            }

            break;
        }

#ifndef STRICT_VCD_ONLY
    case 's':
    case 'S':
        {
            char *d = g_strdup (yytext + 1);

            get_strtoken ();
            v = bsearch_vcd (yytext, yylen);
            if (!v)
            {
                wave_log (TRUE, "Near byte %d, Unknown identifier: '%s'\n", vcdbyteno + (vst - vcdbuf), yytext);
                g_free (d);
            } else
            {
                /* Change this node to a string from real */
                /* FIXME FIXME FIXME, this *really* should go back and change all the
                   previous events on this signal to strings */
                v->node->flags &= ~NODE_FLAGS_TYPE_MASK;
                v->node->flags |= NODE_FLAGS_STRING;

                wave_node_insert_event (v->node, vcd_current_time * wave_config->time_scale, NODE_FLAGS_STRING, 's', (char *) d);
            }

            break;
        }
#endif
    }
}

static void evcd_strcpy (char *dst, char *src)
{
    static char *evcd = "DUNZduLHXTlh01?FAaBbCcf";
    static char *vcd = "01XZ0101XZ0101XZXXXXXXX";

    char ch;
    int i;

    while ((ch = *src))
    {
        for (i = 0; i < 23; i++)
        {
            if (evcd[i] == ch)
            {
                *dst = vcd[i];
                break;
            }
        }
        if (i == 23)
            *dst = 'X';

        src++;
        dst++;
    }

    *dst = 0;                   /* null terminate destination */
}

static void vcd_parse_defs (WaveReader * reader)
{
    int tok;

    for (;;)
    {
        tok = get_token ();
        switch (tok)
        {
        case T_COMMENT:
            sync_end ("COMMENT:");
            break;
        case T_DATE:
            sync_end ("DATE:");
            break;
        case T_VERSION:
            sync_end ("VERSION:");
            break;
        case T_TIMESCALE:
            {
                char prefix = ' ';
                int i, vtok = get_token ();

                if (vtok == T_END || vtok == T_EOF)
                {
                    break;
                }
                wave_config->time_scale = wave_str_to_time_type (yytext, NULL);
                if (!wave_config->time_scale)
                {
                    wave_config->time_scale = 1;
                }
                for (i = 0; i < yylen; i++)
                {
                    if ((yytext[i] < '0') || (yytext[i] > '9'))
                    {
                        prefix = yytext[i];
                        break;
                    }
                }
                if (prefix == ' ')
                {
                    vtok = get_token ();
                    if (vtok == T_END || vtok == T_EOF)
                        break;
                    prefix = yytext[0];
                }
                switch (prefix)
                {
                case ' ':
                case 'm':
                case 'u':
                case 'n':
                case 'p':
                case 'f':
                    wave_config->time_dimension = prefix;
                    break;
                case 's':
                    wave_config->time_dimension = ' ';
                    break;
                default:       /* unknown */
                    wave_config->time_dimension = 'n';
                    break;
                }

                sync_end (NULL);
                break;
            }
        case T_SCOPE:
            T_GET;
            T_GET;
            if (tok == T_STRING)
            {
                StringList *s;

                s = g_new (StringList, 1);

                s->next = NULL;
                s->len = yylen;
                s->str = g_strdup (yytext);

                if (slistcurr)
                {
                    slistcurr->next = s;
                    slistcurr = s;
                } else
                {
                    slistcurr = slistroot = s;
                }

                build_slisthier ();
            }
            sync_end (NULL);
            break;
        case T_UPSCOPE:
            if (slistroot)
            {
                StringList *s = slistroot;

                if (!s->next)
                {
                    g_free (s->str);
                    g_free (s);
                    slistroot = slistcurr = NULL;
                } else
                {
                    for (;; s = s->next)
                    {
                        if (!s->next->next)
                        {
                            g_free (s->next->str);
                            g_free (s->next);
                            s->next = NULL;
                            slistcurr = s;
                            break;
                        }
                    }
                }
                build_slisthier ();
            }
            sync_end (NULL);
            break;
        case T_VAR:
            {
                int vtok;
                char *name = NULL;
                int msi, lsi;
                int nid;
                int size;
                VcdSymbol *v = NULL;

                var_prevch = 0;
                varsplit = NULL;
                vtok = get_vartoken ();
                if (vtok > V_PORT)
                    goto bail;

                v = vcd_new_symbol ();

                v->vartype = vtok;
                msi = lsi = -1;

                if (vtok == V_PORT)
                {
                    vtok = get_vartoken ();
                    if (vtok == V_STRING)
                    {
                        size = wave_str_to_time_type (yytext, NULL);
                        if (!size)
                            size = 1;
                    } else if (vtok == V_LB)
                    {
                        vtok = get_vartoken ();
                        if (vtok == V_END)
                            goto err;
                        if (vtok != V_STRING)
                            goto err;
                        msi = wave_str_to_time_type (yytext, NULL);
                        vtok = get_vartoken ();
                        if (vtok == V_RB)
                        {
                            lsi = msi;
                            size = 1;
                        } else
                        {
                            if (vtok != V_COLON)
                                goto err;
                            vtok = get_vartoken ();
                            if (vtok != V_STRING)
                                goto err;
                            lsi = wave_str_to_time_type (yytext, NULL);
                            vtok = get_vartoken ();
                            if (vtok != V_RB)
                                goto err;

                            if (msi > lsi)
                                size = msi - lsi + 1;
                            else
                                size = lsi - msi + 1;
                        }
                    } else
                        goto err;

                    vtok = get_strtoken ();
                    if (vtok == V_END)
                        goto err;
                    v->id = perm_strdup (yytext);

                    nid = vcdid_hash (yytext, yylen);
                    if (nid < vcd_minid)
                        vcd_minid = nid;
                    if (nid > vcd_maxid)
                        vcd_maxid = nid;

                    vtok = get_vartoken ();
                    if (vtok != V_STRING)
                        goto err;
                    if (slisthier_len)
                    {
                        name = char_alloc (slisthier_len + 1 + yylen + 1);
                        strcpy (name, slisthier);
                        strcpy (name + slisthier_len, vcd_hier_delimiter);
                        copy_change_delim (name + slisthier_len + 1, yytext);
                    } else
                    {
                        name = char_alloc (yylen + 1);
                        copy_change_delim (name, yytext);
                    }

                    if (pv && strcmp (pv->node->nname, name) == 0)
                    {
                        pv->chain = v;
                        v->root = rootv;
                        if (pv == rootv)
                            pv->root = rootv;
                    } else
                        rootv = v;
                    pv = v;
                } else
                {
                    /* regular vcd var, not an evcd port var */
                    vtok = get_vartoken ();
                    if (vtok == V_END)
                        goto err;
                    size = wave_str_to_time_type (yytext, NULL);
                    if (!size)
                        size = 1;
                    vtok = get_strtoken ();
                    if (vtok == V_END)
                        goto err;
                    v->id = perm_strdup (yytext);
                    nid = vcdid_hash (yytext, yylen);
                    if (nid < vcd_minid)
                        vcd_minid = nid;
                    if (nid > vcd_maxid)
                        vcd_maxid = nid;

                    vtok = get_vartoken ();
                    if (vtok != V_STRING)
                        goto err;
                    if (slisthier_len)
                    {
                        name = g_new (char, slisthier_len + 1 + yylen + 1);

                        strcpy (name, slisthier);
                        strcpy (name + slisthier_len, vcd_hier_delimiter);
                        copy_change_delim (name + slisthier_len + 1, yytext);
                    } else
                    {
                        name = g_new (char, yylen + 1);

                        copy_change_delim (name, yytext);
                    }

                    if (pv && strcmp (pv->node->nname, name) == 0)
                    {
                        pv->chain = v;
                        v->root = rootv;
                        if (pv == rootv)
                            pv->root = rootv;
                    } else
                    {
                        rootv = v;
                    }
                    pv = v;

                    vtok = get_vartoken ();
                    if (vtok == V_END)
                        goto dumpv;
                    if (vtok != V_LB)
                        goto err;
                    vtok = get_vartoken ();
                    if (vtok != V_STRING)
                        goto err;
                    msi = wave_str_to_time_type (yytext, NULL);
                    vtok = get_vartoken ();
                    if (vtok == V_RB)
                    {
                        lsi = msi;
                        goto dumpv;
                    }
                    if (vtok != V_COLON)
                        goto err;
                    vtok = get_vartoken ();
                    if (vtok != V_STRING)
                        goto err;
                    lsi = wave_str_to_time_type (yytext, NULL);
                    vtok = get_vartoken ();
                    if (vtok != V_RB)
                        goto err;
                }

              dumpv:
                if (v->vartype == V_REAL ||
                  (convert_to_reals && (v->vartype == V_INTEGER || v->vartype == V_PARAMETER)))
                {
                    v->vartype = V_REAL;
                    size = 1;   /* override any data we parsed in */
                    msi = lsi = -1;
                } else if ((size > 1) && (msi <= 0) && (lsi <= 0))
                {
                    if (v->vartype == V_EVENT)
                    {
                        size = 1;
                    } else
                    {
                        /* any criteria for the direction here? */
                        msi = size - 1;
                        lsi = 0;
                    }
                } else if ((msi > lsi) && ((msi - lsi + 1) != size))
                {
                    if (v->vartype != V_EVENT && v->vartype != V_PARAMETER)
                        goto err;
                    size = msi - lsi + 1;
                } else if ((lsi >= msi) && ((lsi - msi + 1) != size))
                {
                    if (v->vartype != V_EVENT && v->vartype != V_PARAMETER)
                        goto err;
                    size = msi - lsi + 1;
                }

                /* Test for maximum size and keep a value buffer of this size */
                if (size > vcd_max_value_size)
                    vcd_max_value_size = size;

                {
                    NodeFlags node_type;

                    switch (v->vartype)
                    {
                    case V_REAL:
                        node_type = NODE_FLAGS_REAL;
                        break;
                    case V_STRING:
                        node_type = NODE_FLAGS_STRING;
                        break;
                    default:
                        node_type = (size > 1 ? NODE_FLAGS_VECTOR : NODE_FLAGS_DIGITAL);
                        break;
                    }
                    /* Test for maximum size and keep a value buffer of this size */

                    v->node = wave_new_node (name, reader, node_type);
                }

                if (msi != -1 && lsi != -1)
                    v->node->ext = wave_new_ext_node (msi, lsi);

                if (v->vartype == V_EVENT)
                {
                    VcdQueuedEvent *q;

                    v->ev = q = g_new (VcdQueuedEvent, 1);

                    q->sym = v;
                    q->last_event_time = -1;
                    q->next = queuedevents;
                    queuedevents = q;
                }

                if (!vcdsymroot)
                    vcdsymroot = vcdsymcurr = v;
                else
                {
                    vcdsymcurr->next = v;
                    vcdsymcurr = v;
                }
                numsyms++;

                goto bail;
              err:
                if (v)
                    g_free (v);

              bail:
                if (vtok != V_END)
                    sync_end (NULL);
                break;
            }
        case T_ENDDEFINITIONS:
            header_over = TRUE; /* do symbol table management here */
            create_sorted_table ();
            if ((!sorted) && (!indexed))
            {
                fprintf (stderr, "No symbols in VCD file..nothing to do!\n");
                wave_exit (EXIT_FAILURE);
            }
            return;
            break;
        case T_STRING:
            if (!header_over)
            {
                header_over = TRUE; /* do symbol table management here */
                create_sorted_table ();
                if ((!sorted) && (!indexed))
                    break;
            }
            /* catchall for events when header over */
            if (yytext[0] == '#')
            {
                TimeType time;

                time = wave_str_to_time_type (yytext + 1, NULL);

                if (vcd_start_time < 0)
                {
                    vcd_start_time = time;
                }

                vcd_current_time = time;
                if (vcd_end_time < time)
                    vcd_end_time = time; /* in case of malformed vcd files */
            } else
            {
                parse_valuechange ();
            }
            break;
        case T_DUMPALL:        /* dump commands modify vals anyway so */
        case T_DUMPPORTSALL:
            break;              /* just loop through..                 */
        case T_DUMPOFF:
        case T_DUMPPORTSOFF:
            dumping_off = TRUE;
            if ((!blackout_regions) || ((blackout_regions) && (blackout_regions->bstart <= blackout_regions->bend)))
            {
                BlackoutRegion *bt = g_new (BlackoutRegion, 1);

                bt->bstart = vcd_current_time;
                bt->next = blackout_regions;
                blackout_regions = bt;
            }
            break;
        case T_DUMPON:
        case T_DUMPPORTSON:
            dumping_off = FALSE;
            if ((blackout_regions) && (blackout_regions->bstart > blackout_regions->bend))
            {
                blackout_regions->bend = vcd_current_time;
            }
            break;
        case T_DUMPVARS:
        case T_DUMPPORTS:
            if (vcd_current_time < 0)
            {
                vcd_start_time = vcd_current_time = vcd_end_time = 0;
            }
            break;
        case T_VCDCLOSE:
            break;              /* next token will be '#' time related followed by $end */
        case T_END:            /* either closure for dump commands or */
            break;              /* it's spurious                       */
        case T_UNKNOWN_KEY:
            sync_end (NULL);    /* skip over unknown keywords */
            break;
        case T_EOF:
            if ((blackout_regions) && (blackout_regions->bstart > blackout_regions->bend))
            {
                blackout_regions->bend = vcd_current_time;
            }
            return;
            break;
        default:
            break;
        }
    }
}

static void add_verilog_event_events (void)
{
/* dump out any pending events 1st */
    VcdQueuedEvent *q;

    for (q = queuedevents; q; q = q->next)
    {
        VcdSymbol *v = q->sym;

        if (vcd_current_time != (v->ev->last_event_time + 1))
        {
            /* dump degating event */
            wave_node_insert_event (v->node, (v->ev->last_event_time + 1) * wave_config->time_scale, NODE_FLAGS_DIGITAL, '0', NULL);
        }
    }
}

/* vcd_new_symbol_list : new symbol list element containing `val', last in list */
static SymbolList *vcd_new_symbol_list (Symbol * val)
{
    SymbolList *ret = g_new (SymbolList, 1);

    ret->val = val;
    ret->next = NULL;

    return ret;
}

static void vcd_build_symbols (void)
{
    int max_slen = -1;
    Symbol *curnode = NULL;
    SymbolList *sym_chain = NULL, *sym_curr = NULL;
    int duphier = 0;
    char *str = NULL;
    char hashdirty;
    VcdSymbol *vprime, *v;

    v = vcdsymroot;
    while (v)
    {
        int msi = (v->node->ext ? v->node->ext->msi : -1);
        int lsi = (v->node->ext ? v->node->ext->lsi : -1);
        int delta = ((lsi - msi) < 0) ? -1 : 1;
        gboolean substnode = FALSE;
        int slen = strlen (v->node->nname);
        int size = wave_node_size (v->node);

        if (slen > max_slen)
        {
            max_slen = slen;
            str = wave_alloca (slen + 32);
        }
        strcpy (str, v->node->nname);

/* Don't understand this, crop AB FIXME
        if (msi >= 0)
        {
            strcpy (str + slen, vcd_hier_delimiter);
            slen++;
        }
*/

        vprime = bsearch_vcd (v->id, strlen (v->id));
        if (vprime != v)
        {                       /* hash mish means dup net */
            if (size != wave_node_size (vprime->node))
                wave_log (TRUE, "ERROR: Duplicate IDs with differing width: %s %s\n", v->node->nname,
                  vprime->node->nname);
            else
                substnode = TRUE;
        }

        if ((size == 1) && (v->vartype != V_REAL))
        {
            Symbol *s = NULL;

            if (msi >= 0)
                sprintf (str + slen, "[%d]", msi);

            hashdirty = 0;
            if (symfind (str))
            {
                char *dupfix = g_new (char, max_slen + 32);

                hashdirty = 1;

                do
                {
                    sprintf (dupfix, "$DUP%d%s%s", duphier++, vcd_hier_delimiter, str);
                }
                while (symfind (dupfix));

                strcpy (str, dupfix);
                g_free (dupfix);
                duphier = 0;    /* reset for next duplicate resolution */
            }
            /* fallthrough */

            s = symadd (str, (hashdirty ? hash (str) : hashcache));

            s->n = v->node;
            if (substnode)
            {
                Node *n = s->n;
                Node *n2 = vprime->node;

                wave_alias_node (n, n2);
            }

            s->n->nname = s->name;
            if (!firstnode)
                firstnode = curnode = s;
            else
            {
                curnode->next_in_file = s;
                curnode = s;
            }

            wave_num_symbols++;
            msi += delta;

            if (v->root)
            {
                s->vec_root = (Symbol *) v->root; /* these will get patched over */
                s->vec_chain = (Symbol *) v->chain; /* these will get patched over */
                v->sym_chain = s;

                if (!sym_chain)
                {
                    sym_curr = vcd_new_symbol_list (s);
                    sym_chain = sym_curr;
                } else
                {
                    sym_curr->next = vcd_new_symbol_list (s);
                    sym_curr = sym_curr->next;
                }
            }
        } else
        {                       /* atomic vector */
            if (v->vartype != V_REAL)
            {
                sprintf (str + slen, "[%d:%d]", msi, lsi);
            } else
            {
                str[slen] = '\0';
            }

            hashdirty = 0;
            if (symfind (str))
            {
                char *dupfix = g_new (char, max_slen + 32);

                hashdirty = 1;

                do
                {
                    sprintf (dupfix, "$DUP%d%s%s", duphier++, vcd_hier_delimiter, str);
                }
                while (symfind (dupfix));

                strcpy (str, dupfix);
                g_free (dupfix);
                duphier = 0;    /* reset for next duplicate resolution */
            }
            /* fallthrough */
            {
                Symbol *s = symadd (str, hashdirty ? hash (str) : hashcache);

                s->n = v->node;
                if (substnode)
                {
                    Node *n, *n2;

                    n = s->n;
                    n2 = vprime->node;

                    wave_alias_node (n, n2);
                }

                s->n->nname = s->name;
                if (!firstnode)
                {
                    firstnode = curnode = s;
                } else
                {
                    curnode->next_in_file = s;
                    curnode = s;
                }

                wave_num_symbols++;
            }
        }

        v = v->next;
    }

    if (sym_chain)
    {
        sym_curr = sym_chain;
        while (sym_curr)
        {
            sym_curr->val->vec_root = ((VcdSymbol *) sym_curr->val->vec_root)->sym_chain;

            if ((VcdSymbol *) sym_curr->val->vec_chain)
                sym_curr->val->vec_chain = ((VcdSymbol *) sym_curr->val->vec_chain)->sym_chain;

            sym_chain = sym_curr;
            sym_curr = sym_curr->next;
            g_free (sym_chain);
        }
    }
}

static void vcd_sort_symbols (void)
{
    int i;
    Symbol *curnode = firstnode;

    wave_symbols = g_new (Symbol *, wave_num_symbols);

    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;
        int len;

        wave_symbols[i] = curnode;
        if ((len = strlen (subst = curnode->name)) > longestname)
            longestname = len;
        curnode = curnode->next_in_file;
        while ((ch = (*subst)))
        {
            if (ch == hier_delimiter)
            {
                *subst = 0x01;
            }                   /* forces sort at hier boundaries */
            subst++;
        }
    }

    /* quicksort(wave_symbols,0,wave_num_symbols-1); */
    /*
     * quicksort deprecated because it degenerates on sorted traces...
     * badly.  very badly.  (O(n**2) vs. O(n log n))
     */

    wave_heapsort (wave_symbols, wave_num_symbols);

    for (i = 0; i < wave_num_symbols; i++)
    {
        char *subst, ch;

        subst = wave_symbols[i]->name;
        while ((ch = (*subst)))
        {
            if (ch == 0x01)
            {
                *subst = hier_delimiter;
            }                   /* restore back to normal */
            subst++;
        }
    }

    wave_symbols_are_sorted = TRUE;

    init_tree ();
    for (i = 0; i < wave_num_symbols; i++)
        build_tree_from_name (wave_symbols[i]->name, i);

    treeprune (treeroot);
}

static void vcd_cleanup (WaveReader * reader)
{
    StringList *s, *s2;
    VcdSymbol *v, *vt;

    if (indexed)
    {
        g_free (indexed);
        indexed = NULL;
    }

    if (sorted)
    {
        g_free (sorted);
        sorted = NULL;
    }

    v = vcdsymroot;
    while (v)
    {
        if (v->ev)
            g_free (v->ev);
        vt = v;
        v = v->next;
        g_free (vt);
    }
    vcdsymroot = vcdsymcurr = NULL;

    if (slisthier)
    {
        g_free (slisthier);
        slisthier = NULL;
    }
    s = slistroot;
    while (s)
    {
        s2 = s->next;
        if (s->str)
            g_free (s->str);
        g_free (s);
        s = s2;
    }

    slistroot = slistcurr = NULL;
    slisthier_len = 0;
    queuedevents = NULL;        /* deallocated in the symbol stuff */

    FILE_DESC_CLOSE (vcd_handle);

    if (yytext)
    {
        g_free (yytext);
        yytext = NULL;
    }
}

static TimeType vcd_main (WaveReader * reader)
{
    const char *fname = reader->filename;

    /* Check filename */
    if (!fname)
    {
        fprintf (stderr, "cannot read VCD file from stdin\n");
        wave_exit (EXIT_FAILURE);
		return 0;
    }

    pv = rootv = NULL;

    errno = 0;                  /* reset in case it's set for some reason */

    yytext = g_new (char, t_max_str + 1);

    /* Set vcd hierarchy delimiter */
    vcd_hier_delimiter[0] = hier_delimiter;

    if (!hier_was_explicitly_set)
    {                           /* set default hierarchy split char */
        hier_delimiter = '.';
    }

    vcd_handle = file_desc_open (fname, "rt");

    if (!vcd_handle)
    {
        wave_log (TRUE, "Error opening VCD file '%s'.\n", fname);
        wave_exit (EXIT_FAILURE);
		return 0;
    }

    getch_alloc ();             /* alloc membuff for vcd getch buffer */

    /* Parse VCD file */
    build_slisthier ();
    vcd_parse (reader);

    if ((!sorted) && (!indexed))
    {
        fprintf (stderr, "No symbols in VCD file..is it malformed?  Exiting!\n");
        wave_exit (EXIT_FAILURE);
		return 0;
        /* AM : this is no more useful, there are no more trailing dummy signals ... */
    }
    add_verilog_event_events ();

    /* Show the starting and ending times */
    wave_log (FALSE, "[" LL_PRINTF_FORMAT "] start time.\n[" LL_PRINTF_FORMAT "] end time.\n", vcd_start_time,
      vcd_end_time);

    vcd_build_symbols ();
    vcd_sort_symbols ();

    getch_free ();              /* free membuff for vcd getch buffer */

    /* Compute the starting and ending times for the dump */
    vcd_scaled_start_time = vcd_start_time * wave_config->time_scale;
    vcd_scaled_end_time = vcd_end_time * wave_config->time_scale;

    /* If we are in static mode and there is no time span, there is nothing to see : exit */
    if (!(reader->flags & WAVE_READER_DYNAMIC) && (vcd_scaled_start_time == vcd_scaled_end_time
        || vcd_scaled_end_time == 0))
    {
        fprintf (stderr, "VCD times range is equal to zero.  Exiting.\n");
        // wave_exit (EXIT_FAILURE);
    }

    return vcd_scaled_end_time;
}

static void vcd_parse (WaveReader * reader)
{
    int tok;
    TimeType time;

    vcd_parse_defs (reader);

    vcd_value_buffer = g_new (char, vcd_max_value_size + 1);

    for (;;)
    {
        tok = get_token ();
        switch (tok)
        {
        case T_COMMENT:
            sync_end ("COMMENT:");
            break;

        case T_STRING:
            if (yytext[0] == '#')
            {
                time = wave_str_to_time_type (yytext + 1, NULL);

                if (vcd_start_time < 0)
                    vcd_start_time = time;

                if (vcd_end_time < time)
                    vcd_end_time = time;
                /* Set Times */
                vcd_current_time = time;
            } else
            {
                parse_valuechange ();
            }
            /* If last time go to end to calculate end time */

            break;
        case T_DUMPALL:        /* dump commands modify vals anyway so */
        case T_DUMPPORTSALL:
            break;              /* just loop through..                 */
        case T_DUMPOFF:
        case T_DUMPPORTSOFF:
            dumping_off = TRUE;
            if ((!blackout_regions) || ((blackout_regions) && (blackout_regions->bstart <= blackout_regions->bend)))
            {
                BlackoutRegion *bt = g_new (BlackoutRegion, 1);

                bt->bstart = vcd_current_time;
                bt->next = blackout_regions;
                blackout_regions = bt;
            }
            break;
        case T_DUMPON:
        case T_DUMPPORTSON:
            dumping_off = FALSE;
            if ((blackout_regions) && (blackout_regions->bstart > blackout_regions->bend))
            {
                blackout_regions->bend = vcd_current_time;
            }
            break;
        case T_DUMPVARS:
            if (vcd_current_time < 0)
            {
                vcd_start_time = vcd_current_time = vcd_end_time = 0;
            }
        case T_DUMPPORTS:
            if (vcd_current_time < 0)
            {
                vcd_start_time = vcd_current_time = vcd_end_time = 0;
            }
            break;
        case T_VCDCLOSE:
            break;              /* next token will be '#' time related followed by $end */
        case T_END:            /* either closure for dump commands or */
            break;              /* it's spurious                       */
        case T_UNKNOWN_KEY:
            sync_end (NULL);    /* skip over unknown keywords */
            break;
        case T_EOF:
            if ((blackout_regions) && (blackout_regions->bstart > blackout_regions->bend))
            {
                blackout_regions->bend = vcd_current_time;
            }
            vst = vend;
            return;
            break;
        default:
            break;
        }
    }
}

static void vcd_seek_end (WaveReader * reader)
{
    int tok = 0;
    int ch;
    TimeType time;

    ch = FILE_DESC_GETC (vcd_handle);
    if (ch == -1)
        return;

    while (ch != '\n' && ch != EOF)
    {
        ch = FILE_DESC_GETC (vcd_handle); /* Get to "word boundary" */
    }
    getch_alloc ();
    for (;;)
    {
        tok = get_token ();
        switch (tok)
        {
        case T_COMMENT:
            sync_end ("COMMENT:");
            break;
        case T_STRING:
            if (!header_over)
            {
                header_over = TRUE; /* do symbol table management here */
                create_sorted_table ();
                if ((!sorted) && (!indexed))
                    break;
            }
            if (yytext[0] == '#')
            {
                time = wave_str_to_time_type (yytext + 1, NULL);

                if (vcd_start_time < 0)
                    vcd_start_time = time;

                if (vcd_end_time < time)
                    vcd_end_time = time;
                /* Set Times */
                vcd_current_time = time;
            }
            /* If last time go to end to calculate end time */
            break;
        case T_DUMPALL:        /* dump commands modify vals anyway so */
        case T_DUMPPORTSALL:
            break;              /* just loop through..                 */
        case T_DUMPOFF:
        case T_DUMPPORTSOFF:
            dumping_off = TRUE;
            break;
        case T_DUMPON:
        case T_DUMPPORTSON:
            dumping_off = FALSE;
            break;
        case T_DUMPVARS:
            if (vcd_current_time < 0)
            {
                vcd_start_time = vcd_current_time = vcd_end_time = 0;
            }
            break;
        case T_DUMPPORTS:
            if (vcd_current_time < 0)
            {
                vcd_start_time = vcd_current_time = vcd_end_time = 0;
            }
            break;
        case T_VCDCLOSE:
            break;              /* next token will be '#' time related followed by $end */
        case T_END:            /* either closure for dump commands or */
            break;              /* it's spurious                       */
        case T_UNKNOWN_KEY:
            sync_end (NULL);    /* skip over unknown keywords */
            break;
        case T_EOF:
            goto end;
            break;
        default:
            break;
        }
    }
  end:
    tims.last = vcd_scaled_end_time;
    wave_set_end_time (tims.last);

    clicked_zoom_fit (NULL, NULL);
    wave_redraw_trace_pane (wave_window);
    return;
}

static int f_append_vcd_hier (char *str, int offset)
{
    append_vcd_slisthier (str);
    return 0;
}

static int f_convert_to_reals (char *str, int offset)
{
    convert_to_reals = wave_parse_boolean_rc_entry (str);
    return 0;
}

static WaveRcEntry vcd_entries[] = {
    {"vcd_append_hier", f_append_vcd_hier, NULL},
    {"vcd_convert_to_reals", f_convert_to_reals, NULL},
};

WaveReader *vcd_init (void)
{
    WaveReader *ret = wave_new_reader ("vcd", "Verilog Value Change Dump (VCD)",
      WAVE_READER_DYNAMIC | WAVE_READER_COMPRESSED, NULL);

    wave_reader_add_function (ret, "load-declarations", vcd_main);
    wave_reader_add_function (ret, "seek-end", vcd_seek_end);
    wave_reader_add_function (ret, "cleanup", vcd_cleanup);

    wave_reader_add_rc_entries (ret, 2, vcd_entries);

    return ret;
}
